#ifndef _drv_sdd_io_H
#define _drv_sdd_io_H
/*******************************************************************************

                              Copyright (c) 2014
                            Lantiq Deutschland GmbH
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_io.h
   This file contains the defines specific to the SDD driver interface
   and is used by applications.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "ifx_types.h"

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */
#define SDD_IOC_MAGIC 'S'

/**
   Read relevant version information.

   The parameter points to a \ref SDD_IO_Version_t structure

   \code
   SDD_IO_Version_t version;
   int err;

   memset (&version, 0, sizeof(version));

   err = ioctl (fd, FIO_SDD_VERS, &version);
   \endcode
*/
#define FIO_SDD_VERS             _IOR (SDD_IOC_MAGIC,  1, SDD_IO_Version_t)

/**
   Set the driver report levels if the driver is compiled with
   ENABLE_TRACE

   \remarks
   valid arguments are

   \arg 0: off

   \arg 1: low, high output

   \arg 2: normal, general information and warnings

   \arg 3: high, only errors are reported
*/
#define FIO_SDD_REPORT_SET       _IOW (SDD_IOC_MAGIC, 2, IFX_uint32_t)

/**
   Write an SDD command.
   The parameter points to a \ref SDD_IO_MB_CMD_t structure.

   \note
   This interface is only for debugging and testing purposes.
   The use of this interface may disturb the driver operation
*/
#define FIO_SDD_WCMD             _IOW (SDD_IOC_MAGIC, 3, SDD_IO_MB_CMD_t)

/**
   Read command.
   The parameter points to a \ref SDD_IO_MB_CMD_t structure

   \note
   This interface is only for debugging and testing purposes.
   The use of this interface may disturb the driver operation

   \code
   SDD_IO_MB_CMD_t ioCmd;
   int err;

   //read back version
   ioCmd.cmd = 0x8400E608;

   err = ioctl (fd, FIO_SDD_RCMD, &ioCmd);
   \endcode
*/
#define FIO_SDD_RCMD             _IOWR (SDD_IOC_MAGIC, 4, SDD_IO_MB_CMD_t)

/**
   Does a download according to bbd format.
   Parameter is a pointer to a \ref SDD_BBD_Download_t structure.
   \tapiv3
   This structure is a generic bbd library structure.
   \endtapiv3
*/
#define FIO_SDD_BBD_DOWNLOAD     _IOW (SDD_IOC_MAGIC, 5, SDD_BBD_Download_t)

/**
   Does a download of firmware.
   Parameter is a pointer to a \ref SDD_FW_Download_t structure.
*/
#define FIO_SDD_FW_DOWNLOAD      _IOW (SDD_IOC_MAGIC, 6, SDD_FW_Download_t)

/**
   Read command.
   The parameter points to a \ref SDD_IO_MB_CMD_t structure

   \note
   This interface is only for debugging and testing purposes.
   The use of this interface may disturb the driver operation
*/
#define FIO_SDD_MB_READ          _IOWR (SDD_IOC_MAGIC, 7, FIO_SDD_MB_READ_t)

/**
   Read VCODEC register. The parameter points to a \ref FIO_SDD_REG_IO_t
   structure

   \note
   This interface is only for debugging and testing purposes.
   The use of this interface may disturb the driver operation
*/
#define FIO_SDD_RREG             _IOWR (SDD_IOC_MAGIC,  8, FIO_SDD_REG_IO_t)

/**
   Write VCODEC register.
   The parameter points to a \ref FIO_SDD_REG_IO_t structure

   \note
   This interface is only for debugging and testing purposes.
   The use of this interface may disturb the driver operation
*/
#define FIO_SDD_WREG             _IOW  (SDD_IOC_MAGIC,  9, FIO_SDD_REG_IO_t)

/**
   Reset SDD firmware.

   \note
   This interface is only for debugging and testing purposes.
   The use of this interface may disturb the driver operation
*/
#define FIO_SDD_RESET            _IO   (SDD_IOC_MAGIC, 10)

/**
   Enable (1) or disable (0) debug mode. In debug mode events form the VCODEC
   are not proceed by the driver.

   \note
   This interface is only for debugging and testing purposes.
   The use of this interface may disturb the driver operation
*/
#define FIO_SDD_DEBUG            _IOW  (SDD_IOC_MAGIC, 11, IFX_uint32_t)


/* ========================================================================== */
/*                             Type definitions                               */
/* ========================================================================== */

/** Version Io structure */
typedef struct
{
   /** number of supported FXS channels */
   IFX_uint8_t  nFxsCh;
   /** number of supported FXO channels */
   IFX_uint8_t  nFxoCh;
   /** chip revision */
   IFX_uint16_t nHwRev;
   /** asdsp revision */
   IFX_uint16_t nAsdspRev;
   /** FW revision */
   IFX_uint32_t nFwRev;
   /** Device ID */
   IFX_uint16_t nDevID;
   /** driver version */
   IFX_uint32_t nDrvVers;
} SDD_IO_Version_t;

typedef struct
{
   /** Device index. */
   IFX_uint16_t dev;
   /** valid Firmware byte pointer */
   IFX_uint8_t    *pPRAMfw;
   /** size of firmware in bytes */
   IFX_uint32_t   pram_size;
} SDD_FW_Download_t;

#ifdef TAPI_ONE_DEVNODE
typedef struct
{
   /** Device index. */
   IFX_uint16_t dev;
   /** Channel 'module' index. */
   IFX_uint16_t ch;
   /** If IFX_TRUE distribute the BBD on all available channels,
     otherwise the ch field addresses the channel */
   IFX_uint8_t bBroadcast;
   /** block based download buffer,
       big-endian aligned */
   IFX_uint8_t *buf;
   /** size of buffer in bytes */
   IFX_uint32_t size;
} SDD_BBD_Download_t;
#else /* TAPI_ONE_DEVNODE */
/**
   Structure used for SDD bbd download.
*/
typedef struct
{
   /** block based download buffer,
       big-endian aligned */
   IFX_uint8_t *buf;
   /** size of buffer in bytes */
   IFX_uint32_t size;
} SDD_BBD_Download_t;
#endif /* TAPI_ONE_DEVNODE */

typedef struct
{
   /** command 1 according users manual */
   IFX_uint32_t cmd;
   /** read or write data */
   IFX_uint32_t pData [128];
} SDD_IO_MB_CMD_t;

typedef struct
{
   /** SDD event data */
   IFX_uint32_t data [16];
} FIO_SDD_MB_READ_t;

typedef struct
{
   /** offset of register */
   IFX_uint16_t reg;
   /** buffer for data */
   IFX_uint16_t data;
} FIO_SDD_REG_IO_t;

#endif /* _drv_sdd_io_H */
