/******************************************************************************

                            Copyright (c) 2006-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

/**
   \file drv_vmmc_alm.c Implementation of the ALM module.
*/

/* ============================= */
/* Includes                      */
/* ============================= */
#include "drv_api.h"

#ifdef VMMC_FEAT_SLIC

#include "drv_vmmc_api.h"
#include "drv_vmmc_alm.h"
#include "drv_vmmc_pcm.h"
#include "drv_vmmc_con.h"

#include "drv_sdd_api.h"

/* ============================= */
/* Global variables              */
/* ============================= */

/* ============================= */
/* Local Macros & Definitions    */
/* ============================= */

/* ============================= */
/* Local variables and types     */
/* ============================= */
/** Structure for the analog line module channel
    including firmware message cache */
struct VMMC_ALMCH
{
   /* The current NB/WB mode */
   OPMODE_SMPL                   module_mode;
   /* Configuration of the pcm connections on DSP and SDD side. */
   IFX_TAPI_PCM_CFG_t            dsp_pcm_cfg,
                                 sdd_pcm_cfg;
   IFX_boolean_t                 dsp_pcm_enabled,
                                 sdd_pcm_enabled;
   /* Offset for the timeslots in this pcm connection. */
   IFX_uint16_t                  ts_offset;
};

/* ============================= */
/* Local function declaration    */
/* ============================= */

static IFX_int32_t vmmc_alm_VoiceConnect(
                        VMMC_CHANNEL *pCh,
                        IFX_boolean_t bConfUpdate);

static IFX_int32_t vmmc_alm_VoiceDisconnect(
                        VMMC_CHANNEL *pCh);

static IFX_int32_t VMMC_TAPI_LL_ALM_Line_Type_Set (
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_LINE_TYPE_t nType);

static IFX_int32_t VMMC_TAPI_LL_ALM_LEC_Set (
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        TAPI_LEC_DATA_t *pLecConf);

/* ============================= */
/* Global function declaration   */
/* ============================= */

/* ============================= */
/* Function definitions          */
/* ============================= */

 /**
   Configure or check ALM module for given sampling mode.

   \param  pCh             Pointer to the VMMC channel structure.
   \param  action          Action to be executed (set or check).
   \param  sigarray_mode   Signalling array operation mode (16kHz or 8 kHz).
   \param  module_mode     Operation mode of the module (16kHz or 8 kHz).

   \return
   If action is SM_SET: IFX_SUCCESS or IFX_ERROR.
   If action is SM_CHECK: IFX_TRUE when module would do a switch or IFX_FALSE
                          if nothing needs to be done.

*/
IFX_int32_t  VMMC_ALM_SamplingMode (VMMC_CHANNEL *pCh,
                                    SM_ACTION action,
                                    OPMODE_SMPL sigarray_mode,
                                    OPMODE_SMPL module_mode)
{
   VMMC_ALMCH_t *pALM = pCh->pALM;
   IFX_int32_t ret = VMMC_statusOk;

   /* calling function should ensure valid parameters */
   VMMC_ASSERT(pCh != NULL);
   /* Signal array 8 kHz AND module 16 kHz is disallowed */
   VMMC_ASSERT((sigarray_mode != NB_8_KHZ) || (module_mode != WB_16_KHZ));

   /* If action is check return if module needs a switch */
   if (action != SM_SET)
   {
      return (module_mode != pALM->module_mode) ||
             VMMC_PCM_SamplingMode_NoTranslate (
               pCh, action, sigarray_mode, module_mode)
             ? IFX_TRUE : IFX_FALSE;
   }

   /* Action is set */

   if (module_mode != pALM->module_mode)
   {
      /* Remove the existing PCM connection between SDD and DSP. */
      ret = vmmc_alm_VoiceDisconnect(pCh);

      /* Create PCM connection between SDD and DSP with appropriate codec. */
      if(VMMC_SUCCESS(ret))
      {
         TRACE(VMMC, DBG_LEVEL_LOW,
               ("Set ALM channel %u WB = %d\n",
                pCh->nChannel - 1, (module_mode != WB_16_KHZ) ? 0 : 1));

         pALM->dsp_pcm_cfg.nResolution =
         pALM->sdd_pcm_cfg.nResolution =
            (module_mode != WB_16_KHZ) ? IFX_TAPI_PCM_RES_NB_LINEAR_16BIT:
                                         IFX_TAPI_PCM_RES_WB_LINEAR_16BIT;

         ret = vmmc_alm_VoiceConnect(pCh, IFX_FALSE);

         /* In case of an error disconnect all to prevent a half way
            connection. */
         if(!VMMC_SUCCESS(ret))
            vmmc_alm_VoiceDisconnect(pCh);
      }
   }
   else
   {
      TRACE(VMMC, DBG_LEVEL_LOW,
            ("Sampling rate of ALM on channel %u already matching\n",
            pCh->nChannel - 1));
   }

   if(VMMC_SUCCESS(ret))
   {
      ret = VMMC_PCM_SamplingMode_NoTranslate (
               pCh, action, sigarray_mode, module_mode);
   }

   /* Remember the just set mode for the next time. */
   if(VMMC_SUCCESS(ret))
      pALM->module_mode = module_mode;

   return ret;
}


/**
   base ALM configuration

   Use this function where needed to set the base configuration
   of Analog Line Module.

   \param  pCh             Pointer to the VMMC channel structure.

   \return
   - VMMC_statusOk         If successful
*/
IFX_int32_t VMMC_ALM_baseConf (VMMC_CHANNEL *pCh)
{
   IFX_int32_t ret;

   ret = vmmc_alm_VoiceConnect(pCh, IFX_TRUE);

   SDD_ALM_BaseConf(pCh);

   RETURN_STATUS (ret);
}


/**
   Initalize the analog module and the cached firmware messages.

   \param  pCh             Pointer to the VMMC channel structure.
*/
IFX_void_t VMMC_ALM_InitCh (VMMC_CHANNEL *pCh)
{
   VMMC_ALMCH_t *pALM = pCh->pALM;
   IFX_uint8_t ch = pCh->nChannel - 1;

   pALM->dsp_pcm_enabled = pALM->sdd_pcm_enabled = IFX_FALSE;

   memset(&pALM->dsp_pcm_cfg, 0x00, sizeof(pALM->dsp_pcm_cfg));
   memset(&pALM->sdd_pcm_cfg, 0x00, sizeof(pALM->sdd_pcm_cfg));
   /* SDD transmit is received by DSP and vice versa. By definition
      the timeslots for RX and TX are identical for each connection.
      For WB 16-bit linear 4 consecutive timeslots are needed. */
   pALM->dsp_pcm_cfg.nTimeslotRX = pALM->sdd_pcm_cfg.nTimeslotTX = 4 * ch;
   pALM->dsp_pcm_cfg.nTimeslotTX = pALM->sdd_pcm_cfg.nTimeslotRX = 4 * ch;
   pALM->dsp_pcm_cfg.nResolution = pALM->sdd_pcm_cfg.nResolution =
                                          IFX_TAPI_PCM_RES_NB_LINEAR_16BIT;
   pALM->module_mode = NB_8_KHZ;
   pALM->ts_offset = 0;

   /* initialise the structures used for connecting modules */
   VMMC_CON_Init_AlmCh (pCh);
}


/**
   Set the signal inputs of the cached fw message for the given channel

   In this system the ALM modules are emulated by PCM modules on the same
   channel.

   \param  pCh             Pointer to the VMMC channel structure.

   \return
   - VMMC_statusOk if ok
   - Return value from VMMC_PCM_Set_Inputs().
*/
IFX_int32_t VMMC_ALM_Set_Inputs (VMMC_CHANNEL *pCh)
{
   return VMMC_PCM_Set_Inputs(pCh);
}


/**
   Stop ALM on this channel

   \param  pCh             Pointer to the VMMC channel structure.

   \return
   - VMMC_statusOk         If successful
   - VMMC_statusCmdWr      Writing the command has failed
*/
IFX_int32_t VMMC_ALM_ChStop (VMMC_CHANNEL *pCh)
{
   IFX_int32_t ret = VMMC_statusOk;

   /* calling function should ensure valid parameters */
   VMMC_ASSERT(pCh != IFX_NULL);

   if (pCh->pALM != IFX_NULL)
   {
      ret = vmmc_alm_VoiceDisconnect(pCh);
   }

   RETURN_STATUS(ret);
}


/**
   Allocate data structure of the ALM module in the given channel.

   \param  pCh             Pointer to the VMMC channel structure.

   \return
      - VMMC_statusOk if ok
      - VMMC_statusNoMem Memory allocation failed -> struct not created

   \remarks The channel parameter is no longer checked because the calling
   function assures correct values.
*/
IFX_int32_t VMMC_ALM_Allocate_Ch_Structures (VMMC_CHANNEL *pCh)
{
   VMMC_ALM_Free_Ch_Structures (pCh);

   pCh->pALM = VMMC_OS_Malloc (sizeof(VMMC_ALMCH_t));
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Memory allocation failed */
      RETURN_STATUS (VMMC_statusNoMem);
   }
   memset(pCh->pALM, 0, sizeof(VMMC_ALMCH_t));

   return VMMC_statusOk;
}


/**
   Free data structure of the ALM module in the given channel.

   \param  pCh             Pointer to the VMMC channel structure.
*/
IFX_void_t VMMC_ALM_Free_Ch_Structures (VMMC_CHANNEL *pCh)
{
   if (pCh->pALM != IFX_NULL)
   {
      VMMC_OS_Free (pCh->pALM);
      pCh->pALM = IFX_NULL;
   }
}


/**
   Set an offset for the PCM timeslots used for the ALM emulation.

   Note that if the timeslots that result from the offset are occupied
   this function may fail to create the ALM connection again.

   \param  pCh          Pointer to the VMMC channel structure.
   \param  nOffset      Offset for the timeslots to be set.
*/
IFX_int32_t VMMC_ALM_ts_offset_set (VMMC_CHANNEL *pCh, IFX_uint16_t nOffset)
{
   IFX_int32_t ret = VMMC_statusOk;

   /* parameter check */
   VMMC_ASSERT(pCh != IFX_NULL);
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS (VMMC_statusInvalCh);
   }

   if (nOffset != pCh->pALM->ts_offset)
   {
      /* Free the timeslots currently in use for the SLIC connection. */
      ret = vmmc_alm_VoiceDisconnect(pCh);

      if (VMMC_SUCCESS(ret))
      {
         IFX_uint8_t ch = pCh->nChannel - 1;

         /* Remember the offset for use in the check above. */
         pCh->pALM->ts_offset = nOffset;

         /* Calculate the new timeslots with the given offset. */
         pCh->pALM->dsp_pcm_cfg.nTimeslotRX =
            pCh->pALM->sdd_pcm_cfg.nTimeslotTX = 4 * ch + nOffset;
         pCh->pALM->dsp_pcm_cfg.nTimeslotTX =
            pCh->pALM->sdd_pcm_cfg.nTimeslotRX = 4 * ch + nOffset;

         /* Re-establish the SLIC connection on the new timeslots. */
         ret = vmmc_alm_VoiceConnect(pCh, IFX_TRUE);
      }
   }

   RETURN_STATUS (ret);
}


/**
   Create the voice connection via PCM between VCODEC and VoIP FW.

   Configure the PCM of the VCODEC and TDM managed by the VoIP FW to create
   a voice connection between them. The configuration is stored in the private
   data of this channel.

   \param  pCh          Pointer to the VMMC channel structure.
   \param  bConfUpdate  Update of conference sampling rate
                        - IFX_FALSE: no matching of conference sampling rate
                        - IFX_TRUE: do matching of conference sampling rate

   \return
   - VMMC_statusOk         If successful
   - Error codes from PCM ch config or PCM ch enable.
*/
IFX_int32_t vmmc_alm_VoiceConnect(
                        VMMC_CHANNEL *pCh,
                        IFX_boolean_t bConfUpdate)
{
   VMMC_ALMCH_t *pALM = pCh->pALM;
   IFX_int32_t ret;

   /* The VMMC is the PCM master so configure this side first. */
   ret = VMMC_PCM_CH_Cfg (pCh, &pALM->dsp_pcm_cfg);
   if(VMMC_SUCCESS(ret))
      ret = VMMC_PCM_CH_Enable (
               pCh, IFX_ENABLE, &pALM->dsp_pcm_cfg, bConfUpdate);
   if(VMMC_SUCCESS(ret))
      pALM->dsp_pcm_enabled = IFX_TRUE;

   if(VMMC_SUCCESS(ret))
      ret =  SDD_TAPI_LL_PCM_ChCfg (pCh->pSDD, &pALM->sdd_pcm_cfg);
   if(VMMC_SUCCESS(ret))
      ret = SDD_TAPI_LL_PCM_ChEnable(
               pCh->pSDD, IFX_ENABLE, &pALM->sdd_pcm_cfg);
   if(VMMC_SUCCESS(ret))
      pALM->sdd_pcm_enabled = IFX_TRUE;

   RETURN_STATUS(ret);
}


/**
   Remove the voice connection via PCM between VCODEC and VoIP FW.

   \param  pCh             Pointer to the VMMC channel structure.

   \return
   - VMMC_statusOk         If successful
   - Error codes from PCM ch config or PCM ch enable.
*/
IFX_int32_t vmmc_alm_VoiceDisconnect(VMMC_CHANNEL *pCh)
{
   VMMC_ALMCH_t *pALM = pCh->pALM;
   IFX_int32_t ret = VMMC_statusOk;

   if (pALM->dsp_pcm_enabled != IFX_FALSE)
   {
      ret = VMMC_PCM_CH_Enable (
               pCh, IFX_DISABLE, &pALM->dsp_pcm_cfg, IFX_FALSE);
      pALM->dsp_pcm_enabled = IFX_FALSE;
   }

   if (pALM->sdd_pcm_enabled != IFX_FALSE)
   {
      ret = SDD_TAPI_LL_PCM_ChEnable(
               pCh->pSDD, IFX_DISABLE, &pALM->sdd_pcm_cfg);
      pALM->sdd_pcm_enabled = IFX_FALSE;
   }

   RETURN_STATUS(ret);
}


/**
   Set line type to FXS or FXO and the sampling mode operation.

   In case switching to FXS line type, the line mode used is Power Down High
   Impedance (PDH). For FXO the line type FXO is set.
   The sampling rate for FXS can be narrowband, wideband or automatic switching.
   For FXO the sampling rate is always narrowband.

   \param  pLLChannel      Pointer to the VMMC channel structure.
   \param  nType           New line type, can be anything out of
                           IFX_TAPI_LINE_TYPE_t.

   \return
   - VMMC_statusInvalCh    Channel number out of range
   - VMMC_statusParam      The parameters are wrong
   - VMMC_statusOk         If successful
*/
IFX_int32_t VMMC_TAPI_LL_ALM_Line_Type_Set(IFX_TAPI_LL_CH_t *pLLChannel,
                                           IFX_TAPI_LINE_TYPE_t nType)
{
   VMMC_CHANNEL *pCh  = (VMMC_CHANNEL *) pLLChannel;
   IFX_int32_t   ret = VMMC_statusOk;

   /* parameter check */
   VMMC_ASSERT(pCh != IFX_NULL);
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS (VMMC_statusInvalCh);
   }

   /* Check type change with SDD part. */
   ret = SDD_TAPI_LL_ALM_Line_Type_Set(pLLChannel, nType);
   if (!VMMC_SUCCESS (ret))
   {
      RETURN_STATUS (ret);
   }

   /* Set the module's ability to switch between narrowband and wideband
      operation mode. */
   switch (nType)
   {
      case IFX_TAPI_LINE_TYPE_FXO:
      case IFX_TAPI_LINE_TYPE_FXS_NB:
         VMMC_CON_ModuleSamplingModeSet (pCh, VMMCDSP_MT_ALM, VMMC_CON_SMPL_NB);
         break;
      case IFX_TAPI_LINE_TYPE_FXS_WB:
         VMMC_CON_ModuleSamplingModeSet (pCh, VMMCDSP_MT_ALM, VMMC_CON_SMPL_WB);
         break;
      case IFX_TAPI_LINE_TYPE_FXS_AUTO:
         VMMC_CON_ModuleSamplingModeSet (pCh, VMMCDSP_MT_ALM, VMMC_CON_SMPL_AUTO);
         break;
      default:
         /* errmsg: Invalid line type or sampling mode for analog line */
         ret = VMMC_statusInvalAlmTypeSmpl;
         break;
   }

   if (VMMC_SUCCESS (ret))
   {
      /* reevaluate sampling in the conference that this module belongs to */
      ret = VMMC_CON_MatchConfSmplRate(pCh, VMMCDSP_MT_ALM);
   }

   RETURN_STATUS (ret);
}


/**
   Sets the LEC configuration on the analog line.

   \param  pLLChannel   Handle to TAPI low level channel structure.
   \param  pLecConf     Handle to IFX_TAPI_LEC_CFG_t structure.

   \return
   - VMMC_statusInvalCh    Channel number out of range
   - VMMC_statusOk if successful
*/
IFX_int32_t VMMC_TAPI_LL_ALM_LEC_Set (IFX_TAPI_LL_CH_t *pLLChannel,
                                      TAPI_LEC_DATA_t *pLecConf)
{
   VMMC_CHANNEL *pCh  = (VMMC_CHANNEL *) pLLChannel;
   IFX_int32_t   ret;

   /* sanity check */
   VMMC_ASSERT(pCh != IFX_NULL);
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS (VMMC_statusInvalCh);
   }

   ret = VMMC_PCM_LEC_Set (pCh, VMMC_RES_MOD_ALM, pLecConf);

   RETURN_STATUS (ret);
}


/**
   Function that fills in the ALM module function pointers in the driver
   context structure which is passed to HL TAPI during registration.

   \param  pAlm         Pointer to the ALM part in the driver context struct.
*/
IFX_void_t VMMC_ALM_Func_Register (IFX_TAPI_DRV_CTX_ALM_t *pAlm)
{
   /*  SDD specific */
   SDD_ALM_Opmode_Func_Register (pAlm);

   /* Fill the function pointers with the services provided */
   pAlm->Lec_Cfg           = VMMC_TAPI_LL_ALM_LEC_Set;
   /* This overwrites the same function provided by the SDD. Therefore this
      line needs to be kept after SDD_ALM_Opmode_Func_Register(). */
   pAlm->Line_Type_Set     = VMMC_TAPI_LL_ALM_Line_Type_Set;
}

#endif /* VMMC_FEAT_SLIC */
