/******************************************************************************

                            Copyright (c) 2014-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

 ******************************************************************************/

/**
   \file drv_vmmc_loader_handler.c
   This file contains the implementation vmmc loader routines.
 */

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "drv_config.h"
#include "drv_vmmc_loader_version.h"
#include "drv_vmmc_loader_handler.h"
#include "drv_vmmc_loader_linux.h"
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/crc32.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <asm/ltq_vmb.h>
#include <asm/cacheflush.h>

/* ========================================================================== */
/*                            Macro definitions                               */
/* ========================================================================== */
/** Magic number at the end of firmware image part. */
#define FIRMWARE_MAGIC_VALUE 0xCC123456
/** Defence threads */
#define TC_NUM 1
/** Yield resources */
#define YR_NUM 1

/* ========================================================================== */
/*                              Local variables                               */
/* ========================================================================== */
/** Structure holds all settings of loader */
VMMC_LOADER_CONTEXT_t context;

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */
/**
   This function run voice firmware on CPU.

   \param fw_image pointer to VMMC_LOADER_DOWNLOAD_t structure

   \return int32_t VMMC_LOADER_SUCCESS or VMMC_LOADER_ERROR in case of error.
*/
int32_t drv_vmmc_loader_download_firmware(VMMC_LOADER_DOWNLOAD_t* fw_image)
{
   int32_t              ret = 0;
   int32_t              crc32 = 0;
   uint32_t             mem;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4,9,0))
   CPU_launch_t         cpu_launch;
   TC_launch_t          tc_launch[TC_NUM];
#else
   struct CPU_launch_t  cpu_launch;
   struct TC_launch_t   tc_launch[TC_NUM];
#endif /* LINUX_VERSION_CODE check */
   VMMC_LOADER_FOOTER_t fw_footer;

   memset(&fw_footer,  0, sizeof(fw_footer));
   memset(&cpu_launch, 0, sizeof(cpu_launch));
   memset(tc_launch, 0, sizeof(tc_launch));

   if(fw_image == NULL)
   {
      TRACE(TRACE_ERROR, "NULL firmware pointer");
      return VMMC_LOADER_ERROR;
   }

   if(fw_image->data == NULL)
   {
      TRACE(TRACE_ERROR, "NULL image pointer");
      return VMMC_LOADER_ERROR;
   }

   TRACE(TRACE_INFO, "FW pointer: 0x%p length: %d, struct: 0x%p",
          fw_image->data, fw_image->length, &fw_footer);

   /* copy FW footer from user space */
   if (copy_from_user(&fw_footer, fw_image->data +
                      fw_image->length - sizeof(fw_footer),
                      sizeof(fw_footer)) != 0)
   {
      TRACE(TRACE_ERROR, "Firmware copy footer error");
      return VMMC_LOADER_ERROR;
   }

   if(fw_footer.magic != FIRMWARE_MAGIC_VALUE)
   {
      TRACE(TRACE_ERROR, "FW magic value error: 0x%08X != 0x%08X",
             fw_footer.magic, FIRMWARE_MAGIC_VALUE);

      return VMMC_LOADER_ERROR;
   }

   /* Verify the checksum of the memory requirement value. */
   mem = fw_footer.fw_memory;
   if ((0xFF & ~((mem >> 16) + (mem >> 8) + mem)) != (mem >> 24))
   {
      TRACE(TRACE_ERROR, "FW size value error");
      return VMMC_LOADER_ERROR;
   }

   /* Check FW memory usage. In this case FW is started at least once and
      new firmware needs more memory than the current one. */
   if((context.fw_memory_block != NULL) &&
      ((fw_footer.fw_memory & 0x00FFFFFF) > context.fw_size))
   {
      /* Ensure that the firmware is stopped before freeing the memory. */
      if(context.state != NOT_STARTED)
      {
         drv_vmmc_loader_stop_firmware();
      }

      free_pages_exact(context.fw_memory_block, context.fw_memory_block_size);
      context.fw_memory_block = NULL;
   }

   /* Determine the size of the FW from the FW-footer.
      The size is rounded up to the next full 128 kB. */
   context.fw_size = fw_footer.fw_memory & 0x00FFFFFF;
   /* Round up to the next full 128 kB. */
   context.fw_size = ((context.fw_size - 1) & ~(0x20000 - 1)) + 0x20000;

   /* Allocate memory for the VMMC-FW in RAM. The memory needs to be aligned
      to a boundary of 128 kB and physically contiguous.
      To ensure the alignment a memory block which is up to 128 kB larger is
      reserved and the FW is located to the first 128 kB aligned address
      within this memory.
      To make sure that the code written into the memory is seen by the
      core the caches are invalidated. */
   if (context.fw_memory_block == NULL)
   {
      /* Add (128 kB - 1 byte) for the memory block allocation. */
      context.fw_memory_block_size = context.fw_size + (128*1024 - 1);

      context.fw_memory_block =
         (uint32_t*)alloc_pages_exact(context.fw_memory_block_size, GFP_DMA);
      if (context.fw_memory_block == NULL)
      {
         TRACE(TRACE_ERROR, "Memory allocation error");
         return VMMC_LOADER_ERROR;
      }

      dma_cache_inv(
         (uint32_t)context.fw_memory_block,
         context.fw_memory_block_size);

      flush_icache_range(
         (uint32_t)context.fw_memory_block,
         (uint32_t)context.fw_memory_block + context.fw_memory_block_size);

      /* Find the first 128 kB aligned address within the allocated memory
         block and use it as the firmware start address. */
      context.fw_adress = context.fw_memory_block;
      if (((uint32_t)context.fw_adress & (0x20000 - 1)) > 0)
      {
         context.fw_adress =
            (uint32_t*)(
               ((uint32_t)context.fw_adress & ~(0x20000 - 1)) + 0x20000);
      }

      TRACE(TRACE_DEBUG, "VMMC-FW start address: 0x%p", context.fw_adress);
   }

   if (copy_from_user(context.fw_adress, fw_image->data,
                      fw_image->length - sizeof(fw_footer)) != 0)
   {
      TRACE(TRACE_ERROR, "Firmware data copy error");
      return VMMC_LOADER_ERROR;
   }

   crc32 = crc32_be(0xffffffff, (uint8_t*)context.fw_adress,
                    sizeof(uint32_t)*2);

   crc32 = crc32_be(crc32, ((uint8_t *)context.fw_adress) +
                    fw_footer.st_addr_crc,
                    fw_footer.en_addr_crc -
                    fw_footer.st_addr_crc);

   if (((~crc32)^0xffffffff) != fw_footer.crc32)
   {
      TRACE(TRACE_ERROR, "FW checksum error: 0x%08X != 0x%08X",
             fw_footer.crc32, crc32);

      return VMMC_LOADER_ERROR;
   }

   /* The IBL runs in legacy address mode. It requires an uncached unmapped
      address (MIPS KSEG1). So the address needs to be translated into this
      range.
      This works only if in the legacy as well as the EVA address map
      the RAM is located at the same addresses while alloc_pages_exact() for
      the DMA zone will give us addresses below 1GB that can be used in both
      legacy as well as EVA address mode. */
   cpu_launch.start_addr = CKSEG1ADDR(context.fw_adress);

   TRACE(TRACE_DEBUG,
         "FW info - ver: %u.%u.%u.%u.%u, "
         "mem: 0x%08X, len: %u, entry: 0x%p (KSEG1: 0x%X), "
         "crc32: 0x%08X, st_addr_crc: 0x%08X, en_addr_crc: 0x%08X, "
         "enc: 0x%08X, magic: 0x%08X",
         (fw_footer.fw_version >> 24) & 0xff,
         (fw_footer.fw_version >> 16) & 0xff,
         (fw_footer.fw_version >> 12) & 0x0f,
         (fw_footer.fw_version >>  8) & 0x0f,
         (fw_footer.fw_version)       & 0xff,
         context.fw_memory_block, context.fw_memory_block_size,
         context.fw_adress, cpu_launch.start_addr,
         fw_footer.crc32, fw_footer.st_addr_crc, fw_footer.en_addr_crc,
         fw_footer.encryption, fw_footer.magic);

   context.cpun = vmb_cpu_alloc(MAX_CPU, "VOICEFW");
   if (context.cpun < 0)
   {
      TRACE(TRACE_ERROR, "All the CPUs are busy");
      return VMMC_LOADER_ERROR;
   }

   TRACE(TRACE_DEBUG, "Allocated CPU: %d", context.cpun);

   context.tc_num = vmb_tc_alloc(context.cpun);
   if (context.tc_num < 0)
   {
      TRACE(TRACE_ERROR, "TC alloc failed");
      return VMMC_LOADER_ERROR;
   }

   TRACE(TRACE_DEBUG, "Allocated TC: %d", context.tc_num);

   tc_launch[0].tc_num = context.tc_num;

   ret = vmb_cpu_start(context.cpun, cpu_launch, tc_launch, TC_NUM, YR_NUM);
   if (ret != VMB_SUCCESS)
   {
      TRACE(TRACE_ERROR, "Voice FW not loaded on CPU: %d ret: %d",
             context.cpun, ret);

      return VMMC_LOADER_ERROR;
   }
   TRACE(TRACE_DEBUG, "Voice FW loaded on CPU: %d", context.cpun);

   context.state = STARTED;

   return VMMC_LOADER_SUCCESS;
}

/**
   This function stops voice firmware.
*/
void drv_vmmc_loader_stop_firmware(void)
{
   int8_t ret = 0;
   if(context.state == STARTED)
   {
      ret = vmb_cpu_stop(context.cpun);

      if(ret != VMB_SUCCESS)
      {
         TRACE(TRACE_ERROR, "FW stopped with error: %d", ret);
      }
      else
      {
         TRACE(TRACE_DEBUG, "FW stopped");
         context.state = NOT_STARTED;
      }
   }
   else
   {
      TRACE(TRACE_DEBUG, "FW not started");
   }
}

#ifdef WITH_VMB_LOADER
int32_t drv_vmmc_loader_init()
{
   return drv_vmmc_loader_init_module();
}

void drv_vmmc_loader_cleanup()
{
   return drv_vmmc_loader_cleanup_module();
}
#endif /* WITH_VMB_LOADER */

