/******************************************************************************

                            Copyright (c) 2014-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

 ******************************************************************************/

/**
   \file drv_vmmc_loader_linux.c
   This file contains the implementation of linux specific parts of vmmc loader.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "drv_config.h"
#include "drv_vmmc_loader_linux.h"
#include "drv_vmmc_loader_handler.h"
#include "drv_vmmc_loader_version.h"
#include <linux/platform_device.h>
#include <linux/cdev.h>
#include <linux/module.h>
#include <linux/uaccess.h>
#ifdef WITH_PROC_FS
   #include <generated/utsrelease.h>
   #include <linux/proc_fs.h>
   #include <linux/seq_file.h>
#endif /* WITH_PROC_FS */

/* ========================================================================== */
/*                            Macro definitions                               */
/* ========================================================================== */
#define DRV_VMMC_LOADER_DEV_NAME       "vmmc_loader"

/* Do not printout __DATE__ and __TIME__ with GCC version >= 4.9. */
#if !defined(__GNUC__) || \
    (__GNUC__ < 4) || \
    (__GNUC__ == 4 && __GNUC_MINOR__ < 9)
   #define LOADER_FEAT_USE_TIME_DATE
#endif /* GCC version < 4.9 */

/* ========================================================================== */
/*                            Types definitions                               */
/* ========================================================================== */
extern VMMC_LOADER_CONTEXT_t context;

#ifdef WITH_PROC_FS
struct proc_entry
{
   const char* name;
   void* read_function;
   void* write_function;
   struct file_operations ops;
};

typedef void (*drv_vmmc_loader_proc_dump) (struct seq_file *s);
#endif /* WITH_PROC_FS */

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */
#ifdef WITH_PROC_FS
static int32_t drv_vmmc_loader_proc_create(struct proc_dir_entry* proc_dir);
static int32_t drv_vmmc_loader_proc_entry_create(
   struct proc_dir_entry *parent_node, struct proc_entry *proc_entry);
static int32_t drv_vmmc_loader_proc_version_get(struct seq_file *s);
static int32_t drv_vmmc_loader_proc_status_get(struct seq_file *s);
static int32_t drv_vmmc_loader_proc_debug_get(struct seq_file *s);
static int32_t drv_vmmc_loader_proc_debug_set(struct file *file,
                           const char *buffer, unsigned long count, void *data);
static int32_t drv_vmmc_loader_proc_show(struct seq_file *s, void *p);
static int32_t drv_vmmc_loader_proc_open(struct inode* inode,
                                             struct file* file);
static int32_t drv_vmmc_loader_proc_remove(
   struct proc_dir_entry* proc_directory);
/* Configure file should be available only in stand-alone version, when loader
   is compiled with VMMC drv /proc/driver/vmmc/configure should be used. */
#ifndef WITH_VMB_LOADER
static int32_t drv_vmmc_loader_proc_configure_get(struct seq_file *s);
#endif /* WITH_VMB_LOADER */
#endif /* WITH_PROC_FS */

#ifdef WITH_OS_REGISTER
int32_t drv_vmmc_loader_open (struct inode *inode, struct file *file_p);
int32_t drv_vmmc_loader_close (struct inode *inode, struct file *file_p);
long drv_vmmc_loader_ioctl(struct file* file, uint32_t cmd,
                           unsigned long arg);
#endif /* WITH_OS_REGISTER */

/* ========================================================================== */
/*                             Local variables                                */
/* ========================================================================== */
#ifdef WITH_PROC_FS
static struct proc_entry proc_entries[] =
{
   {"version",  drv_vmmc_loader_proc_version_get},
   {"status",   drv_vmmc_loader_proc_status_get},
#ifndef WITH_VMB_LOADER
   {"configure",drv_vmmc_loader_proc_configure_get},
#endif /* WITH_VMB_LOADER */
   {"debug",    drv_vmmc_loader_proc_debug_get, drv_vmmc_loader_proc_debug_set},
};
#endif /* WITH_PROC_FS */

#ifdef WITH_OS_REGISTER
static struct file_operations drv_vmmc_loader_fops =
{
   owner          :  THIS_MODULE,
   poll           :  0,
   unlocked_ioctl :  drv_vmmc_loader_ioctl,
   open           :  drv_vmmc_loader_open,
   release        :  drv_vmmc_loader_close
};
#endif /* WITH_OS_REGISTER */

#if (defined(MODULE) && !defined(WITH_VMB_LOADER))
MODULE_AUTHOR ("Lantiq Deutschland GmbH");
MODULE_DESCRIPTION ("VMMC loader driver for XWAY(TM) GRX500 family");
MODULE_SUPPORTED_DEVICE ("XWAY(TM) GRX500 family");
MODULE_LICENSE ("Dual BSD/GPL");
#endif /* defined(MODULE) && !defined(VMMC_WITH_MPS) */

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */
void TRACE(uint8_t level, const char* message, ...)
{
   va_list args;
   char buffer[256];

   if(level < context.trace_level)
   {
      return;
   }

   va_start(args, message);
   vsnprintf(buffer, 256, message, args);
   va_end(args);

   if(level == TRACE_ERROR)
   {
      printk(KERN_ERR VMMC_LOADER_PREFIX "%s!\n", buffer);
   }

   if(level == TRACE_INFO)
   {
      printk(KERN_INFO VMMC_LOADER_PREFIX "%s\n", buffer);
   }

   if(level == TRACE_DEBUG)
   {
      printk(KERN_DEBUG VMMC_LOADER_PREFIX "%s\n", buffer);
   }
}

#ifdef WITH_PROC_FS
/**
   This function creates the output for the vmmc loader version proc file.

   \param proc_dir pointer to proc_dir_entry struct

   \return int32_t VMMC_LOADER_SUCCESS or VMMC_LOADER_ERROR in case of error
*/
static int32_t drv_vmmc_loader_proc_create(
   struct proc_dir_entry* proc_dir)
{
   uint32_t i;

   TRACE(TRACE_DEBUG, "Create procfs entries");

   proc_dir = proc_mkdir("driver/" DRV_VMMC_LOADER_DEV_NAME, NULL);

   if(proc_dir != NULL)
   {
      for(i = 0; i < sizeof(proc_entries)/sizeof(proc_entries[0]); i++)
      {
         if(drv_vmmc_loader_proc_entry_create(
            proc_dir, &proc_entries[i]) != VMMC_LOADER_SUCCESS)
         {
            TRACE(TRACE_ERROR, "Create procfs entry error: %d", i);
         }
      }
   }
   else
   {
      TRACE(TRACE_ERROR, "Cannot create /proc/" DRV_VMMC_LOADER_DEV_NAME);
      return VMMC_LOADER_ERROR;
   }
   return VMMC_LOADER_SUCCESS;
}

/**
   This function procfs files in kernel.

   \param parent_node pointer to proc_dir_entry struct
   \param proc_entry pointer to proc_entry struct

   \return int32_t VMMC_LOADER_SUCCESS or VMMC_LOADER_ERROR in case of error
*/
static int32_t drv_vmmc_loader_proc_entry_create(
   struct proc_dir_entry *parent_node, struct proc_entry *proc_entry)
{
   memset(&proc_entry->ops, 0, sizeof(struct file_operations));
   proc_entry->ops.owner   = THIS_MODULE;
   proc_entry->ops.open    = drv_vmmc_loader_proc_open;
   proc_entry->ops.read    = seq_read;
   proc_entry->ops.write   = proc_entry->write_function;
   proc_entry->ops.llseek  = seq_lseek;
   proc_entry->ops.release = single_release;

   if(proc_create_data(proc_entry->name, 0, parent_node, &proc_entry->ops,
                     proc_entry->read_function))
   {
      return VMMC_LOADER_SUCCESS;
   }
   else
   {
      return VMMC_LOADER_ERROR;
   }
}

/**
   This function creates the output for the vmmc loader version proc file.

   \param s pointer to seq_file struct

   \return int32_t 0 in all cases
*/
static int32_t drv_vmmc_loader_proc_version_get(struct seq_file *s)
{
   seq_printf(s, VMMC_LOADER_INFO ", version %d.%d.%d.%d\n"
      VMMC_LOADER_COPYRIGHT "\n",
      MAJORSTEP, MINORSTEP, VERSIONSTEP, VERS_TYPE);

   #ifdef LOADER_FEAT_USE_TIME_DATE
      seq_printf(s, "Compiled on %s, %s for Linux kernel %s\n",
         __DATE__, __TIME__, UTS_RELEASE);
   #else
      seq_printf(s, "Compiled for Linux kernel %s\n",
         UTS_RELEASE);
   #endif

   return 0;
}

/**
   This function creates the output for the vmmc loader status proc file.

   \param s pointer to seq_file struct

   \return int32_t 0 in all cases
*/
static int32_t drv_vmmc_loader_proc_status_get(struct seq_file *s)
{
   if(context.state == STARTED)
   {
      seq_printf (s, "Firmware: started on: ");
      seq_printf (s, "CPU: %d, TC: %d\n", context.cpun, context.tc_num);

      seq_printf (s, "FW start adress: 0x%08X\n",
                  (uint32_t)context.fw_adress);
      seq_printf (s, "FW size (bytes): 0x%08X (%d)\n",
                  context.fw_size,
                  context.fw_size);
      seq_printf (s, "FW end adress  : 0x%08X\n",
                  (uint32_t)context.fw_adress + context.fw_size - 1);

      seq_printf (s, "FW memory block start adress: 0x%08X\n",
                  (uint32_t)context.fw_memory_block);
      seq_printf (s, "FW memory block size (bytes): 0x%08X (%d)\n",
                  context.fw_memory_block_size,
                  context.fw_memory_block_size);
      seq_printf (s, "FW memory block end adress  : 0x%08X\n",
                  (uint32_t)context.fw_memory_block +
                     context.fw_memory_block_size - 1);
   }
   else
   {
      seq_printf (s, "Firmware: not started\n");
   }

   return 0;
}

/**
   This function creates the output for the vmmc loader configure proc file.

   \param s pointer to seq_file struct

   \return int32_t 0 in all cases
*/
#ifndef WITH_VMB_LOADER
static int32_t drv_vmmc_loader_proc_configure_get(struct seq_file *s)
{
   seq_printf(s, "configure:" LOADER_CONFIGURE_STR "\n");

   return 0;
}
#endif /* WITH_VMB_LOADER */

/**
   This function creates the output for the vmmc loader debug proc file.

   \param file pointer to seq_file struct
   \param buffer pointer to data buffer
   \param count number of data in buffer
   \param data not used

   \return int32_t -EFAULT or number of readed bytes
*/
static int32_t drv_vmmc_loader_proc_debug_set(struct file *file,
                            const char *buffer, unsigned long count, void *data)
{
   char input_line[1];

   if (count < 1)
   {
      TRACE(TRACE_ERROR, "No input");
      return -EFAULT;
   }

   if (copy_from_user(input_line, buffer, 1) != 0)
   {
      TRACE(TRACE_ERROR, "Copy from user error");
      return -EFAULT;
   }

   /** Look only at the first character. */
   if (input_line[0] > '0' && input_line[0] < '4')
   {
      context.trace_level = input_line[0] - '0';
      TRACE(TRACE_INFO, "Trace level changed to: %d", context.trace_level);
   }
   else
   {
      TRACE(TRACE_ERROR, "Level out of range (1 -- 3)");
      return -EFAULT;
   }

   return count;
}

/**
   This function creates the output for the vmmc loader trace level proc file.

   \param s pointer to seq_file struct

   \return int32_t 0 in all cases
*/
static int32_t drv_vmmc_loader_proc_debug_get(struct seq_file *s)
{
   seq_printf(s, "levels: 1 - all logs, 2 - infos and errors, 3 - off\n");
   seq_printf(s, "current debug level: %d\n", context.trace_level);

   return 0;
}

/**
   This function creates the output for the vmmc loader trace level proc file.

   \param s pointer to seq_file struct
   \param p not used

   \return int32_t 0 in all cases
*/
static int32_t drv_vmmc_loader_proc_show(struct seq_file *s, void *p)
{
   drv_vmmc_loader_proc_dump dump = s->private;

   if (dump != NULL)
      dump(s);

   return 0;
}

/**
   This function is called on procfs open.

   \param inode pointer to inode struct
   \param file pointer to file struct

   \return int32_t Returns value from single_open function.
*/
static int32_t drv_vmmc_loader_proc_open(struct inode* inode,
                                             struct file* file)
{
   return single_open(file, drv_vmmc_loader_proc_show, PDE_DATA(inode));
}

/**
   This function remove procfs entries.

   \param proc_dir pointer to proc_dir_entry struct

   \return int32_t 0 in all cases
*/
static int32_t drv_vmmc_loader_proc_remove(
   struct proc_dir_entry* proc_dir)
{
   remove_proc_entry("driver/"DRV_VMMC_LOADER_DEV_NAME"/version", proc_dir);
   remove_proc_entry("driver/"DRV_VMMC_LOADER_DEV_NAME"/status", proc_dir);
#ifndef WITH_VMB_LOADER
   remove_proc_entry("driver/"DRV_VMMC_LOADER_DEV_NAME"/configure", proc_dir);
#endif /* WITH_VMB_LOADER */
   remove_proc_entry("driver/"DRV_VMMC_LOADER_DEV_NAME"/debug", proc_dir);
   remove_proc_entry("driver/"DRV_VMMC_LOADER_DEV_NAME, proc_dir);
   return 0;
}

#endif /* WITH_PROC_FS */

#ifdef WITH_OS_REGISTER
/**
   This function is called on vmmc loader open deivce.

   \param inode pointer to inode struct
   \param file pointer to file struct

   \return int32_t 0 in all cases
*/
int32_t drv_vmmc_loader_open (struct inode* inode, struct file* file)
{
   TRACE(TRACE_DEBUG, "File opened");
   return 0;
}

/**
   This function is called on vmmc loader ioctl deivce call.

   \param file pointer to file struct
   \param cmd IOCTL value
   \param arg IOCTL arguments

   \return long Returns value from drv_vmmc_loader_download_firmware function.
*/
long drv_vmmc_loader_ioctl(struct file* file, uint32_t cmd, unsigned long arg)
{
   int32_t retvalue = -EINVAL;
   switch (cmd)
   {
      case FIO_LOADER_DOWNLOAD:
      {
         /* Download firmware file */
         VMMC_LOADER_DOWNLOAD_t fw_image;
#ifndef WITH_VMB_LOADER
         if (copy_from_user(&fw_image, (uint32_t*) arg, sizeof (fw_image)) != 0)
         {
            TRACE(TRACE_ERROR, "Copy_from_user error");
         }
#else
         fw_image.data = ((VMMC_LOADER_DOWNLOAD_t*)arg)->data;
         fw_image.length = ((VMMC_LOADER_DOWNLOAD_t*)arg)->length;
#endif /* WITH_VMB_LOADER */
         retvalue = drv_vmmc_loader_download_firmware (&fw_image);
         break;
      }
      default:
      {
         TRACE(TRACE_ERROR, "Unknow IOCTL %d (0x%08X)", cmd, cmd);
      }
   }

   return retvalue;
}

/**
   This function is called on vmmc loader close deivce.

   \param inode pointer to inode struct
   \param file_p pointer to file struct

   \return int32_t 0 in all cases
*/
int32_t drv_vmmc_loader_close (struct inode* inode, struct file* file_p)
{
   TRACE(TRACE_DEBUG, "File closed");
   return 0;
}

/**
   This function registers char device in kernel.

   \param pContext pointer to VMMC_LOADER_CONTEXT_t structure

   \return  -ENOMEM or -EPERM in case of error
   \return  0 on success
*/
int32_t drv_vmmc_loader_os_register (VMMC_LOADER_CONTEXT_t* pContext)
{
   int32_t ret = -ENOMEM;
   dev_t       dev;
   struct cdev* p_cdev = NULL;

   /* dynamic major */
   ret = alloc_chrdev_region(&dev, 1, 1, DRV_VMMC_LOADER_DEV_NAME);
   pContext->major_number = MAJOR(dev);

   if (ret < 0)
   {
      TRACE(TRACE_ERROR, "Allocation device error: %d", ret);
      return ret;
   }

   TRACE(TRACE_DEBUG, "Registered with major: %d", pContext->major_number);

   p_cdev = cdev_alloc();

   if (NULL == p_cdev)
      return -ENOMEM;

   cdev_init(p_cdev, &drv_vmmc_loader_fops);
   p_cdev->owner = THIS_MODULE;

   ret = cdev_add(p_cdev, MKDEV(pContext->major_number, 1), 1);

   if (ret != 0)
   {
      cdev_del (p_cdev);
      return -EPERM;
   }

   return ret;
}

/**
   This function unregisters char device from kernel.

   \param pContext pointer to VMMC_LOADER_CONTEXT_t structure
*/
void drv_vmmc_loader_os_unregister (VMMC_LOADER_CONTEXT_t *pContext)
{
   unregister_chrdev_region(MKDEV(pContext->major_number, 1), 1);
}
#endif /* WITH_OS_REGISTER */

/**
   This function initializes the module.

   \return  VMMC_LOADER_SUCCESS module initialized
   \return  ENOMEM   No memory left for structures
*/
#ifdef WITH_VMB_LOADER
int32_t drv_vmmc_loader_init_module (void)
#else
int __init drv_vmmc_loader_init_module (void)
#endif
{
   memset (&context, 0, sizeof(VMMC_LOADER_CONTEXT_t));

   context.trace_level = TRACE_ERROR;

   TRACE(TRACE_INFO, "version %d.%d.%d.%d, " VMMC_LOADER_COPYRIGHT,
          MAJORSTEP, MINORSTEP, VERSIONSTEP, VERS_TYPE);

#ifdef WITH_PROC_FS
   /* Create proc filesystem entries for driver */
   if(drv_vmmc_loader_proc_create(context.proc_dir) !=
      VMMC_LOADER_SUCCESS)
   {
      return -ENOMEM;
   }
#endif /* WITH_PROC_FS */

   /* Non stand-alone version - registering node is unnecessary */
#ifdef WITH_OS_REGISTER
   if(drv_vmmc_loader_os_register(&context) != VMMC_LOADER_SUCCESS)
   {
      return -ENOMEM;
   }
#endif /* WITH_VMB_LOADER */

   context.state = NOT_STARTED;
   context.fw_adress = NULL;

   return VMMC_LOADER_SUCCESS;
}

/**
   This function cleans up the module.
*/
#ifdef WITH_VMB_LOADER
void drv_vmmc_loader_cleanup_module (void)
#else
void __exit drv_vmmc_loader_cleanup_module (void)
#endif
{
   drv_vmmc_loader_stop_firmware();
   if(context.fw_memory_block != NULL)
   {
      free_pages_exact(context.fw_memory_block, context.fw_memory_block_size);
      context.fw_memory_block = NULL;
   }
#ifdef WITH_OS_REGISTER
   drv_vmmc_loader_os_unregister(&context);
#endif /* WITH_OS_REGISTER */
#ifdef WITH_PROC_FS
   drv_vmmc_loader_proc_remove(context.proc_dir);
#endif /* WITH_PROC_FS */

   TRACE(TRACE_DEBUG, "Module exit");
}

#ifndef WITH_VMB_LOADER
module_init (drv_vmmc_loader_init_module);
module_exit (drv_vmmc_loader_cleanup_module);
#endif /* !WITH_VMB_LOADER */

