/*******************************************************************************

                            Copyright (c) 2013-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_api.c
   This file implements the driver internal api.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "ifx_types.h"
#include "ifxos_interrupt.h"
#include "ifxos_time.h"

#include "drv_sdd_osmap.h"
#include "drv_sdd_global.h"
#include "drv_sdd_fw_cmd.h"

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */
struct msgHead_t {
#if IFXOS_BYTE_ORDER == IFXOS_BIG_ENDIAN
   CMD_HEAD_BE;
#else
   CMD_HEAD_LE;
#endif
};

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */

/**
   Internal: write command to the mailbox

   \param  pDev         Pointer to the device structure.
   \param  pCmd         Pointer to the buffer with the command. It contains the
                        command header and the data to be written.
   \param  nCount       Number of data BYTES to write, without command header.
                        Must not be larger than the maximum mailbox size.

   \return
   - SDD_statusOk       On success.
   - SDD_statusCmdMbWrErr On error.
                        Error can occur if command-inbox space is not
                        sufficient or if low level function/macros fail.

   \remarks
   Protection against tasks and interrupts needs to be done outside.
*/
static IFX_int32_t sdd_cmdWrite(SDD_DEVICE_t *pDev,
                                IFX_uint32_t *pCmd,
                                IFX_uint16_t nCount)
{
   IFX_int32_t err = SDD_statusOk;
   SDD_FIRMWARE_IF_t *p = pDev->pMbxIF;

   /* The command header must not be all zero. */
   SDD_ASSERT ((*pCmd & 0xffff0000) != 0);
   /* The command must be aligned to an even 32 bit address */
   SDD_ASSERT (((IFX_uint32_t)pCmd & 0x3) == 0);

   /* increase count by the length of the command header */
   nCount += CMD_HDR_CNT;
   /* force the number of bytes to be a multiple of 4 - this is required
      for example in case of COP CRAM download */
   if (nCount % 4 != 0)
   {
      nCount += 4 - (nCount%4);
   }

   /* write to the mailbox */
   err = p->cmd_write((IFX_uint8_t *)pCmd, nCount);

   /* error from mailbox driver */
   if (err != SDD_statusOk)
   {
      //SET_DEV_ERROR (VMMC_ERR_MBXWRITE);
      //RETURN_INFO_DEVSTATUS (VMMC_statusCmdWr,pCmd, nCount);
   }

   return err;
}

/**
   Write SDD Command

   The data to be written as well as the length of the data is taken from the
   command that is passed as a pointer to this function.

   \param  pDev         Pointer to the device structure.
   \param  pCmd         Pointer to the buffer with the command. It contains
                        the 2 header command words needed and the data to be
                        written.

   \return
   - SDD_statusOk       On success.
   - SDD_statusCmdMbWrErr On error.
                        Error can occur if command-inbox space is not
                        sufficient or if low level function/macros fail.
*/
IFX_int32_t SDD_CmdWrite(SDD_DEVICE_t *pDev,
                         IFX_uint32_t *pCmd)
{
   struct msgHead_t *pMsgHead = (struct msgHead_t *) pCmd;
   IFX_int32_t ret;

   /* This wrapper just makes sure that the RW bit is set to "write". */

   /* Set fields in command header to defined values. */
   pMsgHead->RW = CMDWRITE;
   pMsgHead->SC = 0;
   pMsgHead->BC = 0;
   pMsgHead->Res00 = 0;

   IFXOS_MutexGet (&pDev->mtxMbxAcc);
   ret = sdd_cmdWrite(pDev, pCmd, pMsgHead->LENGTH);
   IFXOS_MutexRelease (&pDev->mtxMbxAcc);

   return ret;
}

/**
   SDD Read Command (standard)

   \param  pDev         Pointer to the device structure.
   \param  pCmd         Pointer to the buffer with the command. It contains the
                        command header and the data to be written.
   \param  pData        Pointer to a buffer where to store the read command.

   \return
   - SDD_statusOk
   - SDD_statusCmdIbNoSpace
   - SDD_statusCmdMbWrErr
   - SDD_statusCmdObTimeout
   - SDD_statusCmdObRdErr     -  Reading from the event mailbox failed.
   - SDD_statusCmdObDataOvld
*/
IFX_int32_t SDD_CmdRead (SDD_DEVICE_t *pDev,
                         IFX_uint32_t *pCmd,
                         IFX_uint32_t *pData)
{
   IFX_int32_t err;
   struct msgHead_t *pMsgHead = (struct msgHead_t *) pCmd;
   IFX_uint16_t nRdCount;
   SDD_FIRMWARE_IF_t *p = pDev->pMbxIF;

   IFXOS_MutexGet (&pDev->mtxMbxAcc);

   /* Set fields in command header to defined values. */
   pMsgHead->RW = CMDREAD;
   pMsgHead->BC = 0;
   pMsgHead->Res00 = 0;

   /* This is the number of bytes we are going to read. */
   nRdCount = pMsgHead->LENGTH + CMD_HDR_CNT;

   err = p->cmd_read((IFX_uint8_t *)pData, nRdCount);

#if 0

   pDev->bCmdOutBoxData = IFX_FALSE;

   /* Write read command : length 0 because only command header is written. */
   err = sdd_cmdWrite(pDev, pCmd, 0);

   if (err == IFX_SUCCESS)
   {
      /* Read command was written successfully. Now wait for data and read it.
         Even when waiting for the data times-out we try to read the mailbox
         to find if there is something in the mailbox and just the wakeup did
         not work. If there was something in the mailbox this would indicate
         a problem in the OS with blocking or waking up. If there is nothing
         in the mailbox this would indicate that the FW did not answer.
         The following cases are distinguished:
            Wait    Read       Comment
            Ok      Ok         Read successful
            Ok      No data    Error: woke-up but no data?
            Timeout Ok         OS problem with blocking or waking up
            Timeout No data    Timeout - FW did not respond
            -       Error      Read error
      */

      /* Wait for data in the mailbox. */
      ret = SDD_WaitForCmdMbxData(pDev, nRdCount);

      /* Check for CmdReadError.
         The flag is set in IRQ context if CERR irq occurs. */
      if (ret != IFX_SUCCESS)
      {
         printk("Wait error: %d!\n", p->last_err());
         err = IFX_ERROR;
      }
      else
      {
         /* Read data from the mailbox */
         bytes = p->read(pDev->cmdMbx, (IFX_uint8_t *)pData, nRdCount);
         if (bytes > 0)
         {
            /* Read successful */
            err = IFX_SUCCESS;
         }
         else if (bytes == IFX_ERROR && (p->last_err() == -SDD_FW_ERR_ENODATA))
         {
            err = -ENODATA;
         }

         if (err == IFX_SUCCESS)
         {

            if (ret == IFX_ERROR && (p->last_err() == -SDD_FW_ERR_WAIT_TIMEOUT))
            {
               /* OS problem with blocking or waking up. This is not an
                  immediate error but needs investigation. */
               TRACE (SDD, DBG_LEVEL_HIGH, ("Read Cmd MBX overslept.\n"));
            }
         }
         else
         if ((err == -ENODATA) && (ret == IFX_ERROR))
         {
            /* Timeout - FW did not respond */
            TRACE (SDD, DBG_LEVEL_HIGH, ("Read Cmd MBX timeout.\n"));
            //VMMC_DevErrorEvent (pDev, VMMC_statusCmdRdTimeout,
              //                  __LINE__, __FILE__, pCmd, nRdCount);
            /* errmsg: Timeout while waiting for read data. */
            err = SDD_statusCmdRdTimeout;
         }
         else
         if ((err == -ENODATA) && (ret == IFX_SUCCESS))
         {
            /* Error: woke-up but no data? */
            TRACE (SDD, DBG_LEVEL_HIGH,
                   ("Read Cmd MBX spurious wakup - no data.\n"));
            err = IFX_ERROR;
         }
         else
         {
            /* Read error */
            TRACE (SDD, DBG_LEVEL_HIGH, ("Read Cmd MBX error.\n"));
         }
      }
   }
#endif

   IFXOS_MutexRelease (&pDev->mtxMbxAcc);

   return err;
}
