/******************************************************************************

                            Copyright (c) 2013-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

/**
   \file drv_sdd_bbd.c
   This file contains the implementation of the BBD download.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "ifx_types.h"
#include "lib_bbd.h"

#include "drv_sdd_global.h"
#include "drv_sdd_api.h"
#include "drv_sdd_osmap.h"
#include "drv_sdd_global.h"
#include "drv_sdd_fw_cmd.h"
#include "drv_sdd_init.h"
#include "drv_sdd_alm_priv.h"
#include "drv_sdd_calib.h"

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */
/* BBD archive header struct values */
#define BBD_TYPE_GLOBAL_V1    0x05000001  /* type 0x0500, version 0x0001 */
#define BBD_GLOBAL_MAGIC      0x42424435  /* 'BBD5' character sequence */
#define BBD_TOC_SLIC100_IBB   0x00000001
#define BBD_TOC_SLIC100_CIBB  0x00000002
#define BBD_TOC_SLIC100_FXO   0x00008000
#define BBD_TOC_SLIC200_IBB   0x00010000
#define BBD_TOC_SLIC200_CIBB  0x00020000
#define BBD_TOC_SLIC200_IFB   0x00400000
#define BBD_TOC_SLIC200_CIFB  0x00800000

#ifdef S220
   #define BBD_SDD_MAGIC                     0x53323230 /* S220 */
#else /* S220 */
   #define BBD_SDD_MAGIC                     0x47525835 /* GRX5 */
#endif /* S220 */

/** SDD BBD blocks tags */
/* 0x1XXX tags : SDD DC/AC Coefficients */
#define BBD_SDD_CRAM_BLOCK                   0x1001
#define BBD_SDD_RING_CFG_BLOCK               0x1003
#define BBD_SDD_DC_BASIC_CFG_BLOCK           0x1005
#define BBD_SDD_MWL_BLOCK                    0x1008
#define BBD_SDD_DCDC_BLOCK                   0x1009
#define BBD_SDD_FXO_CRAM_BLOCK               0x1010
#define BBD_SDD_CRAM2_BLOCK                  0x1020
#define BBD_SDD_WL_CMD_WRITE                 0x1200

/* maximum expected occurance of a single block type per BBD file,
   can be set also at compile time  */
#ifndef BBD_SDD_BLOCK_MAXDWNLD
#define BBD_SDD_BLOCK_MAXDWNLD               10
#endif /* BBD_SDD_BLOCK_MAXDWNLD */

/* ========================================================================== */
/*                             Type definitions                               */
/* ========================================================================== */

/* Header of BBD archive containing both SLIC100 and SLIC200 BBD files for
   multiple DC/DC converter types. (version 1, big endian) */
struct bbd_archive_header
{
   /* Type and Version identifier */
   IFX_uint32_t   nType;
   /* Length of the payload following this header */
   IFX_uint32_t   nLength;
   /* MAGIC value for endianess checking */
   IFX_uint32_t   nMagic;
   /* Table of contents indicating which BBD files are included.
      Lower 16-bit reserved for SLIC100, Upper 16-bit reserved for SLIC200.
      Bit 0:IBB, 1:CIBB, 2:IB, 3:CIB, 4:BB, 5:CBB, 6:IFB, 7:CIFB,
          8:IBDG, 9: CIBGD, 10:IBVD, 11:CIBVD, 15:FXO */
   IFX_uint32_t   nTOC;
   /* SLIC100 section */
   /* Offset of the BBD for SLIC100 IBB within the payload section */
   IFX_uint32_t   nSlic100IBBOffset;
   /* Length of the BBD for SLIC100 IBB within the payload section */
   IFX_uint32_t   nSlic100IBBLength;
   /* Offset of the BBD for SLIC100 CIBB within the payload section */
   IFX_uint32_t   nSlic100CIBBOffset;
   /* Length of the BBD for SLIC100 CIBB within the payload section */
   IFX_uint32_t   nSlic100CIBBLength;
   /* Offset of the BBD for SLIC100 FXO within the payload section */
   IFX_uint32_t   nSlic100FXOOffset;
   /* Length of the BBD for SLIC100 FXO within the payload section */
   IFX_uint32_t   nSlic100FXOLength;
   /* SLIC200 section */
   /* Offset of the BBD for SLIC200 IBB within the payload section */
   IFX_uint32_t   nSlic200IBBOffset;
   /* Length of the BBD for SLIC200 IBB within the payload section */
   IFX_uint32_t   nSlic200IBBLength;
   /* Offset of the BBD for SLIC200 CIBB within the payload section */
   IFX_uint32_t   nSlic200CIBBOffset;
   /* Length of the BBD for SLIC200 CIBB within the payload section */
   IFX_uint32_t   nSlic200CIBBLength;
   /* Offset of the BBD for SLIC200 IFB within the payload section */
   IFX_uint32_t   nSlic200IFBOffset;
   /* Length of the BBD for SLIC200 IFB within the payload section */
   IFX_uint32_t   nSlic200IFBLength;
   /* Offset of the BBD for SLIC200 CIFB within the payload section */
   IFX_uint32_t   nSlic200CIFBOffset;
   /* Length of the BBD for SLIC200 CIFB within the payload section */
   IFX_uint32_t   nSlic200CIFBLength;
};

/* ========================================================================== */
/*                             Local variables                                */
/* ========================================================================== */

/* registration of supported SDD bbd blocks
   downloadable channelwise */
static const IFX_uint16_t SDD_CH_BBD_Blocks[] =
{
   BBD_SDD_CRAM_BLOCK,
   BBD_SDD_RING_CFG_BLOCK,
   BBD_SDD_DC_BASIC_CFG_BLOCK,
   BBD_SDD_MWL_BLOCK,
   BBD_SDD_DCDC_BLOCK,
   BBD_SDD_FXO_CRAM_BLOCK,
   BBD_SDD_CRAM2_BLOCK,
   BBD_SDD_WL_CMD_WRITE
};

struct SDD_CMD
{
#if IFXOS_BYTE_ORDER == IFXOS_BIG_ENDIAN
   CMD_HEAD_BE;
#else
   CMD_HEAD_LE;
#endif
};

typedef union SDD_Msg
{
   IFX_uint16_t val[32];
   struct SDD_CMD cmd;
} SDD_Msg_t;


/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */
static IFX_int32_t sdd_BBD_SelectAndDownload (
                        SDD_CHANNEL_t *pCh,
                        SDD_BBD_Download_t *pBBD);

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */

/**
   Check for the SDD DC/DC config message and analyse it.

   For a one channel device the flag for the combined DC/DC feature is cleared
   without error because there it makes no sense.
   In combined DC/DC operation mode the DCcontrol of channel 0 supplies also
   channel 1.
   The DC/DC operation mode is stored on a per channel basis. Line operations
   are only possible after configuration of the channel.

   \param  pCh          Pointer to the SDD channel structure.
   \param  pMsg         Pointer to FW command.
*/
static IFX_int32_t sdd_bbd_CheckForSddDcDcConfig (SDD_CHANNEL_t *pCh,
                                                  SDD_Msg_t *pMsg)
{
   SDD_DEVICE_t *pDev = pCh->pDev;
   enum SDD_DCDC_TYPE nNewDcDcType;
   IFX_int32_t ret = SDD_statusOk;

   /* Analyse only SDD DC/DC config commands. */
   if ( (pMsg->cmd.CMD == CMD_SDD) &&
        (pMsg->cmd.MOD == MOD_SDD) &&
        (pMsg->cmd.ECMD == SDD_DcDcConfig_ECMD) &&
        (pMsg->cmd.LENGTH >= SDD_DcDcConfig_LENGTH))
   {
      SDD_DcDcConfig_t *pSddDcDcCfg = (SDD_DcDcConfig_t*)pMsg;
      nNewDcDcType = (enum SDD_DCDC_TYPE)pSddDcDcCfg->DcDcHw +
                     SDD_DCDC_TYPE_IBB;

      /* If the FW does not support combined DC/DC operation mode return
         with error. */
      if ((pSddDcDcCfg->DcDcHw & SDD_DcDcConfig_DcDcHw_Mask_COMBINED) &&
          (pDev->caps.bfw_combDcDc != IFX_TRUE))
      {
         /* errmsg: Desired DC/DC operation mode not supported by FW. */
         return SDD_statusBBDcDcNotSupported;
      }

      /* If the FW does not support buck/boost DC/DC operation mode return
         with error. */
      if (((nNewDcDcType == SDD_DCDC_TYPE_BB) ||
           (nNewDcDcType == SDD_DCDC_TYPE_CBB) ) &&
          (pDev->caps.bfw_BuckBoost != IFX_TRUE))
      {
         /* errmsg: Desired DC/DC operation mode not supported by FW. */
         return SDD_statusBBDcDcNotSupported;
      }

#ifndef S220
      /* Writing of SDD DC/DC config must not be done more than once. So abort
         when configuration was already done. */
      if (pCh->pALM->nDcDcType != SDD_DCDC_TYPE_DEFAULT_IBB)
      {
         TRACE(SDD, DBG_LEVEL_LOW,
               ("SDD INFO: Skipping SDD DC/DC reconfiguration\n"));
         /* errmsg: Ignoring already set SDD DC/DC config in BBD download. */
         return SDD_statusBbdDcDcReconfig;
      }
#endif /* S220 */

      /* Block DC/DC type settings which do not match the HW type. */
      if (nNewDcDcType != pDev->nAllowedDcDcType)
      {
         /* errmsg: DC/DC converter type in the BBD download must match the
            DC/DC HW type. */
         return SDD_statusBbdDcDcHwDiffers;
      }
      /* Remember the type of DC/DC operation mode. In dedicated mode the
         DCcontrol works per channel so the type is stored per channel. */
      pCh->pALM->nDcDcType = nNewDcDcType;
      /* Set flag to start an automatic calibration after BBD download. */
      pCh->pALM->bCalibrationNeeded = IFX_TRUE;

      /* In case of combined mode set the type of DC/DC operation mode
         from channel A also to channel B. */
      if ((pDev->bDcDcHwCombined == IFX_TRUE) &&
          (pSddDcDcCfg->CHAN == CMD_CHAN_A))
      {
         /* The combined DC/DC flag is only set on 2 channel systems.
            So there is always a channel 1. */
         pDev->pChannel[1].pALM->nDcDcType = pCh->pALM->nDcDcType;
         /* Set flag to start an automatic calibration after BBD download. */
         pDev->pChannel[1].pALM->bCalibrationNeeded = IFX_TRUE;

         /* Copy this SDD_DcDcConfig message to channel B to have
            a useful configuration there. */
         pSddDcDcCfg->CHAN = CMD_CHAN_B;
         ret = SDD_CmdWrite (pCh->pDev, (IFX_uint32_t *)pMsg->val);

         /* Correct the channel back to A. */
         pSddDcDcCfg->CHAN = CMD_CHAN_A;
      }
   }

   return ret;
}


/**
   Download CRAM Coefficients on channel.

   \param  pCh          Pointer to the SDD channel structure.
   \param  bbd_cram     Pointer to BBD CRAM block.
   \param  dest         Destination: 0 ASDSP, <>0 ACDC

   \return
   - SDD_statusOk
   - SDD_statusBbdCramErr

   \remark
   - It is assumed that the given CRAM bbd block is valid and that the data
     representation in the bbd block is according to BBD specification,
     as follows:

     offset_16: 0xXX, 0xXX,
     data_16[]: 0xXX, 0xXX,
                0xXX, 0xXX,
                ...
     crc_16   : 0xXX, 0xXX
*/
static IFX_int32_t sdd_bbd_DwldCram  (
                        SDD_CHANNEL_t *pCh,
                        bbd_block_t *bbd_cram,
                        IFX_uint16_t dest)
{
   IFX_int32_t       ret   = SDD_statusBbdCramErr;
   IFX_uint32_t      countWords, posWords = 0;
   IFX_uint8_t       lenMsgBytes, lenCramWords, *pByte;
   IFX_uint16_t      cram_offset;
   struct SDD_Coeff  cramCoef;

   memset (&cramCoef, 0, sizeof(cramCoef));
   /* read offset */
   cpb2w (&cram_offset, &bbd_cram->pData[0], sizeof(cram_offset));
   /* set CRAM payload pointer */
   pByte = &bbd_cram->pData[2];

   cramCoef.CMD   = CMD_SDD;
   cramCoef.CHAN  = pCh->nChannel - 1;
   cramCoef.MOD   = MOD_SDD;
   cramCoef.ECMD  = SDD_Coeff_ECMD;

   /* The block size in 16-bit words without offset and crc bytes. */
   countWords = (bbd_cram->size - 4) >> 1 ;

   /* write CRAM data */
   while (countWords > 0)
   {
      if (countWords > (SDD_Coeff_MAXLENGTH >> 1) - 1)
         lenMsgBytes = SDD_Coeff_MAXLENGTH;
      else
         lenMsgBytes = (countWords + 1 /*offset&length word*/) << 1;

      /* Ensure that the CRAM msg payload length is a multiple of 4 bytes. */
      lenMsgBytes = lenMsgBytes - (lenMsgBytes % 4);

      SDD_ASSERT(((IFX_size_t)(lenMsgBytes-2) <=
         sizeof(cramCoef.Coefficients)));

      /* The amount of words of cram buffer written in this message.
         Note that the amount of data is one 16-bit word less than
         lenMsgBytes, because the message starts with one 16-bit word
         containing block offset and length of coefficient block. */
      lenCramWords = (lenMsgBytes >> 1) - 1 /*offset&length word*/;

      cramCoef.LENGTH = lenMsgBytes;
      /* set offset&length word */
      cramCoef.BlockOffset = cram_offset + posWords;
      cramCoef.BlockLength = lenCramWords;
#ifdef S220
      cramCoef.Dest = (dest == 0) ? SDD_Coeff_Dest_ASDSP : SDD_Coeff_Dest_ACDC;
#else /* S220 */
      SDD_OS_UNUSED (dest);
#endif /* S220 */

      /* copy the cram byte buffer into message while taking care of
         endianess;  */
      cpb2w (&cramCoef.Coefficients[0],
             &pByte[posWords << 1], lenCramWords << 1);
      /* write Data */
      ret = SDD_CmdWrite (pCh->pDev, (IFX_uint32_t *)&cramCoef);
      if (SDD_statusOk != ret)
         break;

      /* increment position */
      posWords += lenCramWords;
      /* Decrement the number of words still to be written. */
      countWords -= lenCramWords;
   }

   if (SDD_statusOk == ret)
   {
      RETURN_STATUS(SDD_statusOk);
   }
   else
   {
      /* errmsg: Download of CRAM BBD block failed. */
      RETURN_STATUS(SDD_statusBbdCramErr);
   }
}


/**
   Write a BBD block on a channel.

   Copies a BBD block into the matching configuration message and writes
   this into the device on the addressed channel. For specific tags the
   command header is generated here. For the white-listed tag the header
   is taken from the BBD block. Despite the name white-listed the function
   allows any command to be written to the device and not only white listed
   commands. For the FXO channel there is actually a blacklist of commands
   which must be skipped.

   \param  pCh          Pointer to the SDD channel structure.
   \param  pBBDblock    Handle to bbd coefficient block.

   \return
   - SDD_statusBBDcDcNotSupported
   - SDD_statusBbdDcDcCfgDiffers
   - SDD_statusOk
*/
static IFX_int32_t sdd_bbd_DwldBlock(
                        SDD_CHANNEL_t *pCh,
                        bbd_block_t  *pBBDblock)
{
   SDD_Msg_t   Msg;
   IFX_int32_t ret = SDD_statusOk;
   IFX_uint16_t j = 0, nDataOffset = 0;
   IFX_uint16_t *pDat = (IFX_uint16_t*)(IFX_void_t*)pBBDblock->pData;

   memset (&Msg, 0, sizeof (SDD_Msg_t));

   switch (pBBDblock->tag)
   {
      case BBD_SDD_WL_CMD_WRITE:
         /* endianess conform header */
         Msg.val[0] = ntohs (pDat[0]);
         Msg.val[1] = ntohs (pDat[1]);
         nDataOffset = 2;
         break;

      case BBD_SDD_RING_CFG_BLOCK:
         Msg.cmd.CMD = CMD_SDD;
         Msg.cmd.MOD = MOD_SDD;
         Msg.cmd.ECMD = SDD_RingConfig_ECMD;
         Msg.cmd.LENGTH = SDD_RingConfig_LENGTH;
         break;

      case BBD_SDD_DC_BASIC_CFG_BLOCK:
         Msg.cmd.CMD = CMD_SDD;
         Msg.cmd.MOD = MOD_SDD;
         Msg.cmd.ECMD = SDD_BasicConfig_ECMD;
         Msg.cmd.LENGTH = SDD_BasicConfig_LENGTH;
         break;

      case BBD_SDD_MWL_BLOCK:
         Msg.cmd.CMD = CMD_SDD;
         Msg.cmd.MOD = MOD_SDD;
         Msg.cmd.ECMD = SDD_MwiConfig_ECMD;
         Msg.cmd.LENGTH = SDD_MwiConfig_LENGTH;
         break;

      case BBD_SDD_DCDC_BLOCK:
         Msg.cmd.CMD = CMD_SDD;
         Msg.cmd.MOD = MOD_SDD;
         Msg.cmd.ECMD = SDD_DcDcConfig_ECMD;
         Msg.cmd.LENGTH = SDD_DcDcConfig_LENGTH;
         break;

      default:
         /* errmsg: Downloading a BBD block failed. */
         RETURN_STATUS(SDD_statusBbdBlockErr);
   }

   /* Set (overwrite) the channel to what is given in pCh. */
   Msg.cmd.CHAN = pCh->nChannel - 1;

   /* endianess aware data copy */
   for (j=0; j < (Msg.cmd.LENGTH / 2); j++)
      Msg.val[(CMD_HDR_CNT / 2) + j] = ntohs (pDat[nDataOffset + j]);

   if (pCh->pALM->line_type_fxs == IFX_TRUE)
   {
      /* FXS line */

      /* Analyse the command and if it contains a SDD DC/DC message certain
         actions are executed. */
      ret = sdd_bbd_CheckForSddDcDcConfig (pCh, &Msg);
   }
   else
   {
      /* FXO line */

      if ((Msg.cmd.CMD == CMD_SDD) &&
          (Msg.cmd.MOD == MOD_SDD) &&
          ((Msg.cmd.ECMD == SDD_MwiConfig_ECMD) ||
           (Msg.cmd.ECMD == SDD_RingConfig_ECMD) ||
           (Msg.cmd.ECMD == SDD_DcDcConfig_ECMD)))
      {
         TRACE (SDD, DBG_LEVEL_NORMAL, ("SDD INFO: ch:%d "
                "skipping BBD block with a cmd unsuitable for FXO line\n",
                pCh->nChannel-1));
         /* Skip the above commands on FXO lines. */
         return SDD_statusOk;
      }
   }

   if (ret == SDD_statusOk)
   {
      ret = SDD_CmdWrite (pCh->pDev, (IFX_uint32_t *)Msg.val);
   }

   /* This is not a real error. Just the CmdWrite needed to be skipped.
      For the entire BBD download this is still a success. */
   if (ret == SDD_statusBbdDcDcReconfig)
      ret = SDD_statusOk;

   RETURN_STATUS(ret);
}


/**
   Handle a BBD block, internally we'll dispatch depending on block type.
   This function is used within the context of a single channel.

   \param  pCh          Pointer to the SDD channel structure.
   \param  pBBDblock    Pointer to a kernel level copy of the bbd block.

   \return
   - SDD_statusOk
   - SDD_statusBbdBlockErr
*/
static IFX_int32_t sdd_bbd_BlockHandler (SDD_CHANNEL_t *pCh,
                                         bbd_block_t *pBBDblock)
{
   IFX_int32_t       ret = SDD_statusOk;

   /*lint -save -e40 */
   TRACE(SDD, DBG_LEVEL_LOW,
         ("SDD INFO: ch:%d BBD block with tag 0x%04X found\n",
          pCh->nChannel - 1, pBBDblock->tag));
   /*lint -restore */

   if (pCh->pALM->line_type_fxs == IFX_TRUE)
   {
      /* FXS line */

      switch (pBBDblock->tag)
      {
         case  BBD_SDD_CRAM_BLOCK:
            ret = sdd_bbd_DwldCram (pCh, pBBDblock, SDD_Coeff_Dest_ASDSP);
            break;

         case  BBD_SDD_CRAM2_BLOCK:
            ret = sdd_bbd_DwldCram (pCh, pBBDblock, SDD_Coeff_Dest_ACDC);
            break;

         case BBD_SDD_WL_CMD_WRITE:
            /*lint -fallthrough */
         case BBD_SDD_RING_CFG_BLOCK:
            /*lint -fallthrough */
         case BBD_SDD_DC_BASIC_CFG_BLOCK:
            /*lint -fallthrough */
         case BBD_SDD_MWL_BLOCK:
            /*lint -fallthrough */
         case BBD_SDD_DCDC_BLOCK:
            ret = sdd_bbd_DwldBlock(pCh, pBBDblock);
            break;

         default:
            TRACE (SDD, DBG_LEVEL_NORMAL, ("SDD INFO: ch:%d "
                   "skipping BBD block tag 0x%04X unsuitable for FXS line\n",
                   pCh->nChannel-1, pBBDblock->tag));
            /* parsing continues */
            break;
      }
   }
#ifdef SDD_FEAT_FXO
   else
   {
      /* FXO line */

      switch (pBBDblock->tag)
      {
         case  BBD_SDD_FXO_CRAM_BLOCK:
            ret = sdd_bbd_DwldCram (pCh, pBBDblock, SDD_Coeff_Dest_ASDSP);
            break;

         case BBD_SDD_WL_CMD_WRITE:
            ret = sdd_bbd_DwldBlock(pCh, pBBDblock);
            break;

         default:
            TRACE (SDD, DBG_LEVEL_NORMAL, ("SDD INFO: ch:%d "
                   "skipping BBD block tag 0x%04X unsuitable for FXO line.\n",
                   pCh->nChannel-1, pBBDblock->tag));
            /* parsing continues */
            break;
      }
   }
#endif /* SDD_FEAT_FXO */

   if (SDD_statusOk != ret)
   {
      /* errmsg: Downloading a BBD block failed. */
      RETURN_STATUS(SDD_statusBbdBlockErr);
   }

   RETURN_STATUS(SDD_statusOk);
}


/**
   Does a BBD download on a given channel.

   This handles the ioctl FIO_SDD_BBD_DOWNLOAD to download a BBD buffer on
   one specified channel.

   \param  pCh          Pointer to the SDD channel structure.
   \param  pBBD         Pointer to userspace BBD buffer.

   \return
   - SDD_statusOk
   - SDD_statusParam
   - SDD_statusInvalCh
   - SDD_statusNoMem
   - SDD_statusBbdCorrupt
   - SDD_statusBbdBlockErr
*/
IFX_int32_t SDD_BBD_Download (SDD_CHANNEL_t *pCh, SDD_BBD_Download_t *pBBD)
{
   IFX_int32_t       ret = SDD_statusOk;
   SDD_DEVICE_t      *pDev = IFX_NULL;
   IFX_boolean_t     bBroadcast = IFX_FALSE;
   IFX_uint32_t      i, j, block_num;
   bbd_block_t       bbd_sdd_block;
   bbd_error_t       bbd_err;
   bbd_format_t      bbd;
   IFX_uint32_t      ch;

   /* in case SDD_BBD_Download is called via device ioctl, parameters
      should be validated */
   if ((IFX_NULL == pCh) || (IFX_NULL == pBBD))
      return SDD_statusParam;

   /* detect channel or device context */
   if (pCh->nChannel == 0)
   {
      /* device context - broadcast on all channels */

      bBroadcast = IFX_TRUE;
      pDev = (SDD_DEVICE_t *)(IFX_void_t *)pCh;

      /* check if there is any channel available for download */
      for (ch=0; ch < SDD_MAX_CH_NR; ch++)
      {
         if (pDev->pChannel[ch].pALM != IFX_NULL)
            break;
      }
      if (ch >= SDD_MAX_CH_NR)
      {
         TRACE(SDD, DBG_LEVEL_NORMAL,
               ("SDD INFO: No analog channels for BBD download found.\n"));

         RETURN_DEVSTATUS(SDD_statusNoChg);
      }
   }
   else
   {
      /* channel context - download on one channel only */

      /* currently the download only covers ALM channels - if the
         application tries to download on a channel which doesn't
         have an ALM we return with error. */
      if (pCh->pALM == IFX_NULL)
      {
         /* errmsg: Resource not valid. Channel number out of range */
         RETURN_STATUS(SDD_statusInvalCh);
      }

      pDev = pCh->pDev;
   }

   /* initializations */
   memset (&bbd, 0, sizeof(bbd));
   memset (&bbd_sdd_block, 0, sizeof (bbd_sdd_block));

   if (pBBD->buf != IFX_NULL)
   {
      /* get local memory for bbd buffer */
      bbd.buf  = OS_MapBuffer(pBBD->buf, pBBD->size);
      if (bbd.buf == NULL)
      {
         /* errmsg: No memory could be allocated. */
         RETURN_DEVSTATUS(SDD_statusNoMem);
      }
      /* set size */
      bbd.size = pBBD->size;
      /* check BBD Buffer integrity */
      bbd_err = bbd_check_integrity (&bbd, BBD_SDD_MAGIC);
      if (bbd_err != BBD_INTG_OK)
      {
         OS_UnmapBuffer(bbd.buf);
         /* errmsg: The BBD download buffer is corrupt. */
         RETURN_DEVSTATUS(SDD_statusBbdCorrupt);
      }
   }

   /* go through bbd buffer and download any channel block of relevance found,
      SDD_CH_BBD_Blocks is a constant array of 16 bit entries defining the
      different supported block types */
   block_num  = (sizeof (SDD_CH_BBD_Blocks) / sizeof (IFX_uint16_t));
   bbd_sdd_block.identifier = BBD_SDD_MAGIC;

   for (i = 0; i < block_num; i++)
   {
      /* the order of block search and handling is defined in
         const SDD_CH_BBD_Blocks array. */
      bbd_sdd_block.tag = SDD_CH_BBD_Blocks [i];

      /* sniff blocks of this tag upto maximum allowed */
      for (j = 0; j < BBD_SDD_BLOCK_MAXDWNLD ; j++)
      {
         /* look at block of this index and download it if available */
         bbd_sdd_block.index = j;
         bbd_get_block (&bbd, &bbd_sdd_block);
         if ((bbd_sdd_block.pData == IFX_NULL) || (bbd_sdd_block.size == 0))
         {
            break;
         }

         if (!bBroadcast)
         {
            /* do the download on a single channel */
            ret = sdd_bbd_BlockHandler(pCh, &bbd_sdd_block);
         }
         else
         {
            /* do the downloads on all ALM channels */

            /* ATTENTION, if we need to extend the download to other channel
               types, checks must be added per block if the selected channel
               has the required channel type! */
            for (ch=0; ch < SDD_MAX_CH_NR; ch++)
            {
               pCh = &pDev->pChannel[ch];

               if (pCh->pALM != IFX_NULL)
               {
                  ret = sdd_bbd_BlockHandler(pCh, &bbd_sdd_block);
               }
               if (SDD_statusOk != ret)
                  break;
            }
         }
         /* stop everything if the previous download went wrong */
         if (SDD_statusOk != ret)
            break;
      }
      /* stop everything if the previous download went wrong */
      if (SDD_statusOk != ret)
      {
         OS_UnmapBuffer(bbd.buf);
         /* errmsg: Downloading a BBD block failed. */
         RETURN_STATUS(SDD_statusBbdBlockErr);
      }
   }
   OS_UnmapBuffer(bbd.buf);

   if (SDD_SUCCESS(ret))
   {
      /* Do automatically a calibration on the channel on which the BBD was
         downloaded to adapt to new coefficients. */
      if (!bBroadcast && !pDev->bDcDcHwCombined)
      {
         if (pCh->pALM->bCalibrationNeeded == IFX_TRUE)
         {
            ret = SDD_ALM_Calibration (pCh);
         }
      }
      else
      {
         for (i=0; (i < pDev->caps.nALI) && SDD_SUCCESS(ret); i++)
         {
            pCh = &pDev->pChannel[i];

            if (pCh->pALM->bCalibrationNeeded == IFX_TRUE)
            {
               ret = SDD_ALM_Calibration (pCh);
            }
            if (!SDD_SUCCESS(ret))
            {
               break;
            }
         }
      }

      if (!SDD_SUCCESS(ret))
      {
         /* errmsg: Automatic calibration after BBD download failed. */
         ret = SDD_statusAutomaticCalibrationFailed;
      }
   }

   /* errmsg: Success, no error occured. */
   RETURN_STATUS(SDD_statusOk);
}


/**
   If the binary is an archive select the matching BBD file otherwise pass
   the given binary to the download function.

   Checks for the archive header and if found locate the BBD file matching
   to this SLIC and DC/DC converter type.

   \param  pDev         Pointer to device data.
   \param  pBBD         Pointer to userspace BBD buffer.

   \return
   - Values from SDD_FwDownload.
   - SDD_statusFwDwldFail when copy of header failed
*/
static IFX_int32_t sdd_BBD_SelectAndDownload (
                        SDD_CHANNEL_t *pCh,
                        SDD_BBD_Download_t *pBBD)
{
   struct bbd_archive_header header;

   memset (&header, 0, sizeof (header));

   if ((pBBD->buf != IFX_NULL) &&
       (pBBD->size > sizeof(header)))
   {
      /* copy archive header from user space */
      if (IFXOS_CpyFromUser(&header,
               pBBD->buf, sizeof(header)) == IFX_NULL)
      {
         TRACE (SDD, DBG_LEVEL_HIGH, ("BBD copy header error.\n"));
         /* errmsg: No memory could be allocated. */
         RETURN_STATUS(SDD_statusNoMem);
      }

      /* Process the archive after verifying the header. */
      if (header.nType == BBD_TYPE_GLOBAL_V1 &&
          header.nMagic == BBD_GLOBAL_MAGIC &&
          header.nLength == header.nSlic100IBBLength +
                            header.nSlic100CIBBLength +
                            header.nSlic100FXOLength +
                            header.nSlic200IBBLength +
                            header.nSlic200CIBBLength +
                            header.nSlic200IFBLength +
                            header.nSlic200CIFBLength)
      {
         SDD_DEVICE_t *pDev = IFX_NULL;
         IFX_boolean_t line_type_fxs = IFX_TRUE;

         /* The BBD archive header (header version 1) is ok. */
         /* Select one of the BBD files depending on SLIC and DC/DC type. */

         /* Find device struct */
         if (pCh->nChannel == 0)
         {
            /* device context */
            pDev = (SDD_DEVICE_t *)(IFX_void_t *)pCh;
         }
         else
         {
            /* channel context */
            pDev = pCh->pDev;

            /* The download only covers ALM channels - if the
               application tries to download on a channel which doesn't
               have an ALM we return with error. */
            if (pCh->pALM == IFX_NULL)
            {
               /* errmsg: Resource not valid. Channel number out of range */
               RETURN_STATUS(SDD_statusInvalCh);
            }

            line_type_fxs = pCh->pALM->line_type_fxs;
         }

         if (line_type_fxs == IFX_TRUE)
         {
            switch (pDev->nAllowedDcDcType)
            {
#ifdef S220
               case SDD_DCDC_TYPE_IBB:
                  if ((header.nTOC & BBD_TOC_SLIC200_IBB) == 0)
                  {
                     /* errmsg: BBD archive does not contain the needed BBD. */
                     RETURN_STATUS(SDD_statusBbdMissingInArchive);
                  }
                  pBBD->buf += sizeof(header) + header.nSlic200IBBOffset;
                  pBBD->size = header.nSlic200IBBLength;
                  break;

               case SDD_DCDC_TYPE_CIBB:
                  if ((header.nTOC & BBD_TOC_SLIC200_CIBB) == 0)
                  {
                     /* errmsg: BBD archive does not contain the needed BBD. */
                     RETURN_STATUS(SDD_statusBbdMissingInArchive);
                  }
                  pBBD->buf += sizeof(header) + header.nSlic200CIBBOffset;
                  pBBD->size = header.nSlic200CIBBLength;
                  break;

               case SDD_DCDC_TYPE_IFB:
                  if ((header.nTOC & BBD_TOC_SLIC200_IFB) == 0)
                  {
                     /* errmsg: BBD archive does not contain the needed BBD. */
                     RETURN_STATUS(SDD_statusBbdMissingInArchive);
                  }
                  pBBD->buf += sizeof(header) + header.nSlic200IFBOffset;
                  pBBD->size = header.nSlic200IFBLength;
                  break;

               case SDD_DCDC_TYPE_CIFB:
                  if ((header.nTOC & BBD_TOC_SLIC200_CIFB) == 0)
                  {
                     /* errmsg: BBD archive does not contain the needed BBD. */
                     RETURN_STATUS(SDD_statusBbdMissingInArchive);
                  }
                  pBBD->buf += sizeof(header) + header.nSlic200CIFBOffset;
                  pBBD->size = header.nSlic200CIFBLength;
                  break;
#else /* S220 */
               case SDD_DCDC_TYPE_DEFAULT_IBB:
                  /*lint -fallthrough */
               case SDD_DCDC_TYPE_IBB:
                  if ((header.nTOC & BBD_TOC_SLIC100_IBB) == 0)
                  {
                     /* errmsg: BBD archive does not contain the needed BBD. */
                     RETURN_STATUS(SDD_statusBbdMissingInArchive);
                  }
                  pBBD->buf += sizeof(header) + header.nSlic100IBBOffset;
                  pBBD->size = header.nSlic100IBBLength;
                  break;

               case SDD_DCDC_TYPE_CIBB:
                  if ((header.nTOC & BBD_TOC_SLIC100_CIBB) == 0)
                  {
                     /* errmsg: BBD archive does not contain the needed BBD. */
                     RETURN_STATUS(SDD_statusBbdMissingInArchive);
                  }
                  pBBD->buf += sizeof(header) + header.nSlic100CIBBOffset;
                  pBBD->size = header.nSlic100CIBBLength;
                  break;
#endif /* S220 */

               default:
                  /* errmsg: BBD archive does not contain the needed BBD. */
                  RETURN_STATUS(SDD_statusBbdMissingInArchive);
            }
         }
#ifndef S220
         else
         {
            /* Only SLIC121 has a FXO port */
            if ((header.nTOC & BBD_TOC_SLIC100_FXO) == 0)
            {
               /* errmsg: BBD archive does not contain the needed BBD. */
               RETURN_STATUS(SDD_statusBbdMissingInArchive);
            }
            pBBD->buf += sizeof(header) + header.nSlic100FXOOffset;
            pBBD->size = header.nSlic100FXOLength;
         }
#endif /* !S220 */

      }
   }

   return SDD_BBD_Download (pCh, pBBD);
}


/**
   Access the BBD download from outside of SDD part.

   \param  pLLDummyCh   Pointer to the SDD channel structure.
   \param  pBBD         Pointer to userspace BBD buffer.

   \return
   Values from SDD_BBD_Download.
*/
IFX_int32_t SDD_BBD_EXT_Download(
                        IFX_TAPI_LL_CH_t *pLLCh,
                        IFX_void_t *pBBD)
{
   return sdd_BBD_SelectAndDownload ((SDD_CHANNEL_t *)pLLCh,
                                     (SDD_BBD_Download_t *)pBBD);
}
