/*******************************************************************************

                            Copyright (c) 2013-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/
/**
   \file drv_sdd_calib.c
   This file implements the ALI calibration functions.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "ifx_types.h"

#include "drv_sdd_osmap.h"
#include "drv_sdd_global.h"
#include "drv_sdd_alm_priv.h"
#include "drv_sdd_init.h"
#include "drv_sdd_opmode.h"

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */
/* Limits for useful values returned by the calibration process. */
#define SDD_CALIBRATE_TXOFFSET_MAX   +983
#define SDD_CALIBRATE_TXOFFSET_MIN   -983
#define SDD_CALIBRATE_MEOFFSET_MAX   +1311
#define SDD_CALIBRATE_MEOFFSET_MIN   -1311
#define SDD_CALIBRATE_IDACGAIN_MAX   +32183
#define SDD_CALIBRATE_IDACGAIN_MIN   +26331
#define SDD_CALIBRATE_VDCDC_MAX      +18842
#define SDD_CALIBRATE_VDCDC_MIN      +13926
#define SDD_CALIBRATE_RINGOFFSET_MAX +3277   /* +5 mA */
#define SDD_CALIBRATE_RINGOFFSET_MIN -3277   /* -5 mA */

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */

static IFX_int32_t sdd_alm_Calibration_Start(
                        SDD_CHANNEL_t *pCh,
                        IFX_boolean_t bInternal);

static IFX_int32_t sdd_alm_Calibration_Stop(
                        SDD_CHANNEL_t *pCh);

static IFX_int32_t sdd_alm_Calibration_Get(
                        SDD_CHANNEL_t *pCh,
                        IFX_TAPI_CALIBRATION_CFG_t *pClbConfig);

static IFX_int32_t sdd_alm_Calibration_Set(
                        SDD_CHANNEL_t *pCh,
                        IFX_TAPI_CALIBRATION_CFG_t const *pClbConfig);

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */
/**
   Start calibration on the given analog line.

   Wrapper to protect the actual start function that runs calibration on
   the analog line. Calibration takes about a second and ends automatically.
   The end is signalled with the event IFX_TAPI_EVENT_CALIBRATION_END.

   \param  pLLChannel   Pointer to the TAPI LL channel structure.

   \return
   - SDD_statusOk                    if successful
   - SDD_statusInvalCh               Channel number out of range
*/
static IFX_int32_t SDD_TAPI_LL_ALM_Calibration_Start(
                        IFX_TAPI_LL_CH_t *pLLChannel)
{
   SDD_CHANNEL_t     *pCh;
   IFX_int32_t       ret = SDD_statusOk;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);

   ret = sdd_alm_Calibration_Start (pCh, IFX_FALSE);

   /* release protection */
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   return ret;
}

/**
   Stop a started calibration immediately.

   \param  pLLChannel   Pointer to the TAPI LL channel structure.

   \return
   - SDD_statusOk                   if successful
   - SDD_statusInvalCh              Channel number out of range
*/
static IFX_int32_t SDD_TAPI_LL_ALM_Calibration_Stop(
                        IFX_TAPI_LL_CH_t *pLLChannel)
{
   SDD_CHANNEL_t     *pCh;
   IFX_int32_t       ret = SDD_statusOk;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);

   ret = sdd_alm_Calibration_Stop (pCh);

   /* release protection */
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   return ret;
}

/**
   Finish calibration process on the analog line.

   This function is called after the firmware indicated that calibration has
   finished. It reads the calibration values and does a plausibility check.
   If values are out of range a reset to defaults is done. In the end the
   line is made available again and a TAPI event is sent to the application
   to inform about the result of the calibration process.

   \param  pLLChannel   Pointer to the TAPI LL channel structure.

   \return
   - SDD_statusOk                   if successful
   - SDD_statusInvalCh              Channel number out of range
*/
static IFX_int32_t SDD_TAPI_LL_ALM_Calibration_Finish(
                        IFX_TAPI_LL_CH_t *pLLChannel)
{
   SDD_CHANNEL_t     *pCh;
   SDD_DEVICE_t      *pDev;
   SDD_Calibrate_t   *pCalibrate,
                     *pLastResults;

   IFX_TAPI_EVENT_t  tapiEvent;
   IFX_int32_t       ret = SDD_statusOk;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* This should actually never occur since this function is only called
         within TAPI HL where the channel is already checked. */
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   pDev = pCh->pDev;
   pCalibrate = &pCh->pALM->fw_sdd_calibrate,
   pLastResults = &pCh->pALM->calibrationLastResults;

   /* protect channel from mutual access */
   if (!pCh->pALM->bCalibrationInternal)
      SDD_OS_MutexGet (&pCh->mtxChAcc);

   /* Read the calibrated values for verification below. */
   ret = SDD_CmdRead(pDev, (IFX_uint32_t *)pCalibrate,
                           (IFX_uint32_t *)pCalibrate);

   if (SDD_SUCCESS (ret))
   {
      /* Remember the just read results for later reading. */
      *pLastResults = *pCalibrate;

#ifdef S220
      TRACE(SDD, DBG_LEVEL_LOW,
            ("Calibration results on channel %d:\n"
             "  TX path offset %i\n"
             "  IDAC gain correction %i\n"
             "  Ring Current Offset Compensation %i\n",
             pCh->nChannel - 1,
             (IFX_int16_t)pCalibrate->TxOffset,
             (IFX_int16_t)pCalibrate->IdacGain,
             (IFX_int16_t)pCalibrate->RingOffset));
#else /* S220 */
      TRACE(SDD, DBG_LEVEL_LOW,
            ("Calibration results on channel %d:\n"
             "  TX path offset %i\n"
             "  Measurement equipment offset %i\n"
             "  IDAC gain correction %i\n",
             pCh->nChannel - 1,
             (IFX_int16_t)pCalibrate->TxOffset,
             (IFX_int16_t)pCalibrate->MeOffset,
             (IFX_int16_t)pCalibrate->IdacGain));
      if (pDev->bDcDcHwCombined == IFX_TRUE)
      {
         TRACE(SDD, DBG_LEVEL_LOW,
               ("  Combined DC/DC voltage correction %i\n",
                (IFX_int16_t)pCalibrate->Vdcdc));
      }
#endif /* S220 */
   }

   /* If reading the values failed or range check results that values are
      out of range reset to zero offset values to prevent that implausible
      values are used. */
   if (!SDD_SUCCESS(ret) ||
       ((IFX_int16_t)pCalibrate->TxOffset > SDD_CALIBRATE_TXOFFSET_MAX) ||
       ((IFX_int16_t)pCalibrate->TxOffset < SDD_CALIBRATE_TXOFFSET_MIN) ||
       ((IFX_int16_t)pCalibrate->IdacGain > SDD_CALIBRATE_IDACGAIN_MAX) ||
       ((IFX_int16_t)pCalibrate->IdacGain < SDD_CALIBRATE_IDACGAIN_MIN) ||
#ifdef S220
       ((IFX_int16_t)pCalibrate->RingOffset > SDD_CALIBRATE_RINGOFFSET_MAX) ||
       ((IFX_int16_t)pCalibrate->RingOffset < SDD_CALIBRATE_RINGOFFSET_MIN)
#else /* S220 */
       ((IFX_int16_t)pCalibrate->MeOffset > SDD_CALIBRATE_MEOFFSET_MAX) ||
       ((IFX_int16_t)pCalibrate->MeOffset < SDD_CALIBRATE_MEOFFSET_MIN) ||
       ((pDev->bDcDcHwCombined == IFX_TRUE) &&
        (((IFX_int16_t)pCalibrate->Vdcdc < SDD_CALIBRATE_VDCDC_MIN) ||
         ((IFX_int16_t)pCalibrate->Vdcdc > SDD_CALIBRATE_VDCDC_MAX))  )
#endif /* S220 */
      )
   {
      /* At least one of the values is out of range -> reset to defaults. */
      pCalibrate->TxOffset       = 0;
      pCalibrate->IdacGain       = 0;
#ifdef S220
      pCalibrate->RingOffset     = 0;
#else /* S220 */
      pCalibrate->MeOffset       = 0;
      pCalibrate->Vdcdc          = 0;
#endif /* S220 */
      ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)pCalibrate);
      /* Set flag that calibration failed */
      pCh->pALM->nCalibrationState = IFX_TAPI_CALIBRATION_STATE_FAILED;
   }
   else
   {
      pCh->pALM->nCalibrationState = IFX_TAPI_CALIBRATION_STATE_DONE;
   }

   /* Calibration automatically returns to Disabled. Update the FW-message.
      This field is also used to lock other linemode settings while calibration
      is running. */
   pCh->pALM->sdd_opmode.OperatingMode = SDD_Opmode_Disabled;

   pCh->pALM->bCalibrationNeeded = IFX_FALSE;

   if (pCh->pALM->bCalibrationInternal)
   {
      /* Wake the waiting call in ALM_Calibration below. */
      SDD_OS_MutexRelease(&pCh->pALM->mtxCalibrationWait);
   }
   else
   {
      /* release protection */
      SDD_OS_MutexRelease (&pCh->mtxChAcc);

      /* Send the calibration-end event for the application */
      memset(&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));
      tapiEvent.ch = pCh->nChannel - 1;
      tapiEvent.id = IFX_TAPI_EVENT_CALIBRATION_END;
      SDD_TAPI_EVENT_MODULE_SET(tapiEvent, IFX_TAPI_MODULE_TYPE_ALM);

      if (pCh->pALM->nCalibrationState == IFX_TAPI_CALIBRATION_STATE_DONE)
      {
         tapiEvent.data.calibration = IFX_TAPI_EVENT_CALIBRATION_SUCCESS;
      }
      else
      {
         tapiEvent.data.calibration = IFX_TAPI_EVENT_CALIBRATION_ERROR_RANGE;
      }
      IFX_TAPI_Event_Dispatch(pCh->pTapiCh, &tapiEvent);
   }

   RETURN_STATUS(ret);
}

/**
   Retrieve calibration data from device

   \param  pLLChannel   Pointer to the TAPI LL channel structure.
   \param  pClbConfig   Result as current calibration data.

   \return
   - SDD_statusOk             if successful
   - SDD_statusInvalCh        Channel number out of range
*/
static IFX_int32_t SDD_TAPI_LL_ALM_Calibration_Get(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_CALIBRATION_CFG_t *pClbConfig)
{
   SDD_CHANNEL_t     *pCh;
   IFX_int32_t       ret = SDD_statusOk;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   SDD_ASSERT(pClbConfig != IFX_NULL);

   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   /* sanity check */
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);

   ret = sdd_alm_Calibration_Get (pCh, pClbConfig);

   /* release protection */
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   return ret;
}

/**
   Set the calibration data.

    \param  pLLChannel   Pointer to the TAPI LL channel structure.
    \param  pClbConfig   New calibration data.

    \return
    - SDD_statusOk             if successful
    - SDD_statusInvalCh        Channel number out of range
*/
static IFX_int32_t SDD_TAPI_LL_ALM_Calibration_Set(
                   IFX_TAPI_LL_CH_t *pLLChannel,
                   IFX_TAPI_CALIBRATION_CFG_t const *pClbConfig)
{
   SDD_CHANNEL_t     *pCh;
   IFX_int32_t       ret = SDD_statusOk;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   SDD_ASSERT(pClbConfig != IFX_NULL);

   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);

   ret = sdd_alm_Calibration_Set (pCh, pClbConfig);

   /* release protection */
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   return ret;
}

/**
   Retrieve the results of the last calibration process

   \param  pLLChannel   Pointer to the TAPI LL channel structure.
   \param  pClbConfig   Result as current calibration data.

   \return
   - SDD_statusOk             if successful
   - SDD_statusInvalCh        Channel number out of range
*/
static IFX_int32_t SDD_TAPI_LL_ALM_Calibration_Results(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_CALIBRATION_CFG_t *pClbConfig)
{
   SDD_CHANNEL_t *pCh;
   SDD_Calibrate_t *pLastResults;

   SDD_ASSERT(pLLChannel != IFX_NULL);

   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   if (pClbConfig == IFX_NULL)
   {
      RETURN_STATUS(SDD_statusParam);
   }

   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      /* errmsg: Operation blocked on FXO line */
      RETURN_STATUS(SDD_statusBlockedFXSOpOnFXO);
   }

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);

   pLastResults = &pCh->pALM->calibrationLastResults;

   /* Set return structure to defined values. */
   /* \todo Review how the calculations are made.
            They are not in synch with documentation */
   memset(pClbConfig, 0x00, sizeof(*pClbConfig));
   pClbConfig->dev = pCh->pDev->nDevNr;
   pClbConfig->ch = pCh->nChannel - 1;
   pClbConfig->nState = pCh->pALM->nCalibrationState;
   pClbConfig->nITransOffset = (IFX_int16_t)(
           ((IFX_int32_t)((IFX_int16_t)pLastResults->TxOffset) * 5000) / 32768);
   pClbConfig->nIdacGain     = (IFX_int16_t)(
  (((IFX_int32_t)((IFX_int16_t)pLastResults->IdacGain) * 1000) / 29257) - 1000);
#ifdef S220
   pClbConfig->nIRingOffset     = (IFX_int16_t)
  (((IFX_int32_t)((IFX_int16_t)pLastResults->RingOffset) * 5000) / 32768);
#else /* S220 */
   pClbConfig->nMeOffset     = (IFX_int16_t)(
          ((IFX_int32_t)((IFX_int16_t)pLastResults->MeOffset) * 25000) / 32768);
   pClbConfig->nVdcdc        = (IFX_int16_t)(
     (((IFX_int32_t)((IFX_int16_t)pLastResults->Vdcdc) * 1000) / 16384) - 1000);
#endif /* S220 */
   /* In dedicated DC/DC mode the Vdcdc value is invalid - mark as unused. */
   if (pCh->pDev->bDcDcHwCombined == IFX_FALSE)
      pClbConfig->nVdcdc = 0;

   /* release protection */
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   return SDD_statusOk;
}

/**
   Start calibration process for analog channel.

   Run calibration mechanism and check new calibration offsets

   \param  pCh          Pointer to the SDD channel structure.
   \param  bInternal    Flag indicating call from external TAPI API (IFX_FALSE)
                        or driver internal (IFX_TRUE).

   \return
   - SDD_statusOk                   if successful
   - SDD_statusNoResource           no resource on given channel
   - SDD_statusCalLineNotDisabled   Invalid current line mode
   - SDD_statusOpModeWrErr          Writing the command has failed
*/
static IFX_int32_t sdd_alm_Calibration_Start(
                        SDD_CHANNEL_t *pCh,
                        IFX_boolean_t bInternal)
{
   SDD_Opmode_t  *pOpmod  = &pCh->pALM->sdd_opmode;
   IFX_int32_t ret = SDD_statusOk;
   SDD_CHANNEL_t     *pOtherCh = IFX_NULL;

   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      /* errmsg: Operation blocked on FXO line */
      RETURN_STATUS(SDD_statusBlockedFXSOpOnFXO);
   }

#ifdef S220
   /* Until the DC/DC converter type is set this operation is not allowed. */
   if (pCh->pALM->nDcDcType == SDD_DCDC_TYPE_NOTSET)
#else /* S220 */
   /* For DC/DC converters other than default a BBD download is required. */
   if ((pCh->pALM->nDcDcType == SDD_DCDC_TYPE_DEFAULT_IBB) &&
       (pCh->pDev->nAllowedDcDcType != SDD_DCDC_TYPE_IBB))
#endif /* S220 */
   {
      /* errmsg: Operation blocked until a BBD with DC/DC configuration
                 matching the connected DC/DC hardware type is downloaded. */
      RETURN_STATUS(SDD_statusBlockedDcDcTypeMissmatch);
   }

   /* if calibration is already active - do nothing and return immediately. */
   if (pOpmod->OperatingMode == SDD_Opmode_Calibrate)
   {
      return SDD_statusOk;
   }

   /* Check if transition is valid */
   if (pOpmod->OperatingMode != SDD_Opmode_Disabled)
   {
      /* errmsg: Current line mode is not DISABLED */
      RETURN_STATUS(SDD_statusCalLineNotDisabled);
   }

   /* For combined DC/DC the other channel must be disabled when starting. */
   if (pCh->pDev->bDcDcHwCombined == IFX_TRUE)
   {
      /* The combined DC/DC flag is only set on 2 channel systems. So there
         is always a neighboring channel. */
      pOtherCh = SDD_GetNeighbourChannel(pCh);
      SDD_ASSERT(pOtherCh != IFX_NULL);

      /* protect other channel from mutual access */
      SDD_OS_MutexGet (&pOtherCh->mtxChAcc);

      if (pOtherCh->pALM->sdd_opmode.OperatingMode != SDD_Opmode_Disabled)
      {
         /* release protection */
         SDD_OS_MutexRelease (&pOtherCh->mtxChAcc);
         /* errmsg: Neighbour line mode is not DISABLED */
         RETURN_STATUS(SDD_statusCalNeighbourLineNotDisabled);
      }
   }

   /* Remember if started from API or driver internal. */
   pCh->pALM->bCalibrationInternal = bInternal;

   /* Start calibration */
   pOpmod->OperatingMode = SDD_Opmode_Calibrate;
   ret = SDD_ALM_OpModeSet (pCh);

   if (!SDD_SUCCESS (ret))
   {
      /* Writing the operation mode failed. Restore the last OP_MODE.
         The only way we could have got here was when line mode was "Disabled".
      */
      pOpmod->OperatingMode = SDD_Opmode_Disabled;
   }

   /* In combined DC/DC mode unlock the other channel. */
   if (pOtherCh != IFX_NULL)
   {
      /* release protection */
      SDD_OS_MutexRelease (&pOtherCh->mtxChAcc);
   }

   RETURN_STATUS(ret);
}

/**
   Stop calibration process for analog channel.

   \param  pCh          Pointer to the SDD channel structure.

   \return
   - SDD_statusOk                    if successful
   - SDD_statusOpModeWrErr           Writing the command has failed
*/
static IFX_int32_t sdd_alm_Calibration_Stop(SDD_CHANNEL_t *pCh)
{
   SDD_Opmode_t  *pOpmod  = &pCh->pALM->sdd_opmode;
   IFX_int32_t ret = SDD_statusOk;

   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      /* errmsg: Operation blocked on FXO line */
      RETURN_STATUS(SDD_statusBlockedFXSOpOnFXO);
   }

   /* check if transition is valid */
   if (pOpmod->OperatingMode != SDD_Opmode_Calibrate)
   {
      /* We are not in calibration state. So stopping is not needed.
         This is what was intended by the user and so not an error. */
      return SDD_statusOk;
   }

   /* Stop calibration by changing operating mode to Disabled */
   pOpmod->OperatingMode = SDD_Opmode_Disabled;
   ret = SDD_ALM_OpModeSet (pCh);
   if (!SDD_SUCCESS (ret))
   {
      /* Writing the operation mode failed. Restore the last OP_MODE.
         The only way we could have got here was when line mode was CALIBRATE.*/
      pOpmod->OperatingMode = SDD_Opmode_Calibrate;
   }

   RETURN_STATUS(ret);
}

/**
   Retrieve calibration data.

   \param  pCh          Pointer to the SDD channel structure.
   \param  pClbConfig   Result as current calibration data.

   \return
   - SDD_statusOk                   if successful
   - SDD_statusNoResource           no resource on given channel
   - SDD_statusCalInProgress        Calibration in progress
   - SDD_statusOpModeWrErr          Writing the command has failed
*/
static IFX_int32_t sdd_alm_Calibration_Get(
                        SDD_CHANNEL_t *pCh,
                        IFX_TAPI_CALIBRATION_CFG_t *pClbConfig)
{
   SDD_DEVICE_t         *pDev    = pCh->pDev;
   SDD_Calibrate_t      *pCalibrate = &pCh->pALM->fw_sdd_calibrate;
   SDD_Opmode_t         *pOpmod  = &pCh->pALM->sdd_opmode;
   IFX_int32_t          ret = SDD_statusOk;

   /* Sanity check */
   if (pClbConfig == IFX_NULL)
   {
      RETURN_STATUS(SDD_statusParam);
   }

   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      /* errmsg: Operation blocked on FXO line */
      RETURN_STATUS(SDD_statusBlockedFXSOpOnFXO);
   }

   /* Set return structure to defined values. */
   memset(pClbConfig, 0x00, sizeof(*pClbConfig));
   pClbConfig->dev = pDev->nDevNr;
   pClbConfig->ch = pCh->nChannel - 1;

   if (pOpmod->OperatingMode == SDD_Opmode_Calibrate)
   {
      /* All returned values are invalid */
      pClbConfig->nState = IFX_TAPI_CALIBRATION_STATE_NO;
      /* errmsg: Current line mode is CALIBRATE */
      RETURN_STATUS(SDD_statusCalInProgress);
   }

   /* Read calibration values from FW */
   ret = SDD_CmdRead (pDev, (IFX_uint32_t *)pCalibrate,
                            (IFX_uint32_t *)pCalibrate);

   if (SDD_SUCCESS (ret))
   {
      /* \todo Review how the calculations are done. */
      pClbConfig->nState = pCh->pALM->nCalibrationState;
      pClbConfig->nITransOffset = (IFX_int16_t)(
             ((IFX_int32_t)((IFX_int16_t)pCalibrate->TxOffset) * 5000) / 32768);
      pClbConfig->nIdacGain     = (IFX_int16_t)(
    (((IFX_int32_t)((IFX_int16_t)pCalibrate->IdacGain) * 1000) / 29257) - 1000);
#ifdef S220
    /* IFX_TAPI_CALIBRATION_CFG_t needs to be mapped */
      pClbConfig->nIRingOffset     = (IFX_int16_t)
    (((IFX_int32_t)((IFX_int16_t)pCalibrate->RingOffset) * 5000) / 32768);
#else /* S220 */
      pClbConfig->nMeOffset     = (IFX_int16_t)(
            ((IFX_int32_t)((IFX_int16_t)pCalibrate->MeOffset) * 25000) / 32768);
      pClbConfig->nVdcdc        = (IFX_int16_t)(
       (((IFX_int32_t)((IFX_int16_t)pCalibrate->Vdcdc) * 1000) / 16384) - 1000);
#endif /* S220 */
      /* In dedicated DC/DC mode the Vdcdc value is invalid - mark as unused. */
      if (pDev->bDcDcHwCombined == IFX_FALSE)
         pClbConfig->nVdcdc = 0;
   }
   else
   {
      /* All returned values are invalid */
      pClbConfig->nState = IFX_TAPI_CALIBRATION_STATE_NO;
   }

   RETURN_STATUS(ret);
}

/**
   Set calibration data.

   \param  pCh          Pointer to the SDD channel structure.
   \param  pClbConfig   New calibration data.

   \return
   - SDD_statusOk                   if successful
   - SDD_statusCalInProgress        Calibration in progress
   - SDD_statusOpModeWrErr          Writing the command has failed
*/
static IFX_int32_t sdd_alm_Calibration_Set(
                        SDD_CHANNEL_t *pCh,
                        IFX_TAPI_CALIBRATION_CFG_t const *pClbConfig)
{
   SDD_DEVICE_t         *pDev    = pCh->pDev;
   SDD_Calibrate_t      *pCalibrate = &pCh->pALM->fw_sdd_calibrate;
   SDD_Opmode_t         *pOpmod  = &pCh->pALM->sdd_opmode;
   IFX_int32_t          ret;
   IFX_int16_t          TxOffset, IdacGain;
#ifdef S220
   IFX_int16_t          RingOffset;
#else /* S220 */
   IFX_int16_t          MeOffset, Vdcdc;
#endif /* S220 */

   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      /* errmsg: Operation blocked on FXO line */
      RETURN_STATUS(SDD_statusBlockedFXSOpOnFXO);
   }

   if (pOpmod->OperatingMode == SDD_Opmode_Calibrate)
   {
      /* errmsg: Current line mode is CALIBRATE */
      RETURN_STATUS(SDD_statusCalInProgress);
   }

   /* calculate SDD values from TAPI */
   TxOffset     = (IFX_int16_t)
                  ((IFX_int32_t)pClbConfig->nITransOffset * 32768 / 5000);
   IdacGain     = (IFX_int16_t)
                  (((IFX_int32_t)pClbConfig->nIdacGain + 1000) * 29257 / 1000);
#ifdef S220
   RingOffset        = (IFX_int16_t)
                  ((IFX_int32_t)pClbConfig->nIRingOffset * 32768 / 5000);
#else
   MeOffset     = (IFX_int16_t)
                  ((IFX_int32_t)pClbConfig->nMeOffset * 32768 / 25000);
   Vdcdc        = (IFX_int16_t)
                  (((IFX_int32_t)pClbConfig->nVdcdc + 1000) * 16384 / 1000);
#endif

      TRACE(SDD, DBG_LEVEL_LOW,
            ("Calibration set values. ch%d:\n"
             "  TX path offset %i\n"
             "  IDAC gain correction %i\n"
#ifdef S220
             "  Ring Current Offset Compensation %i\n"
#else /* S220 */
             "  Measurement equipment offset %i\n"
             "  Combined DC/DC voltage correction %i\n"
#endif /* S220 */
          ,
             pCh->nChannel - 1,
             TxOffset,
             IdacGain,
#ifdef S220
             RingOffset
#else /* S220 */
             MeOffset,
             Vdcdc
#endif /* S220 */
          ));

   /* sanity check */
   if ( (TxOffset > SDD_CALIBRATE_TXOFFSET_MAX) ||
        (TxOffset < SDD_CALIBRATE_TXOFFSET_MIN) ||
        (IdacGain > SDD_CALIBRATE_IDACGAIN_MAX) ||
        (IdacGain < SDD_CALIBRATE_IDACGAIN_MIN) ||
#ifdef S220
        (RingOffset > SDD_CALIBRATE_RINGOFFSET_MAX) ||
        (RingOffset < SDD_CALIBRATE_RINGOFFSET_MIN)
#else /* S220 */
        (MeOffset > SDD_CALIBRATE_MEOFFSET_MAX) ||
        (MeOffset < SDD_CALIBRATE_MEOFFSET_MIN) ||
        ((pDev->bDcDcHwCombined == IFX_TRUE) &&
         ((Vdcdc < SDD_CALIBRATE_VDCDC_MIN) ||
          (Vdcdc > SDD_CALIBRATE_VDCDC_MAX))  )
#endif /* S220 */
      )
   {
      /* errmsg: Calibration set values are out of range  */
      RETURN_STATUS(SDD_statusCalSetValOutOfRange);
   }

   /* prepare command */
    pCalibrate->TxOffset       = (IFX_uint16_t)TxOffset;
    pCalibrate->IdacGain       = (IFX_uint16_t)IdacGain;
#ifdef S220
   pCalibrate->RingOffset     = (IFX_uint16_t)RingOffset;
#else /* S220 */
    pCalibrate->MeOffset       = (IFX_uint16_t)MeOffset;
   pCalibrate->Vdcdc          = (IFX_uint16_t)Vdcdc;
#endif /* S220 */
   ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)pCalibrate);

   RETURN_STATUS(ret);
}


/**
   Used internally to start calibration on the given analog line.

   This function is called after a BBD download was done to calibrate the
   analog channel where the download has just been done. This function blocks
   until the finished event has been received and the validation of the
   calibration was done. This function passes a parameter to prevent the
   sending of the TAPI event IFX_TAPI_EVENT_CALIBRATION_END and at the
   same time initiate the wakeup from the blocking wait.

   \param  pCh          Pointer to the SDD channel structure.

   \return
   - SDD_statusOk       if successful
   - SDD_statusInvalCh  Channel number out of range
*/
IFX_int32_t SDD_ALM_Calibration (SDD_CHANNEL_t *pCh)
{
   IFX_int32_t ret;

   /* sanity check */
   SDD_ASSERT(pCh != IFX_NULL);
   SDD_ASSERT(pCh->pALM != IFX_NULL);

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);

   ret = sdd_alm_Calibration_Start (pCh, IFX_TRUE);

   if (SDD_SUCCESS (ret))
   {
      /* Block until the semaphore is given in the calibration finished
         function or a signal interrupts. */
      ret = SDD_OS_MutexLockInterruptible(&pCh->pALM->mtxCalibrationWait);
      if (ret != 0)
      {
         ret = SDD_statusErr;
      }
   }

   /* release protection */
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   RETURN_STATUS (ret);
}

#ifdef SDD_FEAT_OL_CALIBRATION_SET
/**
   Configures the open loop conductances for each channel.

   \param  pLLChannel   Pointer to the TAPI LL channel structure.

   \return
   - ???SDD_statusOk                   if successful
   - ???SDD_statusCalInProgress        Calibration in progress
   - ???SDD_statusOpModeWrErr          Writing the command has failed
*/
static IFX_int32_t SDD_ALM_OL_Calibration_Set(
                        IFX_TAPI_LL_CH_t *pLLChannel)
{
   IFX_int32_t                    ret = IFX_SUCCESS;
   SDD_DEVICE_t                   *pDev;
   SDD_CHANNEL_t                  *pCh;
   SDD_OpenLoopCalibration_t      *pCalib;
   IFX_TAPI_GR909_RESULT_t        *pResults;
   IFX_uint8_t                    ch;

   /* Sanity check ? */
   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   ch = pCh->nChannel - 1;
   pDev = pCh->pDev;

   /* read values from SDD_GR909ResultsRead*/
   ret = SDD_TAPI_LL_ALM_LT_GR909_Result (pLLChannel, pResults);

   /* calculate grg, gtg, gtr*/
   pCalib->GRG = 2 ^ 16 * 1158 / pResults->RftR2g;
   pCalib->GTG = 2 ^ 16 * 1158 / pResults->RftT2g;
   pCalib->GTR = 2 ^ 16 * 1158 / pResults->RftT2r;

   /* write calibration values */
   pCalib->CMD        = CMD_SDD;
   pCalib->CHAN       = ch;
   pCalib->MOD        = MOD_SDD;
   pCalib->ECMD       = SDD_OpenLoopCalibration_ECMD;
   pCalib->LENGTH     = SDD_OpenLoopCalibration_LENGTH;
   ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)pCalib);
   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Writing the basic configuration failed. */
      RETURN_STATUS(SDD_statusBasicCfgWrErr);
   }

   RETURN_STATUS(ret);
}
#endif /* SDD_FEAT_OL_CALIBRATION_SET */

/* ========================================================================== */
/*                         Function pointer exports                           */
/* ========================================================================== */
/**
   Function called by init_module of device, fills up ALM function pointers
   which are passed to HL TAPI during registration.

   \param  pAlm         Pointer to ALM module.
*/
IFX_void_t SDD_ALM_Calib_Func_Register (IFX_TAPI_DRV_CTX_ALM_t *pAlm)
{
   pAlm->Calibration_Start       = SDD_TAPI_LL_ALM_Calibration_Start;
   pAlm->Calibration_Stop        = SDD_TAPI_LL_ALM_Calibration_Stop;
   pAlm->Calibration_Finish      = SDD_TAPI_LL_ALM_Calibration_Finish;
   pAlm->Calibration_Get         = SDD_TAPI_LL_ALM_Calibration_Get;
   pAlm->Calibration_Set         = SDD_TAPI_LL_ALM_Calibration_Set;
   pAlm->Calibration_Results_Get = SDD_TAPI_LL_ALM_Calibration_Results;
}
