/*******************************************************************************

                            Copyright (c) 2014-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_dwld.c
   This file implements the download functions.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "drv_sdd_dwld.h"
#include "drv_sdd_init.h"

/* ========================================================================== */
/*                            Macro definitions                               */
/* ========================================================================== */
/* FW download header struct values */
#define VCODEC_FW_TYPE_GLOBAL_V1 0x05100001  /* type 0x0510, version 0x0001 */
#define VCODEC_FW_GLOBAL_MAGIC   0x56434F44  /* 'VCOD' character sequence */

/* ========================================================================== */
/*                             Type definitions                               */
/* ========================================================================== */

/* Header of VCODEC binary section containing both SLIC100 and SLIC200
   VCODEC FW binaries (big endian) */
struct vcodec_container_header
{
   /* Type and Version identifier */
   IFX_uint32_t   nType;
   /* Length of the payload following this header */
   IFX_uint32_t   nLength;
   /* MAGIC value for endianess checking */
   IFX_uint32_t   nMagic;
   /* Offset of the SLIC100 VCODEC-FW within the payload section */
   IFX_uint32_t   nSlic100FwOffset;
   /* Length of the SLIC100 VCODEC-FW within the payload section */
   IFX_uint32_t   nSlic100FwLength;
   /* Offset of the SLIC200 VCODEC-FW within the payload section */
   IFX_uint32_t   nSlic200FwOffset;
   /* Length of the SLIC200 VCODEC-FW within the payload section */
   IFX_uint32_t   nSlic200FwLength;
};

/* ========================================================================== */
/*                             Global variables                               */
/* ========================================================================== */

/* ========================================================================== */
/*                              Local variables                               */
/* ========================================================================== */

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */
static IFX_int32_t sdd_FwSelectAndDownload(
                        SDD_DEVICE_t *pDev,
                        SDD_FW_Download_t *pSddFwDwld);

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */

/**
   Ensure the FW contains a SLIC100 and SLIC200 binary and pass one of these
   to the actual download function.

   Checks for the VCODEC section header and if found locates either the
   SLIC100 or SLIC200 VCODEC FW binary and pass this to the download function.

   \param  pDev         Pointer to device data.
   \param  pSddFwDwld   Pointer to struct with parameters.

   \return
   - Values from SDD_FwDownload.
   - SDD_statusFwDwldFail when copy of header failed
*/
static IFX_int32_t sdd_FwSelectAndDownload(
                        SDD_DEVICE_t *pDev,
                        SDD_FW_Download_t *pSddFwDwld)
{
   struct vcodec_container_header header;
   SDD_FIRMWARE_IF_t *p = pDev->pMbxIF;
   IFX_int32_t ret = SDD_statusFwDwldFail;

   memset (&header, 0, sizeof (header));

   if (p == IFX_NULL)
   {
      /*errmsg: The SDD MBX driver is not registered. */
      RETURN_DEVSTATUS (SDD_statusMbxDrvNotRegistered);
   }

   /* copy section header from user space */
   if (IFXOS_CpyFromUser(&header,
            pSddFwDwld->pPRAMfw, sizeof(header)) == IFX_NULL)
   {
      TRACE (SDD, DBG_LEVEL_HIGH, ("VCODEC FW copy header error.\n"));
      /* errmsg: Firmware download failed. */
      RETURN_DEVSTATUS (SDD_statusFwDwldFail);
   }

   /* Process the container after verifying the header. */
   if (header.nType == VCODEC_FW_TYPE_GLOBAL_V1 &&
       header.nLength + sizeof(header) == pSddFwDwld->pram_size &&
       header.nMagic == VCODEC_FW_GLOBAL_MAGIC &&
       header.nLength == header.nSlic100FwLength + header.nSlic200FwLength)
   {
      /* The VCODEC section header (header version 1) is ok. */

      /* Select one of the contained images depending on SLIC configuration. */
#ifdef S220
      pSddFwDwld->pPRAMfw += sizeof(header) + header.nSlic200FwOffset;
      pSddFwDwld->pram_size = header.nSlic200FwLength;
#else /* S220 */
      pSddFwDwld->pPRAMfw += sizeof(header) + header.nSlic100FwOffset;
      pSddFwDwld->pram_size = header.nSlic100FwLength;
#endif /* S220 */

      ret = p->fw_dwld(pSddFwDwld->pPRAMfw, pSddFwDwld->pram_size);

      /* Translate error codes */
      switch (ret)
      {
         case SDD_FW_ERR_NO_ERROR:
            ret = SDD_statusOk;
            break;

         case SDD_FW_ERR_FW_DWLD_TIMEOUT:
            /* errmsg: Firmware download timeout. */
            ret = SDD_statusFwDwldTimeout;
            break;

         default:
            /* errmsg: Firmware download failed. */
            ret = SDD_statusFwDwldFail;
            break;
      }

      return ret;
   }

   /* errmsg: Firmware binary is missing the needed VCODEC section. */
   return SDD_statusFwDwldVcodecMissing;
}

/**
   Access the FW download from outside of SDD part.

   \param  pLLCh        Pointer to the SDD channel structure.
   \param  pSddFwDwld   Pointer to struct with parameters.

   \return
   Values from SDD_FwDownload.
*/
IFX_int32_t SDD_EXT_FwDownload(
                        IFX_TAPI_LL_DEV_t* pLLDev,
                        SDD_FW_Download_t *pSddFwDwld)
{
   return sdd_FwSelectAndDownload((SDD_DEVICE_t *)pLLDev, pSddFwDwld);
}

