/*******************************************************************************

                            Copyright (c) 2013-2015
                            Lantiq Deutschland GmbH
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_event.c
   This file implements the debug info handling functions.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "ifx_types.h"
#include "ifxos_memory_alloc.h"

#include "drv_sdd_osmap.h"
#include "drv_sdd_global.h"

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */
/**
   Extract filename from complete path information.

   \param dst   pointer to character array accepting the filename

   \param src   pointer to constant character array holding the complete path
                information of the file

   \return
   None.
*/
static IFX_void_t SDD_StripPathCpy (IFX_char_t *dst, const IFX_char_t *src)
{
   IFX_uint32_t nMax = strlen (src);
   IFX_uint32_t i = nMax;

   /* only name of file without path */
   while (i > 0 && src[i] != 0x5C && src[i] != '/')
   {
      i--;
   }
   if (i > 0)
   {
      nMax = nMax - i;
      /* ignore / */
      i++;
   }
   if (nMax > (IFX_TAPI_MAX_FILENAME - 1))
   {
      nMax = IFX_TAPI_MAX_FILENAME - 2;
      dst[IFX_TAPI_MAX_FILENAME - 1] = 0;
   }
   memcpy (dst, &src[i], nMax+1);
}


/**
   Report an error to TAPI

   \param  pChOrDev     Pointer to the channel or device context struct.
   \param  err          Error number.
   \param  nLine        Source code line number where the error occured.
   \param  sFile        Source code filename where the error occured as string.
*/
IFX_void_t SDD_ErrorEvent (
                        const IFX_TAPI_LL_CH_t *pChOrDev,
                        const IFX_uint16_t err,
                        const IFX_uint32_t nLine,
                        const IFX_char_t* sFile)
{
   IFX_TAPI_ErrorLine_t errLine;
   IFX_TAPI_EVENT_t tapiEvent;
   SDD_CHANNEL_t *pCh;
   TAPI_CHANNEL *pTapiCh;

   SDD_ASSERT(pChOrDev != IFX_NULL);

   /* Fill the event structure. */
   memset(&tapiEvent, 0, sizeof(tapiEvent));
   memset(&errLine, 0, sizeof(errLine));
   errLine.nLlCode = err;
   errLine.nHlCode = 0;
   errLine.nLine = nLine;
   SDD_StripPathCpy (errLine.sFile, sFile);
   tapiEvent.data.error = &errLine;
   SDD_TAPI_EVENT_MODULE_SET(tapiEvent, IFX_TAPI_MODULE_TYPE_NONE);

   /* Indicate channel context at first. It will be corrected below if
      device context is found. */
   tapiEvent.id = IFX_TAPI_EVENT_FAULT_GENERAL_CHINFO;

   /* For IFX_TAPI_Event_Dispatch a pointer to a TAPI channel is needed.
      Find this from either the SDD channel or SDD device context. */
   pCh = (SDD_CHANNEL_t *)pChOrDev;
   if (pCh->nChannel == 0)
   {
      /* The initial assumption was wrong - we received a device pointer. */
      pCh = ((SDD_DEVICE_t *)pChOrDev)->pChannel;
      /* Indicate that error occured on the device. */
      tapiEvent.id = IFX_TAPI_EVENT_FAULT_GENERAL_DEVINFO;
   }
   pTapiCh = pCh->pTapiCh;

   IFX_TAPI_Event_Dispatch(pTapiCh, &tapiEvent);
}
