/*******************************************************************************

                            Copyright (c) 2013-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_fxo.c
   This file implements the FXO functions.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "drv_sdd_config.h"

#ifdef SDD_FEAT_FXO

#include "drv_sdd_fxo.h"
#include "drv_sdd_fw_cmd.h"
#include "drv_sdd_alm_priv.h"
#include "drv_sdd_opmode.h"

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */

/* ========================================================================== */
/*                             Type definitions                               */
/* ========================================================================== */

/* ========================================================================== */
/*                             Global variables                               */
/* ========================================================================== */

/* ========================================================================== */
/*                              Local variables                               */
/* ========================================================================== */

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */

/**
   Check parameters for FXO functions.

   \param pLLChannel    Handle to TAPI low level channel structure.
   \param ppCh          Pointer to pointer to the SDD channel structure.
   \param ppDev         Pointer to pointer to the SDD device structure.

   \return
   - SDD_statusOk
   - SDD_statusParam
   - SDD_statusInvalCh
   - SDD_statusInvalDev
*/
static IFX_int32_t SDD_ALM_CheckFxoParams(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        SDD_CHANNEL_t **ppCh,
                        SDD_DEVICE_t **ppDev)
{
   SDD_CHANNEL_t       *pCh = IFX_NULL;

   if (pLLChannel == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }

   *ppCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   if (*ppCh == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }

   /* pCh is needed bz RETURN_STATUS() */
   pCh = *ppCh;

   if ((*ppCh)->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   if ((*ppCh)->pDev == IFX_NULL)
   {
      /* errmsg: Resource not valid. Device number out of range */
      RETURN_STATUS(SDD_statusInvalDev);
   }

   *ppDev = (*ppCh)->pDev;

   if ((*ppCh)->pALM->line_type_fxs != IFX_FALSE)
   {
      /* errmsg: FXO service called, line is not FXO */
      RETURN_STATUS(SDD_statusLineNotFXO);
   }

   /* errmsg: Success, no error occured. */
   return SDD_statusOk;
}


/**
   Check state of the FXO channel.

   \param pLLChannel    Handle to TAPI low level channel structure.
   \param pCh           Pointer to pointer to the SDD channel structure.
   \param pDev          Pointer to pointer to the SDD device structure.

   \return IFX_int32_t
   - SDD_statusOk
   - SDD_statusFXOLineDisabled
   - error code from SDD_ALM_CheckFxoParams()
*/
static IFX_int32_t SDD_ALM_CheckFxoState(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        SDD_CHANNEL_t **pCh,
                        SDD_DEVICE_t **pDev)
{
   IFX_int32_t         ret = SDD_statusOk;

   ret = SDD_ALM_CheckFxoParams(pLLChannel, pCh, pDev);
   if(ret != SDD_statusOk)
      RETURN_STATUS(ret);

   if (!((*pCh)->pALM->fxo_flags & (1U << FXO_LINE_MODE)))
      /* errmsg: FXO line disabled. */
      ret = SDD_statusFXOLineDisabled;

   RETURN_STATUS(ret);
}


/**
   Set the new hook state on FXO line.

   \param  pLLChannel   Handle to TAPI low level channel structure.
   \param  hook         Enum type for TAPI FXO hook

   \return
   - SDD_statusOk if successful
   - SDD_statusInvalCh if no ALM resource
   - SDD_statusParam if hook parameter is ambiguous
   - SDD_statusLineNotFXO if line is not FXO
   - error code from SDD_CmdWrite()
*/
static IFX_int32_t SDD_TAPI_LL_ALM_FxoHookSet(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_FXO_HOOK_t hook)
{
   SDD_CHANNEL_t       *pCh = IFX_NULL;
   SDD_DEVICE_t        *pDev = IFX_NULL;
   SDD_FxoHookSwitch_t *pFxoHookSwitch = IFX_NULL;
   IFX_int32_t         ret = SDD_statusOk;

   ret = SDD_ALM_CheckFxoState(pLLChannel, &pCh, &pDev);
   if(ret != SDD_statusOk)
      RETURN_STATUS(ret);

   pFxoHookSwitch = &pCh->pALM->sdd_fxoHookSwitch;

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);
   switch (hook)
   {
      case IFX_TAPI_FXO_HOOK_ONHOOK:
         pFxoHookSwitch->HF = SDD_FxoHookSwitch_HF_ON;
         break;
      case IFX_TAPI_FXO_HOOK_OFFHOOK:
         pFxoHookSwitch->HF = SDD_FxoHookSwitch_HF_OFF;
         break;
      default:
         ret = VMMC_statusParam;
         goto error;
   }

   ret = SDD_CmdWrite(pDev, (IFX_uint32_t*)pFxoHookSwitch);

error:
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   RETURN_STATUS(ret);
}


/**
   Get the current FXO hook state.

   \param  pLLChannel   Handle to TAPI low level channel structure.
   \param  p_hook       Pointer to IFX_TAPI_FXO_HOOK_t enum.

   \return
   - SDD_statusOk if successful
   - SDD_statusInvalCh if no ALM resource
   - SDD_statusLineNotFXO if line is not FXO

   \remarks
   The hook status cannot be read from firmware as SDD_FxoHookSwitch
   command is write-only. The hook state is obtained from SDD_FxoHookSwitch
   command cache in ALM context.
*/
static IFX_int32_t SDD_TAPI_LL_ALM_FxoHookGet(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_FXO_HOOK_t *p_hook)
{
   SDD_FxoHookSwitch_t *pFxoHookSwitch = IFX_NULL;
   SDD_CHANNEL_t       *pCh = IFX_NULL;
   SDD_DEVICE_t        *pDev = IFX_NULL;
   IFX_int32_t         ret = SDD_statusOk;

   ret = SDD_ALM_CheckFxoState(pLLChannel, &pCh, &pDev);
   if(ret != SDD_statusOk)
      return ret;

   pFxoHookSwitch = &pCh->pALM->sdd_fxoHookSwitch;

   switch (pFxoHookSwitch->HF)
   {
      case SDD_FxoHookSwitch_HF_OFF:
         *p_hook = IFX_TAPI_FXO_HOOK_OFFHOOK;
         break;
      default:
         *p_hook = IFX_TAPI_FXO_HOOK_ONHOOK;
         break;
   }

   RETURN_STATUS(ret);
}


/**
   Issue the hook flash on FXO line.

   \param  pLLChannel   Handle to TAPI low level channel structure.

   \return
   - SDD_statusOk if successful
   - SDD_statusInvalCh if no ALM resource
   - SDD_statusLineNotFXO if line is not FXO
   - error code from SDD_CmdWrite()
*/
static IFX_int32_t SDD_TAPI_LL_ALM_FxoFlashSet(IFX_TAPI_LL_CH_t *pLLChannel)
{
   SDD_FxoHookSwitch_t *pFxoHookSwitch;
   IFX_int32_t         ret = SDD_statusOk;
   SDD_CHANNEL_t       *pCh = IFX_NULL;
   SDD_DEVICE_t        *pDev = IFX_NULL;

   ret = SDD_ALM_CheckFxoState(pLLChannel, &pCh, &pDev);
   if(ret != SDD_statusOk)
      RETURN_STATUS(ret);

   pFxoHookSwitch = &pCh->pALM->sdd_fxoHookSwitch;

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);
   pFxoHookSwitch->HF = SDD_FxoHookSwitch_HF_FLASH;
   ret = SDD_CmdWrite(pDev, (IFX_uint32_t*)pFxoHookSwitch);
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   RETURN_STATUS(ret);
}


/**
   Configure the hook flash time.

   \param  pLLChannel   Handle to TAPI low level channel structure.
   \param  p_fhCfg      Pointer to IFX_TAPI_FXO_FLASH_CFG_t structure.

   \return
   - SDD_statusOk if successful
   - SDD_statusInvalCh if no ALM resource
   - error code from SDD_CmdWrite()
*/
static IFX_int32_t SDD_TAPI_LL_ALM_FxoFlashCfg(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_FXO_FLASH_CFG_t const *p_fhCfg)
{
   SDD_FxoConfig_t     *pFxoConfig = IFX_NULL;
   IFX_int32_t         ret = SDD_statusOk;
   SDD_CHANNEL_t       *pCh = IFX_NULL;
   SDD_DEVICE_t        *pDev = IFX_NULL;

   ret = SDD_ALM_CheckFxoParams(pLLChannel, &pCh, &pDev);
   if(ret != SDD_statusOk)
      RETURN_STATUS(ret);

   pFxoConfig = &pCh->pALM->sdd_fxoConfig;

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);
   pFxoConfig->FLASHT = p_fhCfg->nFlashTime >> 1;
   ret = SDD_CmdWrite(pDev, (IFX_uint32_t*)pFxoConfig);
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   RETURN_STATUS(ret);
}


/**
   Configure the maximum detection time for an open switching interval (OSI).

   \param  pLLChannel   Handle to TAPI low level channel structure.
   \param  p_osiCfg     Pointer to IFX_TAPI_FXO_OSI_CFG_t structure.

   \return
   - SDD_statusOk if successful
   - SDD_statusInvalCh if no ALM resource
   - SDD_statusParam if OSI max time <= 70 ms
   - error code from SDD_CmdWrite()
*/
static IFX_int32_t SDD_TAPI_LL_ALM_FxoOsiCfg(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_FXO_OSI_CFG_t const *p_osiCfg)
{
   SDD_FxoConfig_t     *pFxoConfig = IFX_NULL;
   IFX_int32_t         ret = SDD_statusOk;
   SDD_CHANNEL_t       *pCh = IFX_NULL;
   SDD_DEVICE_t        *pDev = IFX_NULL;

   ret = SDD_ALM_CheckFxoParams(pLLChannel, &pCh, &pDev);
   if(ret != SDD_statusOk)
      RETURN_STATUS(ret);

   /* time must be > 70 ms */
   if (p_osiCfg->nOSIMax <= 70)
      RETURN_STATUS(SDD_statusParam);

   pFxoConfig = &pCh->pALM->sdd_fxoConfig;

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);
   pFxoConfig->OSIT = p_osiCfg->nOSIMax >> 1;
   ret = SDD_CmdWrite(pDev, (IFX_uint32_t*)pFxoConfig);
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   RETURN_STATUS(ret);
}


/**
   Get FXO status bit.

   \param  pLLChannel   Handle to TAPI low level channel structure.
   \param  bit_pos      Bit position defined in drv_tapi_ll_interface.h.
                        Can be one of FXO_BATTERY,FXO_APOH,FXO_POLARITY,
                        FXO_RING.
   \param  enable       Pointer to IFX_enDis_t; output value.

   \return
   - SDD_statusOk

   \remarks
   Just a trivial converter from bit value of fxo_flags to IFX_enDis_t enum.
*/
static IFX_int32_t SDD_TAPI_LL_ALM_FxoStatGet(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_uint32_t bit_pos,
                        IFX_enDis_t *enable)
{
   SDD_CHANNEL_t  *pCh = IFX_NULL;
   SDD_DEVICE_t   *pDev = IFX_NULL;
   IFX_int32_t    ret = SDD_statusOk;

   ret = SDD_ALM_CheckFxoState(pLLChannel, &pCh, &pDev);
   if(ret != SDD_statusOk)
      RETURN_STATUS(ret);

   *enable = (pCh->pALM->fxo_flags >> bit_pos & 1) ? IFX_ENABLE : IFX_DISABLE;

   RETURN_STATUS(ret);
}


/**
   FXO line mode set.

   \param  pLLChannel   Handle to TAPI low level channel structure.
   \param  pMode        Pointer to IFX_TAPI_FXO_LINE_MODE_t type

   \return
   - SDD_statusOk
   - SDD_statusParam
*/
static IFX_int32_t SDD_TAPI_LL_ALM_FxoLineModeSet(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_FXO_LINE_MODE_t const *pMode)
{
   SDD_CHANNEL_t  *pCh = IFX_NULL;
   SDD_DEVICE_t   *pDev = IFX_NULL;
   SDD_Opmode_t   *pOpmod = IFX_NULL;
   IFX_uint8_t    curr_opmode;
   IFX_int32_t    ret = SDD_statusOk;

   ret = SDD_ALM_CheckFxoParams(pLLChannel, &pCh, &pDev);
   if(ret != SDD_statusOk)
      RETURN_STATUS(ret);

   pOpmod = &pCh->pALM->sdd_opmode;
   SDD_OS_MutexGet (&pCh->mtxChAcc);

   switch (pMode->mode)
   {
      case IFX_TAPI_FXO_LINE_MODE_DISABLED:
         {
            /* if already disabled - do nothing and exit */
            if (!(pCh->pALM->fxo_flags & (1U << FXO_LINE_MODE)))
               goto exit;

            pOpmod->OperatingMode = SDD_Opmode_Disabled;
            ret = SDD_ALM_OpModeSet (pCh);

            if (SDD_SUCCESS(ret))
            {
               /* update FXO status bit */
               pCh->pALM->fxo_flags &= ~(1U << FXO_LINE_MODE);
            }
         }
         break;

      case IFX_TAPI_FXO_LINE_MODE_ACTIVE:
         {
            /* if already active - do nothing and exit */
            if (pCh->pALM->fxo_flags & (1U << FXO_LINE_MODE))
               goto exit;

            /* Get the current opmode - wait while a opmode change is ongoing.*/
            ret = SDD_ALM_OpmodeGet (pCh, &curr_opmode);
            /* Only exit when interrupted by a signal. When waiting was aborted
               by the timeout assume that no linemode change is pending and try
               to continue. */
            if (ret == VMMC_statusSddEvtWaitInterrupt)
            {
               goto exit;
            }

            /* switching to linemode FXO requires to switch to PDH first;
               if current mode is not PDH - switch to PDH */
            if (curr_opmode != SDD_Opmode_Disabled)
            {
               pOpmod->OperatingMode = SDD_Opmode_Disabled;
               ret = SDD_ALM_OpModeSet (pCh);
               if (!SDD_SUCCESS(ret))
               {
                  goto exit;
               }
            }

            pOpmod->OperatingMode = SDD_Opmode_FXO;
            ret = SDD_ALM_OpModeSet (pCh);
            if (SDD_SUCCESS(ret))
            {
               /* update FXO status bit */
               pCh->pALM->fxo_flags |= (1U << FXO_LINE_MODE);
            }
         }
         break;

      default:
         {
            RETURN_STATUS (SDD_statusParam);
         }
   } /* switch */

exit:
   SDD_OS_MutexRelease (&pCh->mtxChAcc);
   RETURN_STATUS(ret);
}


/**
   Initalize the FXO specific variables in the channel ALM structure.

   \param  pCh          Pointer to the channel structure.
*/
IFX_void_t SDD_ALM_FXO_InitCh (SDD_CHANNEL_t *pCh)
{
   SDD_FxoConfig_t      *pFxoConfig     = IFX_NULL;
   SDD_FxoHookSwitch_t  *pFxoHookSwitch = IFX_NULL;

   /* initialize FXO bits - all 0, polarity 1 (normal polarity) */
   pCh->pALM->fxo_flags = 0;
   pCh->pALM->fxo_flags |= (1 << FXO_POLARITY);

   /* setup SDD_FxoConfig command */
   pFxoConfig = &pCh->pALM->sdd_fxoConfig;
   memset(pFxoConfig, 0, sizeof(*pFxoConfig));
   pFxoConfig->CMD    = CMD_SDD;
   pFxoConfig->CHAN   = pCh->nChannel - 1;
   pFxoConfig->MOD    = MOD_SDD;
   pFxoConfig->ECMD   = SDD_FxoConfig_ECMD;
   pFxoConfig->LENGTH = SDD_FxoConfig_LENGTH;

   /* setup SDD_FxoHookSwitch command */
   pFxoHookSwitch = &pCh->pALM->sdd_fxoHookSwitch;
   memset(pFxoHookSwitch, 0, sizeof(*pFxoHookSwitch));
   pFxoHookSwitch->CMD    = CMD_SDD;
   pFxoHookSwitch->CHAN   = pCh->nChannel - 1;
   pFxoHookSwitch->MOD    = MOD_SDD;
   pFxoHookSwitch->ECMD   = SDD_FxoHookSwitch_ECMD;
   pFxoHookSwitch->LENGTH = SDD_FxoHookSwitch_LEN;
}


/**
   Basic ALM FXO configuration

   Use this function where needed to set the base configuration
   of FXO function in the Analog Line Module.

   \param  pCh          Pointer to the SDD channel structure.

   \return
   - SDD_statusOk if successful
   - error code from SDD_CmdWrite()
*/
IFX_int32_t SDD_ALM_FXO_BaseConf (SDD_CHANNEL_t *pCh)
{
   SDD_DEVICE_t         *pDev = IFX_NULL;
   SDD_Opmode_t         *pOpmod = IFX_NULL;
   IFX_int32_t          ret = SDD_statusOk;

   /* Called internally. Caller must ensure correct channel parameter. */
   SDD_ASSERT(pCh != IFX_NULL);
   SDD_ASSERT(pCh->pALM != IFX_NULL);
   SDD_ASSERT(pCh->pDev != IFX_NULL);

   pDev = pCh->pDev;
   pOpmod = &pCh->pALM->sdd_opmode;

   /* Write FXO config to allow ringing signals with 50 Hz. Ring pulses
      shorter than DEB1 [ms] will be ignored. To allow 50Hz ringing the
      value must be below 20ms. */
   pCh->pALM->sdd_fxoConfig.TACT   = 10;  /* [ms] */
   pCh->pALM->sdd_fxoConfig.DEB2   = 70;  /* [ms] */
   pCh->pALM->sdd_fxoConfig.OSIT   = 100; /* [ms] */
   pCh->pALM->sdd_fxoConfig.FLASHT = 50;  /* [ms] */
   pCh->pALM->sdd_fxoConfig.DEB1   = 16;  /* [ms] */

   ret = SDD_CmdWrite(pDev, (IFX_uint32_t*)&pCh->pALM->sdd_fxoConfig);

   /* activate FXO channel in firmware */
   if (SDD_SUCCESS(ret) &&
       (pCh->pALM->fxo_flags & (1U << FXO_LINE_MODE)))
   {
      pOpmod->OperatingMode = SDD_Opmode_Disabled;
      ret = SDD_ALM_OpModeSet (pCh);
      if (SDD_SUCCESS(ret))
      {
         pOpmod->OperatingMode = SDD_Opmode_FXO;
         ret = SDD_ALM_OpModeSet (pCh);
      }
   }

   RETURN_STATUS(ret);
}


/**
   Function called by init_module of device, fills up ALM FXO function pointers
   which are passed to HL TAPI during registration

   \param  pAlm         Pointer to ALM module.
*/
IFX_void_t SDD_ALM_FXO_Func_Register (IFX_TAPI_DRV_CTX_ALM_t *pAlm)
{
   pAlm->FXO_HookSet       = SDD_TAPI_LL_ALM_FxoHookSet;
   pAlm->FXO_HookGet       = SDD_TAPI_LL_ALM_FxoHookGet;
   pAlm->FXO_FlashSet      = SDD_TAPI_LL_ALM_FxoFlashSet;
   pAlm->FXO_FlashCfg      = SDD_TAPI_LL_ALM_FxoFlashCfg;
   pAlm->FXO_OsiCfg        = SDD_TAPI_LL_ALM_FxoOsiCfg;
   pAlm->FXO_StatGet       = SDD_TAPI_LL_ALM_FxoStatGet;
   pAlm->FXO_LineModeSet   = SDD_TAPI_LL_ALM_FxoLineModeSet;
}

#endif /* SDD_FEAT_FXO */
