#ifndef _drv_sdd_global_H
#define _drv_sdd_global_H
/*******************************************************************************

                            Copyright (c) 2013-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_api.h
   This file contains the declaration of the connection module.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "drv_sdd_config.h"

#include "drv_tapi_ll_interface.h"
#include "drv_vmmc_errno.h"
#include "drv_sdd_io.h"
#include "drv_sdd_osmap.h"
#include "drv_sdd_int_evt.h"
#include "drv_sdd_fw_interface.h"
#include "drv_sdd_fw_cmd.h"

#ifdef EVENT_LOGGER_DEBUG
#include "el_log_macros.h"
#endif /* EVENT_LOGGER_DEBUG */

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */
#define SDD_DEV_NAME "sdd"

/** Number of devices that can be handled by this driver. */
#ifndef SDD_MAX_DEVICES
#define SDD_MAX_DEVICES               1
#endif /* SDD_MAX_DEVICES */

/** Number of channels per device. */
#ifndef SDD_MAX_CH_NR
#define SDD_MAX_CH_NR                 3
#endif /* SDD_MAX_CH_NR */


/* Define trace output
   OFF:    no output
   HIGH:   only important traces, as errors or some warnings
   NORMAL: including traces from high and general proceedings and
           possible problems
   LOW:    all traces and low level traces as basic chip access
           and interrupts, command data
   Traces can be completely switched off with the compiler switch
   ENABLE_TRACE to 0
*/
#define DBG_LEVEL_OFF      4
#define DBG_LEVEL_HIGH     3
#define DBG_LEVEL_NORMAL   2
#define DBG_LEVEL_LOW      1

#ifdef ENABLE_TRACE
   #define CREATE_TRACE_GROUP(name) unsigned int G_nTraceGroup##name = DBG_LEVEL_HIGH
   #define DECLARE_TRACE_GROUP(name) extern unsigned int G_nTraceGroup##name
   #define TRACE_LEVEL_SET(name, level) G_nTraceGroup##name = (level)
   #define TRACE_LEVEL_GET(name, level) (level) = G_nTraceGroup##name
   #define TRACE(name,level,message) do {if(level >= G_nTraceGroup##name) \
         { TRACE_PRINTF message ; } } while(0)
#else
   #define CREATE_TRACE_GROUP(name)
   #define DECLARE_TRACE_GROUP(name)
   #define TRACE_LEVEL_SET(name, level)
   #define TRACE_LEVEL_GET(name, level)
   #define TRACE(name,level,message)
#endif /* ENABLE_TRACE */


/** assert in debug code
\param expr - expression to be evaluated. If expr != TRUE assert is printed
              out with line number */
#ifdef DEBUG
#define SDD_ASSERT(expr) \
   do { \
      /*lint -e(506, 774) */ \
      if(!(expr)) { \
         (void) ASSERT_PRINTF ("\n\r" __FILE__ ":%d: Assertion %s " \
                               "failed!\n\r", __LINE__, #expr); \
      } \
   } while (0)
#else /* DEBUG */
#define SDD_ASSERT(expr)
#endif /* DEBUG */


/** set module field for tapi event */
#define SDD_TAPI_EVENT_MODULE_SET(m_tapiEvent, m_nModule)                  \
   do{                                                                     \
       m_tapiEvent.module = m_nModule;                                     \
   }while(0)

#define RETURN_STATUS(code)                                                \
   do{                                                                     \
      if (SDD_SUCCESS(code) == IFX_FALSE)                                  \
         SDD_ErrorEvent(pCh, code, __LINE__, __FILE__);                    \
      return code;                                                         \
   }while(0)

#define RETURN_DEVSTATUS(code)                                             \
   do{                                                                     \
      if (((IFX_int32_t)code > (IFX_int32_t)TAPI_statusClassWarn) ||       \
           ((IFX_int32_t)code == (IFX_int32_t)IFX_ERROR))                  \
         SDD_ErrorEvent (pDev, code, __LINE__, __FILE__);                  \
      return code;                                                         \
   }while(0)

/* device states */
/* firmware successfully downloaded */
#define   DS_FW_DLD                    0x000001
/* device is up (see system event: UP) */
#define   DS_DEV_UP                    0x000002
/* Basic Init done */
#define   DS_BASIC_INIT                0x000004
/* Tapi Init done */
#define   DS_TAPI_INIT                 0x000008
/** device successfully initialized */
#define   DS_DEV_INIT                  0x000010
/** PCM module enabled  */
#define   DS_PCM_EN                    0x000040
/** SPI interface is active */
#define   DS_SPI_ACTIVE                0x000080
/** SPI interface setup phase */
#define   DS_SPI_SETUP                 0x000100


#define SDD_DEV_TYPE             IFX_TAPI_DEV_TYPE_SDD

#define MAX_CMD_WORD 32


#ifndef SDD_SUCCESS
   #define SDD_SUCCESS TAPI_SUCCESS
#endif /* SDD_SUCCESS */

/* ========================================================================== */
/*                             Type definitions                               */
/* ========================================================================== */
typedef struct _SDD_CHANNEL_t SDD_CHANNEL_t;
typedef struct _SDD_DEVICE_t SDD_DEVICE_t;

struct SDD_PCM_DEV;

/** Type of DC/DC converter */
enum SDD_DCDC_TYPE
{
   /** Default SLIC100: Dedicated Inverting Buck-Boost DC/DC Converter. */
   SDD_DCDC_TYPE_DEFAULT_IBB = 0,
   /** Default SLIC200: No converter set. */
   SDD_DCDC_TYPE_NOTSET = 0,
   /** Dedicated Inverting Buck-Boost DC/DC Converter. */
   SDD_DCDC_TYPE_IBB = 1,
   /** Combined Inverting Buck-Boost DC/DC Converter. */
   SDD_DCDC_TYPE_CIBB = 2,
   /** Dedicated Inverting Boost DC/DC Converter. */
   SDD_DCDC_TYPE_IB = 3,
   /** Combined Inverting Boost DC/DC Converter. */
   SDD_DCDC_TYPE_CIB = 4,
   /* Dedicated Buck-or-Boost DC/DC Converter. */
   SDD_DCDC_TYPE_BB = 5,
   /* Combined Buck-or-Boost DC/DC Converter. */
   SDD_DCDC_TYPE_CBB = 6,
   /* Dedicated Inverting Flyback DC/DC Converter. */
   SDD_DCDC_TYPE_IFB = 7,
   /* Combined Inverting Flyback DC/DC Converter. */
   SDD_DCDC_TYPE_CIFB = 8,
   /* Dedicated Inverting Boost Converter with Gate Driver. */
   SDD_DCDC_TYPE_IBGD = 9,
   /* Combined Inverting Boost Converter with Gate Driver. */
   SDD_DCDC_TYPE_CIBGD = 10,
   /* Dedicated Inverting Boost Converter with Voltage Doubler. */
   SDD_DCDC_TYPE_IBVD = 11,
   /* Combined Inverting Boost Converter with Voltage Doubler. */
   SDD_DCDC_TYPE_CIBVD = 12,
   /* Magic value for lookup operations. */
   SDD_DCDC_TYPE_NOTFOUND = 128
};

/* capabilities structure for internal usage */
typedef struct
{
   /** Number of PCM Channels */
   IFX_uint8_t    nPCM;
   /** Number of Analog Line Channels */
   IFX_uint8_t    nALI;
   /** Number of DTMF receivers */
   IFX_uint8_t    nDTMFR;
   /** Number of FXS Channels */
   IFX_uint8_t    nFXS;
   /** Number of FXO Channels */
   IFX_uint8_t    nFXO;
   /** Flags indicating the type (FXS/FXO) of the first two channels. */
   IFX_boolean_t  FXO0;
   IFX_boolean_t  FXO1;

   /* fw capabilities */
   IFX_boolean_t  bfw_PCM;             /* PCM commands are supported */
   IFX_boolean_t  bfw_BuckBoost;       /* buck/boost DC/DC operation supported*/
   IFX_boolean_t  bfw_WbTsSplit;       /* split wideband timeslots supported */
   IFX_boolean_t  bfw_combDcDc;        /* combined DC/DC operation supported */
   /* Extended line testing support. Note that GR909 is always supported. */
   IFX_boolean_t  bfw_ExtLT;
   /* Capacitance to ground measurement support.
      Note that the capacitance measurement tip2ring is always supported. */
   IFX_boolean_t  bfw_CMeas;
   IFX_boolean_t  bfw_AsDspDwld;       /* ASDSP download is supported */
   IFX_boolean_t  bfw_DTMFR;           /* DTMFR is available */
   IFX_boolean_t  bfw_MWI;             /* MWI is available */
   IFX_boolean_t  bfw_ASPwrSave;       /* FW internal PwrSave handling */
   IFX_boolean_t  b_DrvPwrSaveEn;      /* DRV internal PwrSave handling */

   IFX_boolean_t  bfw_GR909;           /* GR909 is supported by DUSLIC XS and DUSLIC XC */
} SDD_Capabilities_t;


/* Channel structure */
struct _SDD_CHANNEL_t
{
   /* common fields with SDD_CHANNEL_t structure */
   /* Actual  SDD channel, starting with 1, a channel number 0 indicates
      the control device structure SDD_DEVICE */
   IFX_uint8_t                nChannel;

   /* ptr to actual device */
   SDD_DEVICE_t               *pDev;

   /* overall channel protection ( read/write/ioctl level) */
   IFXOS_mutex_t              mtxChAcc;

   /* ALM Channel management structure */
   struct SDD_ALM             *pALM;
   /* PCM Channel management structure */
   struct SDD_PCM             *pPCM;

   TAPI_CHANNEL               *pTapiCh;
};


/* Device Stucture */
struct _SDD_DEVICE_t
{
   /**
      SDD channel. This value is always zero for the device structure.
      Different channel device structure are from type SDD_CHANNEL_t.
      The first field is also a nChannel. This gurantees that the channel
      field is always available, although the cast was wrong. But ensure
      the correct cast after reading the channel field.
   */
   IFX_uint8_t                nChannel;

   /* tracking in use of the device */
   IFX_uint16_t               nInUse;

   /* device error status, never reset. read out for detail
      information if return value of function is not SDD_statusOk */
   IFX_int32_t                nErr;

   /* absolute index of the device, when multiple devices are existing
      For the first device the value is set to zero                        */
   IFX_int32_t                nDevNr;

   /* Status of device, see states defines .
      Must be protected against interrupts and concurrents tasks           */
   IFX_vuint32_t              nDevState;

   /* Device Number */
   IFX_uint32_t nDeviceNr;
   /* Calendar Week */
   IFX_uint8_t nCalendarWeek;
   /* Production Year */
   IFX_uint8_t nYear;

   /* VCODEC FW version */
   IFX_uint8_t                nFwVersion[4];
   IFX_uint32_t               nFwBuildTimestamp;
   /* SmartSLIC version */
   IFX_uint8_t                nSlicDevId;
   IFX_uint8_t                nSlicRevision;
   /* Firmware (DCCtrl) version */
   IFX_uint8_t                nDcDcCtrlVers;
   IFX_uint8_t                nDcDcCtrlStep;
   IFX_uint8_t                nDcDcCtrlHotFix;
   /* ASDSP version */
   IFX_uint16_t               nASDSPVers;
   IFX_uint8_t                nASDSPStep;

   /* DC/DC converter type indicated on the GPIOs of the SLIC. */
   enum SDD_DCDC_TYPE         nAllowedDcDcType;
   /* Indicates combined DC/DC (true) or dedicated DC/DC (false, default) */
   IFX_boolean_t              bDcDcHwCombined;

   /** Capability list for the device  -  Internal data for driver */
   SDD_Capabilities_t         caps;
   IFX_boolean_t              bCapsRead;

   /* PCM interface (device global and not channel specific) */
   struct SDD_PCM_DEV         *pPCM;

   /* Channel specific structures */
   SDD_CHANNEL_t              pChannel[SDD_MAX_CH_NR];

   /* SDD MBX concurent access protection mutex.
      PS: Avoid nested locking of this mutex. It can lead to a deadlock */
   IFXOS_mutex_t              mtxMbxAcc;

   /* cached command inbox length */
   IFX_uint8_t                nMbxCachedCbiLen;

   /* SDD share variables concurent access protection mutex.
      PS: Avoid nested locking of this mutex. It can lead to a deadlock */
   IFXOS_mutex_t              mtxMemberAcc;
#if 0

   /* error flag for cmd read - avoid all kind of compiler optimizations */
   volatile IFX_boolean_t     bCmdReadError;
#endif

#if 0
   /* selection of polling or interrupt event handling mode */
   volatile enum DXT_EventMode   nEventHandlingMode;

   /* variables for shared interrupt handling */
   DXT_IRQ_t                  *pIrq;
#endif
   SDD_DEVICE_t               *pInt_NextDev;

   /* entry to Tapi device, needs protection  */
   TAPI_DEV                   *pTapiDev;

   /* firmware driver API */
   SDD_FIRMWARE_IF_t          *pMbxIF;

   /* Flag indicating an SSI crash */
   IFX_boolean_t              bSSIcrash;
   /* Debug counter for the number of successful SLIC recoveries */
   IFX_uint32_t               nSlicRecoveryCnt;
   /* Flag indicating that SDD device node was opened and all events will be redirected
    * to this interface */
   IFX_boolean_t              bIntercept;
   /* Wakeuplist for the select mechanism */
   IFXOS_drvSelectQueue_t     sdd_wakeup;
   /* Wake up flag */
   IFX_boolean_t              bWakeUp;
   /* Firmware intercepted event */
   FW_Event_t                 evtStore;
   /* Mutex */
   IFXOS_mutex_t              mtxIntercept;
   /* Event to wait read ioctl */
   IFXOS_event_t              readEvent;

#ifdef S220
   /* System Control FW messages */
   struct SYS_CTRL            fwm_sys_ctrl;
#endif /* S220 */
};


/* VMMC Channel structure
   This is needed when SDD is part of VMMC. Then the functions exported in
   the device context will get a pointer to the VMMC channel struct and
   have to follow a pointer to find the SDD channel struct. */
struct VMMC_CHANNEL
{
   /* common fields with VMMC_DEVICE structure */
#ifdef DEBUG
   IFX_uint32_t               magic;
#endif /* DEBUG */
   /* Actual  vmmc channel, starting with 1, a channel number 0 indicates
      the control device structure VMMC_DEVICE */
   IFX_uint8_t                nChannel;
   /* tracking in use of the device */
   IFX_uint16_t               nInUse;

   /* Link to SDD component channel management structure.
      While in the VMMC channel struct this is a generic link here we know
      better and use the SDD local declaration of the SDD channel. */
   SDD_CHANNEL_t              *pSDD;
};

/* VMMC global Device Stucture
   This is needed when SDD is part of VMMC. Then the functions exported in
   the device context will get a pointer to the VMMC device struct and
   have to follow a pointer to find the SDD device struct. */
struct VMMC_DEVICE
{
#ifdef DEBUG
   IFX_uint32_t               magic;
#endif /* DEBUG */
   /* Vmmc channel. This value is always zero for the device structure.
      Different channel device structure are from type VMMC_CHANNEL.
      The first field is also a nChannel. This gurantees that the channel
      field is always available, although the cast was wrong. But ensure
      the correct cast after reading the channel field */
   IFX_uint8_t                nChannel;
   /* tracking in use of the device */
   IFX_uint16_t               nInUse;

   /* Link to SDD component device management structure
      While in the VMMC channel struct this is a generic link here we know
      better and use the SDD local declaration of the SDD device. */
   SDD_DEVICE_t               *pSDD;
};

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */
extern IFX_void_t SDD_ErrorEvent (
                        const IFX_TAPI_LL_CH_t *pChOrDev,
                        const IFX_uint16_t err,
                        const IFX_uint32_t nLine,
                        const IFX_char_t* sFile);

extern IFX_int32_t SDD_CmdWrite(
                        SDD_DEVICE_t *,
                        IFX_uint32_t *);

extern IFX_int32_t SDD_CmdRead (
                        SDD_DEVICE_t *,
                        IFX_uint32_t *,
                        IFX_uint32_t *);

extern IFX_void_t cpb2w (
                        IFX_uint16_t* pWbuf,
                        const IFX_uint8_t const *pBbuf,
                        IFX_uint32_t nB);

extern IFX_int32_t SDD_WaitForSddOpmodeChEvt(
                        SDD_CHANNEL_t *pCh);

/* ========================================================================== */
/*                          Global variable declaration                       */
/* ========================================================================== */
DECLARE_TRACE_GROUP(SDD);

#endif /* _drv_sdd_global_H */
