/******************************************************************************

                            Copyright (c) 2013-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

/**
   \file drv_sdd_gr909.c
   This file implements the GR909 linetesting functions.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "drv_sdd_config.h"
#ifdef SDD_FEAT_GR909

#include "ifx_types.h"

#include "drv_sdd_osmap.h"
#include "drv_sdd_global.h"
#include "drv_sdd_alm_priv.h"
#include "drv_sdd_init.h"
#include "drv_sdd_opmode.h"
#include "drv_sdd_gr909.h"

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */

/* ========================================================================== */
/*                             Type definitions                               */
/* ========================================================================== */

/* ========================================================================== */
/*                             Global variables                               */
/* ========================================================================== */

/* ========================================================================== */
/*                              Local variables                               */
/* ========================================================================== */

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */
/**
   Start selected subset (or all) GR909 tests

   \param  pLLChannel   Pointer to TAPI low level channel structure.
   \param  p_start      Pointer to IFX_TAPI_GR909_START_t structure.

   \return
   IFX_SUCCESS or IFX_ERROR

   \remarks
   GR909 tests can be started only on disabled lines.
*/
static IFX_int32_t SDD_TAPI_LL_ALM_LT_GR909_Start (
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_GR909_START_t const *p_start)
{
   SDD_CHANNEL_t           *pCh;
   SDD_DEVICE_t            *pDev;
   SDD_Opmode_t            *pOpmod;
   IFX_int32_t             ret = IFX_SUCCESS;
   SDD_CHANNEL_t           *pOtherCh = IFX_NULL;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   pDev = pCh->pDev;
   pOpmod = &pCh->pALM->sdd_opmode;

   /* Nothing to test ? */
   if (p_start->test_mask == 0)
   {
      /* errmsg: At least one parameter in function is wrong. */
      RETURN_STATUS(SDD_statusFuncParam);
   }

#ifdef S220
   /* Until the DC/DC converter type is set this operation is not allowed. */
   if (pCh->pALM->nDcDcType == SDD_DCDC_TYPE_NOTSET)
#else /* S220 */
   /* For DC/DC converters other than default a BBD download is required. */
   if ((pCh->pALM->nDcDcType == SDD_DCDC_TYPE_DEFAULT_IBB) &&
       (pCh->pDev->nAllowedDcDcType != SDD_DCDC_TYPE_IBB))
#endif /* S220 */
   {
      /* errmsg: Operation blocked until a BBD with DC/DC configuration
                 matching the connected DC/DC hardware type is downloaded. */
      RETURN_STATUS(SDD_statusBlockedDcDcTypeMissmatch);
   }

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);

   /* check line is in state disabled */
   if (pOpmod->OperatingMode != SDD_Opmode_Disabled)
   {
      /* release channel */
      SDD_OS_MutexRelease (&pCh->mtxChAcc);
      /** errmsg: GR909 might be started only on disabled lines */
      RETURN_STATUS(SDD_statusGR909LineNotDisabled);
   }

   /* For combined DC/DC the other channel must be disabled when starting. */
   if (pCh->pDev->bDcDcHwCombined == IFX_TRUE)
   {
      /* The combined DC/DC flag is only set on 2 channel systems. So there
         is always a neighbouring channel. */
      pOtherCh = SDD_GetNeighbourChannel(pCh);
      SDD_ASSERT(pOtherCh != IFX_NULL);

      /* protect other channel from mutual access */
      SDD_OS_MutexGet (&pOtherCh->mtxChAcc);

      if (pOtherCh->pALM->sdd_opmode.OperatingMode != SDD_Opmode_Disabled)
      {
         /* release protection */
         SDD_OS_MutexRelease (&pOtherCh->mtxChAcc);
         SDD_OS_MutexRelease (&pCh->mtxChAcc);
         /* errmsg: Neighbour line mode is not DISABLED */
         RETURN_STATUS(SDD_statusGR909NeighbourLineNotDisabled);
      }
   }

#ifdef S220
   /* For SLIC200 the FW default limit values are used. */
#else /* S220 */
   /* SLIC100 */
   /* Add selection of testset. The limit values are already set in this
      message. */
   {
      SDD_GR909Config_t       *p_ctrl;

      p_ctrl = &pCh->pALM->sdd_gr909_config;
      p_ctrl->Pdf     = 0;
      p_ctrl->Hpt     = ((p_start->test_mask & IFX_TAPI_GR909_HPT) ? 1 : 0);
      p_ctrl->Femf    = ((p_start->test_mask & IFX_TAPI_GR909_FEMF)? 1 : 0);
      p_ctrl->Rft     = ((p_start->test_mask & IFX_TAPI_GR909_RFT) ? 1 : 0);
      p_ctrl->Roh     = ((p_start->test_mask & IFX_TAPI_GR909_ROH) ? 1 : 0);
      p_ctrl->Rit     = ((p_start->test_mask & IFX_TAPI_GR909_RIT) ? 1 : 0);

      /* If one of these measurements is started, both bits should be set
         because the DC firmware can't handle the ROH measurement separately. */
      if ((p_ctrl->Rft == 1) || (p_ctrl->Roh == 1))
      {
         p_ctrl->Rft = 1;
         p_ctrl->Roh = 1;
      }

      ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)p_ctrl);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing to command mailbox failed. */
         ret = SDD_statusCmdMbWrErr;
         goto error;
      }
   }
#endif /* S220 */

   /* start GR909 measurement */
   pOpmod->OperatingMode = SDD_Opmode_GR909;
   ret = SDD_ALM_OpModeSet (pCh);
   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Writing to command mailbox failed. */
      ret = SDD_statusCmdMbWrErr;
      goto error;
   }

error:
   /* In combined DC/DC mode unlock the other channel. */
   if (pOtherCh != IFX_NULL)
   {
      /* release protection */
      SDD_OS_MutexRelease (&pOtherCh->mtxChAcc);
   }
   /* release channel */
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   RETURN_STATUS(ret);
}

/**
   Stop GR909 tests

   \param  pLLChannel   Pointer to TAPI low level channel structure.

   \return
   IFX_SUCCESS or IFX_ERROR
*/
static IFX_int32_t SDD_TAPI_LL_ALM_LT_GR909_Stop (
                        IFX_TAPI_LL_CH_t *pLLChannel)
{
   SDD_CHANNEL_t           *pCh;
   SDD_Opmode_t            *pOpmod;
   IFX_int32_t             ret = IFX_SUCCESS;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   pOpmod = &pCh->pALM->sdd_opmode;

   /* \todo check if GR909 measurement is running */

#ifdef S220
   /* Until the DC/DC converter type is set this operation is not allowed. */
   if (pCh->pALM->nDcDcType == SDD_DCDC_TYPE_NOTSET)
#else /* S220 */
   /* For DC/DC converters other than default a BBD download is required. */
   if ((pCh->pALM->nDcDcType == SDD_DCDC_TYPE_DEFAULT_IBB) &&
       (pCh->pDev->nAllowedDcDcType != SDD_DCDC_TYPE_IBB))
#endif /* S220 */
   {
      /* errmsg: Operation blocked until a BBD with DC/DC configuration
                 matching the connected DC/DC hardware type is downloaded. */
      RETURN_STATUS(SDD_statusBlockedDcDcTypeMissmatch);
   }

   /* setup GR909 mode now */
   pOpmod->OperatingMode = SDD_Opmode_Disabled;
   ret = SDD_ALM_OpModeSet (pCh);
   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Writing to command mailbox failed. */
      RETURN_STATUS(SDD_statusCmdMbWrErr);
   }
   SDD_ALM_CorrectLinemodeCache(pCh, SDD_Opmode_Disabled);
   return ret;
}

/**
   Read GR909 results

   \param  pLLChannel   Pointer to TAPI low level channel structure.
   \param  pResults     Pointer to result structure.

   \return
   IFX_SUCCESS or IFX_ERROR

   \remarks
   GR909 results are polled by the application after reception of the
   IFX_TAPI_GR909_RDY event
*/
static IFX_int32_t SDD_TAPI_LL_ALM_LT_GR909_Result (
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_GR909_RESULT_t *pResults)
{
   SDD_CHANNEL_t       *pCh;
   IFX_int32_t         ret = IFX_SUCCESS;
   IFX_uint16_t        sdd_res [(SDD_GR909ResultsRead_MAX_LENGTH>>1)+
                                SDD_CMD_HEADER_CNT] = {0};
   IFX_uint32_t        offset  = 0x04; /* index for first 16bit result */
   SDD_GR909ResultsRead_t *pSddRes = (SDD_GR909ResultsRead_t*)sdd_res;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   /* request pass / fail and validity information */
   pSddRes->CMD    = CMD_SDD;
   pSddRes->ECMD   = SDD_GR909ResultsRead_ECMD;
   pSddRes->MOD    = MOD_SDD;

   pSddRes->CHAN   = pCh->nChannel-1;
   pSddRes->LENGTH = SDD_GR909ResultsRead_MAX_LENGTH;

   ret = SDD_CmdRead (pCh->pDev, (IFX_uint32_t *)pSddRes,
                                    (IFX_uint32_t *)pSddRes);
   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Device read access failed. */
      RETURN_STATUS(SDD_statusReadErr);
   }

#ifndef S220
   if (pSddRes->Busy)
   {
      /* errmsg: GR909 measurement ongoing, cannot read results */
      RETURN_STATUS(SDD_statusGR909Busy);
   }
#endif /* S220 */

   /* set validity information */
   pResults->valid  = pSddRes->HptValid  ? IFX_TAPI_GR909_HPT_VALID : 0;
   pResults->valid |= pSddRes->FemfValid ? IFX_TAPI_GR909_FEMF_VALID : 0;
   pResults->valid |= pSddRes->RftValid  ? IFX_TAPI_GR909_RFT_VALID : 0;
   pResults->valid |= pSddRes->RohValid  ? IFX_TAPI_GR909_ROH_VALID : 0;
   pResults->valid |= pSddRes->RitValid  ? IFX_TAPI_GR909_RIT_VALID : 0;

   /* set pass/fail information */
   pResults->passed  = pSddRes->HptPass  ? IFX_TAPI_GR909_HPT : 0;
   pResults->passed |= pSddRes->FemfPass ? IFX_TAPI_GR909_FEMF : 0;
   pResults->passed |= pSddRes->RftPass  ? IFX_TAPI_GR909_RFT : 0;
   pResults->passed |= pSddRes->RohPass  ? IFX_TAPI_GR909_ROH : 0;
   pResults->passed |= pSddRes->RitPass  ? IFX_TAPI_GR909_RIT : 0;

   /* extract results */

   if (pSddRes->HptValid  || pSddRes->FemfValid)
   {  /* copy results to to HPT and FEMF (not measured separately) */
      pResults->HPT_AC_R2G = pResults->FEMF_AC_R2G = sdd_res[offset + 0];
      pResults->HPT_AC_T2G = pResults->FEMF_AC_T2G = sdd_res[offset + 1];
      pResults->HPT_AC_T2R = pResults->FEMF_AC_T2R = sdd_res[offset + 2];
      pResults->HPT_DC_R2G = pResults->FEMF_DC_R2G = sdd_res[offset + 3];
      pResults->HPT_DC_T2G = pResults->FEMF_DC_T2G = sdd_res[offset + 4];
      pResults->HPT_DC_T2R = pResults->FEMF_DC_T2R = sdd_res[offset + 5];
   }
   if (pSddRes->RftValid)
   {
      pResults->RFT_R2G = sdd_res[offset + 6];
      pResults->RFT_T2G = sdd_res[offset + 7];
      pResults->RFT_T2R = sdd_res[offset + 8];
   }
   if (pSddRes->RohValid)
   {
      pResults->ROH_T2R_L = sdd_res[offset + 9];
      pResults->ROH_T2R_H = sdd_res[offset + 10];
   }
   if (pSddRes->RitValid)
   {
      pResults->RIT_RES = sdd_res[offset + 11];
   }

#ifdef S220
   pResults->dev_type = IFX_TAPI_GR909_DEV_S220;
#else /* S220 */
   pResults->dev_type = IFX_TAPI_GR909_DEV_VMMC;
#endif /* S220 */

   pResults->OLR_T2R = pCh->pALM->nlt_ResistanceConfig.fOlResTip2Ring;
   pResults->OLR_T2G = pCh->pALM->nlt_ResistanceConfig.fOlResTip2Gnd;
   pResults->OLR_R2G = pCh->pALM->nlt_ResistanceConfig.fOlResRing2Gnd;

   return(SDD_statusOk);
}

/**
   This function is used to configure the measurement path for line testing
   according to Rmes resistor.
   Used by the IFX_TAPI_NLT_CONFIGURATION_RMES_SET command.

   \param  pLLChannel   Pointer to the TAPI LL channel structure.
   \param  pConfig      Pointer to the IFX_TAPI_NLT_CONFIGURATION_RMES_t

   \return
   - DXT_statusOk or status error
 */
static IFX_int32_t SDD_TAPI_LL_ALM_NLT_RmesConfig_Set(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_NLT_CONFIGURATION_RMES_t *pConfig)
{
   SDD_CHANNEL_t                      *pCh;
   IFX_int32_t                        ret = SDD_statusOk;

   /* calling function should ensure valid parameters */
   SDD_ASSERT(pLLChannel != IFX_NULL);
   SDD_ASSERT(pConfig != IFX_NULL);

   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);

#ifndef S220
   /* Set limit values according to Rmeas resistor. */
   ret = SDD_ALM_GR909_SetLimits(pCh, pConfig->nRmeas);
   if (TAPI_SUCCESS(ret))
#endif /* !S220 */
   {
      pCh->pALM->nRmeas = pConfig->nRmeas;
   }

   /* release channel */
   SDD_OS_MutexRelease (&pCh->mtxChAcc);
   RETURN_STATUS(ret);
}

#ifndef S220
/**
   Set default limits.

   \param  pCh          SDD channel context.
   \param  nRmeas       Rmeas variant connected to the chip.

   \return
   SDD_statusOk or status error

   \note
   For SLIC200 the FW default limit values are used.
*/
IFX_int32_t  SDD_ALM_GR909_SetLimits (SDD_CHANNEL_t *pCh,
                                      IFX_TAPI_NLT_RMEAS_CFG_t nRmeas)
{
   IFX_int32_t            ret = SDD_statusOk;
   SDD_GR909Config_t  *p_ctrl;

   SDD_ASSERT(pCh != IFX_NULL);

   p_ctrl = &pCh->pALM->sdd_gr909_config;

   switch (nRmeas)
   {
      case IFX_TAPI_NLT_RMEAS_1MOHM:
         p_ctrl->HptW2gAcLim  = 0x24EE;
         p_ctrl->HptW2wAcLim  = 0x24EE;
         p_ctrl->HptW2gDcLim  = 0x4738;
         p_ctrl->HptW2wDcLim  = 0x4738;
         p_ctrl->FemfW2gAcLim = 0x0762;
         p_ctrl->FemfW2wAcLim = 0x0762;
         p_ctrl->FemfW2gDcLim = 0x032A;
         p_ctrl->FemfW2wDcLim = 0x032A;
         p_ctrl->RftResLim    = 0x23F2;
         p_ctrl->RohLinLim    = 0x000F;
         p_ctrl->RitLowLim    = 0x0056;
         p_ctrl->RitHighLim   = 0x0996;
         break;

      case IFX_TAPI_NLT_RMEAS_1_5MOHM:
      case IFX_TAPI_NLT_RMEAS_DEFAULT:
         /* 50Vrms */
         p_ctrl->HptW2gAcLim  = 0x18E3;
         /* 50Vrms */
         p_ctrl->HptW2wAcLim  = 0x18E3;
         /* 135V */
         p_ctrl->HptW2gDcLim  = 0x2FC7;
         /* 135V */
         p_ctrl->HptW2wDcLim  = 0x2F83;
         /* 10Vrms */
         p_ctrl->FemfW2gAcLim = 0x04FA;
         /* 10Vrms */
         p_ctrl->FemfW2wAcLim = 0x04FA;
         /* 6V */
         p_ctrl->FemfW2gDcLim = 0x0260;
         /* 6V */
         p_ctrl->FemfW2wDcLim = 0x021D;
         /* 136.36kOhm corresponds to 150kOhm limit to ground and
            140kOhm limit tip to ring  */
         p_ctrl->RftResLim    = 0x20AE;
         /* 15% */
         p_ctrl->RohLinLim    = 0x000F;
         /* 5REN (6930Ohm + 8uF at 20Hz) */
         p_ctrl->RitLowLim    = 0x0056;
         /* 0.175REN (6930Ohm + 8uF at 20Hz) */
         p_ctrl->RitHighLim   = 0x0996;
         break;

      default:
         /** errmsg: Invalid Rmes value */
         ret = SDD_statusInvalidRmes;
         break;
   }

   return ret;
}
#endif /* !S220 */

/**
   Function called by init_module of device, fills up ALM function pointers
   which are passed to HL TAPI during registration

   \param pAlm       pointer to ALM module

   \return
   None.
*/
IFX_void_t SDD_ALM_GR909_Func_Register (IFX_TAPI_DRV_CTX_ALM_t *pAlm)
{
   pAlm->GR909_Start             = SDD_TAPI_LL_ALM_LT_GR909_Start;
   pAlm->GR909_Stop              = SDD_TAPI_LL_ALM_LT_GR909_Stop;
   pAlm->GR909_Result            = SDD_TAPI_LL_ALM_LT_GR909_Result;
   pAlm->NLT_RmesConfigSet       = SDD_TAPI_LL_ALM_NLT_RmesConfig_Set;
}

#endif /* SDD_FEAT_GR909 */
