/*******************************************************************************

  Copyright 2013-2015 Lantiq Deutschland GmbH
  Copyright 2015-2016 Lantiq Beteiligungs-GmbH & Co.KG
  Copyright 2016-2017 Intel Corporation.

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_init.c
   This file implements the initialisation sequence.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "ifx_types.h"

#include "drv_sdd_global.h"
#include "drv_sdd_api.h"
#include "drv_sdd_osmap.h"
#include "drv_sdd_access.h"
#include "drv_sdd_version.h"
#include "drv_sdd_fw_cmd.h"
#include "drv_sdd_opmode.h"
#include "drv_sdd_calib.h"
#include "drv_sdd_pcm.h"
#include "drv_sdd_init.h"
#ifdef EVENT_LOGGER_DEBUG
#include "el_log_macros.h"
#endif /* EVENT_LOGGER_DEBUG */

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */
#ifndef _MKSTR_1
#define _MKSTR_1(x)    #x
#define _MKSTR(x)      _MKSTR_1(x)
#endif

/** device driver version string */
#define DRV_SDD_VER_STR          _MKSTR(MAJORSTEP)    "."   \
                                 _MKSTR(MINORSTEP)    "."   \
                                 _MKSTR(VERSIONSTEP)  "."   \
                                 _MKSTR(VERS_TYPE)

/** what compatible driver version */
#define DRV_SDD_WHAT_STR \
        "@(#)Lantiq SDD device driver, version " DRV_SDD_VER_STR

/* ========================================================================== */
/*                             Global variables                               */
/* ========================================================================== */
#ifdef ENABLE_TRACE
CREATE_TRACE_GROUP(SDD);

#ifdef SDD_STANDALONE
#ifdef DEBUG
IFX_uint32_t debug_level = DBG_LEVEL_LOW;
#else
IFX_uint32_t debug_level = DBG_LEVEL_HIGH;
#endif /* DEBUG */
#else  /* SDD_STANDALONE */
extern IFX_uint32_t debug_level;
#endif /* SDD_STANDALONE */
#endif /* ENABLE_TRACE */


/** what string support, driver version string */
const IFX_char_t DRV_SDD_WHATVERSION[] = DRV_SDD_WHAT_STR;

extern IFX_uint16_t major;
extern IFX_uint16_t minorBase;

/* ========================================================================== */
/*                              Local variables                               */
/* ========================================================================== */
/* static array of device structs */
static SDD_DEVICE_t SddDevices[SDD_MAX_DEVICES];

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */
extern IFX_void_t AddCapability (
                        IFX_TAPI_CAP_t* CapList,
                        IFX_uint8_t *pnCap,
                        IFX_char_t const * description,
                        IFX_int32_t type, IFX_int32_t value);

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */

/**
   Get a pointer to the device struct.

   \param  nr           Number of the device. Counting starts with zero.
   \param  pDev         Returns pointer to a device structure.

   \return
   SDD_statusOk         if ok.
   SDD_statusIfxError   if device number is out of supported range.
*/
IFX_int32_t SDD_GetDevice (IFX_uint16_t nr, SDD_DEVICE_t** pDev)
{
   if (nr >= SDD_MAX_DEVICES)
   {
      *pDev = IFX_NULL;
      return SDD_statusIfxError;
   }

   *pDev = &SddDevices[nr];
   return SDD_statusOk;
}

/**
   Get a pointer to the neighbor channel struct of the given SDD channel.

   By design the SDD has no more than two channels for FXS. When two FXS
   channels are supported they are channel 0 and channel 1. So the neighbour
   channels are always 0 and 1 and if channel 0 is passed channel 1 is
   returned and vice versa.

   \param  pCh          Pointer to the channel.

   \return
   Pointer to the SDD channel struct of the neighbor channel. Also for
   1 channel devices there is a neighboring channel although it will not
   carry any resources.
*/
SDD_CHANNEL_t *SDD_GetNeighbourChannel (SDD_CHANNEL_t *pCh)
{
   IFX_uint8_t nCh;

   /* Note that the channel number stored in the channel is index + 1. */
   nCh = (pCh->nChannel - 1) == 0 ? 1 : 0;

   return &pCh->pDev->pChannel[nCh];
}

/**
   Set all capabilities.

   \param  CapList      Pointer to the capability list (array).
   \param  pnCap        Pointer to next free array index in capability list.
   \param  pLLDev       Pointer to the device structure.
*/
IFX_void_t SDD_AddCaps (IFX_TAPI_CAP_t *CapList,
                        IFX_uint8_t *pnCap,
                        IFX_TAPI_LL_DEV_t* pLLDev)
{
   SDD_DEVICE_t *pDev = (SDD_DEVICE_t *)pLLDev;

   AddCapability (CapList, pnCap, "POTS",
                  IFX_TAPI_CAP_TYPE_PORT, IFX_TAPI_CAP_PORT_POTS);
   AddCapability (CapList, pnCap, "PSTN",
                  IFX_TAPI_CAP_TYPE_PORT, IFX_TAPI_CAP_PORT_PSTN);

   if (pDev->bCapsRead)
   {
      AddCapability (CapList, pnCap, "Phones",
                     IFX_TAPI_CAP_TYPE_PHONES, pDev->caps.nFXS);
      AddCapability (CapList, pnCap, "FXO",
                     IFX_TAPI_CAP_TYPE_FXO, pDev->caps.nFXO);
      /* DC/DC converter type */
      AddCapability (CapList, pnCap, "DC/DC type",
                     IFX_TAPI_CAP_TYPE_DCDC,
                     pDev->nAllowedDcDcType - SDD_DCDC_TYPE_IBB);
   }
}


/**
   Prepare the low level device struct.

   This function clears the device struct, sets the device number and links it
   with the high-level device.

   \param  pTapiDev     Pointer to the high-level device struct.
   \param  devNum       Device number.

   \return
   Pointer to SDD Device or IFX_NULL.
*/
IFX_TAPI_LL_DEV_t* SDD_TAPI_LL_DevicePrepare(
                        TAPI_DEV *pTapiDev,
                        IFX_uint32_t devNum)
{
   SDD_DEVICE_t* pDev;

   /* make sure a valid context is given */
   if (devNum >= SDD_MAX_DEVICES)
   {
      TRACE(SDD, DBG_LEVEL_HIGH,
            ("SDD_LL_Prepare_Dev: SDD device number out of range\n"));
      return IFX_NULL;
   }
   if (pTapiDev == IFX_NULL)
   {
      TRACE(SDD, DBG_LEVEL_HIGH,
            ("SDD_LL_Prepare_Dev: pTapiDev is NULL\n"));
      return IFX_NULL;
   }

   pDev = &SddDevices[devNum];

   /* Clear the device struct (including the channels) */
   memset(pDev, 0, sizeof(SDD_DEVICE_t));

   pDev->nDevNr = devNum;

   /* Store the corresponding HL pointer */
   pDev->pTapiDev = pTapiDev;

   /* Event handling is stopped until it is configured. */
   /* pDev->nEventHandlingMode = SDD_EVENT_STOP; */

   /* Get interface from SDD mailbox driver */
   pDev->pMbxIF = sdd_mbx_interface_get();

   if (pDev->pMbxIF != IFX_NULL)
   {
      TRACE (SDD, DBG_LEVEL_LOW, ("SDD: registering MBX driver instance: %p\n",
                                  pDev->pMbxIF));

      /* Register callback functions */
      pDev->pMbxIF->evt_read_callback = SDD_EventProceed;
      pDev->pMbxIF->error_callback = SDD_ErrorProceed;
   }
   else
   {
      TRACE (SDD, DBG_LEVEL_LOW, ("SDD: registering MBX driver failed!\n"));
   }
#if 0
   sdd_AddCaps (pDev);
#endif

   /* Return the pDev pointer which is stored in the HL device */
   return pDev;
}


/**
   Prepare the low level channel struct.

   \param  pTapiCh      High level channel pointer.
   \param  pLLDev       Pointer to the device structure.
   \param  chNum        Channel number.

   \return
   Pointer to SDD Channel or IFX_NULL.
*/
IFX_TAPI_LL_CH_t* SDD_TAPI_LL_ChannelPrepare(
                        TAPI_CHANNEL *pTapiCh,
                        IFX_TAPI_LL_DEV_t *pLLDev,
                        IFX_uint32_t chNum)
{
   SDD_DEVICE_t  *pDev = (SDD_DEVICE_t *)pLLDev;
   SDD_CHANNEL_t *pCh;

   SDD_ASSERT(pDev != IFX_NULL);

   /* Make sure we have a SDD channel that we can bind to the TAPI channel. */
   if (chNum >= SDD_MAX_CH_NR)
   {
      /* This is a normal case as there are usually more TAPI channels than
         SDD channels in the system. */
      return IFX_NULL;
   }
   if (pTapiCh == IFX_NULL)
   {
      TRACE(SDD, DBG_LEVEL_HIGH,
            ("SDD_TAPI_LL_ChannelPrepare: pTapiCh is NULL\n"));
      return IFX_NULL;
   }

   /* TAPI channel number is identical to the channel index in this driver. */
   pCh  = &(pDev->pChannel[chNum]);

   pCh->nChannel = (IFX_uint8_t)chNum + 1;
   pCh->pDev  = pDev;

   /* Store the corresponding HL TAPI channel pointer */
   pCh->pTapiCh = pTapiCh;

  /* Should return the pCh which should be stored by the HL */
   return pCh;
}


/**
   Frees all resources of the channel.

   Called when the channel is released.

   \param  pCh          Pointer to the channel.
   \param  bChipAccess  Allow or deny chip access in this function.
*/
static IFX_void_t sdd_ChannelExit(SDD_CHANNEL_t *pCh,
                                  IFX_boolean_t bChipAccess)
{
   /* Deactivate all still running algorithms in the FW */
   SDD_ALM_ChStop(pCh, bChipAccess);
   SDD_PCM_ChStop(pCh);

   /* delete channel mutex */
   IFXOS_MutexDelete (&pCh->mtxChAcc);

   /* free the modules */
   SDD_ALM_Free_Ch_Structures (pCh);
   SDD_PCM_ChStructFree (pCh);
}

/**
   Initialise the access to the chip.

   This function allocates all resources needed for access to the chip.

   \param  pDev         Pointer to the device structure.

   \return
   - IFX_SUCCESS        if successful

   \remarks
   Must be the first action after installing the driver and
   before any access to the hardware
*/
IFX_int32_t SDD_ChipAccessInit (SDD_DEVICE_t *pDev)
{
   IFX_int32_t ret = SDD_statusOk;

   /* Do not initialise the chip access more than one time. */
   if (pDev->nDevState & DS_BASIC_INIT)
   {
      return SDD_statusOk;
   }

   /* init mutexes */
   IFXOS_MutexInit(&pDev->mtxMbxAcc);

   ret = SDD_Prepare_Mailbox(pDev, "command");

#ifdef S220
   if (SDD_SUCCESS (ret))
   {
      SDD_FIRMWARE_IF_t *p = pDev->pMbxIF;
      IFX_char_t *pDcDcName;
      enum SDD_DCDC_TYPE nAllowed;

      /* If p is NULL the Prepare_Mailbox above fails and we never get here. */
      SDD_ASSERT(p != IFX_NULL);

      p->dcdc_name(&pDcDcName);
      nAllowed = SDD_ALM_DcDcString2Enum(pDcDcName);

      TRACE(SDD, DBG_LEVEL_LOW,
            ("SDD: found DC/DC name >%s< translated: %d\n",
             pDcDcName, nAllowed));

      if (nAllowed != SDD_DCDC_TYPE_NOTFOUND)
      {
         pDev->nAllowedDcDcType = nAllowed;
         pDev->bDcDcHwCombined =
            (nAllowed == SDD_DCDC_TYPE_CIBB) ||
            (nAllowed == SDD_DCDC_TYPE_CIB) ||
            (nAllowed == SDD_DCDC_TYPE_CBB) ||
            (nAllowed == SDD_DCDC_TYPE_CIFB) ||
            (nAllowed == SDD_DCDC_TYPE_CIBGD) ||
            (nAllowed == SDD_DCDC_TYPE_CIBVD) ? IFX_TRUE : IFX_FALSE;
      }
   }
#endif /* S220 */

   if (SDD_SUCCESS (ret))
   {
      pDev->nDevState |= DS_BASIC_INIT;
   }

   RETURN_DEVSTATUS (ret);
}

/**
   Release the access to the chip.

   This function releases all resources needed for access to the chip.

   \param  pDev         Pointer to the device structure.
*/
static IFX_void_t sdd_ChipAccessExit(SDD_DEVICE_t *pDev)
{
   IFX_int32_t ret = SDD_statusOk;
   SDD_FIRMWARE_IF_t *p = pDev->pMbxIF;

   /* make sure that we really got a device structure */
   SDD_ASSERT((pDev != NULL && pDev->nChannel == 0));

   if (pDev->nDevState & DS_BASIC_INIT)
   {
      /* Device state: basic init is not done */
      pDev->nDevState &= ~DS_BASIC_INIT;
      /* Stop event handling for the device. */
      /*pDev->nEventHandlingMode = SDD_EVENT_STOP;*/

      /* Should there be no mailbox driver skip this without error. */
      if (p != IFX_NULL)
      {
         /* close the mailbox */
         ret = p->close ("command");
         if (SDD_SUCCESS (ret))
         {
            ret = p->close ("event");
            if (!SDD_SUCCESS (ret))
            {
               TRACE(SDD, DBG_LEVEL_HIGH,
                     ("SDD: event mailbox close error (code %08x)\n", ret));
            }
         }
         else
         {
            TRACE(SDD, DBG_LEVEL_HIGH,
                  ("SDD: command mailbox close error (code %08x)\n", ret));
         }
      }

      if (SDD_SUCCESS (ret))
      {
         /* delete the mailbox protection mutex */
         IFXOS_MutexDelete (&pDev->mtxMbxAcc);
      }
   }
}

/**
   Frees all resources of the device.

   This calls also exit on all channels because they are members of the device.
   Called when the device is released.

   \param  pDev         Pointer to the device structure.
   \param  bChipAccess  Allow or deny chip access in this function.
*/
static IFX_void_t sdd_DeviceExit(SDD_DEVICE_t *pDev,
                                 IFX_boolean_t bChipAccess)
{
   IFX_uint16_t i;

   /* call exit on all channels of the device */
   for (i = 0; i < SDD_MAX_CH_NR; i++)
   {
      sdd_ChannelExit(&pDev->pChannel[i], bChipAccess);
   }

   if (bChipAccess != IFX_FALSE)
   {
      /* The PCM interface is device global and can only be stopped after all
         PCM channels are deactivated. */
      SDD_PCM_IF_Stop(pDev);
   }

   /* Fill the capabilities again with initial values.
      To have capabilities available during cleanup this is done last. */
   pDev->bCapsRead = IFX_FALSE;

   /* delete device mutex */
   IFXOS_MutexDelete (&pDev->mtxMemberAcc);

   /* release chip access */
   sdd_ChipAccessExit(pDev);

   /* reset the states variable */
   pDev->nDevState = 0;
}

/**
   Initialise the low level device struct.

   Initialise the member variables of the device structure.

   \param  pLLDev       Pointer to the device structure.

   \return
   - SDD_statusOk       If successful.
   - SDD_statusIfxError In case of an error.
*/
IFX_int32_t SDD_TAPI_LL_DeviceInit(IFX_TAPI_LL_DEV_t* pLLDev)
{
   IFX_int32_t ret = SDD_statusOk;
   SDD_DEVICE_t *pDev = (SDD_DEVICE_t *)pLLDev;

   /* OS and board independent initializations, resets all values */

   /* intially there is no error */
   pDev->nErr = SDD_statusOk;

   /* initialize share variables protection semaphore */
   IFXOS_MutexInit(&pDev->mtxMemberAcc);

   /* default DC/DC type */
   pDev->nAllowedDcDcType = SDD_DCDC_TYPE_IBB;
   /* dedicated DC/DC is the default */
   pDev->bDcDcHwCombined = IFX_FALSE;

   return ret;
}

/**
   Stop the SDD device and free all allocated resources.

   \param  pLLDev       Pointer to the device structure.
   \param  bChipAccess  Allow or deny chip access in this function.
*/
IFX_void_t SDD_TAPI_LL_DeviceExit (IFX_TAPI_LL_DEV_t *pLLDev,
                                          IFX_boolean_t bChipAccess)
{
   SDD_DEVICE_t  *pDev = (SDD_DEVICE_t *)pLLDev;

   /* resource cleanup */
   sdd_DeviceExit(pDev, bChipAccess);
}

/**
   Initialise the low level channel struct.

   \param  pLLCh        Pointer to the channel structure.

   \return
   IFX_SUCCESS
*/
IFX_int32_t SDD_TAPI_LL_ChannelInit (IFX_TAPI_LL_CH_t *pLLCh)
{
   SDD_CHANNEL_t *pCh = (SDD_CHANNEL_t *)pLLCh;

   SDD_ASSERT(pCh != IFX_NULL);

   /* init channel lock */
   IFXOS_MutexInit (&pCh->mtxChAcc);

   return IFX_SUCCESS;
}

/**
   Read the revision information and capabilities from the firmware.

   \param pDev handle to the device structure

   \return
   - SDD_statusOk
   - SDD_statusCmdIbNoSpace
   - SDD_statusCmdMbWrErr
*/
IFX_int32_t sdd_ReadRevision (SDD_DEVICE_t *pDev)
{
   IFX_int32_t ret;
   struct SDD_RevisionRead fwmRevRead;
   struct SYS_ReadCapabilities fwmReadCaps;
   struct SYS_SmartSlicStatus fwmSlicStatus;
   struct SYS_VCODEC_FW_Version fwmFwVersion;

   /* VCODEC FW Version */
   memset (&fwmFwVersion, 0, sizeof(fwmFwVersion));
#ifdef S220
   fwmFwVersion.CMD    = CMD_EOP;
#else /* S220 */
   fwmFwVersion.CMD    = CMD_SDD;
#endif /* S220 */
   fwmFwVersion.MOD    = MOD_SYS;
   fwmFwVersion.ECMD   = SYS_VCODEC_FW_Version_ECMD;
   fwmFwVersion.LENGTH = SYS_VCODEC_FW_Version_LENGTH;

   ret = SDD_CmdRead (pDev, (IFX_uint32_t *)(IFX_void_t *)&fwmFwVersion,
                            (IFX_uint32_t *)(IFX_void_t *)&fwmFwVersion);
   if (!SDD_SUCCESS(ret))
      return ret;

   pDev->nFwVersion[0] = fwmFwVersion.MAJ;
   pDev->nFwVersion[1] = fwmFwVersion.MIN;
   pDev->nFwVersion[2] = fwmFwVersion.HF;
   pDev->nFwVersion[3] = fwmFwVersion.PLA;
   pDev->nFwBuildTimestamp = fwmFwVersion.TIME;

   /* SDD revision */
   memset (&fwmRevRead, 0, sizeof(fwmRevRead));
   fwmRevRead.CMD    = CMD_SDD;
   fwmRevRead.MOD    = MOD_SDD;
   fwmRevRead.ECMD   = SDD_RevisionRead_ECMD;
   fwmRevRead.LENGTH = SDD_RevisionRead_LENGTH;

   ret = SDD_CmdRead (pDev, (IFX_uint32_t *)(IFX_void_t *)&fwmRevRead,
                            (IFX_uint32_t *)(IFX_void_t *)&fwmRevRead);
   if (!SDD_SUCCESS(ret))
      return ret;

   pDev->nSlicDevId = fwmRevRead.DEV;
#ifdef S220
   pDev->nDeviceNr     = fwmRevRead.DEVNR;
   pDev->nCalendarWeek = fwmRevRead.CW;
   pDev->nYear         = fwmRevRead.YEAR;

   /* Number of channels */
   pDev->caps.nALI  = (fwmRevRead.CH == 0) ? 2 : 1;

   /* For unfused devices assume two channels. */
   if ((fwmRevRead.DEVNR == 0) &&
       (fwmRevRead.CW == 0) && (fwmRevRead.YEAR == 0) &&
       (fwmRevRead.DEV == 0) && (fwmRevRead.CH == 0))
   {
      TRACE (SDD, DBG_LEVEL_HIGH,
             ("\nWARNING: UNFUSED device! Assuming 2 channels.\n"));
      pDev->caps.nALI = 2;
   }
#else /* S220 */
   /* SmartSLIC revision */
   pDev->nSlicRevision = fwmRevRead.HardwareRevision;
   /* DCCtrl version */
   pDev->nDcDcCtrlVers = fwmRevRead.DCCtrlVers;
   pDev->nDcDcCtrlStep = fwmRevRead.DCCtrlVersStep;
   pDev->nDcDcCtrlHotFix = fwmRevRead.DCCtrlHotFixRevision;

   /* Number of channels */
   pDev->caps.nALI  = fwmRevRead.CH;
   pDev->caps.FXO0 = fwmRevRead.FXO0;
   pDev->caps.FXO1 = fwmRevRead.FXO1;
#endif /* S220 */
   /* ASDSP version */
   pDev->nASDSPVers = fwmRevRead.ASDSPVers;
   pDev->nASDSPStep = fwmRevRead.ASDSPStep;

   /* VCODEC capabilities */
   memset (&fwmReadCaps, 0, sizeof(fwmReadCaps));

#ifdef S220
   fwmReadCaps.CMD    = CMD_EOP;
#else /* S220 */
   fwmReadCaps.CMD    = CMD_SDD;
#endif /* S220 */
   fwmReadCaps.MOD    = MOD_SYS;
   fwmReadCaps.ECMD   = SYS_ReadCapabilities_ECMD;
   fwmReadCaps.LENGTH = SYS_ReadCapabilities_LENGTH;

   ret = SDD_CmdRead (pDev, (IFX_uint32_t *)(IFX_void_t *)&fwmReadCaps,
                            (IFX_uint32_t *)(IFX_void_t *)&fwmReadCaps);
   if (!SDD_SUCCESS(ret))
      return ret;

#ifdef S220
   pDev->caps.bfw_GR909 = fwmReadCaps.GR909;
   /* SLIC220 supports all these features since the first version. */
   pDev->caps.bfw_PCM = 1;
   pDev->caps.bfw_BuckBoost = 1;
   pDev->caps.bfw_WbTsSplit = 1;
   pDev->caps.bfw_combDcDc = 1;
   pDev->caps.bfw_ExtLT = 1;
   pDev->caps.bfw_CMeas = 1;
   pDev->caps.bfw_AsDspDwld = 1;
   pDev->caps.bfw_DTMFR = 1;
   pDev->caps.bfw_MWI = 1;
   pDev->caps.bfw_ASPwrSave = 1;
#else /* S220 */
   /* FW misses to report PCM capability for VCODEC. Force it here. */
   /* pDev->caps.bfw_PCM = fwmReadCaps.CapPcm; */
   pDev->caps.bfw_PCM = 1;
   pDev->caps.bfw_BuckBoost = fwmReadCaps.CapBB;
   pDev->caps.bfw_WbTsSplit = fwmReadCaps.CapWbTsC;
   pDev->caps.bfw_combDcDc = fwmReadCaps.CapCombDcDc;
   pDev->caps.bfw_ExtLT = fwmReadCaps.CapExtLT;
   pDev->caps.bfw_CMeas = fwmReadCaps.CapCMeas;
   pDev->caps.bfw_AsDspDwld = fwmReadCaps.CapAsDspDL;
   pDev->caps.bfw_DTMFR = fwmReadCaps.CapDTMF;
   pDev->caps.bfw_MWI = fwmReadCaps.CapMWI;
   pDev->caps.bfw_ASPwrSave = fwmReadCaps.CapAS;
#endif /* S220 */

   /* Smart SLIC status */
   memset (&fwmSlicStatus, 0, sizeof(fwmSlicStatus));

#ifdef S220
   fwmSlicStatus.CMD    = CMD_EOP;
#else /* S220 */
   fwmSlicStatus.CMD    = CMD_SDD;
#endif /* S220 */
   fwmSlicStatus.MOD    = MOD_SYS;
   fwmSlicStatus.ECMD   = SYS_SmartSlicStatus_ECMD;
   fwmSlicStatus.LENGTH = SYS_SmartSlicStatus_LENGTH;

   ret = SDD_CmdRead (pDev, (IFX_uint32_t *)(IFX_void_t *)&fwmSlicStatus,
                            (IFX_uint32_t *)(IFX_void_t *)&fwmSlicStatus);
   if (!SDD_SUCCESS(ret))
      return ret;

   if (fwmSlicStatus.SLST == SYS_SmartSlicStatus_SLST_NoSLIC)
   {
      /* No Smart SLIC is connected. */
      pDev->caps.nALI = 0;
   }
   else
   {
      /* Limit the number of channels to what the actual SLIC (HW) can do. */
      if (pDev->caps.nALI > fwmSlicStatus.NALI)
      {
         pDev->caps.nALI = fwmSlicStatus.NALI;
      }
   }

   /* Derived resource count numbers. */

   pDev->caps.nPCM = pDev->caps.bfw_PCM ? pDev->caps.nALI : 0;
   pDev->caps.nDTMFR = pDev->caps.bfw_DTMFR ? pDev->caps.nALI : 0;
   /* Calculate the number of FXO channels. */
   pDev->caps.nFXO = (((pDev->caps.nALI >= 1) && pDev->caps.FXO0) ? 1 : 0) +
                     (((pDev->caps.nALI >= 2) && pDev->caps.FXO1) ? 1 : 0) +
                     (  pDev->caps.nALI >= 3  ? 1 : 0);
   /* The channels which are not FXO are doing FXS. */
   pDev->caps.nFXS = pDev->caps.nALI - pDev->caps.nFXO;

   /* Notify TAPI-HL that FXO is via the SLIC. */
   if (pDev->caps.nFXO > 0)
   {
      TAPI_DEV *pTapiDev = (pDev->pChannel[0].pTapiCh)->pTapiDevice;

      IFX_TAPI_Update_SlicFxo(pTapiDev, IFX_TRUE);
   }

   /* Set flags indicating the mounted DC/DC HW. */

   /* Crosscheck DC/DC HW encoding. */
#ifndef S220
   if ((pDev->caps.nFXS < 2) &&
       (fwmRevRead.DcDcHardware & SDD_RevisionRead_DcDcHardware_Mask_Combined))
   {
      /* The DC/DC HW coding indicates combined DC/DC converter but the SLIC
         does not provide 2 FXS channels. Modify what is reported by the HW
         so that the internal flags are set to dedicated only. */
      fwmRevRead.DcDcHardware &= ~SDD_RevisionRead_DcDcHardware_Mask_Combined;
   }
   /* DC/DC converter type indicated on the SLIC GPIOs. */
   pDev->nAllowedDcDcType = (enum SDD_DCDC_TYPE)
      (fwmRevRead.DcDcHardware + SDD_DCDC_TYPE_IBB);
   /* Quick and easy check if the DC/DC converter is a dedicated type. */
   pDev->bDcDcHwCombined =
      (fwmRevRead.DcDcHardware & SDD_RevisionRead_DcDcHardware_Mask_Combined) ?
      IFX_TRUE : IFX_FALSE;
#endif /* S220 */

   /* Set flag that the capabilities struct is now filled with values. */
   pDev->bCapsRead = IFX_TRUE;

   TRACE(SDD, DBG_LEVEL_NORMAL,
         ("VCODEC %d.%d.%d.%d (SDD: %d.%d.%d / ASDSP %d.%d) buildtime:%08X\n"
          "Resources: ALI %d (FXS:%d/FXO:%d), PCM %d, DTMFR %d\n"
          "Capabilities: ASDSP dwld %d, DTMFR %d, MWI %d, AS %d, CMeas %d\n"
          "              ExtLT %d, CombDcDc %d, WbTsSplit %d, Buck/Boost %d, "
          "PCM %d, GR909 %d\n"
          "SmartSLIC ID 0x%02x, Rev 0x%02x, "
#ifndef S220
          "mounted DC/DC Type %d, "
#endif /* S220 */
          "Device Num %d, Week %d, Year %d \n",
          /* versions */
          pDev->nFwVersion[0], pDev->nFwVersion[1],
          pDev->nFwVersion[2], pDev->nFwVersion[3],
          pDev->nDcDcCtrlVers, pDev->nDcDcCtrlStep, pDev->nDcDcCtrlHotFix,
          pDev->nASDSPVers, pDev->nASDSPStep,
          pDev->nFwBuildTimestamp,
          /* resources */
          pDev->caps.nALI, pDev->caps.nFXS, pDev->caps.nFXO,
          pDev->caps.nPCM, pDev->caps.nDTMFR,
          /* capabilities */
          pDev->caps.bfw_AsDspDwld,
          pDev->caps.bfw_DTMFR,
          pDev->caps.bfw_MWI,
          pDev->caps.bfw_ASPwrSave,
          pDev->caps.bfw_CMeas,
          pDev->caps.bfw_ExtLT,
          pDev->caps.bfw_combDcDc,
          pDev->caps.bfw_WbTsSplit,
          pDev->caps.bfw_BuckBoost,
          pDev->caps.bfw_PCM,
          pDev->caps.bfw_GR909,
          /* SLIC identity */
          pDev->nSlicDevId,
          pDev->nSlicRevision,
#ifndef S220
          fwmRevRead.DcDcHardware, /* show untranslated value from HW */
#endif /* S220 */
          pDev->nDeviceNr,
          pDev->nCalendarWeek,
          pDev->nYear));

   return SDD_statusOk;
}

/**
   Get the number of channels.

   \param  pLLDev       Pointer to the device structure.
*/
IFX_uint8_t SDD_ChannelCountGet (
                        IFX_TAPI_LL_DEV_t* pLLDev)
{
   SDD_DEVICE_t *pDev = (SDD_DEVICE_t *)pLLDev;

   SDD_ASSERT(pDev != IFX_NULL);

   return pDev->caps.nALI;
}

/**
   Report the version information.

   \param  pDev         Pointer to the device structure.
   \param  pVersion     Pointer to the struct where to report the versions.

   \return
   - SDD_statusOk
*/
IFX_int32_t SDD_VersionGet(
                        IFX_TAPI_LL_DEV_t *pLLDev,
                        struct SDD_VERSION_INFO *pVersion)
{
   SDD_DEVICE_t *pDev = (SDD_DEVICE_t *)pLLDev;

   if (pDev == IFX_NULL ||
       pDev->bCapsRead != IFX_TRUE)
      return SDD_statusIfxError;

   pVersion->nVcodecFw[0] = pDev->nFwVersion[0];
   pVersion->nVcodecFw[1] = pDev->nFwVersion[1];
   pVersion->nVcodecFw[2] = pDev->nFwVersion[2];
   pVersion->nVcodecFw[3] = pDev->nFwVersion[3];

   pVersion->nSdd[0] = pDev->nDcDcCtrlVers;
   pVersion->nSdd[1] = pDev->nDcDcCtrlStep;
   pVersion->nSdd[2] = pDev->nDcDcCtrlHotFix;
   pVersion->nSdd[3] = 0;

   pVersion->nASDSP[0] = pDev->nASDSPVers;
   pVersion->nASDSP[1] = pDev->nASDSPStep;
   pVersion->nASDSP[2] = 0;
   pVersion->nASDSP[3] = 0;

   pVersion->nSmartSLIC_HW_rev = pDev->nSlicRevision;
   pVersion->nSmartSLIC_DevId = pDev->nSlicDevId;

#ifdef S220
   pVersion->nSmartSLIC_DeviceNr = pDev->nDeviceNr;
   pVersion->nSmartSLIC_CalendarWeek = pDev->nCalendarWeek;
   pVersion->nSmartSLIC_Year = pDev->nYear;
#endif /* S220 */

   return SDD_statusOk;
}

/**
   Create and initalise the structs handling the firmware modules.

   \param  pDev         Pointer to the device structure.

   \return
   - SDD_statusOk
   - SDD_statusInitFwMsgErr - if FW messages could not be initialized
*/
static IFX_int32_t sdd_FW_DevDataInit(SDD_DEVICE_t *pDev)
{
   IFX_uint8_t i;
   IFX_int32_t ret = SDD_statusOk;
#ifdef S220
   struct SYS_CTRL *pSysCtrl = &pDev->fwm_sys_ctrl;

   memset(pSysCtrl, 0, sizeof(*pSysCtrl));
   pSysCtrl->CMD        = CMD_EOP;
   pSysCtrl->MOD        = MOD_SYS;
   pSysCtrl->ECMD       = SYS_CTRL_ECMD;
   pSysCtrl->LENGTH     = SYS_CTRL_LENGTH;

   /* Enable sleep. */
   pSysCtrl->SE = 1;

   /* Enable/disable the charge pump depending on the allowed DC/DC type. */
   if (pDev->nAllowedDcDcType == SDD_DCDC_TYPE_IFB ||
       pDev->nAllowedDcDcType == SDD_DCDC_TYPE_CIFB ||
       pDev->nAllowedDcDcType == SDD_DCDC_TYPE_IB ||
       pDev->nAllowedDcDcType == SDD_DCDC_TYPE_CIB)
   {
      /* Enable charge pump for all DC/DC types which use a FET transistor. */
      pSysCtrl->CP_EN = 1;
      /* Raise the charge pump voltage in steps up to 5.0V. */
      for (i = SYS_CTRL_CP_VOLT_3_95V;
           SDD_SUCCESS(ret) && i <= SYS_CTRL_CP_VOLT_5_0V; i++)
      {
         pSysCtrl->CP_VOLT = i;
         ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)pSysCtrl);
      }
   }
   else
   {
      /* Do not use the charge pump. */
      pSysCtrl->CP_EN = 0;
      ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)pSysCtrl);
   }

   if (!SDD_SUCCESS(ret))
   {
      pDev->nErr = SDD_statusInitFwMsgErr;
      /* errmsg: Initializing FW messages failed. */
      return SDD_statusInitFwMsgErr;
   }
#endif /* S220 */

   /* allocate and initialize PCM device data */
   ret = SDD_PCM_DevStructAllocate(pDev);
   if (ret != SDD_statusOk)
   {
      pDev->nErr = SDD_statusInitFwMsgErr;
      /* errmsg: Initializing FW messages failed. */
      return SDD_statusInitFwMsgErr;
   }
   else
   {
      SDD_PCM_DevInit(pDev);
   }

   /* allocate and initialize PCM FW messages */
   for (i = 0; (ret == SDD_statusOk) && (i < pDev->caps.nPCM); ++i)
   {
      SDD_CHANNEL_t *pCh = &pDev->pChannel[i];

      ret = SDD_PCM_ChStructAllocate (pCh);
      if (ret != SDD_statusOk)
      {
         pDev->nErr = SDD_statusInitFwMsgErr;
         /* errmsg: Initializing FW messages failed. */
         return SDD_statusInitFwMsgErr;
      }
      else
      {
         SDD_PCM_ChInit (pCh);
      }
   }

   /* allocate and initialize ALM FW messages */
   for (i = 0; (ret == SDD_statusOk) && (i < pDev->caps.nALI); ++i)
   {
      SDD_CHANNEL_t *pCh = &pDev->pChannel[i];

      ret = SDD_ALM_Allocate_Ch_Structures (pCh);
      if (ret != SDD_statusOk)
      {
         pDev->nErr = SDD_statusInitFwMsgErr;
         /* errmsg: Initializing FW messages failed. */
         return SDD_statusInitFwMsgErr;
      }
      else
      {
         SDD_ALM_InitCh (pCh);
      }
   }

   return ret;
}

/**
   Enable the SDD interrupts.

   \param pDev pointer to the device structure

   \return
   - SDD_statusOk

   \remarks
      TODO: Remove this function if masking is not provided by FW
*/
static IFX_int32_t sdd_DefaultEvtEnable(SDD_DEVICE_t *pDev)
{
   IFX_int32_t             err = SDD_statusOk;

   /* TODO: add default mask, if supported by FW */

   return err;
}

/**
   Firmware initialisation function.

   \param  pLLDev       Pointer to the device structure.
   \param  nMode        Enum from IFX_TAPI_INIT_MODE_t specifying the setup.

   \return

   \remarks
*/
IFX_int32_t SDD_TAPI_LL_FW_Init(
                        IFX_TAPI_LL_DEV_t *pLLDev,
                        IFX_uint8_t nMode)
{
   SDD_DEVICE_t         *pDev = (SDD_DEVICE_t*)pLLDev;
   IFX_uint8_t          nMaxRes;
   IFX_int32_t          ret = SDD_statusOk;

   /* do nothing if device initialization was already done */
   if (pDev->nDevState & DS_DEV_INIT)
   {
      return SDD_statusOk;
   }

   ret = SDD_ChipAccessInit(pDev);
   if (ret != SDD_statusOk)
   {
      RETURN_DEVSTATUS (ret);
   }

   /* blocked until the chip access is initialised */
   if (!(pDev->nDevState & DS_BASIC_INIT))
   {
      /* errmsg: Chip access is not setup. */
      return SDD_statusChipAccNotSetup;
   }

   /* Before FW access reset the flags indicating FW error conditions. */
   pDev->bSSIcrash = IFX_FALSE;
   pDev->nSlicRecoveryCnt = 0;

   sdd_ReadRevision (pDev);

   /* Warn if the FW version is older than a defined minimum version. */
   {
      const IFX_uint8_t nMinVers[] = {
                           SDD_MIN_FW_MAJORSTEP,
                           SDD_MIN_FW_MINORSTEP,
                           SDD_MIN_FW_HOTFIXSTEP,
                           SDD_MIN_FW_PLATFORM};
      IFX_uint8_t i;

      /* Iterate over all digits of the version number starting with the
         major. For each digit first check if it is greater than the same
         digit in the required minimum version. If so no further checks are
         needed. Second check if this digit is smaller and if so do a
         warning and end the check. If the digits are equal continue with
         the next digit until all has been checked. */
      for (i=0; (i < sizeof(nMinVers)) &&
                (i < sizeof(pDev->nFwVersion)); i++)
      {
         if (pDev->nFwVersion[i] > nMinVers[i])
         {
            break;
         }
         if (pDev->nFwVersion[i] < nMinVers[i])
         {
            TRACE (SDD, DBG_LEVEL_HIGH,
                   ("\nWARNING: VCODEC-FW version %u.%u.%u.%u too old."
                    " Minimum required VCODEC-FW version is %u.%u.%u.%u\n\n",
                    pDev->nFwVersion[0], pDev->nFwVersion[1],
                    pDev->nFwVersion[2], pDev->nFwVersion[3],
                    nMinVers[0], nMinVers[1], nMinVers[2], nMinVers[3]));
            break;
         }
      }
   }

   TRACE (SDD, DBG_LEVEL_LOW, ("\n"));
   TRACE (SDD, DBG_LEVEL_LOW,
          ("FWCAP: nPCM:%2u  nALM:%2u (FXS:%2u/FXO:%2u)\n",
           pDev->caps.nPCM, pDev->caps.nALI,
           pDev->caps.nFXS, pDev->caps.nFXO));
   TRACE (SDD, DBG_LEVEL_LOW,
          ("FWCAP: nDTMFR:%2u\n",
           pDev->caps.nDTMFR));

   /* Maximum number of resources, is the maximum of: nALI, nPCM. */
   nMaxRes = (pDev->caps.nPCM > pDev->caps.nALI) ?
              pDev->caps.nPCM : pDev->caps.nALI;

   if (nMaxRes > SDD_MAX_CH_NR)
   {
      /* Warn about incorrect definition then limit the counters so that we
         do not crash when trying to initialise non existing resources. */
      TRACE(SDD, DBG_LEVEL_HIGH,
            ("WARNING: Detected maximum resources %d but defined only %d. "
             "Some resources will not be usable\n",
             nMaxRes, SDD_MAX_CH_NR));
      if (pDev->caps.nALI > SDD_MAX_CH_NR)
         pDev->caps.nALI = SDD_MAX_CH_NR;
      if (pDev->caps.nPCM > SDD_MAX_CH_NR)
         pDev->caps.nPCM = SDD_MAX_CH_NR;
   }

   /* set default firmware cache values now, required to allow storage of
      coefficients from BBD Download... */
   if (ret == SDD_statusOk)
   {
      ret = sdd_FW_DevDataInit(pDev);
   }

   if (ret == SDD_statusOk)
   {
      ret = sdd_DefaultEvtEnable(pDev);
      if (ret != SDD_statusOk)
      {
         /* errmsg: Enabling events failed. */
         ret = SDD_statusEvtEnblErr;
      }
   }

   if (SDD_SUCCESS (ret))
   {
      /* Start and initialize kernel thread to handle the interrupt routine. */
      ret = SDD_Prepare_Mailbox(pDev, "event");
      //(void)SDD_interrupt_init();
   }

   if (ret == SDD_statusOk)
   {
      pDev->nDevState |= DS_FW_DLD;
      pDev->nDevState |= DS_DEV_INIT;
   }

   return ret;
}
/*lint +esym(715, nMode) */

/**
   SDD device driver initialization.
   This is the device driver initialization function to call at the system
   startup prior any access to the SDD device driver.
   After the initialization the device driver is ready to be accessed by
   the application.

   \return
   - IFX_SUCCESS        If successful.
   - IFX_ERROR          if registration with HL failed.
*/
IFX_int32_t SDD_DeviceDriverStart(IFX_void_t)
{
   /* Set the default trace level */
   TRACE_LEVEL_SET(SDD, debug_level);

#ifdef LINUX
   printk("%s, (c) 2013-2017 Lantiq Beteiligungs-GmbH & Co.KG\n",
          &DRV_SDD_WHATVERSION [4]);
#else
   printf("%s, (c) 2013-2017 Lantiq Beteiligungs-GmbH & Co.KG\n\r",
          &DRV_SDD_WHATVERSION [4]);
#endif /* LINUX */


   return IFX_SUCCESS;
}

/**
   SDD device driver shutdown.
*/
IFX_void_t SDD_DeviceDriverStop (IFX_void_t)
{
   IFX_uint8_t    i;

   for (i=0; i < SDD_MAX_DEVICES; i++)
   {
      /* resource cleanup */
      sdd_DeviceExit(&SddDevices[i], IFX_TRUE);
   }

   /* Note: the device struct array is a static variable so no free is needed.*/

#ifdef EVENT_LOGGER_DEBUG
   /* Unregister the driver with device name, type and number
      from the eventlogger */
   EL_REG_Unregister(SDD_DEV_NAME,
                     IFX_TAPI_DEV_TYPE_SDD, -1 /* dev num (-1 all) */);
#endif /* EVENT_LOGGER_DEBUG */

   /* Release sdd mailbox interface */
   sdd_mbx_interface_release();
}

