/*******************************************************************************

                            Copyright (c) 2013-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_int.c
   This file contains the implementation of the interrupt handler.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "ifx_types.h"
#include "ifxos_interrupt.h"
#include "ifxos_time.h"

#include "drv_sdd_osmap.h"
#include "drv_sdd_global.h"
#include "drv_sdd_init.h"
#include "drv_sdd_int_evt.h"
#include "drv_sdd_opmode.h"
#include "drv_sdd_fw_cmd.h"
#include "drv_sdd_alm_priv.h"

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */

/**
   Command Error handler.

   \param pDev   reference to device context
   \param pCERR  pointer to FW Command Error event
*/
static IFX_void_t irq_cerr_event (
                        SDD_DEVICE_t *pDev,
                        struct EVT_CERR *pCERR)
{
   IFX_TAPI_EVENT_t tapiEvent;

   /* TODO: reset command mailbox - not yet implemented */

   /* TODO: check if command mailbox is available again - not yet implemented */

   /* send an CERR event including the details to TAPI */
   memset(&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));

   tapiEvent.id = IFX_TAPI_EVENT_DEBUG_CERR;
   tapiEvent.module = IFX_TAPI_MODULE_TYPE_NONE;
   tapiEvent.data.cerr.fw_id = 0x0005; /* SDD FW */
   tapiEvent.data.cerr.reason = pCERR->CMDERR;
   tapiEvent.data.cerr.command = pCERR->CMDHDR;
   IFX_TAPI_Event_DispatchExt(pDev->pChannel[0].pTapiCh, &tapiEvent, IFX_TRUE);
}


/**
   Internal Error handler.

   \param pDev      reference to device context
   \param pIntErr   pointer to FW InternalError event

   \return
   None.
*/
static IFX_void_t irq_internal_err_event (
                        SDD_DEVICE_t *pDev,
                        struct EVT_InternalError *pIntErr)
{
#ifdef S220
   switch (pIntErr->ERREVT)
   {
      case EVT_InternalError_EQO:
         {
            TRACE(SDD, DBG_LEVEL_HIGH, ("Sdd%d EVT: Internal Event Queue Overflow\n", pDev->nDevNr));
         }
         break;
      case EVT_InternalError_CLK_FAIL:
         {
            TRACE(SDD, DBG_LEVEL_HIGH, ("Sdd%d EVT: Clock Fail\n", pDev->nDevNr));
         }
         break;
      case EVT_InternalError_SYNC_FAIL:
         {
            TRACE(SDD, DBG_LEVEL_HIGH, ("Sdd%d EVT: Synchronisation error\n", pDev->nDevNr));
         }
         break;
      case EVT_InternalError_DCDCCFG_FAIL:
         {
            TRACE(SDD, DBG_LEVEL_HIGH, ("Sdd%d EVT: SDD_DcDcConfig command does not fit to hardware\n", pDev->nDevNr));
         }
         break;
      default:
         break;
   }
#else /* S220 */
   if (pIntErr->EBO)
   {
      IFX_TAPI_EVENT_t tapiEvent;

      memset(&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));

      TRACE(SDD, DBG_LEVEL_HIGH, ("Sdd%d EVT: Internal Event-Box Overflow,"
                                  " device should be reset\n", pDev->nDevNr));
      tapiEvent.id = IFX_TAPI_EVENT_FAULT_FW_EBO_OF;
      tapiEvent.module = IFX_TAPI_MODULE_TYPE_NONE;
      IFX_TAPI_Event_DispatchExt(pDev->pChannel[0].pTapiCh, &tapiEvent, IFX_TRUE);
   }

   if (pIntErr->CSE)
   {
      TRACE(SDD, DBG_LEVEL_HIGH, ("Sdd%d EVT: Checksum error\n", pDev->nDevNr));
   }
#endif /* S220 */
}


/**
   SDD Event handler.

   \param pDev    reference to device context
   \param pSddEvt pointer to FW SDD event
*/
static IFX_void_t irq_sdd_event(
                        SDD_DEVICE_t *pDev,
                        struct EVT_SDD *pSddEvt)
{
   SDD_CHANNEL_t   *pCh = pSddEvt->CHAN < pDev->caps.nALI ?
                           &pDev->pChannel[pSddEvt->CHAN] : IFX_NULL;
   TAPI_CHANNEL    *pChannel;
   IFX_TAPI_EVENT_t tapiEvent;

   SDD_ASSERT (pCh != IFX_NULL);

   if (pCh != IFX_NULL)
   {
      pChannel = pCh->pTapiCh;
   }
   else
   {
      /* FW event refers the unsupported driver resource.
         Dropping out. */
      return;
   }

   /* prepare tapi event, channel information is the same for all events */
   memset(&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));
   tapiEvent.ch = pSddEvt->CHAN;

   switch (pSddEvt->EVT)
   {
      case EVT_SDD_EVT_OTEMP:
         {
            irq_SDD_ALM_LineDisable (pCh);

            tapiEvent.id = IFX_TAPI_EVENT_FAULT_LINE_OVERTEMP;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);

            TRACE(SDD, DBG_LEVEL_HIGH,
              ("Sdd%d, Ch%d: IRQ: WARNING: overtemperature,"
               " setting line to power down\n",
               pDev->nDevNr, (pCh->nChannel -1)));
         }
         break;

      case EVT_SDD_EVT_LTEST:
         {
            SDD_ALM_CorrectLinemodeCache(pCh, SDD_Opmode_Disabled);
            tapiEvent.id =
#ifdef SDD_FEAT_CAP_MEASURE
                  (SDD_ALM_CapMeasInProgress(pCh) == IFX_TRUE) ?
                  IFX_TAPI_EVENT_LINE_MEASURE_CAPACITANCE_RDY_INT :
#endif /* SDD_FEAT_CAP_MEASURE */
                  IFX_TAPI_EVENT_LT_GR909_RDY;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

#ifdef S220
      case EVT_SDD_EVT_LT_ABORT:
#ifdef SDD_FEAT_CAP_MEASURE
         if (SDD_ALM_CapMeasInProgress(pCh) == IFX_TRUE)
            SDD_ALM_CapMeasAbort(pCh);
#endif /* SDD_FEAT_CAP_MEASURE */
         break;
#endif /* S220 */

      case EVT_SDD_EVT_GF:
         {
            /* Set the linemode to disabled. If the DCCtrl already did it
               this disabled is the confirmation required by the DCCtrl.
               If the DCCtrl did nothing this will switch off the line.
               In both cases the line is then disabled and may be set to
               other line states by the application. */
            irq_SDD_ALM_LineDisable(pCh);
            /* Inform the application through a TAPI event. */
            tapiEvent.id = IFX_TAPI_EVENT_FAULT_LINE_GK_HIGH;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_GK:
         {
            /* Upon ground key we do nothing. Just a TAPI event is sent to
               the application. */
            tapiEvent.id = IFX_TAPI_EVENT_FAULT_LINE_GK_LOW;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_OPC:
         {
            irq_SDD_ALM_UpdateOpModeAndWakeUp (pCh, pSddEvt->OPMODE);

            /* To detect end of calibration this code looks for a transition
               from opmode calibration to any other opmode. */
            if (pCh->pALM != IFX_NULL &&
                 pSddEvt->OPMODE == SDD_Opmode_Calibrate)
            {
               /* Set flag that now calibration is running. The check below
                  will use this flag to find the end of calibration. */
               pCh->pALM->bCalibrationRunning = IFX_TRUE;
            }
            if (pCh->pALM != IFX_NULL &&
                 pCh->pALM->bCalibrationRunning == IFX_TRUE &&
                   pSddEvt->OPMODE != SDD_Opmode_Calibrate)
            {
               pCh->pALM->bCalibrationRunning = IFX_FALSE;
               tapiEvent.id = IFX_TAPI_EVENT_CALIBRATION_END_INT;
               tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
               IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
            }
         }
         break;

      case EVT_SDD_EVT_OMI:
         {
            irq_SDD_ALM_UpdateOpModeAndWakeUp (pCh, OPMODE_IGNORED);
         }
         break;

      case EVT_SDD_EVT_TTX_FIN:
         {
            tapiEvent.id = IFX_TAPI_EVENT_METERING_END;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_ONH:
         {
            tapiEvent.id = IFX_TAPI_EVENT_FXS_ONHOOK_INT;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            tapiEvent.data.hook_int.nTime =
               SDD_ALM_ElapsedTimeSinceLastHook(pCh, pSddEvt->TIMESTAMP);
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_OFFH:
         {
            tapiEvent.id = IFX_TAPI_EVENT_FXS_OFFHOOK_INT;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            tapiEvent.data.hook_int.nTime =
               SDD_ALM_ElapsedTimeSinceLastHook(pCh, pSddEvt->TIMESTAMP);
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_GF_FIN:
         {
            tapiEvent.id = IFX_TAPI_EVENT_FAULT_LINE_GK_HIGH_END;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_GK_FIN:
         {
            tapiEvent.id = IFX_TAPI_EVENT_FAULT_LINE_GK_LOW_END;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_OTEMP_FIN:
         {
            tapiEvent.id = IFX_TAPI_EVENT_FAULT_LINE_OVERTEMP_END;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_SSI_CRASH:
         {
            /* Report only the first occurance from any of the channels. */
            if (pDev->bSSIcrash == IFX_FALSE)
            {
               /* Mark in the device struct that SSI has crashed. */
               pDev->bSSIcrash = IFX_TRUE;

               /* Send an TAPI event to the application. */
               tapiEvent.id = IFX_TAPI_EVENT_FAULT_HW_SSI_ERROR_INT;
               IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
               /* Report this fatal error on the console. */
               TRACE(SDD, DBG_LEVEL_HIGH,
                       ("FATAL: SSI bus crash on SDD device %d, ch %d\n",
                        pDev->nDevNr, (pCh->nChannel - 1)));
               }
         }
         break;

      case EVT_SDD_EVT_SSI_CRASH_FIN:
         {
            /* Report only the first occurance from any of the channels. */
            if (pDev->bSSIcrash == IFX_TRUE)
            {
               /* Mark in the device struct that SSI has recovered. */
               pDev->bSSIcrash = IFX_FALSE;

               /* Send an TAPI event to the application. */
               tapiEvent.id = IFX_TAPI_EVENT_FAULT_HW_SSI_FIXED_INT;
               IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
            }
         }
         break;

#ifdef SDD_FEAT_FXO
      case EVT_SDD_EVT_FXO_RING_ON:
         {
            pCh->pALM->fxo_flags |= (1 << FXO_RING);
            tapiEvent.id = IFX_TAPI_EVENT_FXO_RING_START;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_FXO_RING_OFF:
         {
            pCh->pALM->fxo_flags &= ~(1 << FXO_RING);
            tapiEvent.id = IFX_TAPI_EVENT_FXO_RING_STOP;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_FXO_BATT_ON:
         {
            pCh->pALM->fxo_flags |= (1 << FXO_BATTERY);
            tapiEvent.id = IFX_TAPI_EVENT_FXO_BAT_FEEDED;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_FXO_BATT_OFF:
         {
            pCh->pALM->fxo_flags &= ~(1 << FXO_BATTERY);
            tapiEvent.id = IFX_TAPI_EVENT_FXO_BAT_DROPPED;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_FXO_OSI_END:
         {
            tapiEvent.id = IFX_TAPI_EVENT_FXO_OSI;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_FXO_APOH_ON:
         {
            pCh->pALM->fxo_flags |= (1 << FXO_APOH);
            tapiEvent.id = IFX_TAPI_EVENT_FXO_APOH;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_FXO_APOH_OFF:
         {
            pCh->pALM->fxo_flags &= ~(1 << FXO_APOH);
            tapiEvent.id = IFX_TAPI_EVENT_FXO_NOPOH;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_FXO_POL_REV:
         {
            pCh->pALM->fxo_flags &= ~(1 << FXO_POLARITY);
            tapiEvent.id = IFX_TAPI_EVENT_FXO_POLARITY;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            tapiEvent.data.fxo_polarity =
               IFX_TAPI_EVENT_DATA_FXO_POLARITY_REVERSED;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;

      case EVT_SDD_EVT_FXO_POL_NOR:
         {
            pCh->pALM->fxo_flags |= (1 << FXO_POLARITY);
            tapiEvent.id = IFX_TAPI_EVENT_FXO_POLARITY;
            tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
            tapiEvent.data.fxo_polarity =
               IFX_TAPI_EVENT_DATA_FXO_POLARITY_NORMAL;
            IFX_TAPI_Event_DispatchExt(pChannel, &tapiEvent, IFX_TRUE);
         }
         break;
#endif /* SDD_FEAT_FXO */

#ifdef S220
      case EVT_SDD_EVT_OPM_DIS:
         break;

      case EVT_SDD_EVT_AC_LM_FIN:
         break;

      case EVT_SDD_EVT_DATALINK_ERR:
         /* Print for debugging and development but otherwise ignore event. */
         TRACE(SDD, DBG_LEVEL_LOW,
               ("Sdd%d, Ch%d: SDD event: Datalink error at SSI.\n",
               pDev->nDevNr, (pCh->nChannel - 1)));
         break;

      case EVT_SDD_EVT_DCDCCFG_FAIL:
         /* Print for debugging and development but otherwise ignore event. */
         TRACE(SDD, DBG_LEVEL_LOW,
               ("Sdd%d, Ch%d: SDD event: SDD_DcDcConfig does not fit to HW.\n",
               pDev->nDevNr, (pCh->nChannel - 1)));
         break;
#endif /* S220 */

      default:
         TRACE(SDD, DBG_LEVEL_HIGH,
               ("Sdd%d, Ch%d: IRQ: unsupported SDD event 0x%02X\n",
               pDev->nDevNr, (pCh->nChannel - 1), pSddEvt->EVT));
         break;
   }
}

/**
   Sdd error handler.

   \param error number

   \return IFX_int32_t IFX_SUCCESS
*/
IFX_int32_t SDD_ErrorProceed(IFX_int32_t error)
{
   IFX_TAPI_EVENT_t tapiEvent;
   SDD_DEVICE_t *pDev;

   /* Device 0 is hardcoded and cannot fail. */
   (void)SDD_GetDevice (0, &pDev);

   /* send an ERR event including the details to TAPI */
   memset(&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));

   /* We receive negative value */
   switch(error)
   {
      case SDD_FW_ERR_CMD_MBX_UFL:
         tapiEvent.id = IFX_TAPI_EVENT_FAULT_FW_CBO_UF;
         break;

      case SDD_FW_ERR_CMD_MBX_OFL:
         tapiEvent.id = IFX_TAPI_EVENT_FAULT_FW_CBI_OF;
         break;

      case SDD_FW_ERR_EVT_MBX_UFL:
         tapiEvent.id = IFX_TAPI_EVENT_FAULT_FW_EBO_UF;
         break;

      case SDD_FW_ERR_EVT_MBX_OFL:
         tapiEvent.id = IFX_TAPI_EVENT_FAULT_FW_EBO_OF;
         break;

      default:
         tapiEvent.id = IFX_TAPI_EVENT_FAULT_HW_SPI_ACCESS;
         break;
   }

   tapiEvent.module = IFX_TAPI_MODULE_TYPE_NONE;
   tapiEvent.data.cerr.fw_id = 0x0005; /* SDD FW */
   IFX_TAPI_Event_DispatchExt(pDev->pChannel[0].pTapiCh, &tapiEvent, IFX_TRUE);

   return IFX_SUCCESS;
}

/**
   SDD event handler.

   \param pData pointer to the buffer
   \param len length of the event

   \return IFX_int32_t
*/
IFX_int32_t SDD_EventProceed(IFX_uint8_t *pData, IFX_uint16_t len)
{
   FW_Event_t evt, *pEvt = &evt;
   IFX_int32_t ret;
   SDD_DEVICE_t *pDev;

#ifdef S220
   struct SYS_CMD_ERR_ACK SYS_CmdErrAck;
#endif

   /* Device 0 is hardcoded and cannot fail. */
   (void)SDD_GetDevice (0, &pDev);

   if((pData != IFX_NULL) && (len > 0))
   {
      memcpy(&evt, pData, len);
   }
   else
   {
      TRACE(SDD, DBG_LEVEL_HIGH, ("Empty event!\n"));
       return IFX_SUCCESS;
   }

   /* Do not proceed events when device is not initialized. */
   if (!(pDev->nDevState & DS_BASIC_INIT))
   {
      TRACE(SDD, DBG_LEVEL_HIGH, ("Event skipped - slic is not initialized!\n"));
      return IFX_SUCCESS;
   }

   /*
    * For debug events investigation device node can be opened
    * by user application add events are intercepted
    * and not passed to up layer.
    */
   if (pDev->bIntercept == IFX_TRUE)
   {
      IFXOS_MutexGet (&pDev->mtxIntercept);
      /* Copy event */
      memcpy ( &pDev->evtStore, &evt, sizeof (evt));
      pDev->bWakeUp = IFX_TRUE;
      IFXOS_MutexRelease (&pDev->mtxIntercept);
      /* Wake up to notify about data reception */
      IFXOS_DrvSelectQueueWakeUp (&(pDev->sdd_wakeup), 0);
      ret = 0;
      IFXOS_EventWait (&pDev->readEvent, 6000, &ret);
      if (ret != 0)
      {
         /* event not received skip event */
         TRACE(SDD, DBG_LEVEL_HIGH,
                ("SDD ERR: Event SKIPPED: 0x%08X\n", pEvt->val[0]));

      }
      return IFX_SUCCESS;
   }

   if (pEvt->hdr.CMD == EVT_CMD_EVT &&
       pEvt->hdr.MOD == EVT_MOD_SYS)
   {
      /* system event, check ECMD */
      if (pEvt->hdr.ECMD == EVT_CERR_ECMD)
      {
         /* Command Error */
         irq_cerr_event(pDev, &pEvt->cerr);
#ifdef S220
               /* Send Command Error Event */
               SYS_CmdErrAck.CMD    = CMD_EOP;
               SYS_CmdErrAck.ECMD   = SYS_CMD_ERR_ACK_ECMD;
               SYS_CmdErrAck.MOD    = MOD_SYS;
               SYS_CmdErrAck.LENGTH = SYS_CMD_ERR_ACK_LENGTH;
               ret = SDD_CmdWrite(pDev, (IFX_uint32_t *)&SYS_CmdErrAck);
               if (!SDD_SUCCESS (ret))
               {
                  TRACE(SDD, DBG_LEVEL_HIGH,
                        ("SDD ERR: PANIC CmdErrAck failed!\n"));
               }
#endif /* S220 */
      }
      else
      if (pEvt->hdr.ECMD == EVT_InternalError_ECMD)
      {
         /* Internal Error */
         irq_internal_err_event(pDev, &pEvt->int_err);
      }
   }
   else
   if (pEvt->hdr.CMD == EVT_CMD_EVT &&
       pEvt->hdr.MOD == EVT_MOD_ALI &&
       pEvt->hdr.ECMD == EVT_SDD_ECMD)
   {
      /* ALI event */
      irq_sdd_event(pDev, &pEvt->sdd);
   }
   else
   {
      TRACE(SDD, DBG_LEVEL_HIGH,
             ("Unknown event: 0x%08X\n", pEvt->val[0]));
   }

   return IFX_SUCCESS;
}

