/******************************************************************************

                            Copyright (c) 2013-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_ioctl.c
   This file implements the device access functions.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "ifx_types.h"

#include "drv_sdd_osmap.h"
#include "drv_sdd_global.h"
#include "drv_sdd_version.h"
#include "drv_sdd_api.h"
#include "drv_sdd_bbd.h"
#include "drv_sdd_init.h"

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */

/* ========================================================================== */
/*                             Type definitions                               */
/* ========================================================================== */

/* ========================================================================== */
/*                             Global variables                               */
/* ========================================================================== */

/* ========================================================================== */
/*                              Local variables                               */
/* ========================================================================== */

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */
static IFX_int32_t SDD_Version (
                        SDD_DEVICE_t *pDev,
                        SDD_IO_Version_t *pInd);

static IFX_int32_t SDD_Read_Cmd (
                        SDD_DEVICE_t *pDev,
                        SDD_IO_MB_CMD_t *pCmd);

static IFX_int32_t SDD_Write_Cmd (
                        SDD_DEVICE_t *pDev,
                        SDD_IO_MB_CMD_t *pCmd);

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */

IFX_int32_t SDD_Version (
                        SDD_DEVICE_t *pDev,
                        SDD_IO_Version_t *pInd)
{
   /* errmsg: Device not yet initialized. */
   IFX_int32_t  ret = SDD_statusNotInitialized;

   /* validate parameter */
   if (IFX_NULL == pInd)
      RETURN_DEVSTATUS(SDD_statusParam);

   if (pDev->nDevState & DS_DEV_INIT)
   {
      ret = sdd_ReadRevision (pDev);
      if(ret != SDD_statusOk)
      {
         return ret;
      }

      pInd->nFxsCh   = pDev->caps.nFXS;
      pInd->nFxoCh   = pDev->caps.nFXO;

      /* Device ID */
      pInd->nDevID = pDev->nSlicDevId;
      /* chip revision */
      pInd->nHwRev = pDev->nSlicRevision;

      /* Driver version */
      pInd->nDrvVers =  MAJORSTEP << 24 |
                        MINORSTEP << 16 |
                        VERSIONSTEP << 8 |
                        VERS_TYPE;
      /* VCODEC FW version */
      pInd->nFwRev =    (IFX_uint32_t)pDev->nFwVersion[0] << 24 |
                        (IFX_uint32_t)pDev->nFwVersion[1] << 16 |
                        (IFX_uint32_t)pDev->nFwVersion[2] <<  8 |
                        (IFX_uint32_t)pDev->nFwVersion[3];
      /* ASDSP version */
      pInd->nAsdspRev = (IFX_uint16_t)pDev->nASDSPVers << 8 |
                        (IFX_uint16_t)pDev->nASDSPStep;

      ret = SDD_statusOk;
   }

   RETURN_DEVSTATUS(ret);
}


/**
   Write command to read data

   \param pDev pointer to the device interface

   \param pCmd pointer to command data

   \return
    IFX_SUCCESS if no error, otherwise IFX_ERROR (return value of CmdRead)
*/
static IFX_int32_t SDD_Read_Cmd (SDD_DEVICE_t *pDev, SDD_IO_MB_CMD_t *pCmd)
{
   IFX_int32_t err;

   err = SDD_CmdRead (pDev, (IFX_uint32_t*)((IFX_void_t *)pCmd),
                            (IFX_uint32_t*)((IFX_void_t *)pCmd));

   /* if read failed set the cmd length to 0 */
   if (err != IFX_SUCCESS)
      pCmd->cmd &= 0xffffff00;

   return err;
}


/**
   Write SDD command

   \param pDev pointer to the device interface

   \param pCmd pointer to command data

   \return
    IFX_SUCCESS if no error, otherwise IFX_ERROR (return value of SDD_CmdWrite)
*/
static IFX_int32_t SDD_Write_Cmd (SDD_DEVICE_t *pDev, SDD_IO_MB_CMD_t *pCmd)
{
   IFX_int32_t  ret;
   IFX_uint32_t *pData = (IFX_uint32_t*)((IFX_void_t *)pCmd);

   ret = SDD_CmdWrite (pDev, pData);

   /* if read failed set the cmd length to 0 */
   if (ret != IFX_SUCCESS)
      pCmd->cmd &= 0xffffff00;

   return ret;
}


/**
   SDD device specific ioctl handling

   \param  pLLDummyCh   Pointer to IFX_TAPI_LL_CH_t (could be also a pointer
                        to a SDD_DEVICE_t structure - so we have to check).
   \param  iocmd        IOCTL identifier.
   \param  ioarg        IOCTL argument.

   \return
   - SDD_statusOk
   - SDD_statusNotInitialized
*/
IFX_int32_t SDD_Dev_Spec_Ioctl (
                        IFX_TAPI_LL_CH_t *pLLDummyCh,
                        IFX_uint32_t iocmd,
                        IFX_ulong_t  ioarg)
{
   IFX_int32_t ret = SDD_statusOk;
   SDD_CHANNEL_t *pCh;
   SDD_DEVICE_t  *pDev;

   /* validate pointer */
   if (IFX_NULL == pLLDummyCh)
      return SDD_statusParam;

#ifndef TAPI_ONE_DEVNODE
   pCh = (SDD_CHANNEL_t *)pLLDummyCh;

   /* distinguish device / channel context */
   if (pCh->nChannel == 0)
   {
      /* initial assumption is wrong - we received a device pointer */
      pDev = (SDD_DEVICE_t *)pLLDummyCh;
      pCh  = IFX_NULL;
   }
   else
   {
      /* initial assumption was correct so we only need to get the
         corresponding device pointer */
      pDev = pCh->pDev;
   }
#else /* TAPI_ONE_DEVNODE */
   /* by default use first device and set channel to null */
   SDD_GetDevice(0, &pDev);
   pCh = IFX_NULL;
#endif /* TAPI_ONE_DEVNODE */

   /* getting context for specific ioctls */
   switch (iocmd)
   {
      case FIO_SDD_VERS:
         {
            SDD_IO_Version_t *pVers = IFXOS_BlockAlloc(sizeof(*pVers));
            SDD_ASSERT (pVers != IFX_NULL);
            ret = SDD_Version (pDev, pVers);
            IFXOS_CpyToUser((IFX_uint8_t*)ioarg, pVers, sizeof (*pVers));
            IFXOS_BlockFree(pVers);
         }
         break;

      /* Drv specific ioctls */
      case FIO_SDD_REPORT_SET:
         if ((ioarg > DBG_LEVEL_HIGH) || (ioarg < DBG_LEVEL_LOW))
         {
            TRACE_LEVEL_SET(SDD, DBG_LEVEL_OFF);
         }
         else
         {
            TRACE_LEVEL_SET(SDD, ioarg);
         }
         break;

      /* Dev specific ioctls */
      case FIO_SDD_FW_DOWNLOAD:
         {
            SDD_FW_Download_t   ioSddFwDwld;

            IFXOS_CpyFromUser ((char *)&ioSddFwDwld,
                               (char *)ioarg, sizeof(ioSddFwDwld));

            ret = SDD_EXT_FwDownload(pDev, &ioSddFwDwld);
         }
         break;

      /* channel specific ioctls - currently only one */
      case FIO_SDD_BBD_DOWNLOAD:
         {
            SDD_BBD_Download_t *pBbdArg;

            /* simple, not perfect, argument validation */
            if (ioarg <= 255)
            {
               RETURN_DEVSTATUS(SDD_statusParam);
            }

            /* get data from user space */
            pBbdArg = OS_MapBuffer((IFX_void_t*) ioarg,
                                   sizeof(SDD_BBD_Download_t));
            if (NULL == pBbdArg)
            {
               RETURN_DEVSTATUS(SDD_statusNoMem);
            }
#ifdef TAPI_ONE_DEVNODE
            if (SDD_statusOk != SDD_GetDevice(pBbdArg->dev, &pDev))
            {
               OS_UnmapBuffer(pBbdArg);
               RETURN_DEVSTATUS(SDD_statusDeviceIdErr);
            }
            /* if broadcast, then use this BBD for all channels */
            if (IFX_TRUE == pBbdArg->bBroadcast)
            {
               pCh = (SDD_CHANNEL_t *) pDev;
            }
            /* otherwise use only specified channel */
            else
            {
               /* get specific channels */
               if (SDD_MAX_CH_NR <= pBbdArg->ch)
               {
                  OS_UnmapBuffer(pBbdArg);
                  RETURN_DEVSTATUS(SDD_statusInvalCh);
               }
               pCh = &pDev->pChannel[pBbdArg->ch];
            }
#endif /* TAPI_ONE_DEVNODE */
            ret = SDD_BBD_Download (pCh, pBbdArg);
            OS_UnmapBuffer(pBbdArg);
         }
         break;

      case FIO_SDD_WCMD:
         {
            SDD_IO_MB_CMD_t *pCmd = IFXOS_BlockAlloc(sizeof(SDD_IO_MB_CMD_t));
            SDD_ASSERT (pCmd != IFX_NULL);
             IFXOS_CpyFromUser((char*)pCmd,
                              (IFX_uint8_t*)ioarg, sizeof (SDD_IO_MB_CMD_t));
            ret = SDD_Write_Cmd (pDev, pCmd);
            IFXOS_BlockFree(pCmd);
         }
         break;

      case FIO_SDD_RCMD:
         {
            SDD_IO_MB_CMD_t *pCmd = IFXOS_BlockAlloc(sizeof(SDD_IO_MB_CMD_t));
            SDD_ASSERT (pCmd != IFX_NULL);
            IFXOS_CpyFromUser((char*)pCmd,
                              (IFX_uint8_t*)ioarg, sizeof (SDD_IO_MB_CMD_t));
            ret = SDD_Read_Cmd (pDev, pCmd);
            if (ret != IFX_SUCCESS)
            {
               TRACE(SDD, DBG_LEVEL_HIGH,
                      ("Command read error (code 0x%08x)\n", ret));
            }
            else
            {
               IFXOS_CpyToUser((IFX_uint8_t*)ioarg,
                                 pCmd, sizeof (SDD_IO_MB_CMD_t));
            }

            IFXOS_BlockFree(pCmd);
         }
         break;

      case FIO_SDD_MB_READ:
         {
            SDD_FIRMWARE_IF_t *p = pDev->pMbxIF;

            if (p == IFX_NULL)
            {
               /*errmsg: The SDD MBX driver is not registered. */
               RETURN_DEVSTATUS (SDD_statusMbxDrvNotRegistered);
            }
            p->ioctl(SDD_MBX_EVT_READ, (IFX_void_t *)ioarg);
         }
         break;

      case FIO_SDD_RREG:
         {
            SDD_FIRMWARE_IF_t *p = pDev->pMbxIF;

            if (p == IFX_NULL)
            {
               /*errmsg: The SDD MBX driver is not registered. */
               RETURN_DEVSTATUS (SDD_statusMbxDrvNotRegistered);
            }
            p->ioctl(SDD_MBX_REG_READ, (IFX_void_t *)ioarg);
         }
         break;

      case FIO_SDD_WREG:
         {
            SDD_FIRMWARE_IF_t *p = pDev->pMbxIF;

            if (p == IFX_NULL)
            {
               /*errmsg: The SDD MBX driver is not registered. */
               RETURN_DEVSTATUS (SDD_statusMbxDrvNotRegistered);
            }
            p->ioctl(SDD_MBX_REG_WRITE, (IFX_void_t *)ioarg);
         }
         break;

      case FIO_SDD_RESET:
         {
            SDD_FIRMWARE_IF_t *p = pDev->pMbxIF;

            if (p == IFX_NULL)
            {
               /*errmsg: The SDD MBX driver is not registered. */
               RETURN_DEVSTATUS (SDD_statusMbxDrvNotRegistered);
            }
            p->ioctl(SDD_MBX_RESET, IFX_NULL);
         }
         break;

      case FIO_SDD_DEBUG:
         {
            SDD_FIRMWARE_IF_t *p = pDev->pMbxIF;

            if (p == IFX_NULL)
            {
               /*errmsg: The SDD MBX driver is not registered. */
               RETURN_DEVSTATUS (SDD_statusMbxDrvNotRegistered);
            }
            p->ioctl(SDD_MBX_DEBUG_MODE, (IFX_void_t *)ioarg);
         }
         break;

      default:
         /* errmsg: Ioctl not supported. */
         RETURN_DEVSTATUS(SDD_statusNotSupported);
         break;
   }

   return ret;
}
