/*******************************************************************************

                            Copyright (c) 2014-2015
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_linux.c
   This file contains the implementation of linux specific driver functions.

*/

#ifdef LINUX
/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "drv_sdd_config.h"

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/wait.h>
#include <linux/sched.h>
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,39))
#include <linux/smp_lock.h>
#endif
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,33))
      #include <linux/utsrelease.h>
#else
      #include <generated/utsrelease.h>
#endif
#ifdef SDD_FEAT_PROCFS
   #include <linux/proc_fs.h>
   #include <linux/seq_file.h>
#endif /* SDD_FEAT_PROCFS */
#ifdef SDD_FEAT_DEVFS
   #include <linux/cdev.h>
#endif /* SDD_FEAT_DEVFS */

#include "ifx_types.h"

#include "drv_sdd_osmap.h"
#include "drv_sdd_global.h"
#include "drv_sdd_init.h"
#include "drv_sdd_fw_cmd.h"
#include "drv_sdd_opmode.h"
#include "drv_sdd_alm_priv.h"
#include "drv_sdd_pcm.h"
#include "drv_sdd_api.h"

/* ========================================================================== */
/*                             Global variables                               */
/* ========================================================================== */

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */

#if 0
MODULE_DESCRIPTION("SDD device driver");
MODULE_AUTHOR("Lantiq Beteiligungs-GmbH & Co.KG - www.lantiq.com");
MODULE_SUPPORTED_DEVICE("SDD");
MODULE_LICENSE("Dual BSD/GPL");
#endif

#ifdef SDD_FEAT_DEVFS
/* first minor number */
#define LQ_SDD_FIRST_MINOR 1
/* total file descriptor number */
#define LQ_SDD_TOTAL_FD    1
/* device name */
#define LQ_SDD_DEV_NAME "ltq_sdd"
#endif /* SDD_FEAT_DEVFS */

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */
#ifdef SDD_FEAT_DEVFS
IFX_int32_t sdd_open (struct inode * inode, struct file * file_p);
IFX_int32_t sdd_close (struct inode * inode, struct file * file_p);
static IFX_uint32_t sdd_poll (struct file *file_p, poll_table * wait);
long sdd_ioctl (struct file *file_p, IFX_uint32_t nCmd, IFX_ulong_t arg);
#endif /* SDD_FEAT_DEVFS */

/* ========================================================================== */
/*                             Local variables                                */
/* ========================================================================== */
#ifdef SDD_FEAT_DEVFS
static ushort sdd_major_id = 0;

/* the driver callbacks */
static struct file_operations sdd_fops = {
   owner: THIS_MODULE,
   poll: sdd_poll,
   unlocked_ioctl: sdd_ioctl,
   open: sdd_open,
   release: sdd_close
};
#endif /* SDD_FEAT_DEVFS */

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */
#ifdef SDD_FEAT_DEVFS
/**
 * Open SDD device.
 * Open the device from user mode (e.g. application). An inode
 * value of 0 can be used only.
 *
 * \param   inode   Pointer to device inode
 * \param   file_p  Pointer to file descriptor
 * \return  0       IFX_SUCCESS, device opened
 * \return  EMFILE  Device already open
 * \return  EINVAL  Invalid minor ID
 * \ingroup API
 */
IFX_int32_t sdd_open (struct inode * inode, struct file * file_p)
{
   SDD_DEVICE_t* pDev;

   if ( SDD_GetDevice (MINOR(inode->i_rdev), &pDev) != SDD_statusOk )
      return -EINVAL;
   if (pDev == IFX_NULL)
      return -EINVAL;

   pDev->bIntercept = IFX_TRUE;
   if (IFX_SUCCESS !=
       IFXOS_DrvSelectQueueInit (&pDev->sdd_wakeup))
   {
      TRACE (SDD, DBG_LEVEL_HIGH,
             ("%s(): init_waitqueue_head error\r\n", __FUNCTION__));
   }

   /* use file_p->private_data to point to the device data */
   file_p->private_data = pDev;

   /* initialize events intercepter mutex */
   IFXOS_MutexInit(&pDev->mtxIntercept);
   IFXOS_EventInit (&pDev->readEvent);

   return 0;
}


/**
 * Close SDD device.
 * Close the device from user mode (e.g. application). An inode
 * value of 0 used only.
 *
 * \param   inode   Pointer to device inode
 * \param   file_p  Pointer to file descriptor
 * \return  0       IFX_SUCCESS, device closed
 * \return  ENODEV  Device invalid
 * \return  EINVAL  Invalid minor ID
 * \ingroup API
 */
IFX_int32_t sdd_close (struct inode * inode, struct file * file_p)
{
   SDD_DEVICE_t* pDev;

   if ( SDD_GetDevice (MINOR(inode->i_rdev), &pDev) != SDD_statusOk )
      return -EINVAL;
   if (pDev == IFX_NULL)
      return -EINVAL;

   pDev->bIntercept = IFX_FALSE;

   /* delete mutex */
   IFXOS_MutexDelete (&pDev->mtxIntercept);
   IFXOS_EventDelete (&pDev->readEvent);

   if ((IFX_int32_t) inode > 0)
   {
      return -EINVAL;
   }
   else
   {
      return 0;
   }
}


/**
 * Poll handler.
 * The select function of the driver. A user space program may sleep until
 * the driver wakes it up.
 *
 * \param   file_p  File structure of device
 * \param   wait    Internal table of poll wait queues
 * \return  mask    If new data is available the POLLPRI bit is set,
 *                  triggering an exception indication. If the device pointer
 *                  is null POLLERR is set.
 * \ingroup API
 */
static IFX_uint32_t sdd_poll (struct file *file_p, poll_table * wait)
{
   SDD_DEVICE_t* pDev = file_p->private_data;
   IFX_uint32_t mask;

   /* add to poll queue */
   IFXOS_DrvSelectQueueAddTask ((IFXOS_drvSelectOSArg_t *) file_p,
                                &(pDev->sdd_wakeup),
                                (IFXOS_drvSelectTable_t *) wait);
   mask = 0;
   if (pDev->bWakeUp == IFX_TRUE)
   {
      mask = POLLPRI;
   }
   return mask;
}

/**
 * SDD IOCTL handler.
 * \param   inode        Inode of device
 * \param   file_p       File structure of device
 * \param   nCmd         IOCTL command
 * \param   arg          Argument for some IOCTL commands
 * \return  0            Setting the LED bits was successfull
 * \return  -EINVAL      Invalid minor ID
 * \return  -ENOIOCTLCMD Invalid command
 * \ingroup API
 */
long sdd_ioctl (struct file *file_p, IFX_uint32_t nCmd, IFX_ulong_t arg)
{
   IFX_int32_t retvalue;
   SDD_DEVICE_t* pDev = file_p->private_data;

   retvalue = SDD_Dev_Spec_Ioctl ((IFX_TAPI_LL_CH_t *)pDev, nCmd, arg);
   return retvalue;
}
#endif /* SDD_FEAT_DEVFS */

/**
   This function allocates a temporary kernel buffer
      and copies the user buffer contents to it.

   \param  p_buffer     Pointer to the user buffer.
   \param  size         Size of the buffer.

   \return
   Pointer to kernel buffer or NULL.

   \remarks
   Buffer has to be released with \ref OS_UnmapBuffer.
*/
IFX_void_t* OS_MapBuffer(IFX_void_t* p_buffer, IFX_size_t size)
{
   IFX_void_t* kernel_buffer;

   if (p_buffer == IFX_NULL)
   {
      return IFX_NULL;
   }
   kernel_buffer = IFXOS_BlockAlloc(size);
   if (kernel_buffer == IFX_NULL)
   {
      return IFX_NULL;
   }

   if (IFXOS_CpyFromUser(kernel_buffer, p_buffer, size) == IFX_NULL)
   {
      IFXOS_BlockFree(kernel_buffer);
      return IFX_NULL;
   }

   return kernel_buffer;
}

/**
   This function releases a temporary kernel buffer.

   \param  p_buffer     Pointer to the kernel buffer.

   \remarks
   Counterpart of \ref OS_MapBuffer.
*/
IFX_void_t OS_UnmapBuffer(IFX_void_t* p_buffer)
{
   if (p_buffer != IFX_NULL)
   {
      IFXOS_BlockFree(p_buffer);
      /* p_buffer = IFX_NULL; */
   }
}

/**
   Waiting for completion of SDD Operating Mode change or timeout.

   \param  pCh          Pointer to SDD channel context .

   \return  >0      wakeup occured, remaining jiffies
            0       timeout
            <0      interrupted by a signal
*/
IFX_int32_t SDD_WaitForSddOpmodeChEvt(SDD_CHANNEL_t *pCh)
{
   SDD_FIRMWARE_IF_t *p = pCh->pDev->pMbxIF;

   return wait_event_interruptible_timeout(pCh->pALM->sdd_event.object,
                                           !pCh->pALM->bOpmodeChangePending,
                                           (SDD_EVT_TIMEOUT_MS * HZ / 1000) +
                                           p->path_delay());
}

#ifdef SDD_FEAT_PROCFS
/* ========================================================================== */
/*                           Proc filesystem support                          */
/* ========================================================================== */

/**
   Read the version information from the driver.

   \param   s        Pointer to seq_file struct.
   \return  0 on success
*/
static IFX_int32_t sdd_proc_VersionGet (struct seq_file *s)
{
   IFX_uint8_t          dev;
   SDD_DEVICE_t         *pDev    =  IFX_NULL;

   seq_printf (s, "%s\n",
                  &DRV_SDD_WHATVERSION[4]);
   #ifdef TAPI_FEAT_USE_TIME_DATE
      seq_printf(s, "Compiled on %s, %s for Linux kernel %s\n",
         __DATE__, __TIME__, UTS_RELEASE);
   #else
      seq_printf(s, "Compiled for Linux kernel %s\n",
         UTS_RELEASE);
   #endif

   seq_printf (s, "Supporting only %s family devices\n",
#ifdef S220
                  "SLIC 200"
#else /* S220 */
                  "SLIC 100"
#endif /* S220 */
              );

   for (dev = 0; dev < SDD_MAX_DEVICES; dev++)
   {
      SDD_GetDevice(dev, &pDev);

      if ((pDev != IFX_NULL) &&
          (pDev->nDevState & DS_BASIC_INIT) &&
          (pDev->nDevState & DS_DEV_INIT))
      {
         seq_printf(s, "Device #%d: "
                     "SmartSLIC ID 0x%02x, Rev 0x%02x, mounted DC/DC Type %d "
                     "FW %d.%d.%d.%d\n",
                     dev,
                     pDev->nSlicDevId,
                     pDev->nSlicRevision,
                     pDev->nAllowedDcDcType - SDD_DCDC_TYPE_DEFAULT_IBB,
                     pDev->nFwVersion[0], pDev->nFwVersion[1],
                     pDev->nFwVersion[2], pDev->nFwVersion[3]);
      }
   }

   return 0;
}

/**
   Read the device status from the driver.

   \param  s            Pointer to seq_file struct.

   \return  0 on success
*/
static IFX_int32_t sdd_proc_StatusGet (
                        struct seq_file *s)
{
   IFX_uint8_t          nDev;

   /* Loop over all devices */
   for (nDev = 0; nDev < SDD_MAX_DEVICES; nDev++)
   {
      SDD_DEVICE_t *pDev = IFX_NULL;

      if (SDD_GetDevice (nDev, &pDev) == IFX_SUCCESS)
      {
         /* Headline */
         seq_printf(s, "-- Device: #%d --\n", nDev);

         SDD_ALM_proc_StatusGet (s, pDev);
         SDD_PCM_proc_StatusGet (s, pDev);
      }
   }

   return 0;
}

typedef void (*vmmc_dump) (struct seq_file *s);

static int sdd_proc_show ( struct seq_file *s, void *p )
{
   vmmc_dump dump = s->private;

   if (dump != NULL)
      dump(s);

   return 0;
}

static int sdd_proc_open ( struct inode *inode, struct file *file )
{
   return single_open (file, sdd_proc_show, PDE_DATA(inode));
}

struct proc_entry
{
   const char *name;
   void *read_function;
   void *write_function;
   struct file_operations fops;
};

static struct proc_entry proc_entries[] =
{
   { "version", sdd_proc_VersionGet },
   { "status", sdd_proc_StatusGet },
};

static void sdd_proc_EntryCreate (
                        struct proc_dir_entry *parent_node,
                        struct proc_entry *proc_entry)
{
   memset(&proc_entry->fops, 0, sizeof(struct file_operations));
   proc_entry->fops.owner   = THIS_MODULE;
   proc_entry->fops.open    = sdd_proc_open;
   proc_entry->fops.read    = seq_read;
   proc_entry->fops.write   = proc_entry->write_function;
   proc_entry->fops.llseek  = seq_lseek;
   proc_entry->fops.release = single_release;

   proc_create_data ( proc_entry->name, 0, parent_node,
            &proc_entry->fops, proc_entry->read_function);
}


/**
   Initialize and install the /proc filesystem entries.

   \return
   IFX_SUCCESS or IFX_ERROR
*/
static int sdd_proc_EntriesInstall (void)
{
   struct proc_dir_entry *driver_proc_node;

   /* install the proc entry */
   TRACE(SDD, DBG_LEVEL_NORMAL, ("sdd: using proc fs\n"));

   driver_proc_node = proc_mkdir( "driver/" SDD_DEV_NAME, NULL);
   if (driver_proc_node != NULL)
   {
      IFX_uint32_t i;

      for(i=0; i<sizeof(proc_entries)/sizeof(proc_entries[0]); i++) {
         sdd_proc_EntryCreate (driver_proc_node, &proc_entries[i]);
      }
   }
   else
   {
      TRACE(SDD, DBG_LEVEL_HIGH,("sdd: cannot create proc entry\n"));
      return IFX_ERROR;
   }

   return IFX_SUCCESS;
}


/**
   Remove proc filesystem entries.
*/
static void sdd_proc_EntriesRemove(void)
{
   remove_proc_entry("driver/" SDD_DEV_NAME "/version", 0);
   remove_proc_entry("driver/" SDD_DEV_NAME "/status", 0);
   remove_proc_entry("driver/" SDD_DEV_NAME , 0);

   return;
}

#endif /* SDD_FEAT_PROCFS */

/**
   Stop a kernel thread. Called by the removing instance.

   \param  pThrCntrl    Pointer to a thread control struct.
*/
IFX_void_t SDD_OS_ThreadKill(IFXOS_ThreadCtrl_t *pThrCntrl)
{
   if ((pThrCntrl) &&
       (IFXOS_THREAD_INIT_VALID(pThrCntrl) == IFX_TRUE) &&
       (pThrCntrl->thrParams.bRunning == 1))
   {
      /* signal the thread routine to shutdown */
      pThrCntrl->thrParams.bShutDown = IFX_TRUE;
      mb();
      /* Send a signal to wake the process when it is blocked but
         interruptible. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
      kill_proc(pThrCntrl->tid, SIGKILL, 1);
#else
#if IFXOS_VERSION_CHECK_L_THAN(1,5,19)
      kill_pid(find_vpid(pThrCntrl->tid), SIGKILL, 1);
#else
      kill_pid(get_task_pid(pThrCntrl->tid, PIDTYPE_PID), SIGKILL, 1);
#endif /* IFXOS_VERSION_CHECK_L_THAN(1, 5, 19) */
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28) */
      wait_for_completion (&pThrCntrl->thrCompletion);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
      /* Now that we are sure the thread is in zombie state.
         We notify keventd to clean the process up. */
      kill_proc(2, SIGCHLD, 1);
#endif /* kernel 2.6.23 */
   }
}


/* ========================================================================== */
/*                            Linux module support                            */
/* ========================================================================== */

/**
   Initialize the module.

   \return 0 if successful or an negative error code otherwise.

   \remarks
   Called by the kernel.
*/
IFX_int32_t SDD_module_init(void)
{
   IFX_int32_t ret;
#ifdef SDD_FEAT_DEVFS
   dev_t       dev;
#endif /* SDD_FEAT_DEVFS */

   ret = SDD_DeviceDriverStart();

#ifdef SDD_FEAT_PROCFS
   if (ret == 0)
   {
      sdd_proc_EntriesInstall();
   }
#endif /* SDD_FEAT_PROCFS */

#ifdef SDD_FEAT_DEVFS
   if (sdd_major_id)
   {
      dev = MKDEV(sdd_major_id, LQ_SDD_FIRST_MINOR);
      ret = register_chrdev_region(dev, LQ_SDD_TOTAL_FD, LQ_SDD_DEV_NAME);
   }
   else
   {
      /* dynamic major */
      ret = alloc_chrdev_region(&dev, LQ_SDD_FIRST_MINOR, LQ_SDD_TOTAL_FD, LQ_SDD_DEV_NAME);
      sdd_major_id = MAJOR(dev);
   }
   if (ret < 0)
   {
      TRACE (SDD, DBG_LEVEL_HIGH,
             ("IFX_SDD: can't get major %d\n", sdd_major_id));
      return ret;
   } else {
      struct cdev *p_cdev = cdev_alloc();

      if (NULL == p_cdev)
         return -ENOMEM;

      cdev_init(p_cdev, &sdd_fops);
      p_cdev->owner = THIS_MODULE;

      ret = cdev_add(p_cdev, MKDEV(sdd_major_id, 0), 1);
      if (ret != 0)
      {
         cdev_del (p_cdev);
         return -EPERM;
      }
   }
#endif /* SDD_FEAT_DEVFS */

   /* Report error codes as negative numbers. */
   return (0 - ret);
}

/**
   Clean up the module when unloaded.

   \remarks
   Called by the kernel.
*/
IFX_void_t SDD_module_exit(void)
{
#ifdef SDD_FEAT_PROCFS
   sdd_proc_EntriesRemove();
#endif /* SDD_FEAT_PROCFS */

   SDD_DeviceDriverStop();

   TRACE (SDD, DBG_LEVEL_LOW, ("Cleaned up %s module.\n", SDD_DEV_NAME));
}
#endif /* LINUX */
