/*******************************************************************************

  Copyright (c) 2013-2015 Lantiq Deutschland GmbH
  Copyright (c) 2015 Lantiq Beteiligungs-GmbH & Co.KG
  Copyright 2016, Intel Corporation.

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

*******************************************************************************/

/**
   \file drv_sdd_opmode.c
   This file contains the implementations of tapi low level functions for
   the ALM module.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */
#include "drv_sdd_osmap.h"
#include "drv_sdd_global.h"
#include "drv_sdd_alm_priv.h"
#include "drv_sdd_init.h"
#include "drv_sdd_opmode.h"
#ifdef SDD_FEAT_GR909
   #include "drv_sdd_gr909.h"
#endif /* SDD_FEAT_GR909 */
#include "drv_sdd_fxo.h"
#include "drv_sdd_calib.h"


/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */

/**
   Represents the minimal allowed value for programming the ALM Rx relative
   gain, due to TAPI limitation (device allows up to -89dB)
*/
#define ALM_RX_GAIN_MIN         -24

/**
   Represents the maximum allowed value for programming the ALM Rx relative gain
   0.55 dB rounded to 1 dB
*/
#define ALM_RX_GAIN_MAX         1

/**
   Represents the minimal allowed value for programming the ALM Tx relative gain
*/
#define ALM_TX_GAIN_MIN         -4

/**
   Represents the maximum allowed value for programming the ALM Tx relative
   gain due to TAPI limitation (device allows up to 86dB)
*/
#define ALM_TX_GAIN_MAX         24

/* ========================================================================== */
/*                              Local variables                               */
/* ========================================================================== */
/** Used to get right table index for dB -> HEX conversion. */
static const IFX_int32_t SDD_ALM_TXGAIN_OFFSET = 4;

/**
   Table with calculated ALM TX gain coefficients. It
   represents values between -4dB and 24dB, in steps of 1dB
   calculated with the formula below.

   TxGain = 20224 * 10^(-Lx[dB]/20)
*/
static const IFX_uint16_t SDD_AlmTxGain[] =
{
   0x7D35, 0x6F97, 0x6375, 0x58A4, 0x4F00,   /* -4dB  -3dB  -2dB  -1dB  0dB   */
   0x4669, 0x3EC0, 0x37ED, 0x31D8, 0x2C6D,   /*  1dB   2dB   3dB   4dB  5dB   */
   0x2798, 0x234A, 0x1F73, 0x1C08, 0x18FB,   /*  6dB   7dB   8dB   9dB 10dB   */
   0x1644, 0x13D8, 0x11B0, 0x0FC3, 0x0E0C,   /* 11dB  12dB  13dB  14dB 15dB   */
   0x0C85, 0x0B29, 0x09F2, 0x98DD, 0x07E6,   /* 16dB  17dB  18dB  19dB 20dB   */
   0x070A, 0x0646, 0x0598, 0x04FC            /* 21dB  22dB  23dB  24dB        */
};

/** Used to get right table index for dB -> HEX conversion. */
static const IFX_int32_t SDD_ALM_RXGAIN_OFFSET = 24;

/**
   Table with calculated ALM RX gain coefficients. It
   represents values between -24dB and 1dB, in steps of 1dB
   calculated with the formula below.

   TxGain = 30720 * 10^(Lr[dB]/20)
*/
static const IFX_uint16_t SDD_AlmRxGain[] =
{
   0x0792, 0x087F, 0x0988, 0x0AB2, 0x0C00,   /* -24dB -23dB -22dB -21dB -20dB */
   0x0D77, 0x0F1B, 0x10F3, 0x1305, 0x1557,   /* -19dB -18dB -17dB -16dB -15dB */
   0x17F1, 0x1ADD, 0x1E25, 0x21D2, 0x25F3,   /* -14dB -13dB -12dB -11dB -10dB */
   0x2A94, 0x2FC6, 0x359A, 0x3C24, 0x437B,   /*  -9dB  -8dB  -7dB  -6dB  -5dB */
   0x4BB7, 0x54F4, 0x5F52, 0x6AF3, 0x7800,   /*  -4dB  -3dB  -2dB  -1dB   0dB */
   0x7FFF                                    /*   1dB                         */
};

/* Type of one element in the array below. */
struct DcDcTranslation
{
   const char * const name;
   enum SDD_DCDC_TYPE type;
   const char * const description;
};

/* Translation table for DC/DC converter type strings into corresponding
   enum value or full description. */
static const struct DcDcTranslation DcDcTypes[] =
{
   {"IBB",   SDD_DCDC_TYPE_IBB,
    "Inverting Buck-Boost Converter"},
   {"CIBB",  SDD_DCDC_TYPE_CIBB,
    "Combined Inverting Buck-Boost Converter"},
   {"IB",    SDD_DCDC_TYPE_IB,
    "Dedicated Inverting Boost Converter"},
   {"CIB",   SDD_DCDC_TYPE_CIB,
    "Combined Inverting Boost Converter"},
   {"BB",    SDD_DCDC_TYPE_BB,
    "Dedicated Buck-or-Boost Converter"},
   {"CBB",   SDD_DCDC_TYPE_CBB,
    "Combined Buck-or-Boost Converter"},
   {"IFB",   SDD_DCDC_TYPE_IFB,
    "Dedicated Inverting Flyback Converter"},
   {"CIFB",  SDD_DCDC_TYPE_CIFB,
    "Combined Inverting Flyback Converter"},
   {"IBGD",  SDD_DCDC_TYPE_IBGD,
    "Dedicated Inverting Boost Converter with Gate Driver"},
   {"CIBGD", SDD_DCDC_TYPE_CIBGD,
    "Combined Inverting Boost Converter with Gate Driver"},
   {"IBVD",  SDD_DCDC_TYPE_IBVD,
    "Dedicated Inverting Boost Converter with Voltage Doubler"},
   {"CIBVD", SDD_DCDC_TYPE_CIBVD,
    "Combined Inverting Boost Converter with Voltage Doubler"},
#ifdef S220
   {"",      SDD_DCDC_TYPE_NOTSET,
    "NONE (default)"}
#else /* S220 */
   {"",      SDD_DCDC_TYPE_DEFAULT_IBB,
    "DEFAULT (Inverting Buck-Boost Converter)"}
#endif /* S220 */
};

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */
static IFX_int32_t sdd_alm_Line_Mode_Set(
                        SDD_CHANNEL_t *pCh,
                        IFX_int32_t nMode);

static IFX_int32_t SDD_TAPI_LL_ALM_Line_Mode_Set (
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_int32_t nMode,
                        IFX_uint8_t nTapiLineMode);

static IFX_int32_t SDD_TAPI_LL_ALM_Line_Mode_Get (
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_LINE_FEED_t *pFeed);

#ifdef SDD_FEAT_PROCFS
static const char *sdd_alm_DcDcEnum2Description (
                        enum SDD_DCDC_TYPE dcdc_type);
#endif /* SDD_FEAT_PROCFS */

static IFX_int32_t SDD_TAPI_LL_ALM_SlicRecovery (
                        IFX_TAPI_LL_DEV_t *pLLDev);

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */

/**
   Get the current opmode but defer while a opmode change is in progress.

   \param  pCh             Pointer to the SDD channel structure.
   \param  pCurrentOpmode  Pointer to variable returning the current opmode.

   \return
   - SDD_statusOk                   if successful
   - SDD_statusSddEvtWaitTmout      timeout waiting for an event from SDD
   - SDD_statusSddEvtWaitInterrupt  waiting for event from SDD is interrupted
                                    by a signal
*/
IFX_int32_t SDD_ALM_OpmodeGet (SDD_CHANNEL_t *pCh,
                               IFX_uint8_t *pCurrentOpmode)
{
   IFX_int32_t timeout;

   /* Wait while a linemode change is pending to avoid returning an outdated
      linemode. When the waiting was aborted by a signal or a timeout occured
      while waiting inform about this in the return code. */
   timeout = SDD_WaitForSddOpmodeChEvt(pCh);

   /* Return the current opmode regardless of the result of the waiting. */
   *pCurrentOpmode = pCh->pALM->curr_opmode;

   if (timeout > 0)
   {
      /* Positive values indicate the wait ended because the expected event
         occured. */
      return SDD_statusOk;
   }
   else if (timeout == 0)
   {
      /* A return value of zero indicates that the timeout expired. */
      /* errmsg: Timeout waiting for an event from SDD. */
      return SDD_statusSddEvtWaitTmout;
   }

   /* All negative values indicate that the waiting was interrupted by a
      signal from the OS. */
   /* errmsg: Waiting for an event from SDD interrupted by a signal. */
   return SDD_statusSddEvtWaitInterrupt;
}


/**
   Write the SDD_Opmode command and wait for opmode change event from
   firmware. Possible waiting results are - success, timeout and interrupted
   wait. Success means the complete opmode transition within SDD.

   \param  pCh             Pointer to the SDD channel structure.

   \return
   - SDD_statusCmdWr               writing the command has failed
   - SDD_statusSddEvtWaitTmout     timeout waiting for an event from SDD
   - SDD_statusSddEvtWaitInterrupt waiting for event from SDD is interrupted
                                    by a signal
   - SDD_statusOk                  if successful

*/
IFX_int32_t SDD_ALM_OpModeSet (SDD_CHANNEL_t *pCh)
{
   SDD_Opmode_t *pOpmod = &pCh->pALM->sdd_opmode;
   IFX_int32_t ret;
   IFX_uint8_t curr_opmode;

   /* Get the current opmode - wait while a opmode change is ongoing. */
   ret = SDD_ALM_OpmodeGet (pCh, &curr_opmode);
   /* Only exit when interrupted by a signal. When waiting was aborted by the
      timeout assume that no linemode change is pending and try to continue
      with the opmode change that was requested. */
   if (ret == SDD_statusSddEvtWaitInterrupt)
   {
      return ret;
   }
   /* do nothing if parameters didn't change */
   if (pOpmod->OperatingMode != SDD_Opmode_Disabled &&
       pOpmod->OperatingMode == curr_opmode
#ifndef S220
       &&
       pCh->pALM->curr_polarity == pOpmod->Polarity &&
       pCh->pALM->curr_howler == pOpmod->Howler &&
       SDD_Opmode_TTX_NO == pOpmod->TtxBurst &&
       pCh->pALM->curr_mwi_lamp == pOpmod->MwiLamp
#endif /* S220 */
      )
   {
      return SDD_statusOk;
   }

   /* Only when the opmode changes wait for the opmode change event. */
   if (pOpmod->OperatingMode != curr_opmode)
   {
      pCh->pALM->bOpmodeChangePending = IFX_TRUE;
   }
   else
   {
      pCh->pALM->bOpmodeChangePending = IFX_FALSE;
   }

#ifdef S220
   ret = SDD_CmdWrite(pCh->pDev, (IFX_uint32_t *)pOpmod);
#else /* S220 */
   if (pOpmod->OperatingMode == SDD_Opmode_Disabled)
   {
      /* clear other fields of SDD_Opmode */
      pOpmod->Polarity = SDD_Opmode_Pol_Normal;
      pOpmod->Howler = SDD_Opmode_HT_Speech;
      /* pOpmod->TtxBurst = SDD_Opmode_TTX_NO; handled below */
   }

   if (pOpmod->OperatingMode != SDD_Opmode_Ringing)
   {
      /* Allow MWL only with linefeed ringing. */
      pOpmod->MwiLamp = SDD_Opmode_MWL_Off;
   }

   if (pOpmod->OperatingMode != SDD_Opmode_Active)
   {
      /* Allow TTX only with linefeed active. */
      pOpmod->TtxBurst = SDD_Opmode_TTX_NO;
   }

   ret = SDD_CmdWrite(pCh->pDev, (IFX_uint32_t *)pOpmod);

   /* Automatically reset the TTX flag. */
   pOpmod->TtxBurst = SDD_Opmode_TTX_NO;

   if (SDD_SUCCESS(ret))
   {
      /* store current status - pCh->pALM->curr_opmode will be updated by ISR */
      pCh->pALM->curr_polarity = pOpmod->Polarity;
      pCh->pALM->curr_howler = pOpmod->Howler;
      pCh->pALM->curr_mwi_lamp = pOpmod->MwiLamp;
   }
#endif /* S220 */

   return ret;
}

/** Set the phone volume

   Gain Parameter are given in 'dB'. The range is -24dB ... 24dB.
   This function enables the ALM in firmware. There will be
   no error reported if it is not enabled before (IFX_TAPI_CH_INIT done).

   \param pLLChannel Handle to TAPI low level channel structure

   \param pVol       Handle to IFX_TAPI_LINE_VOLUME_t structure

   \return
   - SDD_statusOk          if successful
   - SDD_statusTxGainErr   Wrong parameters passed. This code is returned
      when any gain parameter is lower than -24 dB or higher 24 dB than
   - SDD_statusRxGainErr   Wrong parameters passed. This code is returned
      when any gain parameter is lower than -24 dB or higher 24 dB than
   - SDD_statusAlmVolErr   Writing the settings failed.
*/
static IFX_int32_t SDD_TAPI_LL_ALM_Volume_Set(
                   IFX_TAPI_LL_CH_t *pLLChannel,
                   IFX_TAPI_LINE_VOLUME_t const *pVol)
{
   IFX_int32_t          err   = SDD_statusOk;
   SDD_CHANNEL_t        *pCh;
   SDD_DEVICE_t         *pDev;
   SDD_TxRxGain_t       *pSddTxRxGain;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   if ((pVol->nGainRx > ALM_RX_GAIN_MAX) ||
       (pVol->nGainRx < ALM_RX_GAIN_MIN))
   {
      /* parameter is out of supported range */
      TRACE (SDD, DBG_LEVEL_HIGH,
             ("SDD_ERR: RxGain Level out of range (%d), "
              "(allowed range %d..%d dB)\n",
              pVol->nGainRx, ALM_RX_GAIN_MIN, ALM_RX_GAIN_MAX));
      /* errmsg: ALM RxGain parameter out of range.*/
      RETURN_STATUS(SDD_statusRxGainErr);
   }

   if ((pVol->nGainTx > ALM_TX_GAIN_MAX) ||
       (pVol->nGainTx < ALM_TX_GAIN_MIN))
   {
      /* parameter is out of supported range */
      TRACE (SDD, DBG_LEVEL_HIGH,
             ("SDD_ERR: TxGain Level out of range (%d), "
              "(allowed range %d..%d dB)\n",
              pVol->nGainTx, ALM_TX_GAIN_MIN, ALM_TX_GAIN_MAX));
      /* errmsg: ALM TxGain parameter out of range.*/
      RETURN_STATUS(SDD_statusTxGainErr);
   }

   /* protect channel from mutual access */
   SDD_OS_MutexGet (&pCh->mtxChAcc);

   pDev = pCh->pDev;
   pSddTxRxGain = &pCh->pALM->sdd_txrx_gain;

   /* get the values for RX and TX gain from table */
   pSddTxRxGain->RxGain = SDD_AlmRxGain[pVol->nGainRx + SDD_ALM_RXGAIN_OFFSET];
   pSddTxRxGain->TxGain = SDD_AlmTxGain[pVol->nGainTx + SDD_ALM_TXGAIN_OFFSET];
   err = SDD_CmdWrite (pDev, (IFX_uint32_t *)pSddTxRxGain);

   /* release channel lock */
   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   if (!SDD_SUCCESS (err))
   {
      /* errmsg: Writing the ALM volume failed.*/
      RETURN_STATUS(SDD_statusAlmVolErr);
   }
   return SDD_statusOk;
}

/**
   This service enables or disables a high level path of a phone channel.

   It is intended for phone channels only and must be used in
   combination with IFX_TAPI_PHONE_VOLUME_SET or IFX_TAPI_PCM_VOLUME_SET
   to set the max. level (IFX_TAPI_LINE_VOLUME_HIGH) or to restore level

   \param  pLLCh        Pointer to the TAPI LL channel structure.
   \param  bEnable      The parameter represent a boolean value of
                        \ref IFX_TAPI_LINE_LEVEL_t.
                        - 0: IFX_TAPI_LINE_LEVEL_DISABLE, disable the
                             high level path.
                        - 1: IFX_TAPI_LINE_LEVEL_ENABLE, enable the
                             high level path.
   \return
   - SDD_statusCmdWr Writing the command failed
   - SDD_statusOk if successful
*/
static IFX_int32_t SDD_TAPI_LL_ALM_Volume_High_Level (IFX_TAPI_LL_CH_t *pLLCh,
                                                      IFX_int32_t bEnable)
{
   SDD_CHANNEL_t     *pCh;
   SDD_Opmode_t      *pOpmod;
   IFX_int32_t       ret;

   SDD_ASSERT(pLLCh != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLCh)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   pOpmod = &pCh->pALM->sdd_opmode;

#ifdef S220
   switch (pCh->pALM->curr_opmode)
   {
      case SDD_Opmode_Active:                   /*lint -fallthrough */
      case SDD_Opmode_Active_HOWL:
         pOpmod->OperatingMode =
            bEnable != IFX_TAPI_LINE_LEVEL_DISABLE ?
            SDD_Opmode_Active_HOWL : SDD_Opmode_Active;
         break;

      case SDD_Opmode_Active_REVPOL:            /*lint -fallthrough */
      case SDD_Opmode_Active_HOWL_REVPOL:
         pOpmod->OperatingMode =
            bEnable != IFX_TAPI_LINE_LEVEL_DISABLE ?
            SDD_Opmode_Active_HOWL_REVPOL : SDD_Opmode_Active_REVPOL;
         break;

      default:
         if (bEnable != IFX_TAPI_LINE_LEVEL_DISABLE)
         {
            /* errmsg: Setting of high level output requires active feeding */
            RETURN_STATUS(SDD_statusHighLevelNotActive);
         }
         break;
   }
#else /* S220 */
   if (bEnable != IFX_TAPI_LINE_LEVEL_DISABLE)
   {
      /* Howler tone can only be set when linefeeding is active. */
      if (pCh->pALM->curr_opmode != SDD_Opmode_Active)
      {
         /* errmsg: Setting of high level output requires active feeding */
         RETURN_STATUS(SDD_statusHighLevelNotActive);
      }
      /* howler-tone on */
      pOpmod->Howler = SDD_Opmode_HT_Howler;
   }
   else
   {
      /* howler-tone off */
      pOpmod->Howler = SDD_Opmode_HT_Speech;
   }
#endif /* S220 */

   ret = SDD_ALM_OpModeSet (pCh);

   RETURN_STATUS(ret);
}

/**
   Prepare parameters and call the target configuration function to switch the
   line mode.

   Checks on valid transitions are done. Nevertheless they should be avoided by
   the user application.

   \param  pCh          Pointer to SDD channel structure.
   \param  nMode        Linefeed mode.

   \return
   - SDD_statusOk             if successful
   - SDD_statusInvalLMSwitch  Trying to do an invalid line mode change.
   - SDD_statusOpModeWrErr    Writing the command has failed
*/
static IFX_int32_t sdd_alm_Line_Mode_Set(
                        SDD_CHANNEL_t *pCh,
                        IFX_int32_t nMode)
{
   SDD_Opmode_t *pOpmod = &pCh->pALM->sdd_opmode;
   IFX_uint8_t curr_opmode;
   IFX_int32_t ret;

   /* Check the mode parameter. */
   switch (nMode)
   {
      /* These are the supported linemodes. */
      case IFX_TAPI_LINE_FEED_DISABLED:         /*lint -fallthrough */
      case IFX_TAPI_LINE_FEED_STANDBY:          /*lint -fallthrough */
      case IFX_TAPI_LINE_FEED_ACTIVE:           /*lint -fallthrough */
      case IFX_TAPI_LINE_FEED_ACTIVE_REV:       /*lint -fallthrough */
      case IFX_TAPI_LINE_FEED_RING_BURST:       /*lint -fallthrough */
      case IFX_TAPI_LINE_FEED_GROUND_START_TIP_OPEN:  /*lint -fallthrough */
      case IFX_TAPI_LINE_FEED_GROUND_START_TIP2GND:
         break;

      /* For backward compatibility translate the following old linemodes. */

      case IFX_TAPI_LINE_FEED_RING_PAUSE:
#ifdef S220
         switch (pOpmod->OperatingMode)
         {
            case SDD_Opmode_Active_REVPOL:      /*lint -fallthrough */
            case SDD_Opmode_Ringing_REVPOL:
               nMode = IFX_TAPI_LINE_FEED_ACTIVE_REV;
               break;
            default:
               nMode = IFX_TAPI_LINE_FEED_ACTIVE;
               break;
         }
#else /* S220 */
         nMode = (pOpmod->Polarity == SDD_Opmode_Pol_Normal) ?
                 IFX_TAPI_LINE_FEED_ACTIVE : IFX_TAPI_LINE_FEED_ACTIVE_REV;
#endif /* S220 */
         break;

      case IFX_TAPI_LINE_FEED_NORMAL_AUTO:      /*lint -fallthrough */
      case IFX_TAPI_LINE_FEED_ACTIVE_LOW:       /*lint -fallthrough */
      case IFX_TAPI_LINE_FEED_ACTIVE_BOOSTED:
         nMode = IFX_TAPI_LINE_FEED_ACTIVE;
         break;

      case IFX_TAPI_LINE_FEED_REVERSED_AUTO:
         nMode = IFX_TAPI_LINE_FEED_ACTIVE_REV;
         break;

      case IFX_TAPI_LINE_FEED_PARKED_REVERSED:
         nMode = IFX_TAPI_LINE_FEED_STANDBY;

      /* All remaining linemodes are not supported by this SW/FW/HW. */
      default:
         RETURN_STATUS(SDD_statusNotSupported);
   }

   /* Get the current opmode - wait while an opmode change is ongoing. */
   ret = SDD_ALM_OpmodeGet (pCh, &curr_opmode);
   /* When waiting is interrupted exit here with error. */
   if (ret == SDD_statusSddEvtWaitInterrupt)
   {
      /* errmsg: Waiting for an event from SDD interrupted by a signal. */
      RETURN_STATUS(SDD_statusSddEvtWaitInterrupt);
   }

   /* Check if the transition is valid then prepare the command. */

#ifdef S220
   switch (nMode)
   {
      case IFX_TAPI_LINE_FEED_DISABLED:
         pOpmod->OperatingMode = SDD_Opmode_Disabled;
         break;

      case IFX_TAPI_LINE_FEED_STANDBY:
         switch (curr_opmode)
         {
            case SDD_Opmode_Disabled:           /*lint -fallthrough */
            case SDD_Opmode_Standby:            /*lint -fallthrough */
            case SDD_Opmode_Active:             /*lint -fallthrough */
            case SDD_Opmode_Active_REVPOL:      /*lint -fallthrough */
            case SDD_Opmode_Ringing:            /*lint -fallthrough */
            case SDD_Opmode_Ringing_REVPOL:
               pOpmod->OperatingMode = SDD_Opmode_Standby;
               break;
            default:
               /* errmsg: Line mode switch is invalid.
                          Not every transition is valid. */
               RETURN_STATUS(SDD_statusInvalLMSwitch);
         }
         break;

      case IFX_TAPI_LINE_FEED_ACTIVE:
         switch (curr_opmode)
         {
            case SDD_Opmode_Disabled:           /*lint -fallthrough */
            case SDD_Opmode_Standby:            /*lint -fallthrough */
            case SDD_Opmode_Active:             /*lint -fallthrough */
            case SDD_Opmode_Active_REVPOL:      /*lint -fallthrough */
            case SDD_Opmode_Ringing:            /*lint -fallthrough */
            case SDD_Opmode_Ground_Start:       /*lint -fallthrough */
            case SDD_Opmode_Ground_Start_T2G:   /*lint -fallthrough */
               pOpmod->OperatingMode = SDD_Opmode_Active;
               break;
            default:
               /* errmsg: Line mode switch is invalid.
                          Not every transition is valid. */
               RETURN_STATUS(SDD_statusInvalLMSwitch);
         }
         break;

      case IFX_TAPI_LINE_FEED_ACTIVE_REV:
         switch (curr_opmode)
         {
            case SDD_Opmode_Disabled:           /*lint -fallthrough */
            case SDD_Opmode_Standby:            /*lint -fallthrough */
            case SDD_Opmode_Active:             /*lint -fallthrough */
            case SDD_Opmode_Active_REVPOL:      /*lint -fallthrough */
            case SDD_Opmode_Ringing_REVPOL:     /*lint -fallthrough */
            case SDD_Opmode_Ground_Start:       /*lint -fallthrough */
            case SDD_Opmode_Ground_Start_T2G:
               pOpmod->OperatingMode = SDD_Opmode_Active_REVPOL;
               break;
            default:
               /* errmsg: Line mode switch is invalid.
                          Not every transition is valid. */
               RETURN_STATUS(SDD_statusInvalLMSwitch);
         }
         break;

      case IFX_TAPI_LINE_FEED_RING_BURST:
         switch (curr_opmode)
         {
            case SDD_Opmode_Disabled:           /*lint -fallthrough */
            case SDD_Opmode_Standby:            /*lint -fallthrough */
            case SDD_Opmode_Active:             /*lint -fallthrough */
            case SDD_Opmode_Ringing:            /*lint -fallthrough */
            case SDD_Opmode_Ground_Start:       /*lint -fallthrough */
            case SDD_Opmode_Ground_Start_T2G:
               pOpmod->OperatingMode = SDD_Opmode_Ringing;
               break;
            case SDD_Opmode_Ringing_REVPOL:     /*lint -fallthrough */
            case SDD_Opmode_Active_REVPOL:
               pOpmod->OperatingMode = SDD_Opmode_Ringing_REVPOL;
               break;
            default:
               /* errmsg: Line mode switch is invalid.
                          Not every transition is valid. */
               RETURN_STATUS(SDD_statusInvalLMSwitch);
         }
         break;

      case IFX_TAPI_LINE_FEED_GROUND_START_TIP_OPEN:
         switch (curr_opmode)
         {
            case SDD_Opmode_Disabled:           /*lint -fallthrough */
            case SDD_Opmode_Active:             /*lint -fallthrough */
            case SDD_Opmode_Active_REVPOL:      /*lint -fallthrough */
            case SDD_Opmode_Ground_Start:       /*lint -fallthrough */
            case SDD_Opmode_Ground_Start_T2G:
               pOpmod->OperatingMode = SDD_Opmode_Ground_Start;
               break;
            default:
               /* errmsg: Line mode switch is invalid.
                          Not every transition is valid. */
               RETURN_STATUS(SDD_statusInvalLMSwitch);
         }
         break;

      case IFX_TAPI_LINE_FEED_GROUND_START_TIP2GND:
         switch (curr_opmode)
         {
            case SDD_Opmode_Ground_Start:       /*lint -fallthrough */
            case SDD_Opmode_Ground_Start_T2G:
               pOpmod->OperatingMode = SDD_Opmode_Ground_Start_T2G;
               break;
            default:
               /* errmsg: Line mode switch is invalid.
                          Not every transition is valid. */
               RETURN_STATUS(SDD_statusInvalLMSwitch);
         }
         break;

      default:
         /* Never used when the switch statement before this one is correct. */
         RETURN_STATUS(SDD_statusNotSupported);
   }

   {
      struct TTX_GEN_CTRL *pTtxGenCtrl = &pCh->pALM->ttx_gen_ctrl;
      /* always clear TTX bit */
      pTtxGenCtrl->EN = TTX_GEN_CTRL_DIS;
   }

#else /* S220 */

   switch (nMode)
   {
      case IFX_TAPI_LINE_FEED_DISABLED:
         pOpmod->OperatingMode = SDD_Opmode_Disabled;
         pOpmod->Polarity = SDD_Opmode_Pol_Normal;
         break;

      case IFX_TAPI_LINE_FEED_STANDBY:
         switch (curr_opmode)
         {
            case SDD_Opmode_Disabled:           /*lint -fallthrough */
            case SDD_Opmode_Standby:            /*lint -fallthrough */
            case SDD_Opmode_Active:             /*lint -fallthrough */
            case SDD_Opmode_Ringing:
               pOpmod->OperatingMode = SDD_Opmode_Standby;
               pOpmod->Polarity = SDD_Opmode_Pol_Normal;
               break;
            default:
               /* errmsg: Line mode switch is invalid.
                          Not every transition is valid. */
               RETURN_STATUS(SDD_statusInvalLMSwitch);
         }
         break;

      case IFX_TAPI_LINE_FEED_ACTIVE:           /*lint -fallthrough */
      case IFX_TAPI_LINE_FEED_ACTIVE_REV:       /*lint -fallthrough */
      case IFX_TAPI_LINE_FEED_RING_PAUSE:
         switch (curr_opmode)
         {
            case SDD_Opmode_Disabled:           /*lint -fallthrough */
            case SDD_Opmode_Standby:            /*lint -fallthrough */
            case SDD_Opmode_Active:             /*lint -fallthrough */
            case SDD_Opmode_Ringing:
               pOpmod->OperatingMode = SDD_Opmode_Active;
               break;
            default:
               /* errmsg: Line mode switch is invalid.
                          Not every transition is valid. */
               RETURN_STATUS(SDD_statusInvalLMSwitch);
         }
         switch (nMode)
         {
            case IFX_TAPI_LINE_FEED_ACTIVE_REV:
               pOpmod->Polarity = SDD_Opmode_Pol_Reverse;
               break;
            case IFX_TAPI_LINE_FEED_RING_PAUSE:
               /* leave reverse polarity bit as it is */
               break;
            default:
               pOpmod->Polarity = SDD_Opmode_Pol_Normal;
               break;
         }
         break;

      case IFX_TAPI_LINE_FEED_RING_BURST:
         switch (curr_opmode)
         {
            case SDD_Opmode_Disabled:           /*lint -fallthrough */
            case SDD_Opmode_Standby:            /*lint -fallthrough */
            case SDD_Opmode_Active:             /*lint -fallthrough */
            case SDD_Opmode_Ringing:
               pOpmod->OperatingMode = SDD_Opmode_Ringing;
               /* leave reverse polarity bit as it is */
               break;
            default:
               /* errmsg: Line mode switch is invalid.
                          Not every transition is valid. */
               RETURN_STATUS(SDD_statusInvalLMSwitch);
         }
         break;

      default:
         /* Never used when the switch statement before this one is correct. */
         RETURN_STATUS(SDD_statusNotSupported);
   }

   /* The MWL will be (implicitly) deactivated on any linemode change (!),
      it's up to the application to reactivate the MWL e.g. after a
      ring sequence... */
   if (1 == pOpmod->MwiLamp)
   {
      SDD_DEVICE_t      *pDev  = pCh->pDev;
      SDD_BasicConfig_t *pBasicCfg = &pCh->pALM->sdd_basic_config_tmp_mwl;
      SDD_RingConfig_t  *pRingCfg = &pCh->pALM->sdd_ring_config_tmp_mwl;

      /* restore basic configuration */
      ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)pBasicCfg);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing the basic configuration failed. */
         RETURN_STATUS(SDD_statusBasicCfgWrErr);
      }

      /* restore ring configuration */
      ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)pRingCfg);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing the ring configuration failed. */
         RETURN_STATUS(SDD_statusRingCfgWrErr);
      }

      pOpmod->MwiLamp = 0;
   }

   /* Automatically deactivate howler tone with any linefeeding which is not
      active. */
   if ((pOpmod->Howler == SDD_Opmode_HT_Howler) &&
       (pOpmod->OperatingMode != SDD_Opmode_Active))
   {
      pOpmod->Howler = SDD_Opmode_HT_Speech;
   }

   /* always clear TTX bit */
   pOpmod->TtxBurst = SDD_Opmode_TTX_NO;
#endif /* S220 */

   ret = SDD_ALM_OpModeSet (pCh);

   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Writing the operation mode failed.*/
      RETURN_STATUS(SDD_statusOpModeWrErr);
   }

   return SDD_statusOk;
}

/**
   Set line type to FXS or FXO.

   \param  pLLChannel      Pointer to the SDD channel structure.
   \param  nType           New line type, can be anything out of
                           IFX_TAPI_LINE_TYPE_t.

   \return
   - SDD_statusOk             If successful.
   - SDD_statusInvalCh        No analog line on this channel.
*/
IFX_int32_t SDD_TAPI_LL_ALM_Line_Type_Set (
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_LINE_TYPE_t nType)
{
   SDD_CHANNEL_t    *pCh;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   /* check if setting is valid */
   switch (nType)
   {
      case IFX_TAPI_LINE_TYPE_FXS_NB:
      case IFX_TAPI_LINE_TYPE_FXS_WB:
      case IFX_TAPI_LINE_TYPE_FXS_AUTO:
         if (pCh->pALM->line_type_fxs == IFX_FALSE)
         {
            /* SmartSLIC has fixed channel types. */
            /* errmsg: Line type change not allowed */
            RETURN_STATUS(SDD_statusLineTypChNotAll);
         }
         break;

      case IFX_TAPI_LINE_TYPE_FXO:
         if (pCh->pALM->line_type_fxs != IFX_FALSE)
         {
            /* SmartSLIC has fixed channel types. */
            /* errmsg: Line type change not allowed */
            RETURN_STATUS(SDD_statusLineTypChNotAll);
         }
         break;

      default:
         /* errmsg: Invalid line type or sampling mode for analog line */
         RETURN_STATUS(SDD_statusInvalAlmTypeSmpl);
   }

   RETURN_STATUS(SDD_statusOk);
}

/** Prepare parameters and call the target configuration function to switch the
   line mode.

   A check is done, whether the channel supports an analog line. If not
   SDD_statusInvalCh is returned.
   Checks on valid transitions are done. Nevertheless they should be avoided by
   the user application.

   \param pLLChannel Handle to TAPI low level channel structure

   \param nMode Linefeed mode

   \param nTapiLineMode the currently set line mode. Usage in discussion

   \return
   - SDD_statusOk             if successful
   - SDD_statusInvalCh        The parameters are wrong
*/
static IFX_int32_t SDD_TAPI_LL_ALM_Line_Mode_Set (IFX_TAPI_LL_CH_t *pLLChannel,
                                                  IFX_int32_t nMode,
                                                  IFX_uint8_t nTapiLineMode)
{
   SDD_CHANNEL_t    *pCh;
   SDD_CHANNEL_t    *pOtherCh = IFX_NULL;
   IFX_int32_t       ret;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   if ((pCh->pALM->line_type_fxs == IFX_FALSE) &&
       (nMode != IFX_TAPI_LINE_FEED_DISABLED))
   {
      /* PDH is allowed, e.g. for IFX_TAPI_DEV_STOP */
      /* errmsg: Operation blocked on FXO line */
      RETURN_STATUS(SDD_statusBlockedFXSOpOnFXO);
   }

#ifdef S220
   /* Until the DC/DC converter type is set this operation is not allowed. */
   if ((pCh->pALM->line_type_fxs == IFX_TRUE) &&
       (pCh->pALM->nDcDcType == SDD_DCDC_TYPE_NOTSET))
#else /* S220 */
   /* For DC/DC converters other than default a BBD download is required. */
   if ((pCh->pALM->line_type_fxs == IFX_TRUE) &&
       (pCh->pALM->nDcDcType == SDD_DCDC_TYPE_DEFAULT_IBB) &&
       (pCh->pDev->nAllowedDcDcType != SDD_DCDC_TYPE_IBB))
#endif /* S220 */
   {
      /* errmsg: Operation blocked until a BBD with DC/DC configuration
                 matching the connected DC/DC hardware type is downloaded */
      RETURN_STATUS(SDD_statusBlockedDcDcTypeMissmatch);
   }

   /* In combined DC/DC mode mode changing the linemode to anything other
      than disabled is only allowed if the neighbor channel is not doing
      calibration or linetesting */
   if ((pCh->pDev->bDcDcHwCombined == IFX_TRUE) &&
       (nMode != IFX_TAPI_LINE_FEED_DISABLED))
   {
      /* The combined DC/DC flag is only set on 2 channel systems. So there
         is always a neighboring channel. */
      pOtherCh = SDD_GetNeighbourChannel(pCh);
      SDD_ASSERT(pOtherCh != IFX_NULL);

      if ((pOtherCh->pALM->sdd_opmode.OperatingMode == SDD_Opmode_GR909) ||
          (pOtherCh->pALM->sdd_opmode.OperatingMode == SDD_Opmode_Calibrate))
      {
         /* errmsg: Linefeed change not allowed while neighbour line is doing
                    linetesting. */
         RETURN_STATUS(SDD_statusNeighbourLineBlocksLMSwitch);
      }
   }

   ret = sdd_alm_Line_Mode_Set (pCh, nMode);

   RETURN_STATUS(ret);
}

/**
   Read the current operating mode and report to TAPI.
   Reports IFX_TAPI_LINE_FEED_DISABLED if called during PDH,
   FXO, Calibration or GR909.

   \param  pLLChannel      Pointer to the VMMC channel structure.
   \param  pFeed           Pointer to TAPI IFX_TAPI_LINE_FEED_t variable.

   \return
   - VMMC_statusOk         always
*/
static IFX_int32_t SDD_TAPI_LL_ALM_Line_Mode_Get (
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_LINE_FEED_t *pFeed)
{
   SDD_CHANNEL_t *pCh;
   IFX_TAPI_LINE_MODE_t nMode;
   IFX_uint8_t curr_opmode;
   IFX_int32_t ret;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   /* Get the current opmode - wait while a opmode change is ongoing. */
   ret = SDD_ALM_OpmodeGet (pCh, &curr_opmode);
   /* Only exit when interrupted by a signal. When waiting was aborted by the
      timeout assume that no linemode change is pending and try to continue. */
   if (ret == SDD_statusSddEvtWaitInterrupt)
   {
      RETURN_STATUS (ret);
   }

   switch (curr_opmode)
   {
#ifdef S220
      case SDD_Opmode_Active:                   /*lint -fallthrough */
      case SDD_Opmode_Active_HOWL:              /*lint -fallthrough */
      case SDD_Opmode_Active_MWL:
         nMode = IFX_TAPI_LINE_FEED_ACTIVE;
         break;

      case SDD_Opmode_Active_REVPOL:            /*lint -fallthrough */
      case SDD_Opmode_Active_HOWL_REVPOL:
         nMode = IFX_TAPI_LINE_FEED_ACTIVE_REV;
         break;

      case SDD_Opmode_Ringing_REVPOL:
         nMode = IFX_TAPI_LINE_FEED_RING_BURST;
         break;

      case SDD_Opmode_Ground_Start:             /*lint -fallthrough */
      case SDD_Opmode_Ground_Start_T2G:
         nMode = IFX_TAPI_LINE_FEED_STANDBY;
         break;
#else /* S220 */
      case SDD_Opmode_Active:
         if (pCh->pALM->curr_polarity == SDD_Opmode_Pol_Normal)
            nMode = IFX_TAPI_LINE_FEED_ACTIVE;
         else
            nMode = IFX_TAPI_LINE_FEED_ACTIVE_REV;
         break;
#endif /* S220 */

      case SDD_Opmode_Standby:
         nMode = IFX_TAPI_LINE_FEED_STANDBY;
         break;

      case SDD_Opmode_Ringing:
         nMode = IFX_TAPI_LINE_FEED_RING_BURST;
         break;

      /* report disabled status in case of PDH,
         FXO, Calibration and GR909 */
      default:
         nMode = IFX_TAPI_LINE_FEED_DISABLED;
         break;
   }

#ifdef TAPI_ONE_DEVNODE
   pFeed->lineMode = nMode;
#else  /* TAPI_ONE_DEVNODE */
   *pFeed = nMode;
#endif /* TAPI_ONE_DEVNODE */

   return SDD_statusOk;
}

/**
   Sends one metering burst

   \param  pCh          Pointer to the SDD channel structure.
   \param  nPulseNum    Number of pulses in burst. 0 to abort a burst.

   \return
   - SDD_statusParm  Wrong parameters passed. This code is returned
                     when the nPulseNum parameter has an invalid or
                     unsupported value.
   - SDD_statusCmdWr Writing the command has failed
   - SDD_statusOk    if successful

   \remarks
   The SDD FW allows only one pulse to be sent with each call. Therefore the
   nPulseNum parameter must always be set to 1 only.
*/
static IFX_int32_t sdd_alm_MeteringBurst (SDD_CHANNEL_t *pCh,
                                          IFX_uint32_t nPulseNum)
{
   IFX_uint8_t       curr_opmode;
   IFX_int32_t       ret;
#ifdef S220
   struct TTX_GEN_CTRL *pTtxGenCtrl = &pCh->pALM->ttx_gen_ctrl;
#else /* S220 */
   SDD_Opmode_t *pOpmod = &pCh->pALM->sdd_opmode;
#endif /* S220 */

   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      /* errmsg: Operation blocked on FXO line */
      RETURN_STATUS(SDD_statusBlockedFXSOpOnFXO);
   }
#ifdef S220
   /* Until the DC/DC converter type is set this operation is not allowed. */
   if (pCh->pALM->nDcDcType == SDD_DCDC_TYPE_NOTSET)
#else /* S220 */
   /* For DC/DC converters other than default a BBD download is required. */
   if ((pCh->pALM->nDcDcType == SDD_DCDC_TYPE_DEFAULT_IBB) &&
       (pCh->pDev->nAllowedDcDcType != SDD_DCDC_TYPE_IBB))
#endif /* S220 */
   {
      /* errmsg: Operation blocked until a BBD with DC/DC configuration
                 matching the connected DC/DC hardware type is downloaded. */
      RETURN_STATUS(SDD_statusBlockedDcDcTypeMissmatch);
   }

   /* only one single pulse or zero to abort are supported */
   if (nPulseNum > 1)
   {
      RETURN_STATUS(SDD_statusParam);
   }

   /* Abort if the line feeding is anything but active or active reversed. */
   /* Get the current opmode - wait while a opmode change is ongoing. */
   ret = SDD_ALM_OpmodeGet (pCh, &curr_opmode);
   if (ret == SDD_statusSddEvtWaitInterrupt)
   {
      /* Exit because interrupted by a signal. */
      RETURN_STATUS(ret);
   }
   if ((curr_opmode != SDD_Opmode_Active)
#ifdef S220
       && (curr_opmode != SDD_Opmode_Active_REVPOL)
#endif /* S220 */
      )
   {
      /* errmsg: Metering pulse requires active line feeding. */
      RETURN_STATUS(SDD_statusMeteringNotActive);
   }

   /* activate metering burst */
#ifndef S220
   /* The opmode is already active - no need to set it again. Add only TTX. */
   pOpmod->TtxBurst = nPulseNum ? SDD_Opmode_TTX_BURST : SDD_Opmode_TTX_NO;
   ret = SDD_ALM_OpModeSet (pCh);
#else /* S220 */
   pTtxGenCtrl->EN = nPulseNum ? TTX_GEN_CTRL_EN : TTX_GEN_CTRL_DIS;
   ret = SDD_CmdWrite(pCh->pDev, (IFX_uint32_t *)pTtxGenCtrl);
#endif /* S220 */

   RETURN_STATUS(ret);
}

/**
   Sends one metering burst

   \param  pLLChannel   Pointer to TAPI low level channel structure.
   \param  nPulseNum    Number of pulses in burst.

   \return
   - SDD_statusInvalCh  Channel not valid or channel number out of range
   - SDD_statusParm     Wrong parameters passed. This code is returned
                        when the nPulseNum parameter has an invalid or
                        unsupported value.
   - SDD_statusCmdWr    Writing the command has failed
   - SDD_statusOk       if successful
*/
static IFX_int32_t SDD_TAPI_LL_ALM_MeteringStart (IFX_TAPI_LL_CH_t *pLLChannel,
                                                  IFX_uint32_t nPulseNum)
{
   SDD_CHANNEL_t *pCh;
   IFX_int32_t ret = SDD_statusOk;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   ret = sdd_alm_MeteringBurst (pCh, nPulseNum);

   RETURN_STATUS(ret);
}

/** This service controls an 8 kHz loop switching in the device for testing.
    If switched on, signals that are played to the subscriber are looped back
    to the receiving side.

\param pLLChannel Pointer to TAPI low level channel structure.
\param pLoop Pointer to the test loop details
         - if pLoop->bAnalog == 0x1 - loop is switched on
         - if pLoop->bAnalog == 0x0 - loop is switched off
\return
   - SDD_statusCmdWrErr Writing the command has failed
   - SDD_statusInvalCh The resource is not valid
   - SDD_statusOk if successful
*/
static IFX_int32_t SDD_TAPI_LL_ALM_TestLoop (IFX_TAPI_LL_CH_t *pLLChannel,
                                             IFX_TAPI_TEST_LOOP_t const *pLoop)
{
   SDD_CHANNEL_t *pCh;
   SDD_DEVICE_t *pDev;
   IFX_int32_t ret;
   SDD_En8kLoop_t ctrl8kLoopCfg;
   IFX_uint8_t ch;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   pDev = pCh->pDev;
   ch = pCh->nChannel - 1;

   if (pCh->pALM != IFX_NULL)
   {
      memset (&ctrl8kLoopCfg, 0, sizeof (SDD_En8kLoop_t));
      ctrl8kLoopCfg.CMD = CMD_SDD;
      ctrl8kLoopCfg.MOD = MOD_SDD;
      ctrl8kLoopCfg.ECMD = SDD_En8kLoop_ECMD;
      ctrl8kLoopCfg.CHAN = ch;
      ctrl8kLoopCfg.LENGTH = SDD_En8kLoop_LENGTH;

      /* depending on the request enable or disable the test loop */
      if (pLoop->bAnalog)
      {
         /* close digital loop at 8kHz */
         ctrl8kLoopCfg.En8kLoop = SDD_En8kLoop_ENABLE;
      }
      else
      {
         /* normal operation, digital loop at 8 kHz open */
         ctrl8kLoopCfg.En8kLoop = SDD_En8kLoop_DISABLE;
      }

      /* write message contents */
      ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)&ctrl8kLoopCfg);

      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing a command to channel failed. */
         RETURN_STATUS(SDD_statusCmdWrErr);
      }

      return SDD_statusOk;
   }
   else
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }
}


/** Gets the activation status of the message waiting lamp.

   \param pCh           Pointer to SDD channel structure.
   \param pActivation   Pointer to \ref IFX_TAPI_MWL_ACTIVATION_t structure.

   \return
   - SDD_statusOk          if successful
   - SDD_statusNotSupported
   - SDD_statusReadErr
*/
static IFX_int32_t sdd_ALM_MWL_Activation_Get(SDD_CHANNEL_t *pCh,
                                 IFX_TAPI_MWL_ACTIVATION_t *pActivation)
{
   IFX_int32_t        ret     = SDD_statusOk;
   SDD_DEVICE_t      *pDev    = pCh->pDev;

   if (!pDev->caps.bfw_MWI)
   {
      RETURN_STATUS(SDD_statusNotSupported);
   }

#ifdef S220
   if (pCh->pALM->curr_opmode == SDD_Opmode_Active_MWL)
#else /* S220 */
   if (SDD_Opmode_Ringing == pCh->pALM->curr_opmode &&
       SDD_Opmode_MWL_On == pCh->pALM->curr_mwi_lamp )
#endif /* S220 */
   {
      pActivation->nActivation = IFX_ENABLE;
   }
   else
   {
      pActivation->nActivation = IFX_DISABLE;
   }

   RETURN_STATUS(ret);
}


/**
   Gets the activation status of the message waiting lamp.

   \param pLLChannel    Pointer to TAPI low level channel structure.
   \param pActivation   Pointer to \ref IFX_TAPI_MWL_ACTIVATION_t structure.

   \return
   - SDD_statusOk          if successful
   - SDD_statusInvalCh
   - SDD_statusReadErr
*/
static IFX_int32_t SDD_TAPI_LL_ALM_MWL_Activation_Get(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_MWL_ACTIVATION_t *pActivation)
{
   IFX_int32_t        ret     = SDD_statusOk;
   SDD_CHANNEL_t     *pCh;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   SDD_ASSERT(pActivation != IFX_NULL);

   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   ret = sdd_ALM_MWL_Activation_Get(pCh, pActivation);

   RETURN_STATUS(ret);
}


/**
   Activate/deactivates the message waiting lamp.

   \param pCh           Pointer to SDD channel structure.
   \param pActivation   Pointer to \ref IFX_TAPI_MWL_ACTIVATION_t structure.

   \return
   - SDD_statusOk             if successful
   - SDD_statusNotSupported
   - SDD_statusParam
   - SDD_statusMwlNotStandby
   - SDD_statusOpModeWrErr
*/
static IFX_int32_t sdd_ALM_MWL_Activation_Set(
                   SDD_CHANNEL_t *pCh,
                   IFX_TAPI_MWL_ACTIVATION_t const *pActivation)
{
   IFX_int32_t            ret      = SDD_statusOk;
   SDD_DEVICE_t          *pDev     = pCh->pDev;
   SDD_Opmode_t          *pOpmod   = &pCh->pALM->sdd_opmode;
   IFX_uint8_t           curr_opmode;
#ifndef S220
   SDD_RingConfig_t      *pRingCfg = &pCh->pALM->sdd_ring_config_tmp_mwl,
                         sdd_ring_config_mwi;
   SDD_BasicConfig_t     *pBasicCfg = &pCh->pALM->sdd_basic_config_tmp_mwl,
                         sdd_basic_config_mwi;
#endif /* S220 */

   if (!pDev->caps.bfw_MWI)
   {
      RETURN_STATUS(SDD_statusNotSupported);
   }

   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      /* errmsg: Operation blocked on FXO line */
      RETURN_STATUS(SDD_statusBlockedFXSOpOnFXO);
   }
#ifdef S220
   /* Until the DC/DC converter type is set this operation is not allowed. */
   if (pCh->pALM->nDcDcType == SDD_DCDC_TYPE_NOTSET)
#else /* S220 */
   /* For DC/DC converters other than default a BBD download is required. */
   if ((pCh->pALM->nDcDcType == SDD_DCDC_TYPE_DEFAULT_IBB) &&
       (pCh->pDev->nAllowedDcDcType != SDD_DCDC_TYPE_IBB))
#endif /* S220 */
   {
      /* errmsg: Operation blocked until a BBD with DC/DC configuration
                 matching the connected DC/DC hardware type is downloaded. */
      RETURN_STATUS(SDD_statusBlockedDcDcTypeMissmatch);
   }
   /* MWI and combined DC/DC together is not possible. */
   if (pDev->bDcDcHwCombined == IFX_TRUE)
   {
      /* errmsg: MWL not possible in combined DC/DC mode */
      RETURN_STATUS(SDD_statusNoMwlAndCombinedDcDc);
   }

   /* Get the current opmode - wait while an opmode change is ongoing. */
   ret = SDD_ALM_OpmodeGet (pCh, &curr_opmode);
   /* When waiting is interrupted exit here with error. */
   if (ret == SDD_statusSddEvtWaitInterrupt)
   {
      /* errmsg: Waiting for an event from SDD interrupted by a signal. */
      RETURN_STATUS(SDD_statusSddEvtWaitInterrupt);
   }

#ifdef S220
   if (IFX_ENABLE == pActivation->nActivation)
   {
      /* Enable MWL SLIC200 */

      if (curr_opmode != SDD_Opmode_Standby)
      {
         /* errmsg: MWL can be started only in linemode standby */
         RETURN_STATUS(SDD_statusMwlNotStandby);
      }

      /* MWL can only be entered from active. So implicitly go from standby
         to active to have the same behaviour as on SLIC100 family. */
      pOpmod->OperatingMode = SDD_Opmode_Active;
      ret = SDD_ALM_OpModeSet (pCh);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing the operation mode failed. */
         RETURN_STATUS(SDD_statusOpModeWrErr);
      }

      pOpmod->OperatingMode = SDD_Opmode_Active_MWL;
   }
   else
   {
      /* Disable MWL SLIC200 */

      if (curr_opmode != SDD_Opmode_Active_MWL)
      {
         /* errmsg: MWL is currently not active  */
         RETURN_STATUS(SDD_statusMwlNotActive);
      }

      /* MWL can only exit to active. So implicitly go to active and then
         to standby to have the same behaviour as on SLIC100 family. */
      pOpmod->OperatingMode = SDD_Opmode_Active;
      ret = SDD_ALM_OpModeSet (pCh);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing the operation mode failed. */
         RETURN_STATUS(SDD_statusOpModeWrErr);
      }

      pOpmod->OperatingMode = SDD_Opmode_Standby;
   }

   ret = SDD_ALM_OpModeSet (pCh);

   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Writing the operation mode failed. */
      RETURN_STATUS(SDD_statusOpModeWrErr);
   }
#else /* S220 */
   if (IFX_ENABLE == pActivation->nActivation)
   {
      /* Enable MWL SLIC100 */

      if (curr_opmode != SDD_Opmode_Standby)
      {
         /* errmsg: MWL can be started only in linemode standby */
         RETURN_STATUS(SDD_statusMwlNotStandby);
      }

      /* read back the basic configuration */
      ret = SDD_CmdRead (pDev, (IFX_uint32_t *)(IFX_void_t *)pBasicCfg,
                               (IFX_uint32_t *)(IFX_void_t *)pBasicCfg);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Reading the basic configuration failed. */
         RETURN_STATUS(SDD_statusBasicCfgRdErr);
      }

      /* read back the ring configuration */
      ret = SDD_CmdRead (pDev, (IFX_uint32_t *)(IFX_void_t *)pRingCfg,
                               (IFX_uint32_t *)(IFX_void_t *)pRingCfg);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Reading the ring configuration failed. */
         RETURN_STATUS(SDD_statusRingCfgRdErr);
      }

      /* clear GsEn bit for MWL mode */
      sdd_basic_config_mwi = *pBasicCfg;
      sdd_basic_config_mwi.GsEn = SDD_BasicConfig_DISABLE;
      ret = SDD_CmdWrite(pDev,
                         (IFX_uint32_t *)(IFX_void_t *)&sdd_basic_config_mwi);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing the basic configuration failed. */
         RETURN_STATUS(SDD_statusBasicCfgWrErr);
      }

      /* reconfigure ring for MWL mode */
      sdd_ring_config_mwi = *pRingCfg;
      sdd_ring_config_mwi.RingMode      = 0; /* always */
      sdd_ring_config_mwi.RingSignal    = 0; /* sinewave ringing signal */
      sdd_ring_config_mwi.RingCrestFact = 4; /* near sinewave ring */
      ret = SDD_CmdWrite(pDev,
                         (IFX_uint32_t *)(IFX_void_t *)&sdd_ring_config_mwi);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing the ring configuration failed. */
         RETURN_STATUS(SDD_statusRingCfgWrErr);
      }

      pOpmod->OperatingMode = SDD_Opmode_Ringing;
      pOpmod->MwiLamp = SDD_Opmode_MWL_On;
      ret = SDD_ALM_OpModeSet (pCh);

      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing the operation mode failed. */
         RETURN_STATUS(SDD_statusOpModeWrErr);
      }
   }
   else
   {
      /* Disable MWL SLIC100 */

      if ((pOpmod->MwiLamp == SDD_Opmode_MWL_Off) ||
          (curr_opmode != SDD_Opmode_Ringing))
      {
         /* errmsg: MWL is currently not active */
         RETURN_STATUS(SDD_statusMwlNotActive);
      }
      pOpmod->MwiLamp = SDD_Opmode_MWL_Off;
      pOpmod->OperatingMode = SDD_Opmode_Standby;
      ret = SDD_ALM_OpModeSet (pCh);

      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing the operation mode failed. */
         RETURN_STATUS(SDD_statusOpModeWrErr);
      }

      /* restore basic configuration */
      ret = SDD_CmdWrite(pDev, (IFX_uint32_t *)(IFX_void_t *)pBasicCfg);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing the basic configuration failed. */
         RETURN_STATUS(SDD_statusBasicCfgWrErr);
      }

      /* restore ring configuration */
      ret = SDD_CmdWrite(pDev, (IFX_uint32_t *)(IFX_void_t *)pRingCfg);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Writing the ring configuration failed. */
         RETURN_STATUS(SDD_statusRingCfgWrErr);
      }
   }
#endif /* S220 */

   RETURN_STATUS(ret);
}


/**
   Sets the activation status of the message waiting lamp.

   \param pLLChannel    Pointer to TAPI low level channel structure.
   \param pActivation   Pointer to \ref IFX_TAPI_MWL_ACTIVATION_t structure.

   \return
   - SDD_statusOk          if successful
   - SDD_statusInvalCh
   - SDD_statusReadErr
*/
static IFX_int32_t SDD_TAPI_LL_ALM_MWL_Activation_Set(
                   IFX_TAPI_LL_CH_t *pLLChannel,
                   IFX_TAPI_MWL_ACTIVATION_t const *pActivation)
{
   IFX_int32_t       ret      = SDD_statusOk;
   SDD_CHANNEL_t     *pCh;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   SDD_ASSERT(pActivation != IFX_NULL);

   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }
   if ((IFX_ENABLE != pActivation->nActivation) &&
       (IFX_DISABLE != pActivation->nActivation))
   {
      RETURN_STATUS(SDD_statusParam);
   }

   ret = sdd_ALM_MWL_Activation_Set(pCh, pActivation);

   RETURN_STATUS(ret);
}


#ifdef SDD_FEAT_CONT_MEASUREMENT
/**
   Request continuous measurement results.

   This function does nothing on firmware interface. It dispatches an
   event IFX_TAPI_EVENT_CONTMEASUREMENT in a task context. Added for
   compatibility with TAPI_V4 interface.

   \param  pLLChannel   Handle to TAPI low level channel structure.

   \return
      - SDD_statusInvalCh
      - SDD_statusFXSCallOnFXO
      - SDD_statusOk
*/
static IFX_int32_t SDD_TAPI_LL_ALM_ContMeas_Req (IFX_TAPI_LL_CH_t *pLLChannel)
{
   SDD_CHANNEL_t *pCh;
   TAPI_CHANNEL    *pChannel;
   IFX_TAPI_EVENT_t tapiEvent;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if ((pCh->pALM == IFX_NULL) ||
       (pCh->pALM->line_type_fxs == IFX_FALSE))
   {
      /* errmsg: Resource not valid. Channel number out of range or
                 line type is FXO */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   pChannel = pCh->pTapiCh;

   memset (&tapiEvent, 0, sizeof(IFX_TAPI_EVENT_t));
   tapiEvent.id = IFX_TAPI_EVENT_CONTMEASUREMENT;
   tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;

   IFX_TAPI_Event_Dispatch(pChannel, &tapiEvent);
   RETURN_STATUS(SDD_statusOk);
}
#endif /* SDD_FEAT_CONT_MEASUREMENT */


#ifdef SDD_FEAT_CONT_MEASUREMENT_RESET
/**
   Reset continuous measurement results.

   \param  pLLChannel   Handle to TAPI low level channel structure.

   \return
      - SDD_statusInvalCh
      - SDD_statusFXSCallOnFXO
      - SDD_statusCmdWr Writing the command has failed
      - SDD_statusOk if successful
*/
static IFX_int32_t SDD_TAPI_LL_ALM_ContMeas_Reset (IFX_TAPI_LL_CH_t *pLLChannel)
{
   SDD_CHANNEL_t *pCh;
   SDD_DEVICE_t  *pDev;
   SDD_ContMeasClear_t *pContMeasClear;
   IFX_int32_t   ret;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if ((pCh->pALM == IFX_NULL) ||
       (pCh->pALM->line_type_fxs == IFX_FALSE))
   {
      /* errmsg: Resource not valid. Channel number out of range or
                 line type is FXO */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   pDev  = pCh->pDev;
   pContMeasClear = &pCh->pALM->fw_sdd_contMeasClear;


   pContMeasClear->ClearCMD = SDD_ContMeasClear_CLEAR;
   ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)(IFX_void_t *)pContMeasClear);

   RETURN_STATUS(ret);
}
#endif /* SDD_FEAT_CONT_MEASUREMENT_RESET */


#ifdef SDD_FEAT_CONT_MEASUREMENT
/**
   Return the stored continuous measure results.

   \param  pLLChannel   Handle to TAPI low level channel structure.
   \param  pContMeas    Pointer to struct where to return the results.

   \return
      - SDD_statusInvalCh
      - SDD_statusFXSCallOnFXO
      - IFX_ERROR CmdRead error
      - SDD_statusOk if successful
*/
static IFX_int32_t SDD_TAPI_LL_ALM_ContMeas_Get (IFX_TAPI_LL_CH_t *pLLChannel,
                                      IFX_TAPI_CONTMEASUREMENT_GET_t *pContMeas)
{
   SDD_CHANNEL_t *pCh;
   SDD_DEVICE_t *pDev;
   SDD_ContMeasRead_t *pContMeasRead;
   IFX_int32_t   ret;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      return SDD_statusInvalCh;
   }
   if ((pCh->pALM == IFX_NULL) ||
       (pCh->pALM->line_type_fxs == IFX_FALSE))
   {
      /* errmsg: Resource not valid. Channel number out of range or
                 line type is FXO */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   pDev = pCh->pDev;
   pContMeasRead = &pCh->pALM->fw_sdd_contMeasRead;

   ret = SDD_CmdRead (pDev, (IFX_uint32_t *)(IFX_void_t *)pContMeasRead,
                            (IFX_uint32_t *)(IFX_void_t *)pContMeasRead);

   if (SDD_SUCCESS (ret))
   {
      /* Typecasts are mandatory because of signed values. The double cast
         is needed to do the sign extension when going to 32-bit. The cast
         to the target size is done to show that this is intended. */
      pContMeas->nILine = (IFX_int16_t)
         (((IFX_int32_t)(IFX_int16_t)pContMeasRead->Itrans * 10000) >> 15);
      pContMeas->nVLineDesired = (IFX_int16_t)
         (((IFX_int32_t)(IFX_int16_t)pContMeasRead->Vlim * 14400) >> 15);
#ifdef S220
      pContMeas->nILineRingPeak = (IFX_int16_t)
         (((IFX_int32_t)(IFX_int16_t)pContMeasRead->Iring * 7071) >> 15);
      pContMeas->nVLineRingPeak = (IFX_int16_t)
         (((IFX_int32_t)(IFX_int16_t)pContMeasRead->Vring * 10182000) >> 15);
#else /* S220 */
      pContMeas->nILineRingPeak = (IFX_int16_t)
         (((IFX_int32_t)(IFX_int16_t)pContMeasRead->Iring * 10000) >> 15);
      pContMeas->nVLineRingPeak = (IFX_int16_t)
         (((IFX_int32_t)(IFX_int16_t)pContMeasRead->Vring * 14400) >> 15);
      pContMeas->nVLineWireRing = (IFX_int16_t)
         (((IFX_int32_t)(IFX_int16_t)pContMeasRead->VlineRing * 14400) >> 15);
      pContMeas->nVLineWireTip = (IFX_int16_t)
         (((IFX_int32_t)(IFX_int16_t)pContMeasRead->VlineTip * 14400) >> 15);
      pContMeas->nILineLong = (IFX_int16_t)
         (((IFX_int32_t)(IFX_int16_t)pContMeasRead->Ilong * 10000) >> 15);
      pContMeas->nTtxMeterReal = (IFX_int16_t)pContMeasRead->TtxReal;
      pContMeas->nTtxMeterImag = (IFX_uint16_t)pContMeasRead->TtxImag;
      pContMeas->nTtxMeterLen = (IFX_uint16_t)pContMeasRead->TtxLen;
      pContMeas->nITtxMeter = (IFX_int16_t)pContMeasRead->Ittx;
      pContMeas->nVTtxMeter = (IFX_int16_t)pContMeasRead->Vttx;

      pContMeas->nVBat = (IFX_int16_t)
         (((IFX_int32_t)(IFX_int16_t)pContMeasRead->Vbat * 14400) >> 15);

      pContMeas->nSlicTemp = (IFX_int16_t)pContMeasRead->SlicTemp;
#endif /* S220 */
   }

   RETURN_STATUS(ret);
}
#endif /* SDD_FEAT_CONT_MEASUREMENT */


/**
   Change the DUP timer configuration between normal and override mode.

   \param  pLLChannel   Pointer to SDD channel structure.
   \param  nStandbyTime DUP_OVERRIDE_OFF to clear override or time in ms

   \return
   - SDD_statusOk or status error

   \remarks
   Do not use time values below 2ms. There the hook detection behaves instable.
*/
static IFX_int32_t SDD_TAPI_LL_ALM_CfgDupTimer_Override (
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_uint8_t nStandbyTime)
{
   SDD_CHANNEL_t *pCh;
   SDD_DEVICE_t *pDev;
   SDD_BasicConfig_t *pSDD_BasicConfig, OverrideSddBasicConfig;
   IFX_int32_t  ret = SDD_statusOk;

   /* calling function should ensure valid parameters */
   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;
   SDD_ASSERT(pCh->pALM != IFX_NULL);
   pDev = pCh->pDev;
   /* DUP times below 2ms behave instable. */
   SDD_ASSERT((nStandbyTime >= 2) || (nStandbyTime == DUP_OVERRIDE_OFF));
   /* 15 ms is the upper limit of the standby DUP time */
   SDD_ASSERT(nStandbyTime <= 15);
   SDD_ASSERT(pDev != IFX_NULL);

   /* do nothing if state is already set */
   if (pCh->pALM->nStdbyDupTimeOverride == nStandbyTime)
   {
      return SDD_statusOk;
   }

   pSDD_BasicConfig = &pCh->pALM->fw_sdd_basic_config;

   if (nStandbyTime == DUP_OVERRIDE_OFF)
   {
      /* Normal mode */
      ret = SDD_CmdWrite (pDev, (IFX_uint32_t *)(IFX_void_t *)pSDD_BasicConfig);
   }
   else
   {
      /* Override mode */
      ret = SDD_CmdRead (pDev, (IFX_uint32_t *)(IFX_void_t *)pSDD_BasicConfig,
                               (IFX_uint32_t *)(IFX_void_t *)pSDD_BasicConfig);
      if (!SDD_SUCCESS (ret))
      {
         /* errmsg: Reading the basic configuration failed. */
         RETURN_STATUS(SDD_statusBasicCfgRdErr);
      }

      OverrideSddBasicConfig = *pSDD_BasicConfig;
#ifdef S220
      OverrideSddBasicConfig.ActiveDup = (IFX_uint32_t)(nStandbyTime & 0xF);
#else /* S220 */
      OverrideSddBasicConfig.OnhookDup = (IFX_uint32_t)(nStandbyTime & 0xF);
#endif /* S220 */
      ret = SDD_CmdWrite (pDev,
                         (IFX_uint32_t *)(IFX_void_t *)&OverrideSddBasicConfig);
   }

   /* Remember the value for the entrance check. */
   pCh->pALM->nStdbyDupTimeOverride = nStandbyTime;

   RETURN_STATUS(ret);
}


/**
   Disables the line in case of an emergency.

   This function is used directly from interrupt context to disable the line
   feeding in case of ground fault or SLIC overtemperature.

   \param  pCh             Pointer to the SDD channel structure.

   \remarks
   This function is called from interrupt context and is thus not protected.
*/
IFX_void_t irq_SDD_ALM_LineDisable (SDD_CHANNEL_t *pCh)
{
   SDD_ASSERT (pCh->pALM != IFX_NULL);

   if (pCh->pALM != IFX_NULL)
   {
      SDD_Opmode_t *pOpmod = &pCh->pALM->sdd_opmode;
#ifdef S220
      struct TTX_GEN_CTRL *pTtxGenCtrl = &pCh->pALM->ttx_gen_ctrl;
#endif /* S220 */

      pOpmod->OperatingMode = SDD_Opmode_Disabled;
#ifdef S220
      pTtxGenCtrl->EN = TTX_GEN_CTRL_DIS;
#else /* S220 */
      pOpmod->Polarity = SDD_Opmode_Pol_Normal;
      pOpmod->Howler = SDD_Opmode_HT_Speech;
      pOpmod->TtxBurst = SDD_Opmode_TTX_NO;
      pOpmod->MwiLamp = SDD_Opmode_MWL_Off;
#endif /* S220 */

      SDD_CmdWrite (pCh->pDev, (IFX_uint32_t *)pOpmod);

      pCh->pALM->curr_opmode = pOpmod->OperatingMode;
#ifndef S220
      pCh->pALM->curr_polarity = pOpmod->Polarity;
      pCh->pALM->curr_howler = pOpmod->Howler;
      pCh->pALM->curr_mwi_lamp = pOpmod->MwiLamp;
#endif /* S220 */
   }
}


/**
   Calculate the elapsed time since the last hook event.

   This function is to be called with each hook event. It calculates the
   elapsed time since the last hook event.

   The SDD provides a timestamp with each FW event. However, this timestamp
   has a wraparound. So an addtional timer is used to prevent calculation
   after a wraparound. This function wraps the calculation and also restarts
   the timer.

   For the SLIC 1xx family firmware the timestamp in the FW event has a
   granularity of 1/8 ms and the wraparound occurs after 8192ms.
   For the SLIC 2xx family firmware the timestamp in the FW event has a
   granularity of 1 ms and the wraparound occurs after 65536 ms.
   The timer to detect the wraparound is set to 7000ms as the largest usable
   value for both cases.

   \param  pCh          Pointer to SDD channel structure.
   \param  current_timestamp  Current timestamp value from the FW event.

   \return
   Elapsed time since the last hook event in 1/8 ms steps.
   If the last event happened too long ago 0xFFFF is retuned.
*/
IFX_uint16_t SDD_ALM_ElapsedTimeSinceLastHook (SDD_CHANNEL_t *pCh,
                                               IFX_uint16_t current_timestamp)
{
   IFX_uint16_t   elapsedTime;

   if(pCh == IFX_NULL)
      return 0;

   if(pCh->pALM == IFX_NULL)
      return 0;

   if (pCh->pALM->bHookWindow)
   {
      /* Calculate elapsed time since the saved timestamp. In case the 16-bit
         counter had a wraparound since the saved timestamp the calculation
         will still be correct because the calculation is done with 16-bit
         unsigned data type. This will always result in the positive distance
         between the two timestamps. */
      elapsedTime = current_timestamp - pCh->pALM->nLastHookEvtTimestamp;
#ifdef S220
      /* Multiply by 8 to get from 1 ms to 1/8 ms. 3 MSB are 0 before the shift,
       * otherwise bHookWindow would be FALSE. */
      elapsedTime <<= 3;
#endif /* S220 */
   }
   else
   {
      elapsedTime = 0xFFFF;
   }

   /* Start timer, length of timer = 7000 ms. This time needs to be below
      the wraparound of the HW timestamp. */
   if (pCh->pALM->nHookWindowTimerId != 0)
   {
      pCh->pALM->bHookWindow = IFX_TRUE;
      TAPI_SetTime_Timer (pCh->pALM->nHookWindowTimerId,
                          7000, IFX_FALSE, IFX_TRUE);
   }
   /* store the timestamp from this event */
   pCh->pALM->nLastHookEvtTimestamp = current_timestamp;

   return elapsedTime;
}


/*******************************************************************************
Description:
   Prepare parameters and call the Target Configuration Function to set
   the ring configuration.
Arguments:
   pChannel        - handle to TAPI_CHANNEL structure
   pRingConfig     - pointer to ring config structure
Return:
   IFX_SUCCESS or IFX_ERROR
*******************************************************************************/
IFX_int32_t SDD_TAPI_LL_ALM_Ring_Cfg_Set (IFX_TAPI_LL_CH_t *pLLChannel,
   IFX_TAPI_RING_CFG_t const *pRingConfig)
{
   IFX_int32_t    ret      =  IFX_SUCCESS;
   SDD_CHANNEL_t  *pCh;
   SDD_RingConfig_t SDD_RingConfig;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   
   memset(&SDD_RingConfig, 0, sizeof(SDD_RingConfig_t));

   /* Before modifying any registers check if all parameters are valid */
   if (pRingConfig->nMode != IFX_TAPI_RING_CFG_MODE_INTERNAL_BALANCED)
   {
      TRACE (SDD, DBG_LEVEL_HIGH,
             ("SDD_ERR: Invalid ring mode (%d)\n",
                pRingConfig->nMode));
      /* errmsg: Invalid ring mode.*/
      RETURN_STATUS(SDD_statusRingModeErr);
   }

   if (pRingConfig->nSubmode == IFX_TAPI_RING_CFG_SUBMODE_AC_RNG_TRIP_FAST)
   {
      TRACE (SDD, DBG_LEVEL_HIGH,
             ("SDD_ERR: Invalid ring submode (%d)\n",
                pRingConfig->nMode));
      /* errmsg: Invalid ring submode.*/
      RETURN_STATUS(SDD_statusRingSubmodeErr);
   }

   SDD_RingConfig.CMD    = CMD_SDD;
   SDD_RingConfig.ECMD   = SDD_RingConfig_ECMD;
   SDD_RingConfig.MOD    = MOD_SDD;
   SDD_RingConfig.CHAN   = pCh->nChannel-1;
   SDD_RingConfig.LENGTH = SDD_RingConfig_LENGTH;
   ret = SDD_CmdRead (pCh->pDev, (IFX_uint32_t *)&SDD_RingConfig,
                                    (IFX_uint32_t *)&SDD_RingConfig);
   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Reading the ring configuration failed. */
      RETURN_STATUS(SDD_statusRingCfgRdErr);
   }

   switch (pRingConfig->nSubmode)
   {
   case IFX_TAPI_RING_CFG_SUBMODE_DC_RNG_TRIP_STANDARD:
      SDD_RingConfig.RingTripType = SDD_RingConfig_DC_Ring_Trip;
      break;
   case IFX_TAPI_RING_CFG_SUBMODE_DC_RNG_TRIP_FAST:
      SDD_RingConfig.RingTripType = SDD_RingConfig_Fast_Ring_Trip;
      break;
   case IFX_TAPI_RING_CFG_SUBMODE_AC_RNG_TRIP_STANDARD:
      SDD_RingConfig.RingTripType = SDD_RingConfig_AC_Ring_Trip;
      break;
   default:
      break;
   }
   ret = SDD_CmdWrite (pCh->pDev, (IFX_uint32_t *)&SDD_RingConfig);
   if (!SDD_SUCCESS (ret))
   {
      RETURN_STATUS(SDD_statusRingCfgWrErr);
   }
   return ret;
}


/*******************************************************************************
Description:
   Get the ring configuration.
Arguments:
   pChannel        - handle to TAPI_CHANNEL structure
   pRingConfig     - pointer to ring config structure where to store the result
Return:
   IFX_SUCCESS
*******************************************************************************/
IFX_int32_t SDD_TAPI_LL_ALM_Ring_Cfg_Get (IFX_TAPI_LL_CH_t *pLLChannel,
                                          IFX_TAPI_RING_CFG_t *pRingConfig)
{
   IFX_int32_t    ret      =  IFX_SUCCESS;
   SDD_CHANNEL_t  *pCh;
   SDD_RingConfig_t SDD_RingConfig = {0};

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }
   
   memset(&SDD_RingConfig, 0, sizeof(SDD_RingConfig_t));

   SDD_RingConfig.CMD    = CMD_SDD;
   SDD_RingConfig.ECMD   = SDD_RingConfig_ECMD;
   SDD_RingConfig.MOD    = MOD_SDD;
   SDD_RingConfig.CHAN   = pCh->nChannel-1;
   SDD_RingConfig.LENGTH = SDD_RingConfig_LENGTH;
   ret = SDD_CmdRead (pCh->pDev, (IFX_uint32_t *)&SDD_RingConfig,
                                    (IFX_uint32_t *)&SDD_RingConfig);
   if (!SDD_SUCCESS (ret))
   {
       /* errmsg: Reading the ring configuration failed. */
      RETURN_STATUS(SDD_statusRingCfgRdErr);
   }

   pRingConfig->nMode = IFX_TAPI_RING_CFG_MODE_INTERNAL_BALANCED;
   switch (SDD_RingConfig.RingTripType)
   {
   case SDD_RingConfig_DC_Ring_Trip:
      pRingConfig->nSubmode = IFX_TAPI_RING_CFG_SUBMODE_DC_RNG_TRIP_STANDARD;
      break;
   case SDD_RingConfig_Fast_Ring_Trip:
      pRingConfig->nSubmode = IFX_TAPI_RING_CFG_SUBMODE_DC_RNG_TRIP_FAST;
      break;
   case SDD_RingConfig_AC_Ring_Trip:
      pRingConfig->nSubmode = IFX_TAPI_RING_CFG_SUBMODE_AC_RNG_TRIP_STANDARD;
      break;
   default:
      break;
   }
   return ret;
}

/**
   Timer return function to supervise the hook event window.

   To prevent calculation after a wraparound this timer resets the HookWindow
   flag.

   \param  Timer        Timer context.
   \param  arg          Pointer to a SDD channel structure.
*/
static IFX_void_t sdd_alm_HookWindow_OnTimer (Timer_ID Timer,
                                              IFX_ulong_t arg)
{
   SDD_CHANNEL_t       *pCh        = (SDD_CHANNEL_t *)arg;
   SDD_OS_UNUSED (Timer);

   if ((pCh != IFX_NULL) && (pCh->pALM != IFX_NULL))
   {
      pCh->pALM->bHookWindow = IFX_FALSE;
   }
}

/**
   Free data structure of the ALM module in the given channel.

   \param  pCh             Pointer to the SDD channel structure.
*/
IFX_void_t SDD_ALM_Free_Ch_Structures (SDD_CHANNEL_t *pCh)
{
   if (pCh->pALM != IFX_NULL)
   {
      IFXOS_BlockFree(pCh->pALM);
      pCh->pALM = IFX_NULL;
   }
}

/**
   Allocate data structure of the ALM module in the given channel.

   \param  pCh             Pointer to the SDD channel structure.

   \return
      - SDD_statusOk if ok
      - SDD_statusNoMem Memory allocation failed -> struct not created

   \remarks The channel parameter is no longer checked because the calling
   function assures correct values.
*/
IFX_int32_t SDD_ALM_Allocate_Ch_Structures (SDD_CHANNEL_t *pCh)
{
   SDD_ALM_Free_Ch_Structures (pCh);

   pCh->pALM = IFXOS_BlockAlloc(sizeof(*pCh->pALM));
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: No memory could be allocated. */
      RETURN_STATUS(SDD_statusNoMem);
   }
   memset(pCh->pALM, 0, sizeof(*pCh->pALM));

   return SDD_statusOk;
}

/**
   Initalize the cached firmware message for the analog line / Smart
   Device Driver.

   \param  pCh          Pointer to the channel structure.

   \return
   None.

   \remark
   The SDD_Opmode message is used to write and read all parameters that are
   needed for setting the desired operating mode of the Analog Line / Smart
   Device Driver
*/
IFX_void_t SDD_ALM_InitCh (SDD_CHANNEL_t *pCh)
{
   SDD_Opmode_t              *pOpmode        = IFX_NULL;
   SDD_TxRxGain_t            *pSddTxRxGain   = IFX_NULL;
   SDD_BasicConfig_t         *pBasicCfg      = IFX_NULL;
   SDD_RingConfig_t          *pRingCfg       = IFX_NULL;
   SDD_Calibrate_t           *pCalibrate     = IFX_NULL;
#ifdef S220
   struct TTX_GEN_CTRL       *pTtxGenCtrl    = IFX_NULL;
#endif /* S220 */
#ifdef SDD_FEAT_GR909
   #ifndef S220
      SDD_GR909Config_t      *pGR909Config   = IFX_NULL;
   #endif /* !S220 */
#endif /* SDD_FEAT_GR909 */
#ifdef SDD_FEAT_CAP_MEASURE
   #ifdef S220
      SDD_CapacitanceMeasurementRead_t
                             *pCapMeas       = IFX_NULL;
   #else /* S220 */
      SDD_GR909PhoneDetection_t
                             *pCapMeas       = IFX_NULL;
   #endif /* S220 */
#endif /* SDD_FEAT_CAP_MEASURE */
#ifdef SDD_FEAT_AC_LEVEL_METERING
   SDD_AcLevelMeterControl_t *pAcLMControl   = IFX_NULL;
   SDD_AcLevelMeterControl_t *pAcLMConfig    = IFX_NULL;
   SDD_AcLevelMeterControl_t *pAcLMResult    = IFX_NULL;
#endif /* SDD_FEAT_AC_LEVEL_METERING */
   IFX_uint8_t               ch = pCh->nChannel - 1;

   /* Called internally. Caller must ensure correct channel parameter. */
   SDD_ASSERT(pCh != IFX_NULL);
   SDD_ASSERT(pCh->pALM != IFX_NULL);

   memset(pCh->pALM, 0, sizeof(*pCh->pALM));

   /* default DC/DC type at startup */
#ifdef S220
   pCh->pALM->nDcDcType = SDD_DCDC_TYPE_NOTSET;
#else /* S220 */
   pCh->pALM->nDcDcType = SDD_DCDC_TYPE_DEFAULT_IBB;
#endif /* S220 */


   /* set the FXS / FXO flag */
   if (((ch == 0) && !pCh->pDev->caps.FXO0) ||
       ((ch == 1) && !pCh->pDev->caps.FXO1))
      pCh->pALM->line_type_fxs = IFX_TRUE;
   else
      pCh->pALM->line_type_fxs = IFX_FALSE;

   if (pCh->pALM->line_type_fxs == IFX_TRUE)
   {
      /* create timer to supervise elapsed time since the last hook event */
      pCh->pALM->nHookWindowTimerId =
         TAPI_Create_Timer((TIMER_ENTRY)sdd_alm_HookWindow_OnTimer,
                           (IFX_ulong_t)pCh);
   }

   /* begin with disabled line feeding */
   pCh->pALM->curr_opmode = SDD_Opmode_Disabled;
#ifndef S220
   pCh->pALM->curr_polarity = SDD_Opmode_Pol_Normal;
   pCh->pALM->curr_howler = SDD_Opmode_HT_Speech;
   pCh->pALM->curr_mwi_lamp = SDD_Opmode_MWL_Off;
#endif /* S220 */
   pCh->pALM->bOpmodeChangePending = IFX_FALSE;

#ifdef S220
   /* setup Metering Pulse Control */
   pTtxGenCtrl = &pCh->pALM->ttx_gen_ctrl;
   memset (pTtxGenCtrl, 0, sizeof (*pTtxGenCtrl));
   pTtxGenCtrl->CMD         = CMD_EOP;
   pTtxGenCtrl->CHAN        = ch;
   pTtxGenCtrl->MOD         = MOD_SIG_GEN;
   pTtxGenCtrl->ECMD        = TTX_GEN_CTRL_ECMD;
   pTtxGenCtrl->LENGTH      = TTX_GEN_CTRL_LENGTH;
   pTtxGenCtrl->EN          = TTX_GEN_CTRL_DIS;
#endif /* S220 */

   /* setup opmode control cmd */
   pOpmode = &pCh->pALM->sdd_opmode;
   memset (pOpmode, 0, sizeof (*pOpmode));
   pOpmode->CMD         = CMD_SDD;
   pOpmode->CHAN        = ch;
   pOpmode->MOD         = MOD_SDD;
   pOpmode->ECMD        = SDD_Opmode_ECMD;
   pOpmode->LENGTH      = SDD_Opmode_LENGTH;
#ifndef S220
   pOpmode->Polarity    = SDD_Opmode_Pol_Normal;
   pOpmode->Howler      = SDD_Opmode_HT_Speech;
   pOpmode->TtxBurst    = SDD_Opmode_TTX_NO;
#endif /* S220 */
   pOpmode->OperatingMode = SDD_Opmode_Disabled;

   /* setup gain cmd */
   pSddTxRxGain = &pCh->pALM->sdd_txrx_gain;
   memset (pSddTxRxGain, 0, sizeof (*pSddTxRxGain));
   pSddTxRxGain->CMD    = CMD_SDD;
   pSddTxRxGain->CHAN   = ch;
   pSddTxRxGain->MOD    = MOD_SDD;
   pSddTxRxGain->ECMD   = SDD_TxRxGain_ECMD;
   pSddTxRxGain->LENGTH = SDD_TxRxGain_LENGTH;

   /* setup basic configuration command */
   pBasicCfg = &pCh->pALM->fw_sdd_basic_config;
   memset (pBasicCfg, 0, sizeof (*pBasicCfg));
   pBasicCfg->CMD       = CMD_SDD;
   pBasicCfg->CHAN      = ch;
   pBasicCfg->MOD       = MOD_SDD;
   pBasicCfg->ECMD      = SDD_BasicConfig_ECMD;
   pBasicCfg->LENGTH    = SDD_BasicConfig_LENGTH;

   /* Setup copy of basic config for MWL feature. */
   pCh->pALM->sdd_basic_config_tmp_mwl = *pBasicCfg;

   /* setup ring configuration command */
   pRingCfg = &pCh->pALM->sdd_ring_config_tmp_mwl;
   memset (pRingCfg, 0, sizeof (*pRingCfg));
   pRingCfg->CMD        = CMD_SDD;
   pRingCfg->CHAN       = ch;
   pRingCfg->MOD        = MOD_SDD;
   pRingCfg->ECMD       = SDD_RingConfig_ECMD;
   pRingCfg->LENGTH     = SDD_RingConfig_LENGTH;

   /* Reading calibration data */
   pCalibrate = &pCh->pALM->fw_sdd_calibrate;
   memset (pCalibrate, 0, sizeof (*pCalibrate));
   pCalibrate->CMD      = CMD_SDD;
   pCalibrate->CHAN     = ch;
   pCalibrate->MOD      = MOD_SDD;
   pCalibrate->ECMD     = SDD_Calibrate_ECMD;
   pCalibrate->LENGTH   = SDD_Calibrate_LENGTH;

#ifdef SDD_FEAT_CAP_MEASURE
   /* setup SDD_CapacitanceMeas command */
   pCapMeas = &pCh->pALM->fw_sdd_capacitance_meas;
   memset (pCapMeas, 0, sizeof(*pCapMeas));
   pCapMeas->CMD        = CMD_SDD;
   pCapMeas->CHAN       = ch;
   #ifdef S220
      pCapMeas->MOD     = MOD_SDD;
      pCapMeas->ECMD    = SDD_CapacitanceMeasurementRead_ECMD;
      pCapMeas->LENGTH  = SDD_CapacitanceMeasurementRead_LENGTH;
   #else /* S220 */
      pCapMeas->MOD     = MOD_SDD1;
      pCapMeas->ECMD    = SDD_GR909PhoneDetection_ECMD;
      pCapMeas->LENGTH  = SDD_GR909PhoneDetection_LENGTH;
   #endif /* S220 */
#endif /* SDD_FEAT_CAP_MEASURE */

#ifdef SDD_FEAT_AC_LEVEL_METERING
   /* Setup SDD_AcLevelMeterControl command */
   pAcLMControl = &pCh->pALM->fw_sdd_acLevelMeterControl;
   memset (pAcLMControl, 0, sizeof (pAcLMControl));
   pAcLMControl->CMD  = CMD_SDD;
   pAcLMControl->CHAN = ch;
   pAcLMControl->MOD  = MOD_SDD;
   pAcLMControl->ECMD = SDD_AcLevelMeterControl_ECMD;
   pAcLMControl->LENGTH = SDD_AcLevelMeterControl_LENGTH;
   /* Setup SDD_AcLevelMeterConfig command */
   pAcLMConfig = &pCh->pALM->fw_sdd_acLevelMeterControl;
   memset (pAcLMConfig, 0, sizeof (pAcLMConfig));
   pAcLMConfig->CMD  = CMD_SDD;
   pAcLMConfig->CHAN = ch;
   pAcLMConfig->MOD  = MOD_SDD;
   pAcLMConfig->ECMD = SDD_AcLevelMeterConfig_ECMD;
   pAcLMConfig->LENGTH = SDD_AcLevelMeterConfig_LENGTH;
   /* Setup SDD_AcLevelMeterResult command */
   pAcLMResult = &pCh->pALM->fw_sdd_acLevelMeterControl;
   memset (pAcLMResult, 0, sizeof (pAcLMResult));
   pAcLMResult->CMD  = CMD_SDD;
   pAcLMResult->CHAN = ch;
   pAcLMResult->MOD  = MOD_SDD;
   pAcLMResult->ECMD = SDD_AcLevelMeterResult_ECMD;
   pAcLMResult->LENGTH = SDD_AcLevelMeterResult_LENGTH;
#endif /* SDD_FEAT_AC_LEVEL_METERING */

#ifdef SDD_FEAT_CONT_MEASUREMENT
   {
      SDD_ContMeasRead_t *pContMeasRead = IFX_NULL;

      /* setup SDD_ContMeasRead command */
      pContMeasRead  = &pCh->pALM->fw_sdd_contMeasRead;
      memset (pContMeasRead, 0, sizeof (*pContMeasRead));
      pContMeasRead->CMD   = CMD_SDD;
      pContMeasRead->CHAN  = ch;
      pContMeasRead->MOD   = MOD_SDD;
      pContMeasRead->ECMD  = SDD_ContMeasRead_ECMD;
      pContMeasRead->LENGTH = SDD_ContMeasRead_LENGTH;
   }
#endif /* SDD_FEAT_CONT_MEASUREMENT */
#ifdef SDD_FEAT_CONT_MEASUREMENT_RESET
   {
      SDD_ContMeasClear_t *pContMeasClear = IFX_NULL;

      /* setup SDD_ContMeasClear command */
      pContMeasClear = &pCh->pALM->fw_sdd_contMeasClear;
      memset (pContMeasClear, 0, sizeof (*pContMeasClear));
      pContMeasClear->CMD  = CMD_SDD;
      pContMeasClear->CHAN = ch;
      pContMeasClear->MOD  = MOD_SDD;
      pContMeasClear->ECMD = SDD_ContMeasClear_ECMD;
      pContMeasClear->LENGTH = SDD_ContMeasClear_LENGTH;
   }
#endif /* SDD_FEAT_CONT_MEASUREMENT_RESET */

#ifdef SDD_FEAT_NLT
   /* setup the default configuration of the measurement path for
      line testing.*/
   pCh->pALM->nRmeas = IFX_TAPI_NLT_RMEAS_1MOHM;

#ifndef S220
   /* Set default values for resistance open loop calibration factors. */
   pCh->pALM->nlt_ResistanceConfig.fOlResTip2Ring = -5000000;
   pCh->pALM->nlt_ResistanceConfig.fOlResTip2Gnd = -1500000;
   pCh->pALM->nlt_ResistanceConfig.fOlResRing2Gnd = -1500000;
#endif /* S220 */

   pCh->pALM->nlt_CapacitanceConfig.fOlCapTip2Ring = 7.5;
   pCh->pALM->nlt_CapacitanceConfig.fOlCapTip2Gnd = 15;
   pCh->pALM->nlt_CapacitanceConfig.fOlCapRing2Gnd = 15;
#endif /* SDD_FEAT_NLT */

#ifdef SDD_FEAT_GR909
#ifndef S220
   /* setup SDD_GR909Config command */
   pGR909Config = &pCh->pALM->sdd_gr909_config;
   memset(pGR909Config, 0, sizeof(*pGR909Config));
   pGR909Config->CMD    = CMD_SDD;
   pGR909Config->CHAN   = ch;
   pGR909Config->MOD    = MOD_SDD;
   pGR909Config->ECMD   = SDD_GR909Config_ECMD;
   pGR909Config->LENGTH = SDD_GR909Config_LENGTH;

   /* Set default limits. */
   SDD_ALM_GR909_SetLimits(pCh, pCh->pALM->nRmeas);
#endif /* !S220 */
#endif /* SDD_FEAT_GR909 */

#ifdef SDD_FEAT_CAP_MEASURE
   /* Initial state of capacitance measurement */
   pCh->pALM->eCapMeasState = SDD_CapMeasState_Inactive;
   /* Capacitance measurement is not running */
   pCh->pALM->bCapMeasInProgress = IFX_FALSE;
#endif /* SDD_FEAT_CAP_MEASURE */

   /* Initialise the semaphore to wait for calibration to finish. */
   SDD_OS_MutexInit(&pCh->pALM->mtxCalibrationWait);
   SDD_OS_MutexGet(&pCh->pALM->mtxCalibrationWait);
   /* No calibration was done yet. */
   pCh->pALM->nCalibrationState = IFX_TAPI_CALIBRATION_STATE_NO;
   /* Currently calibration is not running. */
   pCh->pALM->bCalibrationRunning = IFX_FALSE;
   /* No calibration is needed - flag will be set during BBD download. */
   pCh->pALM->bCalibrationNeeded = IFX_FALSE;
   /* Clear the cache with the last calibration results. */
   memset(&pCh->pALM->calibrationLastResults, 0,
          sizeof(pCh->pALM->calibrationLastResults));

   /* initialize SDD event */
   IFXOS_EventInit (&pCh->pALM->sdd_event);

#ifdef SDD_FEAT_FXO
   SDD_ALM_FXO_InitCh (pCh);
#endif /* SDD_FEAT_FXO */
}


/**
   Basic ALM configuration

   Use this function where needed to set the base configuration
   of FXO function in the Analog Line Module.

   \param  pCh          Pointer to the SDD channel structure.

   \return
   - SDD_statusOk if successful
   - error code from SDD_CmdWrite()
*/
IFX_int32_t SDD_ALM_BaseConf(IFX_TAPI_LL_CH_t *pLLChannel)
{
   IFX_int32_t    ret = SDD_statusOk;
   SDD_CHANNEL_t  *pCh = IFX_NULL;;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   if (pCh == IFX_NULL)
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;

   if (pCh->pALM == IFX_NULL)
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);

#ifdef SDD_FEAT_FXO
   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      ret = SDD_ALM_FXO_BaseConf(pCh);
   }
#endif /* SDD_FEAT_FXO */

   RETURN_STATUS(ret);
}


/**
   Stop ALM on this channel

   \param  pCh             Pointer to the SDD channel structure.
   \param  bChipAccess

   \return
*/
IFX_int32_t SDD_ALM_ChStop (SDD_CHANNEL_t *pCh,
                            IFX_boolean_t bChipAccess)
{
   IFX_int32_t          ret   = SDD_statusOk;

   /* calling function should ensure valid parameters */
   SDD_ASSERT(pCh != IFX_NULL);
   SDD_ASSERT(pCh->pTapiCh != IFX_NULL);
   SDD_ASSERT(pCh->pTapiCh->pLLChannel != IFX_NULL);

   if (pCh->pALM != IFX_NULL)
   {
      IFXOS_MutexGet (&pCh->mtxChAcc);

      if (pCh->pALM->nHookWindowTimerId != 0)
      {
         TAPI_Delete_Timer (pCh->pALM->nHookWindowTimerId);
      }

      if (bChipAccess != IFX_FALSE)
      {
         /* set the line to disabled (this implicitly writes the fw message) */
         ret = SDD_TAPI_LL_ALM_Line_Mode_Set (pCh->pTapiCh->pLLChannel,
                                              IFX_TAPI_LINE_FEED_DISABLED, 0);
      }

      /* Reset of the DC/DC type to default allows to download the BBD again. */
#ifdef S220
      pCh->pALM->nDcDcType = SDD_DCDC_TYPE_NOTSET;
#else /* S220 */
      pCh->pALM->nDcDcType = SDD_DCDC_TYPE_DEFAULT_IBB;
#endif /* S220 */

      IFXOS_MutexRelease (&pCh->mtxChAcc);

      /* Delete the semaphore to wait for calibration to finish. */
      SDD_OS_MutexRelease(&pCh->pALM->mtxCalibrationWait);
      SDD_OS_MutexDelete(&pCh->pALM->mtxCalibrationWait);
   }

   RETURN_STATUS(ret);
}


/**
   Correct the cached linemode inside the low level device driver.

   \param  pCh          Pointer to SDD channel structure.
   \param  lm           New linemode.
*/
IFX_void_t SDD_ALM_CorrectLinemodeCache(SDD_CHANNEL_t *pCh, IFX_uint16_t lm)
{
   SDD_ASSERT (pCh->pALM != IFX_NULL);

   if (pCh->pALM != IFX_NULL)
   {
      pCh->pALM->sdd_opmode.OperatingMode = lm;
   }
}

/**
   Updates current line operating mode.

   This function is used directly from interrupt context to update
   the current line operating mode of an analog channel and wake up
   the thread waiting on sdd_event. The sdd_event informs the waiting
   thread of complete operating mode transition.

   \param  pCh             Pointer to the SDD channel structure.
   \param  lm              Current line mode reported by firmware.
*/
IFX_void_t irq_SDD_ALM_UpdateOpModeAndWakeUp (SDD_CHANNEL_t *pCh,
                                              IFX_uint8_t lm)
{
   SDD_ASSERT (pCh->pALM != IFX_NULL);

   if (pCh->pALM != IFX_NULL)
   {
      if (lm != OPMODE_IGNORED)
      {
         pCh->pALM->curr_opmode = lm;
      }

      pCh->pALM->bOpmodeChangePending = IFX_FALSE;
      IFXOS_EventWakeUp (&pCh->pALM->sdd_event);
   }
}


#ifdef SDD_FEAT_CAP_MEASURE
IFX_boolean_t  SDD_ALM_CapMeasInProgress (SDD_CHANNEL_t *pCh)
{
   SDD_ASSERT ((pCh != IFX_NULL) && (pCh->pALM != IFX_NULL));

   return pCh->pALM->bCapMeasInProgress;
}
#endif /* SDD_FEAT_CAP_MEASURE */


#ifdef SDD_FEAT_PROCFS
/**
   Linux procfs read the status of PCM.

   \param  s            Pointer to seq_file struct.
   \param  pDev         Pointer to the device structure.
   \return  0 on success
*/
IFX_int32_t SDD_ALM_proc_StatusGet (
                        struct seq_file *s,
                        SDD_DEVICE_t *pDev)
{
   IFX_uint8_t nCh;
   SDD_CHANNEL_t *pCh  = IFX_NULL;
   const char *pConverterName;

   seq_printf(s, "  SSI status : %s (recoveries: %d)\n",
                 pDev->bSSIcrash ? "CRASHED!" : "ok",
                 pDev->nSlicRecoveryCnt);

   /* Attached DC/DC converter type */
   seq_printf(s, "  Allowed DC/DC type: %s\n",
              sdd_alm_DcDcEnum2Description (pDev->nAllowedDcDcType));

   /* SDD channel status */
   for (nCh = 0; nCh < pDev->caps.nALI; nCh++)
   {
      pCh = &pDev->pChannel[nCh];

      if (pCh->pALM)
      {
         if (pCh->pALM->line_type_fxs == IFX_FALSE)
         {
            pConverterName = "n/a";
         }
         else
         {
            pConverterName =
               sdd_alm_DcDcEnum2Description (pCh->pALM->nDcDcType);
         }

         seq_printf(s, "  Channel %d: %s - %s\n",
                    nCh,
                    pCh->pALM->line_type_fxs ? "FXS" : "FXO",
                    pConverterName);
      }
   } /* for all ALI channels */

   return 0;
}
#endif /* SDD_FEAT_PROCFS */


#ifdef SDD_FEAT_PROCFS
/**
   Translate the DC/DC converter enum into the description string.

   \param  dcdc_type    Value from enum SDD_DCDC_TYPE.

   \return C-style string describing the given enum value.
*/
static const char *sdd_alm_DcDcEnum2Description (
                        enum SDD_DCDC_TYPE dcdc_type)
{
   IFX_uint8_t i;

   for (i=0; i < (sizeof(DcDcTypes)/sizeof(struct DcDcTranslation)); i++)
   {
      if (dcdc_type == DcDcTypes[i].type)
      {
         return DcDcTypes[i].description;
      }
   }

   /* Nothing found - this can never happen but is handled here anyway. */
   return "UNKNOWN";
}
#endif /* SDD_FEAT_PROCFS */


/**
   Translate the DC/DC converter type string into an enum type.

   \param  string       Pointer to C-style character string.

   \return Enum value corresponding to the string.
*/
enum SDD_DCDC_TYPE SDD_ALM_DcDcString2Enum (
                        const char *string)
{
   char string_copy[6] = {0};
   IFX_uint8_t i;

   /* To strip voltage digits from the string copy only uppercase characters
      from the beginning of the string. */
   for (i=0; (i < 6) && (string[i] >= 'A') && (string[i] <= 'Z'); i++)
   {
      string_copy[i] = string[i];
   }

   /* Lookup any non zero length string in the table. */
   if (strlen(string_copy) > 0)
   {
      for (i=0; i < (sizeof(DcDcTypes)/sizeof(struct DcDcTranslation)); i++)
      {
         if (strcmp(string_copy, DcDcTypes[i].name) == 0)
         {
            return DcDcTypes[i].type;
         }
      }
   }

   /* Nothing found */
   return SDD_DCDC_TYPE_NOTFOUND;
}


#ifdef SDD_FEAT_CAP_MEASURE
/**
   Check capacitance measurement support

   This function is used by phone detection to find if the chip supports the
   tip to ring capacitance measurement.

   \param  pLLChannel   Pointer to TAPI low level channel structure.
   \param  pSupported   Returns information whether capacitance measurement
                        is supported (IFX_TRUE) or not (IFX_FALSE).

   \return
   - SDD_statusOk or status error from SDD_CmdWrite
*/
IFX_int32_t SDD_TAPI_LL_ALM_LT_CheckCapMeasSup (IFX_TAPI_LL_CH_t *pLLChannel,
                                                IFX_boolean_t *pSupported)
{
   SDD_CHANNEL_t         *pCh;
   SDD_DEVICE_t          *pDev;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }

   pDev = pCh->pDev;

   /* Using the combined DC/DC operation the capacitance measurement requires
      that the other channel is disabled. In this case report that there is
      no capacitance measurement. */
   *pSupported = (pDev->bDcDcHwCombined != IFX_TRUE) ? IFX_TRUE : IFX_FALSE;

   RETURN_STATUS(SDD_statusOk);
}
#endif /* SDD_FEAT_CAP_MEASURE */


#ifdef SDD_FEAT_CAP_MEASURE
/**
   Start an analog line capacitance measurement.

   \param  pCh pointer to SDD_CHANNEL structure

   \return

   \remarks
      This function configures GR909 for phone detection
      and starts the GR909 measurement in FW.
*/
static IFX_int32_t SDD_alm_Capacitance_Measurement_Start(SDD_CHANNEL_t *pCh)
{
   IFX_int32_t            ret;

#ifdef SDD_FEAT_NLT
   /* clear previous results */
   memset (&(pCh->pALM->t2r_cap_result), 0, sizeof(pCh->pALM->t2r_cap_result));
   memset (&(pCh->pALM->l2g_cap_result), 0, sizeof(pCh->pALM->l2g_cap_result));
#endif /* SDD_FEAT_NLT */

#ifdef S220
   pCh->pALM->sdd_opmode.OperatingMode =
      pCh->pALM->bCapMeasTip2RingOnly ? SDD_Opmode_PDF : SDD_Opmode_CAP_MEAS;
   ret = SDD_ALM_OpModeSet (pCh);
#else /* S220 */
   {
      SDD_GR909Config_t *pSddGR909Config = &pCh->pALM->sdd_gr909_config;
      SDD_DEVICE_t *pDev = pCh->pDev;

      /* configure SDD to start line capacitance measurement */
      pSddGR909Config->Pdf = 1;
      /* clear all GR909 tests */
      pSddGR909Config->Hpt =
      pSddGR909Config->Femf =
      pSddGR909Config->Rft =
      pSddGR909Config->Roh =
      pSddGR909Config->Rit = 0;

      ret = SDD_CmdWrite(pDev, (IFX_uint32_t *)pSddGR909Config);
   }

   if (SDD_SUCCESS(ret))
   {
      pCh->pALM->sdd_opmode.OperatingMode = SDD_Opmode_GR909;
      ret = SDD_ALM_OpModeSet (pCh);
   }
#endif /* S220 */

   return ret;
}
#endif /* SDD_FEAT_CAP_MEASURE */


#ifdef SDD_FEAT_CAP_MEASURE
/**
   Start an analog line capacitance measurement session.

   \param  pLLChannel   Handle to TAPI low level channel structure.
   \param  bTip2RingOnly  not used.

   \return

   \remarks
      This function enables Opmode change events from FW,
      then sets the line Opmode to PDH. If the line mode already
      was PDH, the function starts the capacitance measurement.
      Otherwise the capacitance measurement start will be
      triggered by Opmode change event (PDH) from FW.
      The reason for this solution is that GR909 measurement
      can only be started in PDH line mode, so the driver has
      to be sure that FW has completed the switch to PDH.
*/
static IFX_int32_t SDD_TAPI_LL_ALM_CapMeasStart(IFX_TAPI_LL_CH_t *pLLChannel,
                                                 IFX_boolean_t bTip2RingOnly)
{
   SDD_CHANNEL_t         *pCh;
   IFX_uint8_t  curr_opmode;
   IFX_int32_t  ret = SDD_statusOk;

   SDD_OS_UNUSED(bTip2RingOnly);
   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   SDD_ASSERT(pCh != IFX_NULL);
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS (SDD_statusInvalCh);
   }

   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      /* errmsg: Protection from calling on FXO line */
      RETURN_STATUS (SDD_statusFXSCallOnFXO);
   }

#ifdef S220
   /* Until the DC/DC converter type is set this operation is not allowed. */
   if (pCh->pALM->nDcDcType == SDD_DCDC_TYPE_NOTSET)
#else /* S220 */
   /* For DC/DC converters other than default a BBD download is required. */
   if ((pCh->pALM->nDcDcType == SDD_DCDC_TYPE_DEFAULT_IBB) &&
       (pCh->pDev->nAllowedDcDcType != SDD_DCDC_TYPE_IBB))
#endif /* S220 */
   {
      /* errmsg: Operation blocked until a BBD with DC/DC configuration
                 matching the connected DC/DC hardware type is downloaded. */
      RETURN_STATUS(SDD_statusBlockedDcDcTypeMissmatch);
   }

   if (pCh->pALM->eCapMeasState != SDD_CapMeasState_Inactive)
   {
      /* errmsg: Capacitance measurement already in progress */
      RETURN_STATUS (SDD_statusCapMeasStartWhileActive);
   }

   pCh->pALM->eCapMeasState = SDD_CapMeasState_Started;
   pCh->pALM->bCapMeasTip2RingOnly = bTip2RingOnly;
   pCh->pALM->bCapMeasInProgress = IFX_TRUE;

   /* Get the current opmode - wait while an opmode change is ongoing. */
   ret = SDD_ALM_OpmodeGet (pCh, &curr_opmode);
   /* Only exit when interrupted by a signal. When waiting was aborted by
      timeout assume that no linemode change is pending and try to continue. */
   if (ret != SDD_statusSddEvtWaitInterrupt)
   {
      if (curr_opmode != SDD_Opmode_Disabled)
      {
         /* set PDH line mode */
         pCh->pALM->sdd_opmode.OperatingMode = SDD_Opmode_Disabled;
         ret = SDD_ALM_OpModeSet (pCh);
      }

      if (SDD_SUCCESS(ret))
      {
         ret = SDD_alm_Capacitance_Measurement_Start(pCh);
      }
   }

   RETURN_STATUS(ret);
}
#endif /* SDD_FEAT_CAP_MEASURE */


#ifdef SDD_FEAT_CAP_MEASURE
/**
   Stop any running analog line capacitance measurement session.

   \param  pLLChannel   Handle to TAPI low level channel structure.

   \return

*/
static IFX_int32_t SDD_TAPI_LL_ALM_CapMeasStop(IFX_TAPI_LL_CH_t *pLLChannel)
{
   SDD_CHANNEL_t         *pCh;
   IFX_int32_t  ret = IFX_SUCCESS;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   SDD_ASSERT(pCh != IFX_NULL);
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS (SDD_statusInvalCh);
   }

   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      /* errmsg: Protection from calling on FXO line */
      RETURN_STATUS (SDD_statusFXSCallOnFXO);
   }

   pCh->pALM->sdd_opmode.OperatingMode = SDD_Opmode_Disabled;
   ret = SDD_ALM_OpModeSet (pCh);

   if (SDD_SUCCESS(ret))
   {
      pCh->pALM->eCapMeasState = SDD_CapMeasState_Inactive;
      pCh->pALM->bCapMeasInProgress = IFX_FALSE;
   }

   RETURN_STATUS(ret);
}
#endif /* SDD_FEAT_CAP_MEASURE */


#ifdef SDD_FEAT_CAP_MEASURE
/**
   Used to update the driver status and the application when the FW reports
   that the capacitance measurement was aborted.

   \param  pCh          Pointer to SDD_CHANNEL structure.

   \return

*/
IFX_void_t SDD_ALM_CapMeasAbort(SDD_CHANNEL_t *pCh)
{
   IFX_TAPI_EVENT_t tapiEvent;

   /* sanity check */
   SDD_ASSERT(pCh != IFX_NULL);

   memset(&tapiEvent, 0, sizeof(tapiEvent));
   tapiEvent.ch = pCh->nChannel - 1;
   tapiEvent.id = IFX_TAPI_EVENT_LINE_MEASURE_CAPACITANCE_RDY;
   tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
   tapiEvent.data.lcap.nReturnCode = IFX_ERROR;
   IFX_TAPI_Event_Dispatch(pCh->pTapiCh, &tapiEvent);

   if (pCh->pALM != IFX_NULL)
   {
      pCh->pALM->eCapMeasState = SDD_CapMeasState_Inactive;
      pCh->pALM->bCapMeasInProgress = IFX_FALSE;

      if (pCh->pALM->bCapMeasTip2RingOnly == IFX_FALSE)
      {
         /* Capacitance measurement was started by user.
            Report that capacitance measurement is finished and results can be
            read by the IFX_TAPI_NLT_CAPACITANCE_RESULT_GET */
         tapiEvent.id = IFX_TAPI_EVENT_NLT_END;
         tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
         IFX_TAPI_Event_Dispatch(pCh->pTapiCh, &tapiEvent);
      }
   }
}
#endif /* SDD_FEAT_CAP_MEASURE */


#ifdef SDD_FEAT_CAP_MEASURE
/**
   Read capacitance measurement result.

   \param  pLLChannel   Handle to TAPI low level channel structure.

   \return

*/
static IFX_int32_t SDD_TAPI_LL_ALM_CapMeasResult(IFX_TAPI_LL_CH_t *pLLChannel)
{
   SDD_CHANNEL_t         *pCh;
   SDD_DEVICE_t          *pDev;

   IFX_TAPI_EVENT_t tapiEvent;
   IFX_int32_t  ret;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* next three sanity checks are paranoid, as this function is driver
      internal API and thus the channel resources should be correct.
      The checks may be hit only in case if FW reports wrong channel in
      a capacitance measurement ready event. */
   SDD_ASSERT(pCh != IFX_NULL);
   if (pCh->pALM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS (SDD_statusInvalCh);
   }

   if (pCh->pALM->line_type_fxs == IFX_FALSE)
   {
      /* errmsg: Protection from calling on FXO line */
      RETURN_STATUS (SDD_statusFXSCallOnFXO);
   }

   pDev = pCh->pDev;

   memset(&tapiEvent, 0, sizeof(tapiEvent));
   tapiEvent.ch = pCh->nChannel - 1;
   tapiEvent.id = IFX_TAPI_EVENT_LINE_MEASURE_CAPACITANCE_RDY;
   tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;

#ifdef S220
   {
      SDD_CapacitanceMeasurementRead_t *pCapMeasurement;

      pCapMeasurement = &pCh->pALM->fw_sdd_capacitance_meas;
      ret = SDD_CmdRead(pDev,
                        (IFX_uint32_t *)pCapMeasurement,
                        (IFX_uint32_t *)pCapMeasurement);

      if (SDD_SUCCESS(ret))
      {
         /* zero capacitance is an error */
         tapiEvent.data.lcap.nReturnCode =
            (pCapMeasurement->CapT2R == 0) ? IFX_ERROR : ret;
         tapiEvent.data.lcap.nCapacitance = pCapMeasurement->CapT2R;

#ifdef SDD_FEAT_NLT
         /* Store result from measurement.
            Needed for IFX_TAPI_NLT_CAPACITANCE_RESULT_GET. */
         pCh->pALM->t2r_cap_result.bValidTip2Ring =
            (pCapMeasurement->CapT2R == 0) ? IFX_FALSE : IFX_TRUE;
         pCh->pALM->t2r_cap_result.nCapTip2Ring = pCapMeasurement->CapT2R;
         /* Line to ground is new in SLIC220. */
         pCh->pALM->l2g_cap_result.bValidLine2Gnd =
            (pCapMeasurement->CapR2G == 0) || (pCapMeasurement->CapT2G == 0) ?
            IFX_FALSE : IFX_TRUE;
         pCh->pALM->l2g_cap_result.nCapRing2Gnd = pCapMeasurement->CapR2G;
         pCh->pALM->l2g_cap_result.nCapTip2Gnd = pCapMeasurement->CapT2G;
#endif /* SDD_FEAT_NLT */
      }
   }
#else /* S220 */
   {
      SDD_GR909PhoneDetection_t *pGR909PhoneDetection;

      pGR909PhoneDetection = &pCh->pALM->fw_sdd_capacitance_meas;
      ret = SDD_CmdRead(pDev,
                        (IFX_uint32_t *)pGR909PhoneDetection,
                        (IFX_uint32_t *)pGR909PhoneDetection);

      if (SDD_SUCCESS(ret))
      {
         /* zero capacitance is an error */
         tapiEvent.data.lcap.nReturnCode =
                   (pGR909PhoneDetection->Capacitance == 0) ? IFX_ERROR : ret;
         tapiEvent.data.lcap.nCapacitance = pGR909PhoneDetection->Capacitance;

#ifdef SDD_FEAT_NLT
         /* Store result from measurement. Needed for
               IFX_TAPI_NLT_CAPACITANCE_RESULT_GET */
         pCh->pALM->t2r_cap_result.bValidTip2Ring =
                   (pGR909PhoneDetection->Capacitance == 0) ? IFX_FALSE : IFX_TRUE;
         pCh->pALM->t2r_cap_result.nCapTip2Ring = pGR909PhoneDetection->Capacitance;
#endif /* SDD_FEAT_NLT */
      }
   }
#endif /* S220 */

   /* clear progress indicator */
   pCh->pALM->eCapMeasState = SDD_CapMeasState_Inactive;
   pCh->pALM->bCapMeasInProgress = IFX_FALSE;

   if (SDD_SUCCESS(ret))
   {
      IFX_TAPI_Event_Dispatch(pCh->pTapiCh, &tapiEvent);

      if (pCh->pALM->bCapMeasTip2RingOnly == IFX_FALSE)
      {
         /* Capacitance measurement was started by user.
            Report that capacitance measurement is finished and results can be
            read by the IFX_TAPI_NLT_CAPACITANCE_RESULT_GET */
         tapiEvent.id = IFX_TAPI_EVENT_NLT_END;
         tapiEvent.module = IFX_TAPI_MODULE_TYPE_ALM;
         IFX_TAPI_Event_Dispatch(pCh->pTapiCh, &tapiEvent);
      }
   }

   RETURN_STATUS(ret);
}
#endif /* SDD_FEAT_CAP_MEASURE */


#ifdef SDD_FEAT_NLT
/**
   Read the results of the capacitance measurement.

   Used by the linetesting library to read the results of the capacitance
   measurement. The result also returns the stored open loop calibration
   values. This function serves IFX_TAPI_NLT_CAPACITANCE_RESULT_GET which
   is an internal IOCTL between the linetesting library and this driver.

   \param  pLLChannel   Pointer to the TAPI LL channel structure.
   \param  pResult      Pointer to the IFX_TAPI_NLT_CAPACITANCE_RESULT_t.

   \return
   - SDD_statusOk
   - SDD_statusInvalCh
*/
static IFX_int32_t SDD_TAPI_LL_ALM_NLT_Cap_Result(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_NLT_CAPACITANCE_RESULT_t *pResult)
{
   SDD_CHANNEL_t *pCh;

   /* Internal IOCTL. Caller must ensure valid parameters. */
   SDD_ASSERT(pLLChannel != IFX_NULL);
   SDD_ASSERT(pResult != IFX_NULL);

   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }

   SDD_OS_MutexGet (&pCh->mtxChAcc);

   /* Copy measurement results */
   pResult->bValidTip2Ring = pCh->pALM->t2r_cap_result.bValidTip2Ring;
   pResult->nCapTip2Ring = pCh->pALM->t2r_cap_result.nCapTip2Ring;
   pResult->bValidLine2Gnd = pCh->pALM->l2g_cap_result.bValidLine2Gnd ;
   pResult->nCapTip2Gnd = pCh->pALM->l2g_cap_result.nCapTip2Gnd;
   pResult->nCapRing2Gnd = pCh->pALM->l2g_cap_result.nCapRing2Gnd;
   pResult->fOlCapTip2Ring = pCh->pALM->nlt_CapacitanceConfig.fOlCapTip2Ring;
   pResult->fOlCapTip2Gnd = pCh->pALM->nlt_CapacitanceConfig.fOlCapTip2Gnd;
   pResult->fOlCapRing2Gnd = pCh->pALM->nlt_CapacitanceConfig.fOlCapRing2Gnd;

   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   RETURN_STATUS (SDD_statusOk);
}
#endif /* SDD_FEAT_NLT */


#ifdef SDD_FEAT_NLT
/**
   Store the open loop calibration factors in the channel.

   This function serves IFX_TAPI_NLT_CONFIGURATION_OL_SET an internal IOCTL
   between the linetesting library and this driver.

   \param  pLLChannel   Pointer to the TAPI LL channel structure.
   \param  pConfig      Pointer to the IFX_TAPI_NLT_CONFIGURATION_OL_t.

   \return
   - always SDD_statusOk
*/
static IFX_int32_t SDD_TAPI_LL_ALM_NLT_OLConfig_Set(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_NLT_CONFIGURATION_OL_t *pConfig)
{
   SDD_CHANNEL_t *pCh;

   /* Internal IOCTL. Caller must ensure valid parameters. */
   SDD_ASSERT(pLLChannel != IFX_NULL);
   SDD_ASSERT(pConfig != IFX_NULL);

   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }

   SDD_OS_MutexGet (&pCh->mtxChAcc);

   /* Store configuration */
   pCh->pALM->nlt_ResistanceConfig.fOlResTip2Ring = pConfig->fOlResTip2Ring;
   pCh->pALM->nlt_ResistanceConfig.fOlResTip2Gnd = pConfig->fOlResTip2Gnd;
   pCh->pALM->nlt_ResistanceConfig.fOlResRing2Gnd = pConfig->fOlResRing2Gnd;
   pCh->pALM->nlt_CapacitanceConfig.fOlCapTip2Ring = pConfig->fOlCapTip2Ring;
   pCh->pALM->nlt_CapacitanceConfig.fOlCapTip2Gnd = pConfig->fOlCapTip2Gnd;
   pCh->pALM->nlt_CapacitanceConfig.fOlCapRing2Gnd = pConfig->fOlCapRing2Gnd;

   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   RETURN_STATUS (SDD_statusOk);
}
#endif /* SDD_FEAT_NLT */


#ifdef SDD_FEAT_NLT
/**
   Return the open loop calibration factors which are stored in the channel.

   This function serves IFX_TAPI_NLT_CONFIGURATION_OL_GET an internal IOCTL
   between the linetesting library and this driver.

   \param  pLLChannel   Pointer to the TAPI LL channel structure.
   \param  pConfig      Pointer to the IFX_TAPI_NLT_CONFIGURATION_OL_t.

   \return
   - always SDD_statusOk
*/
static IFX_int32_t SDD_TAPI_LL_ALM_NLT_OLConfig_Get(
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_NLT_CONFIGURATION_OL_t *pConfig)
{
   SDD_CHANNEL_t *pCh;

   /* Internal IOCTL. Caller must ensure valid parameters. */
   SDD_ASSERT(pLLChannel != IFX_NULL);
   SDD_ASSERT(pConfig != IFX_NULL);

   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   /* sanity check */
   if (pCh == IFX_NULL)
   {
      /* RETURN_STATUS needs a valid pointer in pCh to send a TAPI event. */
      /* errmsg: Resource not valid. Channel number out of range */
      return SDD_statusInvalCh;
   }

   SDD_OS_MutexGet (&pCh->mtxChAcc);

   /* Get configuration */
   pConfig->fOlResTip2Ring = pCh->pALM->nlt_ResistanceConfig.fOlResTip2Ring;
   pConfig->fOlResTip2Gnd = pCh->pALM->nlt_ResistanceConfig.fOlResTip2Gnd;
   pConfig->fOlResRing2Gnd = pCh->pALM->nlt_ResistanceConfig.fOlResRing2Gnd;
   pConfig->fOlCapTip2Ring = pCh->pALM->nlt_CapacitanceConfig.fOlCapTip2Ring;
   pConfig->fOlCapTip2Gnd = pCh->pALM->nlt_CapacitanceConfig.fOlCapTip2Gnd;
   pConfig->fOlCapRing2Gnd = pCh->pALM->nlt_CapacitanceConfig.fOlCapRing2Gnd;

#ifdef S220
   pConfig->dev_type = IFX_TAPI_GR909_DEV_S220;
#else /* S220 */
   pConfig->dev_type = IFX_TAPI_GR909_DEV_VMMC;
#endif /* S220 */

   SDD_OS_MutexRelease (&pCh->mtxChAcc);

   RETURN_STATUS (SDD_statusOk);
}
#endif /* SDD_FEAT_NLT */


/**
   Bring SLIC back into operation after fault end was reported.

   The SDD puts the SLIC in emergency shutdown mode once a fault (SSI crash)
   was detected. To exit the emergency shutdown mode linefeed 'disabled'
   needs to be set on all channels. After this the last known linefeed is
   programmed back into the line. After each programming do a delay of at
   least 2ms in order not to overwrite the previous linefeed request before
   the SDD can process it. Because the OPC event is not sent after disabled
   in emergency shutdown mode the delay is the only way to handle this.

   \param  pLLDev       Pointer to SDD device structure.

   \return
   - always SDD_statusOk
*/
static IFX_int32_t SDD_TAPI_LL_ALM_SlicRecovery (
                        IFX_TAPI_LL_DEV_t *pLLDev)
{
   SDD_DEVICE_t *pDev = ((struct VMMC_DEVICE *)pLLDev)->pSDD;
   struct SDD_ALM *pALM;
   IFX_uint16_t i;

   /* Set linefeed disabled on all channels of the device in order to get
      out of the emergency shutdown mode. */
   for (i = 0; i < SDD_MAX_CH_NR; i++)
   {
      pALM = pDev->pChannel[i].pALM;

      if (pALM != IFX_NULL)
      {
         /* Backup the FW message before setting disabled.
            The last linemode which was confirmed by the FW is in curr_opmode.
            Update the command with it but when an opmode change is pending
            repeat the last set opmode command as it is. */
         pALM->fw_last_sdd_opmode = pALM->sdd_opmode;
         if (!pALM->bOpmodeChangePending)
         {
            pALM->fw_last_sdd_opmode.OperatingMode = pALM->curr_opmode;
         }
         /* Exit the emergency shutdown with linefeed disabled. */
         irq_SDD_ALM_LineDisable(&pDev->pChannel[i]);
      }
   }

   /* Allow the SDD to process the disabled commands on each of the channels. */
   IFXOS_MSecSleep(5);

   /* Restore the previous linefeeding on each of the channels of the device. */
   for (i = 0; i < SDD_MAX_CH_NR; i++)
   {
      pALM = pDev->pChannel[i].pALM;

      if (pALM != IFX_NULL)
      {
         SDD_Opmode_t *pOpmode = &pALM->sdd_opmode;

         /* Restore the last opmode command. */
         *pOpmode = pALM->fw_last_sdd_opmode;

         SDD_CmdWrite(pDev, (IFX_uint32_t *)pOpmode);

         pALM->curr_opmode = pOpmode->OperatingMode;
#ifndef S220
         pALM->curr_polarity = pOpmode->Polarity;
         pALM->curr_howler = pOpmode->Howler;
#endif /* S220 */
      }
   }

   /* Allow the SDD to process the opmode commands on each of the channels. */
   IFXOS_MSecSleep(5);

   pDev->nSlicRecoveryCnt++;

   return SDD_statusOk;
}


#ifdef SDD_FEAT_AC_LEVEL_METERING
// \todo
// 'SDD_TAPI_LL_ALM_AcLevelMeterControl' defined but not used [-Wunused-function]
// 'SDD_TAPI_LL_ALM_AcLevelMeterConfig_Get' defined but not used [-Wunused-function]
// 'SDD_TAPI_LL_ALM_AcLevelMeterConfig_Set' defined but not used [-Wunused-function]
// 'SDD_TAPI_LL_ALM_AcLevelMeterResult_Get' defined but not used [-Wunused-function]
/**
   Function that starts and stops the AC level meter measurement.

   \param  pLLChannel   Handle to TAPI low level channel structure.
           enable       Used to start and stop the measurements
   \return

*/
static IFX_int32_t SDD_TAPI_LL_ALM_AcLevelMeterControl (
                   IFX_TAPI_LL_CH_t *pLLChannel, IFX_uint8_t enable)
{
   SDD_CHANNEL_t         *pCh;
   IFX_int32_t           ret;
   SDD_AcLevelMeterControl_t   *pAcLMControl;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   pAcLMControl = &pCh->pALM->fw_sdd_acLevelMeterControl;
   memset(&pAcLMControl, 0, sizeof(pAcLMControl));

   pAcLMControl->EN = enable;

   ret = SDD_CmdWrite(pCh->pDev, (IFX_uint32_t *)pAcLMControl);
   if (!SDD_SUCCESS(ret))
   {
      RETURN_STATUS(ret);
   }
   return ret;
}
#endif /* SDD_FEAT_AC_LEVEL_METERING */


#ifdef SDD_FEAT_AC_LEVEL_METERING
static IFX_int32_t SDD_TAPI_LL_ALM_AcLevelMeterConfig_Get (
                   IFX_TAPI_LL_CH_t *pLLChannel,
                   SDD_AcLevelMeterConfig_t* pConfig)
{
   SDD_CHANNEL_t         *pCh;
   IFX_int32_t           ret;
   SDD_AcLevelMeterConfig_t   *pAcLMConfig;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   pAcLMConfig = &pCh->pALM->fw_sdd_acLevelMeterConfig;
   memset(&pAcLMConfig, 0, sizeof(pAcLMConfig));

   ret = SDD_CmdRead (pCh->pDev, (IFX_uint32_t *)&pAcLMConfig,
                                    (IFX_uint32_t *)&pAcLMConfig);
   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Reading failed. */
     /* \todo How to handle the error? New error code?
      RETURN_STATUS(SDD_statusCmdRdErr); */
   }
   memcpy(&pConfig, &pAcLMConfig, sizeof(pConfig));
   return ret;

}
#endif /* SDD_FEAT_AC_LEVEL_METERING */


#ifdef SDD_FEAT_AC_LEVEL_METERING
static IFX_int32_t SDD_TAPI_LL_ALM_AcLevelMeterConfig_Set (
                   IFX_TAPI_LL_CH_t *pLLChannel,
                   const SDD_AcLevelMeterConfig_t* pConfig)
{
   SDD_CHANNEL_t         *pCh;
   IFX_int32_t           ret;
   SDD_AcLevelMeterConfig_t   *pAcLMConfig;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   pAcLMConfig = &pCh->pALM->fw_sdd_acLevelMeterConfig;
   memset(&pAcLMConfig, 0, sizeof(pAcLMConfig));


   ret = SDD_CmdRead (pCh->pDev, (IFX_uint32_t *)&pAcLMConfig,
                                    (IFX_uint32_t *)&pAcLMConfig);
   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Reading failed. */
     /* \todo How to handle the error? New error code?
      RETURN_STATUS(SDD_statusCmdRdErr); */
   }

   pAcLMConfig->BW   = pConfig->BW;
   pAcLMConfig->CF   = pConfig->CF;
   pAcLMConfig->BP   = pConfig->BP;
   pAcLMConfig->TH   = pConfig->TH;
   pAcLMConfig->Del  = pConfig->Del;
   pAcLMConfig->Int  = pConfig->Int;

   ret = SDD_CmdWrite (pCh->pDev, (IFX_uint32_t *)pAcLMConfig);
   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Writting failed. */
      RETURN_STATUS(SDD_statusCmdWrErr);
   }
   return ret;
}
#endif /* SDD_FEAT_AC_LEVEL_METERING */


#ifdef SDD_FEAT_AC_LEVEL_METERING
static IFX_int32_t SDD_TAPI_LL_ALM_AcLevelMeterResult_Get (
                   IFX_TAPI_LL_CH_t *pLLChannel,
                   SDD_AcLevelMeterResult_t *pResult)
{
   SDD_CHANNEL_t         *pCh;
   IFX_int32_t           ret;
   SDD_AcLevelMeterResult_t   *pAcLMResult;

   SDD_ASSERT(pLLChannel != IFX_NULL);
   pCh  = ((struct VMMC_CHANNEL *)pLLChannel)->pSDD;

   pAcLMResult = &pCh->pALM->fw_sdd_acLevelMeterResult;
   memset(&pAcLMResult, 0, sizeof(pAcLMResult));


   ret = SDD_CmdRead (pCh->pDev, (IFX_uint32_t *)&pAcLMResult,
                                    (IFX_uint32_t *)&pAcLMResult);
   if (!SDD_SUCCESS (ret))
   {
      /* errmsg: Reading failed. */
     /* \todo How to handle the error? New error code?
      RETURN_STATUS(SDD_statusCmdRdErr); */
   }
   memcpy(&pResult, &pAcLMResult, sizeof(SDD_AcLevelMeterResult_t));
   return ret;
}
#endif /* SDD_FEAT_AC_LEVEL_METERING */


/**
   Function that fills in the ALM module function pointers in the driver
   context structure which is passed to HL TAPI during registration.

   \param  pAlm         Pointer to the ALM part in the driver context struct.
*/
IFX_void_t SDD_ALM_Opmode_Func_Register (IFX_TAPI_DRV_CTX_ALM_t *pAlm)
{
   /* volume handling */
   pAlm->Volume_Set        = SDD_TAPI_LL_ALM_Volume_Set;
   pAlm->Volume_High_Level = SDD_TAPI_LL_ALM_Volume_High_Level;

   pAlm->Ring_Cfg_Set = SDD_TAPI_LL_ALM_Ring_Cfg_Set;
   pAlm->Ring_Cfg_Get = SDD_TAPI_LL_ALM_Ring_Cfg_Get;

   /* line type and line feeding */
   pAlm->Line_Type_Set     = SDD_TAPI_LL_ALM_Line_Type_Set;
   pAlm->Line_Mode_Set     = SDD_TAPI_LL_ALM_Line_Mode_Set;
   pAlm->Line_Mode_Get     = SDD_TAPI_LL_ALM_Line_Mode_Get;

#ifdef SDD_FEAT_FXO
   /* FXO Callbacks register */
   SDD_ALM_FXO_Func_Register (pAlm);
#endif /* SDD_FEAT_FXO */

   /* metering pulse generation */
   pAlm->Metering_Start    = SDD_TAPI_LL_ALM_MeteringStart;

   /* test functions */
   pAlm->TestLoop          = SDD_TAPI_LL_ALM_TestLoop;

   /* message waiting lamp handling */
   pAlm->MWL_Activation_Get   = SDD_TAPI_LL_ALM_MWL_Activation_Get;
   pAlm->MWL_Activation_Set   = SDD_TAPI_LL_ALM_MWL_Activation_Set;

#ifdef SDD_FEAT_CONT_MEASUREMENT
   /* continuous measurement handling */
   pAlm->ContMeasReq             = SDD_TAPI_LL_ALM_ContMeas_Req;
   pAlm->ContMeasGet             = SDD_TAPI_LL_ALM_ContMeas_Get;
#endif /* SDD_FEAT_CONT_MEASUREMENT */
#ifdef SDD_FEAT_CONT_MEASUREMENT_RESET
   pAlm->ContMeasReset           = SDD_TAPI_LL_ALM_ContMeas_Reset;
#endif /* SDD_FEAT_CONT_MEASUREMENT_RESET */

   pAlm->CfgDupTimer_Override    = SDD_TAPI_LL_ALM_CfgDupTimer_Override;

#ifdef SDD_FEAT_GR909
   SDD_ALM_GR909_Func_Register (pAlm);
#endif /* SDD_FEAT_GR909 */

   SDD_ALM_Calib_Func_Register (pAlm);

#ifdef SDD_FEAT_CAP_MEASURE
   pAlm->CheckCapMeasSup         = SDD_TAPI_LL_ALM_LT_CheckCapMeasSup;
   pAlm->CapMeasStart            = SDD_TAPI_LL_ALM_CapMeasStart;
   pAlm->CapMeasStop             = SDD_TAPI_LL_ALM_CapMeasStop;
   pAlm->CapMeasResult           = SDD_TAPI_LL_ALM_CapMeasResult;
#endif /* SDD_FEAT_CAP_MEASURE */

#ifdef SDD_FEAT_NLT
   pAlm->NLT_capacitance_result_get = SDD_TAPI_LL_ALM_NLT_Cap_Result;
   pAlm->NLT_OLConfigSet      = SDD_TAPI_LL_ALM_NLT_OLConfig_Set;
   pAlm->NLT_OLConfigGet      = SDD_TAPI_LL_ALM_NLT_OLConfig_Get;
#endif /* SDD_FEAT_NLT */

   pAlm->SlicRecovery         = SDD_TAPI_LL_ALM_SlicRecovery;

#ifdef SDD_FEAT_AC_LEVEL_METERING
   /* \todo Mapping needed in IFX_TAPI_DRV_CTX_ALM_t ? */
   pAlm->AcLevelMeterControl     = SDD_TAPI_LL_ALM_AcLevelMeterControl;
   pAlm->AcLevelMeterConfig_Set  = SDD_TAPI_LL_ALM_AcLevelMeterConfig_Set;
   pAlm->AcLevelMeterConfig_Get  = SDD_TAPI_LL_ALM_AcLevelMeterConfig_Get;
   pAlm->AcLevelMeterResult      = SDD_TAPI_LL_ALM_AcLevelMeterResult_Get;
#endif /* SDD_FEAT_AC_LEVEL_METERING */
}
