/******************************************************************************

                            Copyright (c) 2013-2016
                        Lantiq Beteiligungs-GmbH & Co.KG
                             http://www.lantiq.com

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

/**
   \file drv_sdd_pcm.c
   This file implements the SDD PCM module.
*/

/* ========================================================================== */
/*                                 Includes                                   */
/* ========================================================================== */

#include "drv_sdd_osmap.h"
#include "drv_sdd_global.h"
#include "drv_sdd_api.h"
#include "drv_sdd_pcm.h"
#include "drv_sdd_init.h"
#include "drv_sdd_fw_cmd.h"
#ifdef SDD_FEAT_PROCFS
   #include <linux/seq_file.h>
#endif /* SDD_FEAT_PROCFS */

/* ========================================================================== */
/*                             Macro definitions                              */
/* ========================================================================== */
#define SDD_PCM_HIGHWAY             1
#define SDD_PCM_MAX_TS              127
#define SDD_PCM_TS_ARRAY            ((SDD_PCM_MAX_TS + 3) / 32)

/* ========================================================================== */
/*                             Type definitions                               */
/* ========================================================================== */

struct SDD_PCM_DEV
{
   /* FW message for the PCM interface configuration. */
   struct SDD_PCM_IF_CTRL     fwm_PcmIfCtrl;
   /* Timeslot reservation bitfield. */
   IFX_uint32_t               PcmRxTs[SDD_PCM_HIGHWAY][SDD_PCM_TS_ARRAY];
   IFX_uint32_t               PcmTxTs[SDD_PCM_HIGHWAY][SDD_PCM_TS_ARRAY];
   /* Actual number of PCM timeslots resulting from the DCL clock setting. */
   IFX_uint16_t               nMaxTimeslot;
};

struct SDD_PCM
{
   /* FW message for PCM channel control. */
   struct SDD_PCM_CH_CTRL     fwm_PcmChCtrl;
};

/* ========================================================================== */
/*                             Global variables                               */
/* ========================================================================== */

/* ========================================================================== */
/*                           Function prototypes                              */
/* ========================================================================== */

/* ========================================================================== */
/*                         Function implementation                            */
/* ========================================================================== */

/**
   Allocate data structure for PCM on the device.

   \param  pDev         Pointer to the device structure.

   \return
   - SDD_statusOk
   - SDD_statusNoMem    in case the stucture could not be created

   \remarks The device parameter is no longer checked because the calling
   function assures correct values.
*/
IFX_int32_t SDD_PCM_DevStructAllocate (
                        SDD_DEVICE_t *pDev)
{
   SDD_PCM_DevStructFree (pDev);

   pDev->pPCM = IFXOS_BlockAlloc(sizeof(*pDev->pPCM));
   if (pDev->pPCM == IFX_NULL)
   {
      /* errmsg: No memory could be allocated. */
      return SDD_statusNoMem;
   }
   memset(pDev->pPCM, 0, sizeof(*pDev->pPCM));

   return SDD_statusOk;
}


/**
   Free data structure for PCM on the device.

   \param  pDev         Pointer to the device structure.
*/
IFX_void_t SDD_PCM_DevStructFree (
                        SDD_DEVICE_t *pDev)
{
   if (pDev->pPCM != IFX_NULL)
   {
      IFXOS_BlockFree(pDev->pPCM);
      pDev->pPCM = IFX_NULL;
   }
}


/**
   Initialise PCM device settings

   \param  pDev         Pointer to the device structure.
*/
IFX_void_t SDD_PCM_DevInit (
                        SDD_DEVICE_t *pDev)
{
   struct SDD_PCM_DEV *pPcmDev = pDev->pPCM;
   struct SDD_PCM_IF_CTRL *pPCM_CtrlCmd = &pPcmDev->fwm_PcmIfCtrl;

   memset (pPCM_CtrlCmd, 0, sizeof (*pPCM_CtrlCmd));
   pPCM_CtrlCmd->CMD          = CMD_EOP;
   pPCM_CtrlCmd->MOD          = MOD_PCM;
   pPCM_CtrlCmd->ECMD         = SDD_PCM_IF_CTRL_ECMD;
   pPCM_CtrlCmd->LENGTH       = SDD_PCM_IF_CTRL_LENGTH;

   pPcmDev->nMaxTimeslot = 0;
}


/**
   Configure and enable PCM interface.

   \param  pDev         Pointer to the device structure.
   \param  pCfg         Pointer to the coefficient structure.

   \return
   - SDD_statusOk
   - SDD_statusInvalDev - Device number out of range
   - SDD_statusIfxError -  PCM interface has already been enabled
   - SDD_statusParam -  One of the parameters is not supported
*/
IFX_int32_t SDD_TAPI_LL_PCM_IF_Cfg (
                        IFX_TAPI_LL_DEV_t* pLLDev,
                        const IFX_TAPI_PCM_IF_CFG_t *pCfg)
{
   SDD_DEVICE_t *pDev = (SDD_DEVICE_t *)pLLDev;
   IFX_int32_t ret;

   /* parameter check */
   if (pDev == IFX_NULL)
   {
      return SDD_statusInvalDev;
   }

   ret = SDD_PCM_IF_Cfg(pDev, pCfg);

   return ret;
}


/**
   Configure and enable PCM interface.

   \param  pDev         Pointer to the device structure.
   \param  pCfg         Pointer to the coefficient structure.

   \return
   - SDD_statusOk
   - SDD_statusIfxError -  PCM interface has already been enabled
   - SDD_statusParam -  One of the parameters is not supported
*/
IFX_int32_t SDD_PCM_IF_Cfg (
                        SDD_DEVICE_t *pDev,
                        const IFX_TAPI_PCM_IF_CFG_t *pCfg)
{
   struct SDD_PCM_IF_CTRL *pPCM_IfCtrl  = &pDev->pPCM->fwm_PcmIfCtrl;
   IFX_int32_t ret = SDD_statusIfxError;
   static IFX_TAPI_PCM_IF_CFG_t pcm_if_cfg_defaults;

   if (pCfg == IFX_NULL)
   {
      /* set static defaults for PCM i/f */
      pcm_if_cfg_defaults.nDoubleClk = IFX_DISABLE;
      pcm_if_cfg_defaults.nSlopeTX   = IFX_TAPI_PCM_IF_SLOPE_RISE;
      pcm_if_cfg_defaults.nSlopeRX   = IFX_TAPI_PCM_IF_SLOPE_FALL;
      pcm_if_cfg_defaults.nOffsetTX  = IFX_TAPI_PCM_IF_OFFSET_NONE;
      pcm_if_cfg_defaults.nOffsetRX  = IFX_TAPI_PCM_IF_OFFSET_NONE;
      pcm_if_cfg_defaults.nDrive     = IFX_TAPI_PCM_IF_DRIVE_ENTIRE;
      pcm_if_cfg_defaults.nShift     = IFX_DISABLE;
      pcm_if_cfg_defaults.nDCLFreq   = IFX_TAPI_PCM_IF_DCLFREQ_2048;

      pCfg = &pcm_if_cfg_defaults;
   }

   TRACE(SDD, DBG_LEVEL_LOW, ("VCODEC PCM if cfg\n"));
   TRACE(SDD, DBG_LEVEL_LOW, ("PCM mode slave (always)\n"));
   TRACE(SDD, DBG_LEVEL_LOW, ("PCM clk  %d (double %d)\n",
         pCfg->nDCLFreq, pCfg->nDoubleClk));
   TRACE(SDD, DBG_LEVEL_LOW, ("PCM slope  tx %s\n",
        (pCfg->nSlopeTX==IFX_TAPI_PCM_IF_SLOPE_RISE)?"rising":"falling"));
   TRACE(SDD, DBG_LEVEL_LOW, ("PCM slope  rx %s\n",
        (pCfg->nSlopeRX==IFX_TAPI_PCM_IF_SLOPE_RISE)?"rising":"falling"));
   TRACE(SDD, DBG_LEVEL_LOW, ("PCM offset tx %d\n", pCfg->nOffsetTX));
   TRACE(SDD, DBG_LEVEL_LOW, ("PCM offset rx %d\n", pCfg->nOffsetRX));

   /* If the interface is already active switch it off before changing
      the parameters. */
   if (!(pDev->nDevState & DS_PCM_EN))
   {
      /* Disable the PCM Interface */
      pPCM_IfCtrl->EN = SDD_PCM_IF_CTRL_DISABLE;

      ret = SDD_CmdWrite (pDev, (IFX_uint32_t*)(IFX_void_t*)pPCM_IfCtrl);
      if (!SDD_SUCCESS(ret))
      {
         RETURN_DEVSTATUS(ret);
      }

      /* pDev->nDevState &= ~DS_PCM_EN; *//* changes in the next command */
   }

   /* Enable the PCM Interface */
   pPCM_IfCtrl->EN   = SDD_PCM_IF_CTRL_ENABLE;
   pPCM_IfCtrl->XOFF = pCfg->nOffsetTX;
   pPCM_IfCtrl->DBL  = pCfg->nDoubleClk;
   pPCM_IfCtrl->XS   = pCfg->nSlopeTX;
   pPCM_IfCtrl->RS   = !pCfg->nSlopeRX;
   pPCM_IfCtrl->DRV0 = pCfg->nDrive;
   pPCM_IfCtrl->SH   = pCfg->nShift;
   pPCM_IfCtrl->ROFF = pCfg->nOffsetRX;

   ret = SDD_CmdWrite (pDev, (IFX_uint32_t*)(IFX_void_t*)pPCM_IfCtrl);

   if (SDD_SUCCESS(ret))
   {
      /* Write was successful. Mark the interface as active. */
      pDev->nDevState |= DS_PCM_EN;

      /* For checking the timeslots we need the maximum number of timeslots
         that can be used. However as slave interface we do not set the DCL
         frequency but only synchronise to it. As there is no way to know
         the actual frequency to which the interface has synchronised the
         frequency given in the IOCTL is used instead. */
      switch (pCfg->nDCLFreq)
      {
         case IFX_TAPI_PCM_IF_DCLFREQ_512:
            pDev->pPCM->nMaxTimeslot = 8;
            break;
         case IFX_TAPI_PCM_IF_DCLFREQ_1024:
            pDev->pPCM->nMaxTimeslot = 16;
            break;
         case IFX_TAPI_PCM_IF_DCLFREQ_1536:
            pDev->pPCM->nMaxTimeslot = 24;
            break;
         case IFX_TAPI_PCM_IF_DCLFREQ_2048:
            pDev->pPCM->nMaxTimeslot = 32;
            break;
         case IFX_TAPI_PCM_IF_DCLFREQ_4096:
            pDev->pPCM->nMaxTimeslot = 64;
            break;
         case IFX_TAPI_PCM_IF_DCLFREQ_8192:
            pDev->pPCM->nMaxTimeslot = 128;
            break;
         case IFX_TAPI_PCM_IF_DCLFREQ_16384:
         default:
            TRACE(SDD, DBG_LEVEL_LOW, ("VCODEC: unsupported PCM clock\n"));
            break;
      }
      /* Double clocking offers only half the timeslots. */
      if (pPCM_IfCtrl->DBL)
      {
         pDev->pPCM->nMaxTimeslot /= 2;
      }
   }

   RETURN_DEVSTATUS(ret);
}


/**
   Stop the PCM interface

   \param  pDev         Pointer to the device structure.

   \return
   - SDD_statusOk         If successful
   - SDD_statusCmdWr      Writing the command has failed
*/
IFX_int32_t SDD_PCM_IF_Stop (
                        SDD_DEVICE_t *pDev)
{
   struct SDD_PCM_IF_CTRL *pPCM_IfCtrl  = &pDev->pPCM->fwm_PcmIfCtrl;
   IFX_int32_t ret = SDD_statusOk;

   /* calling function should ensure valid parameters */
   SDD_ASSERT(pDev != IFX_NULL);


   /* Check that all PCM channels are stopped is not done as this function
      is only called when stopping the driver. */

   if (pDev->nDevState & DS_PCM_EN)
   {
      /* Disable the PCM Interface */
      pPCM_IfCtrl->EN = SDD_PCM_IF_CTRL_DISABLE;

      ret = SDD_CmdWrite (pDev, (IFX_uint32_t*)(IFX_void_t*)pPCM_IfCtrl);

      /* reset PCM timeslot management flags */
      memset(pDev->pPCM->PcmRxTs, 0, sizeof(pDev->pPCM->PcmRxTs));
      memset(pDev->pPCM->PcmTxTs, 0, sizeof(pDev->pPCM->PcmTxTs));
   }

   RETURN_DEVSTATUS(ret);
}


/**
   Prepare and check PCM channel parameters.

   Performs error checking according to the underlying device capability.
   This function just checks the configuration. No firmware message is sent.

   \param  pCh          Pointer to the channel structure.
   \param  pPCMConfig   Contains the new configuration for PCM channel.

   \return
   - SDD_statusOk
   - SDD_statusNotSupported
   - SDD_statusFuncParam
   - SDD_statusPcmTsInvalid
   - SDD_statusPcmHwInvalid
*/
IFX_int32_t SDD_TAPI_LL_PCM_ChCfg (
                        IFX_TAPI_LL_CH_t *pLLChannel,
                        IFX_TAPI_PCM_CFG_t const *pPCMConfig)
{
   SDD_CHANNEL_t *pCh = (SDD_CHANNEL_t *)pLLChannel;
   IFX_int32_t   ret;

   /* parameter check */
   if (pCh == IFX_NULL)
   {
      return SDD_statusInvalCh;
   }

   ret = SDD_PCM_ChCfg (pCh, pPCMConfig);

   return ret;
}


/**
   Prepare and check PCM channel parameters.

   Performs error checking according to the underlying device capability.
   This function just checks the configuration. No firmware message is sent.

   \param  pCh          Pointer to the channel structure.
   \param  pPCMConfig   Contains the new configuration for PCM channel.

   \return
   - SDD_statusOk
   - SDD_statusNotSupported
   - SDD_statusFuncParam
   - SDD_statusPcmTsInvalid
   - SDD_statusPcmHwInvalid
*/
IFX_int32_t SDD_PCM_ChCfg (
                        SDD_CHANNEL_t *pCh,
                        IFX_TAPI_PCM_CFG_t const *pPCMConfig)
{
   IFX_int32_t    ret   = SDD_statusOk;
   SDD_DEVICE_t   *pDev = pCh->pDev;
   IFX_uint32_t   timeslots = 1,
                  split = 0;

   /* check if interface needs to be enabled (with defaults) */
   if (!(pDev->nDevState & DS_PCM_EN))
   {
      /* nDevState DS_PCM_EN is set inside (if successful) */
      ret = SDD_PCM_IF_Cfg (pDev, IFX_NULL);
   }
   if (SDD_statusOk != ret)
   {
      return ret;
   }

   /* check PCM resolution for supported values */
   switch (pPCMConfig->nResolution)
      {
      case IFX_TAPI_PCM_RES_NB_ALAW_8BIT:
      case IFX_TAPI_PCM_RES_NB_ULAW_8BIT:
         timeslots = 1;
         break;
      case IFX_TAPI_PCM_RES_NB_LINEAR_16BIT:
         timeslots = 2;
         break;
      case IFX_TAPI_PCM_RES_WB_LINEAR_16BIT:
         timeslots = 4;
         break;
      case IFX_TAPI_PCM_RES_WB_LINEAR_SPLIT_16BIT:
         if (pDev->caps.bfw_WbTsSplit != 1)
         {
            /* FW does not support split timeslots for WB mode. */
            /* errmsg: Requested PCM resolution not supported. */
            return SDD_statusPcmResolutionNotSupported;
         }
          /* 2nd group has 2 timeslots */
         timeslots = 2;
          /* offset between the 1st and  2nd group */
         split = pDev->pPCM->nMaxTimeslot >> 1;
         break;
      case IFX_TAPI_PCM_RES_WB_ALAW_8BIT:
      case IFX_TAPI_PCM_RES_WB_ULAW_8BIT:
         /* errmsg: Requested PCM resolution not supported. */
         return SDD_statusPcmResolutionNotSupported;
      default:
         return SDD_statusFuncParam;
   }

   /* Here we check that the PCM sample has enough space on the given timeslot.
      Because we have also coders that use 2 and 4 consecutive timeslots we
      set above the number of timeslots used by one sample and then check here
      that the maximum timeslot is not exceeded.
      Note: Timeslots start counting from 0 */
   if (((pPCMConfig->nTimeslotRX + timeslots + split) > pDev->pPCM->nMaxTimeslot) ||
       ((pPCMConfig->nTimeslotTX + timeslots + split) > pDev->pPCM->nMaxTimeslot))
   {
      /* errmsg: PCM timeslot given out of range. */
      return SDD_statusPcmTsInvalid;
   }

   if (pPCMConfig->nHighway >= SDD_PCM_HIGHWAY)
   {
      /* errmsg: PCM Highway number out of range. */
      return SDD_statusPcmHwInvalid;
   }

   return SDD_statusOk;
}


/**
   Activate or deactivate a PCM channel.

   The configuration must be set previously with the low level function
   SDD_TAPI_LL_PCM_ChCfg.
   Resource availability check is done. The function returns with error when
   the resource or the timeslot is not available.

   No message will be sent and no error returned, when the setting has
   not been changed. The driver checks for the cached values.

   When deactivating the channel a possibly running LEC and ES are automatically
   deactivated and released. No error is reported in this case.

   \param  pLLChannel   Pointer to channel structure.
   \param  nMode        Activation mode
                        - 1: timeslot activated
                        - 0: timeslot deactivated
   \param  pPcmCfg      Pointer to the current PCM configuration

   \return
   Return value from SDD_PCM_ChEnable().
*/
IFX_int32_t SDD_TAPI_LL_PCM_ChEnable (
                        IFX_TAPI_LL_CH_t *pLLCh,
                        IFX_uint32_t nMode,
                        IFX_TAPI_PCM_CFG_t *pPcmCfg)
{
   SDD_CHANNEL_t *pCh = (SDD_CHANNEL_t *)pLLCh;
   IFX_int32_t   ret;

   /* parameter check */
   if (pCh == IFX_NULL)
   {
      return SDD_statusInvalCh;
   }

   ret = SDD_PCM_ChEnable(pCh, nMode, pPcmCfg);

   return ret;
}


/**
   Activate the PCM channel with the given parameters.

   The configuration must have been done previously. The resource availability
   check is done in this function. The function returns with error when
   the resource or the timeslot is not available.

   No message will be send and no error returned, when the setting has
   not been changed. The driver checks for the cached values.

   \param  pCh          Pointer to the channel structure.
   \param  nMode        Activation mode
                        - 1: timeslot activated
                        - 0: timeslot deactivated.
   \param  pPcmCfg      Pointer to the current PCM configuration.

   \return
   - IFX_SUCCESS
   - IFX_ERROR
*/
IFX_int32_t SDD_PCM_ChEnable(
                        SDD_CHANNEL_t *pCh,
                        IFX_uint32_t nMode,
                        IFX_TAPI_PCM_CFG_t *pPcmCfg)
{
   IFX_int32_t ret   = SDD_statusOk;
   SDD_DEVICE_t *pDev = pCh->pDev;
   IFX_uint32_t IdxRx, BitValueRx;
   IFX_uint32_t IdxTx, BitValueTx;
   IFX_uint32_t i, j, split = 0, timeslots = 1;
   struct SDD_PCM_CH_CTRL *pPcmCh;

   if (pCh->pPCM == IFX_NULL)
   {
      /* errmsg: Resource not valid. Channel number out of range */
      RETURN_STATUS(SDD_statusInvalCh);
   }

   if ((pDev->nDevState & DS_PCM_EN) == IFX_FALSE)
   {
      /* errmsg: PCM interface is not initialized. */
      RETURN_STATUS(SDD_statusPcmNotInitialized);
   }

   pPcmCh = &pCh->pPCM->fwm_PcmChCtrl;

   /* Activate Timeslots and set highway */
   switch (nMode)
   {
      case  0:
         if (pPcmCh->EN == SDD_PCM_CH_CTRL_ENABLE)
         {
            /*
               The PCM device variables PcmRxTs, PcmTxTs are global for all
               channels and must be protected against concurent tasks access.
            */
            SDD_OS_MutexGet (&pDev->mtxMemberAcc);
            switch (pPcmCh->COD)
            {
               case SDD_PCM_CH_CTRL_COD_LINEAR_16BIT:
                  timeslots = (pPcmCh->WIDE == SDD_PCM_CH_CTRL_WIDE_16KHZ) ? 4 : 2;
                  break;
               case SDD_PCM_CH_CTRL_COD_ALAW_8BIT:
               case SDD_PCM_CH_CTRL_COD_ULAW_8BIT:
                  timeslots = 1;
                  break;
               default:
                  /* nothing, just for the compiler */
                  break;
            }

            if (pPcmCh->WBTSC == SDD_PCM_CH_CTRL_WBTSC_SPLIT)
            {
               /* The offset between 1st and 2nd group is half the number of
                  maximum timeslots. */
               split = pDev->pPCM->nMaxTimeslot >> 1;
            }

            /* free the timeslots in the map */
            for (j = 0;  j < timeslots; j++)
            {
               /* Translate the loop parameter so that for split timeslots
                  the index of the 2nd group includes the offset of half the
                  possible timeslots. For consecutive timeslots i=j is used. */
               i = ((split != 0) && (j >= 2)) ? ((j - 2) + split) : j;

               IdxRx = ((pPcmCh->RTS & 0x7F) + i) >> 5;
               BitValueRx = 1 << ((pPcmCh->RTS + i) & 0x1F);
               IdxTx = ((pPcmCh->XTS & 0x7F) + i) >> 5;
               BitValueTx = 1 << ((pPcmCh->XTS + i) & 0x1F);

               pDev->pPCM->PcmRxTs[pPcmCfg->nHighway][IdxRx] &= ~BitValueRx;
               pDev->pPCM->PcmTxTs[pPcmCfg->nHighway][IdxTx] &= ~BitValueTx;
            }

            /* disable channel: This will deactivate the timeslots */
            pPcmCh->EN = SDD_PCM_CH_CTRL_DISABLE;

            /* write coefficients back */
            ret = SDD_CmdWrite (pDev, (IFX_uint32_t*)(IFX_void_t*)pPcmCh);
            /* release share variables lock */
            SDD_OS_MutexRelease (&pDev->mtxMemberAcc);
         }
         break;

      case 1:
         /*
            The PCM device variables PcmRxTs, PcmTxTs are global for all
            channels and must be protected against concurent tasks access.
         */
         SDD_OS_MutexGet (&pDev->mtxMemberAcc);
         /* check timeslot allocation */
         switch (pPcmCfg->nResolution)
         {
            case IFX_TAPI_PCM_RES_NB_LINEAR_16BIT:
               /* for 16 Bit mode, the next higher time slot has to be reserved too */
               timeslots = 2;
               break;
            case IFX_TAPI_PCM_RES_WB_LINEAR_16BIT:
               timeslots = 4;
               break;
            case IFX_TAPI_PCM_RES_WB_LINEAR_SPLIT_16BIT:
               timeslots = 4;
               /* The offset between 1st and 2nd group is half the number of
                  maximum timeslots. */
               split = pDev->pPCM->nMaxTimeslot >> 1;
               break;
            case IFX_TAPI_PCM_RES_NB_ALAW_8BIT:
            case IFX_TAPI_PCM_RES_NB_ULAW_8BIT:
               timeslots = 1;
               break;
            default:
               SDD_OS_MutexRelease (&pDev->mtxMemberAcc);
               /* errmsg: Requested PCM resolution not supported. */
               return SDD_statusPcmResolutionNotSupported;
         }

         for (j = 0;  j < timeslots; j++)
         {
            /* Translate the loop parameter so that for split timeslots
               the index of the 2nd group includes the offset of half the
               possible timeslots. For consecutive timeslots i=j is assigned. */
            i = ((split != 0) && (j >= 2)) ? ((j - 2) + split) : j;
            /* Get the index in the time slot array. Each bit reflects one time
            slot. Modulo 32 separate the index of the array and the bit inside
            the field. */
            IdxRx = (IFX_uint32_t)(pPcmCfg->nTimeslotRX + i) >> 5;
            BitValueRx = 1 << ((pPcmCfg->nTimeslotRX + i) & 0x1F);
            IdxTx = (IFX_uint32_t)(pPcmCfg->nTimeslotTX + i) >> 5;
            BitValueTx = 1 << ((pPcmCfg->nTimeslotTX + i) & 0x1F);

            /* RX Slot checking */
            if (pDev->pPCM->PcmRxTs[pPcmCfg->nHighway][IdxRx] & BitValueRx)
            {
               TRACE (SDD, DBG_LEVEL_HIGH,
                   ("DRV_ERROR: Rx Slot %u is already in use\n",
                    pPcmCfg->nTimeslotRX + i));
               /* errmsg: Requested PCM timeslot is already in use. */
               ret = SDD_statusPcmRequestedTsInUse;
#ifndef ENABLE_TRACE
               /* Without trace we break here to save time. */
               break;
#endif /* ENABLE_TRACE */
            }

            /* TX Slot checking */
            if (pDev->pPCM->PcmTxTs[pPcmCfg->nHighway][IdxTx] & BitValueTx)
            {
               TRACE (SDD,DBG_LEVEL_HIGH,
                   ("DRV_ERROR: Tx Slot %u is already in use\n",
                    pPcmCfg->nTimeslotTX + i));
               /* errmsg: Requested PCM timeslot is already in use. */
               ret = SDD_statusPcmRequestedTsInUse;
#ifndef ENABLE_TRACE
               /* Without trace we break here to save time. */
               break;
#endif /* ENABLE_TRACE */
            }
         }

         if (!SDD_SUCCESS(ret))
         {
            SDD_OS_MutexRelease (&pDev->mtxMemberAcc);
            return ret;
         }

         for (j = 0;  j < timeslots; j++)
         {
            /* For comments see in the block above. */
            i = ((split != 0) && (j >= 2)) ? ((j - 2) + split) : j;
            IdxRx = (IFX_uint32_t)(pPcmCfg->nTimeslotRX + i) >> 5;
            BitValueRx = 1 << ((pPcmCfg->nTimeslotRX + i) & 0x1F);
            IdxTx = (IFX_uint32_t)(pPcmCfg->nTimeslotTX + i) >> 5;
            BitValueTx = 1 << ((pPcmCfg->nTimeslotTX + i) & 0x1F);
            /* reserve the time slots now, because the parameter are OK */
            pDev->pPCM->PcmRxTs[pPcmCfg->nHighway][IdxRx] |= BitValueRx;
            pDev->pPCM->PcmTxTs[pPcmCfg->nHighway][IdxTx] |= BitValueTx;
         }

         /* activate Timeslots */
         switch (pPcmCfg->nResolution)
         {
            case IFX_TAPI_PCM_RES_NB_ALAW_8BIT:
               pPcmCh->WIDE = SDD_PCM_CH_CTRL_WIDE_8KHZ;
               pPcmCh->COD  = SDD_PCM_CH_CTRL_COD_ALAW_8BIT;
               break;
            case IFX_TAPI_PCM_RES_NB_ULAW_8BIT:
               pPcmCh->WIDE = SDD_PCM_CH_CTRL_WIDE_8KHZ;
               pPcmCh->COD  = SDD_PCM_CH_CTRL_COD_ULAW_8BIT;
               break;
            case IFX_TAPI_PCM_RES_WB_LINEAR_16BIT:
               pPcmCh->WIDE = SDD_PCM_CH_CTRL_WIDE_16KHZ;
               pPcmCh->COD  = SDD_PCM_CH_CTRL_COD_LINEAR_16BIT;
               pPcmCh->WBTSC= SDD_PCM_CH_CTRL_WBTSC_CONSECUTIVE;
               break;
            case IFX_TAPI_PCM_RES_WB_LINEAR_SPLIT_16BIT:
               pPcmCh->WIDE = SDD_PCM_CH_CTRL_WIDE_16KHZ;
               pPcmCh->COD  = SDD_PCM_CH_CTRL_COD_LINEAR_16BIT;
               pPcmCh->WBTSC= SDD_PCM_CH_CTRL_WBTSC_SPLIT;
               break;
            case IFX_TAPI_PCM_RES_NB_LINEAR_16BIT:
            default:
               pPcmCh->WIDE = SDD_PCM_CH_CTRL_WIDE_8KHZ;
               pPcmCh->COD  = SDD_PCM_CH_CTRL_COD_LINEAR_16BIT;
               break;
         }

         pPcmCh->XTS  = pPcmCfg->nTimeslotTX;
         pPcmCh->RTS  = pPcmCfg->nTimeslotRX;
#ifndef S220
         pPcmCh->X_HW = pPcmCfg->nHighway;
         pPcmCh->R_HW = pPcmCfg->nHighway;
#endif /* S220 */
         pPcmCh->EN   = SDD_PCM_CH_CTRL_ENABLE;

         ret = SDD_CmdWrite (pDev, (IFX_uint32_t*)(IFX_void_t*)pPcmCh);
         SDD_OS_MutexRelease (&pDev->mtxMemberAcc);
         break;
      default:
         /* nothing, just for the compiler */
         break;
   }

   return ret;
}


/**
   Allocate data structures of the PCM module for the given channel.

   \param  pCh          Pointer to the channel structure.

   \return
   - SDD_statusOk
   - SDD_statusNoMem    in case the stucture could not be created

  \remarks The channel parameter is no longer checked because the calling
   function assures correct values.
*/
IFX_int32_t SDD_PCM_ChStructAllocate (
                        SDD_CHANNEL_t *pCh)
{
   SDD_PCM_ChStructFree (pCh);

   pCh->pPCM = IFXOS_BlockAlloc(sizeof(*pCh->pPCM));
   if (pCh->pPCM == IFX_NULL)
   {
      /* errmsg: No memory could be allocated. */
      return SDD_statusNoMem;
   }
   memset(pCh->pPCM, 0, sizeof(*pCh->pPCM));

   return SDD_statusOk;
}


/**
   Free data structure of the PCM module in the given channel.

   \param  pCh          Pointer to the channel structure.
*/
IFX_void_t SDD_PCM_ChStructFree (
                        SDD_CHANNEL_t *pCh)
{
   if (pCh->pPCM != IFX_NULL)
   {
      IFXOS_BlockFree(pCh->pPCM);
      pCh->pPCM = IFX_NULL;
   }
}


/**
   Initalize the PCM module and the cached firmware messages

   \param  pCh          Pointer to the channel structure.
*/
IFX_void_t SDD_PCM_ChInit (
                        SDD_CHANNEL_t *pCh)
{
   struct SDD_PCM_CH_CTRL  *pPcmCh;
   IFX_uint8_t             ch = pCh->nChannel - 1;

   /* PCM ch message */
   pPcmCh = &pCh->pPCM->fwm_PcmChCtrl;
   memset (pPcmCh, 0, sizeof (*pPcmCh));
   pPcmCh->CMD    = CMD_EOP;
   pPcmCh->CHAN   = ch;
   pPcmCh->MOD    = MOD_PCM;
   pPcmCh->ECMD   = SDD_PCM_CH_CTRL_ECMD;
   pPcmCh->LENGTH = SDD_PCM_CH_CTRL_LENGTH;
   pPcmCh->EN     = SDD_PCM_CH_CTRL_DISABLE;
   pPcmCh->COD    = SDD_PCM_CH_CTRL_COD_LINEAR_16BIT;
   pPcmCh->WIDE   = SDD_PCM_CH_CTRL_WIDE_8KHZ;
}


/**
   Stop PCM on this channel

   \param  pCh          Pointer to the channel structure.

   \return
   - SDD_statusOk
   - SDD_statusCmdWr    Writing the command has failed
*/
IFX_int32_t SDD_PCM_ChStop (
                        SDD_CHANNEL_t *pCh)
{
   struct SDD_PCM_CH_CTRL *pPcmCh;
   SDD_DEVICE_t *pDev = pCh->pDev;
   IFX_int32_t ret   = SDD_statusOk;

   /* calling function should ensure valid parameters */
   SDD_ASSERT(pCh != IFX_NULL);

   if (pCh->pPCM != IFX_NULL)
   {
      /* protect fw msg */
      SDD_OS_MutexGet (&pCh->mtxChAcc);

      /* get pointer to cached fw message */
      pPcmCh = &pCh->pPCM->fwm_PcmChCtrl;

      /* PCM module deactivation is needed if the module is enabled */
      if ((pPcmCh->EN != SDD_PCM_CH_CTRL_DISABLE))
      {
         /* disable PCM channel */
         pPcmCh->EN = SDD_PCM_CH_CTRL_DISABLE;

         /* write PCM channel command */
         ret = SDD_CmdWrite (pDev, (IFX_uint32_t*)(IFX_void_t*)pPcmCh);
      }

      SDD_OS_MutexRelease (&pCh->mtxChAcc);
   }

   return ret;
}


#ifdef SDD_FEAT_PROCFS
/**
   Linux procfs read the status of PCM.

   \param  s            Pointer to seq_file struct.
   \param  pDev         Pointer to the device structure.
   \return  0 on success
*/
IFX_int32_t SDD_PCM_proc_StatusGet (
                        struct seq_file *s,
                        SDD_DEVICE_t *pDev)
{

   if ((pDev->pPCM != IFX_NULL) &&
       (pDev->nDevState & DS_PCM_EN))
   {
      IFX_uint32_t   i, j, Idx, BitValue;

      /* Print the timeslot allocation bitmap in groups of 8 bits. */

      seq_printf(s, "  PCM TS RX: ");
      for (i = j = 0;  i < pDev->pPCM->nMaxTimeslot; i++)
      {
         Idx = i >> 5;
         BitValue = 1 << (i & 0x1F);
         seq_putc(s,(pDev->pPCM->PcmRxTs[0][Idx] & BitValue) ? '1' : '0');
         if ((++j % 8) == 0)
         {
            seq_putc(s, ' ');
         }
      }
      seq_putc(s, '\n');

      seq_printf(s, "  PCM TS TX: ");
      for (i = j = 0;  i < pDev->pPCM->nMaxTimeslot; i++)
      {
         Idx = i >> 5;
         BitValue = 1 << (i & 0x1F);
         seq_putc(s,(pDev->pPCM->PcmTxTs[0][Idx] & BitValue) ? '1' : '0');
         if ((++j % 8) == 0)
         {
            seq_putc(s, ' ');
         }
      }
      seq_putc(s, '\n');
   }

   return 0;
}
#endif /* SDD_FEAT_PROCFS */
